#!/usr/bin/env python

from IGenerateOrder import *
from generate_order_7 import *
from Ordering_IDD import *
import sys


# Global information of the module
#  - all_t   ==> MAP List of transition where key is transition :
# 0 => Transition is not fired
# 1 => Transition is fired (do not use it)
# 2 => Transition is in the TSS Vector, and will be used later 
#  - all_p   ==> MAP List of place where key is place :
# 0 => Place is not ordered
# 1 => Place is ordered
#  - fires   ==> The total number of transition fired during the ordering variable
#  - tss     ==> Array vector of transitions which can be fired. From this vector, we must
# calculate weight of each transition and choose the transition to fired
#  - tps2    ==> Variable time calculation
all_t = {}
fires = 0
tss = []
all_p = {}
    

# Ordering Generation Strategy 8 :
# We use a Firing Model Heuristic for generate an FLAT Order, then we cut in into X Y Hierarchy
# This strategy use the firing technical of E. Paviot-Adet, and it consist to :
#  - consider all place marking as True or False (ignore marking)
#  - consider all arc valuation as True (ignore valuation)
#  - After a firing, we get all transition which can be fired only on the firing modifications places
#  - One transition can be fired only one time
#  - One transition must impacted the ordering vector
# ADVANTAGE :
#  - Very Quick firing
#  - No infinite loop of fires (transition can be fire only one time)
# Another optimization are implemented on the kind of data structure used (MAP List instead of Vector array)
class GenerateOrder_8(GenerateOrder_7):
    
    # Constructor :
    # net        : The model
    # path_conf  : The directory where we generate the JSON File configuration
    # api        : Cami API Communication
    def __init__(self,net,path_conf,api):
        GenerateOrder_7.__init__(self,net,path_conf,api)
        # Public Attribute Function, can be replace by another strategy function
        self.function_weight = self.weight
        self.function_fire = self.fire
        self.function_chooseFire = self.chooseFire
        self.function_checkFire = self.checkFire
        self.function_canFire = self.canFire
        self.function_chooseNTransition = self.chooseNTransition
        self.function_pre_firing_order = self.pre_firing_order
        
    
    # Print Information Order into XLS information
    def __str__(self):
        res = ''
        res += str(self.time)   + '\t' 
        res += str(self.places_count) + '\t' 
        res += str(self.transitions_count)  + '\t' 
        res += str(self.X)  + '\t' 
        res += str(self.Y) + '\t'
        res += 'Pre-Fired (PF-2) Model Flat + naive XY-Hierarchy'
        return res

    # Help command
    # Print help about option of this generation order
    def help(self):
        str = "Generation order 8 : Pre-Firing variable ordering 2 (PF-2) : optimization of the PF-1 + Hierarchy build with N-Cut-Naive \n"
        str += " --flat : desactivate the hierarchy (flat order)\n"
        str +="Parameters for N-Cut-Naive Hierarchy\n"
        str +="   --X <value> : Maximum height of structure\n"
        str +="   --Y <value> : Number of element per list (Y=2 by default)\n"
        str +="If X is not set, X is automatically calculated from X(Y)=Log[Y](count_places)"
        
        return str

    # Function weight :
    # Description :
    # Calculation of the transition weight and return all transition with the MAX Weight calculation
    def weight(self,order,ts):
        min = -20.0
        min_w = []
        global all_p
        # For each transition, calculate the Weight of the transition
        for t in ts:
            # Calculate W(T)
            w = 0.0
            places = set() 
            for p, tmp in t.predecessors():places.add(p)
            for p, tmp in t.successors():places.add(p)
            for p in places:
                #if p in order:
                if all_p[p] != None:
                    w += 1
            # If W == T(PRED) + T(SUCC)
            if len(places) == w:
                w = -1
            else:
                w = (w / len(places))
    
            # Build the list of transition with W = MIN
            if min < w:
                min_w = [] # Reseting
                min_w.append(t)
                min = w
            elif min == w:
                min_w.append(t)
    
        return min_w

    # Function chooseFire :
    # Description :
    # At each time of pre-firing transition, some cases can appears :
    #  - 0 transition can be fire : stop the algorithm
    #  - 1 transition can be fire : choose this one
    #  - N transitions can be fire
    #        - From N transitions, groups all which have the greatest Weight Calculation Transition (N[MAX_W])
    #        - From N[MAX_W] transitions and from the MAP List Transition Counter
    #            - Keep the one which have not recently fired, or the first one if thera are many transitions (TransitionChoose)
    #        - From the TransitionChoose, check if transition is not a indeterminism transition
    #            - If yes, fire the other and keep TransitionChoose
    #        - Then, check in no infinite loop are detected, if yes break the loop and return None
    #        - Return the TransitionChoose
    # Calculation of the weight :
    #  W ==> The number of place in PRED and SUCC already in the order vector
    #  if W ==  T(PRED) + T(SUCC) ==> All place are already in vector, W = -1
    #  W = -1 ==> All place of the transition are already in vector, so we prefere to fire
    #  another transition
    #        
    def chooseFire(self,order , ts , torder):
        min_w = self.function_weight(order,ts)
        # Choose the transition
        if len(min_w) == 0:
            # No transition can be fired : Break ALGO
            transitionChoose = None
        elif len(min_w) == 1:
            # Only one transition can be choose : choose it
            transitionChoose = min_w[0]
        else:
            # We must choose one transition among N one
            # First initialize the indeterminism transition table
            links = {}
            for t in ts:
                for p, tmp in t.predecessors():
                    if not links.has_key(p): 
                        links[p] = []
                    if not t in links[p]:
                        links[p].append(t)        
            
            # Among the N transition, choose Strategy
            transitionChoose = self.function_chooseNTransition(min_w)
            
            
            # At the end, check if the transition choose is no indeterminism, if yes
            # We must do a pre-fired all one which not the current transition
            for pla,arc in transitionChoose.predecessors():
                if len(links[pla]) > 1:
                    # We have found one indeterminism transition
                    for ttt in links[pla]:
                        if transitionChoose != ttt:
                            self.function_fire(order,ttt,torder)
                                
        return transitionChoose
    
    
    # Define function for choose one transition of N
    # It can be happening at the beginning firing model
    def chooseNTransition(self,min_w):
        # Among the N transition, choose the first one
        return min_w[0]
    
    
    # Function fire :
    # Description :
    # Apply the firing transition, only the POST Firing (PRE are not apply).
    # This is a wrong firing transition, because we don't want to generate one possibility
    # of the space state system, but we want to ensure that we fire the most transition as much as possible.
    # And so cover most of the place's model, directly touch by the firing transitions.
    # Flat Ordering precision :
    # At each fire, we put :
    #  - All predecessors places of the current transition if not exist in ordering vector
    #  - All successors places of the current transition if not exist in ordering vector
    #
    # OPTIMIZATION :
    # 1) New Firing strategy (avoid loop of all transition for know which transition firing)
    # After firing one transition, from all successors places impacting by the current one,
    # we must recalculate the current TSS vector for know which transitions is now firing
    # 2) Avoiding repetition transition :
    # After one firing, we must check in the current vector TSS if there are any transitions
    # that its fired will no add new place in Ordering vector. So we avoid the firing of Transition where all place are already in vector
    #
    # GLOBAL :
    #  --> Fires : Count the total number of fire
    #  --> all_t : For switch off some transition of the firing strategy because :
    #    - One transition can be fired only one time
    #    - One transition must impact the ordering vector, else that's mean all places
    # are already in vector, so transition is not to be used
    #  --> all_p : For know and note that one place is in order vector
    #  --> tss :  - After firing current transition :
    #     - calculate the new transition which can be fired
    #     - remove all transition that will not impacted the ordering vector 
    #
    # Parameter :
    #  - order : Variable Flat oder vector
    #  - transition : the current transition to fire
    #  - torder : the Transition vector ordering
    def fire(self,order,transition,torder):
        # Global used
        global fires,all_t,all_p,tss,tps2
        # Increment firing counter
        fires += 1
        # Check of transition was not removed from a previous firing
        # It can be happened when algorithme have fired some transition in structural conflict
        if not transition in tss:
            # Avoid function
            return
        
        # Add the Predecessors places in vector
        for place, arc in transition.predecessors():
            if all_p[place] == None:
                order.append(place)
                all_p[place] = 1
        # Add the Successors places in vector
        for place, arc in transition.successors():
            place.marking = 1 # Now we work only in True, False Value on place marking
            if all_p[place] == None:
                order.append(place)
                all_p[place] = 1
                
        # Note that current transition fired, and remove it from TSS vector
        all_t[transition] = 1
        if transition in tss:
            tss.remove(transition)
        
        # After a fired, check if all transition in TSS vector will impact
        # the ordering vector, of not remove, removed it and note it
        ttttt = []
        for t in tss:
            if not self.function_checkFire(t):
                all_t[t] = 1
                ttttt.append(t)
        for t in ttttt:
            tss.remove(t)
                
            
        # After a fired, update the new transition will can be fire
        # Check only transition impacted by the current successors places 
        for place, arc in transition.successors():
            for T,arc in place.successors():
                # only if we can use the transition t
                if all_t[T] == None:
                    if self.function_canFire(T):
                        # Say that current transition is in vector
                        tss.append(T)
                        all_t[T] = 2
    
    # Function checkFire :
    # Description :
    # This function is used for check if all places PRED and SUCC of <transition>
    # are note in ordering vector
    def checkFire(self,transition):
        global all_p
        for place, arc in transition.successors():
            if all_p[place] == None:
                return True
        for place, arc in transition.predecessors():
            if all_p[place] == None:
                return True
        return False

    # Function canFire :
    # Description :
    # Check if on transition can be fire and if yes,
    # we assure that the transition will impact the ordering vector
    # Parameter :
    #  - transition : the current transition to check
    def canFire(self,transition):
        for place, arc in transition.predecessors():
            if int(place.marking) == 0:
                return False
        return self.function_checkFire(transition)
    
    def pre_firing_order(self):
          # Ordering Variable consist into Pre-firing Net
          # and list an order on the first Variable Firing
          # Ordering Transition by Pre-Firing order Net
          order = Hierarchy()
          torder = Hierarchy()
          
          # Total of place and transition and global initialization
          # We must initialize TSS vector with the current firing transition
          global all_t,all_p,tss
          for p in self.net.places():
              all_p[p] = None # Not in ordering vector
          for t in self.net.transitions():
              all_t[t] = None # Not fired and not used
              if self.function_canFire(t):
                tss.append(t)
                all_t[t] = 2 # In used
          self.api.write('Number of places into model : ' + str(self.places_count))
          self.api.write('Number of transitions into model : ' + str(self.transitions_count))
    
          # While we can fired any transitions
          while len(tss) !=  0 and len(order) != self.places_count:
              if len(tss) != 0:
                  tt = self.function_chooseFire(order,tss,torder)
                  if tt != None:
                      self.function_fire(order,tt,torder)
          
          # If all places is not in ordering vector, replacing it by First Predecessors heuristics
          diffP = self.places_count - order.size()        
          if diffP > 0:
              for p in self.net.places():
                  if not p in order:
                    for tra,arc in p.predecessors():
                        for pla,arc in tra.predecessors():
                            if pla in order and not (p in order):
                                order.insert( order.index(pla),p)
                  # At the end, if P not in the end, adding it to the end of vector
                  if not p in order:
                      order.append(p)     
          global fires
          self.api.write('')
          self.api.write('')
          self.api.write('Statistic of the Pre-Fired Heuristic')
          self.api.write('Total of transition firing                         : '+ str(fires))
          self.api.write('Total of place touched by firing                   : ' + str(self.places_count-diffP) + '/' + str(self.places_count))
          self.api.write('Total of place replacing by first predecessor      : '+ str(diffP))
          self.api.write('Total time of the ordering                         : '+str(self.time) + ' s')
          self.api.write( '')
          self.api.write('')
          
          return order,torder
    
    # Strategy Operation for generate one Configuration
    def generate_order(self):
    
      order,torder = self.function_pre_firing_order()
      
      # Calculate Automatic or a User definition of XY-Hierarchy
      self.X , self.Y = self.calculate_XY(len(order))
      # Naive XY Build Hierarchy
      order = self.write_partition(order,self.X,self.Y)
      if '--flat' in sys.argv:
          order =flating(order)
      # Write JSON File
      conf = []
      conf.append(self.path_conf)
      write_conf(self.path_conf,order)
    
      return conf
          

