import functools

############################################################################

class Base(object):

  def __init__(self,net,identifier):
    self.net = net
    self.id  = identifier

  def __getattr__(self,name):
    
    if type(self) in attributes_mapping:
      if name in attributes_mapping[type(self)]:
        return attributes_mapping[type(self)][name]
      else:
        raise Exception( "'" + str(name) + "' attribute not found in mapping " + str(attributes_mapping))
    else:      
      raise Exception(str(type(self)) + " does not exist")

  def __setattr__(self,name,value):
    
    #if type(self) in attributes_mapping:
    #  if name in attributes_mapping[type(self)]:
    #    value = type(attributes_mapping[type(self)][name])(value)
    
    # to avoid recursive call, we directly work on __dict__ so as to avoid 
    # to call __setattr__ again
    self.__dict__[name] = value

############################################################################

class Arc(Base):
  
  def __init__(self,net,identifier,src,dst):
    if src == None:
      raise Exception("Source is undefined")
    if dst == None:
      raise Exception("Destination is undefined")
    super(Arc,self).__init__(net,identifier)
    self.source      = src
    self.destination = dst
    self.source.add_arc(self)
    self.destination.add_arc(self)    

  def __str__(self):
    return str(self.source) + " -> " + str(self.destination)

############################################################################

class Node(Base):

  """
  Define the common behavior of Places and Transisitions.
  A node has a set of predecessors. Arcs between a node and
  its predecessors hold valuations.
  """

  def __init__(self,net,identifier):
    super(Node,self).__init__(net,identifier)
    self.succ = {}
    self.pred = {}

  def predecessors(self):
    """
    Iterate over predecessors of a node. Each call
    returns a pair < predecessor , valuation >.
    """
    for x , arc in self.pred.items():
      yield x , arc


  def successors(self):
    """
    Iterate over successors of a node. Each call
    returns a pair < successor , valuation >.
    """
    for x , arc in self.succ.items():
      yield x , arc


  def add_arc(self,arc):
    
    if arc.source == self:

      self.check_arc(arc.destination)
      self.succ[arc.destination] = arc

    elif arc.destination == self:

      self.check_arc(arc.source)
      self.pred[arc.source] = arc
      
    else:
      
      raise Exception( "Node " + node + " is not an end of arc " + arc )

############################################################################

class Place(Node):
  
  def __init__(self,net,identifier):
    super(Place,self).__init__(net,identifier)

  def check_arc(self,end):
    if not isinstance(end,Transition):
      raise Exception( end + " is not a transition")

  def __str__(self):
    return self.name + "(" + str(self.marking) + ")"

############################################################################

class Transition(Node):
  
  def __init__(self,net,identifier):
    super(Transition,self).__init__(net,identifier)    

  def check_arc(self,end):
    if not isinstance(end,Place):
      raise Exception( end + " is not a place")

  def __str__(self):
    return self.name

############################################################################

class Net(Base):
  
  def __init__(self):

    super(Net,self).__init__(self,1)
    
    self._nodes = {}
    self._arcs  = {}
    self._unprocessed = {}

    # dynamically create methods for iteration on arcs or nodes
    for ( cami_type , elt_type , method_name ) in cami_mapping.mapping:
      
      if elt_type != None:
      
        if issubclass(elt_type,Node):
          self.__dict__[method_name] = functools.partial(self.iter_nodes_or_arcs,self._nodes,elt_type)
          
        elif issubclass(elt_type,Arc):
          self.__dict__[method_name] = functools.partial(self.iter_nodes_or_arcs,self._arcs,elt_type)


  def create_node(self,cami_node_type,identifier):

    if cami_node_type == "net":
      return self

    else:

      # get the type of the class that stores a node_type
      pn_type = cami_mapping.cami_as_pn(cami_node_type)

      n = pn_type(self,identifier)
      self._nodes[identifier] = n
      return n



  def create_arc(self,cami_arc_type,identifier,src,dst):
    
    # get the type of the class that stores a node_type
    pn_type = cami_mapping.cami_as_pn(cami_arc_type)

    a = pn_type(self,identifier,src,dst)
    self._arcs[identifier] = a
    return a
    

  def all_nodes(self):
    return self._nodes
  
    
  def all_arcs(self):
    return self._arcs
    
    
  def get(self,identifier):

    if identifier == self.id:
      return self

    elif identifier in self._nodes:
      return self._nodes[identifier]

    elif identifier in self._arcs:
      return self._arcs[identifier]
        
    else:
      raise Exception("No node with id " + str(identifier))
      
      
  def iter_nodes_or_arcs(self,container,node_type):
    
    for n in container.values():
      if type(n) == node_type:
        yield n
      else:
        continue

############################################################################

attributes_mapping = {
  Arc        : { 'valuation' : '1' } 
, Place      : { 'name' : "" , 'marking' : '0' , 'domain' : None }
, Transition : { 'name' : "" }
, Net        : { 'authors' : "" , 'declaration' : None }
}

############################################################################

class cami_mapping(object):

  mapping = [
    ( "arc"            , Arc        , "arcs"        )
  , ( "inhibitor"      , Arc       ,  "arcs"          )
  , ( "test"      , Arc       ,  "arcs"          )
  , ( "reset"      , Arc       ,  "arcs"          )
  , ( "place"          , Place      , "places"      )
  , ( "queue"          , None       , None          )
  , ( "transition"     , Transition , "transitions" )
  , ( "immediate"      , None       , None          )
  ]

  @staticmethod
  def cami_as_pn(cami_type):

    for ( cami_t , pn_t , method_name ) in cami_mapping.mapping:
      if cami_type == cami_t:
        return pn_t

    raise Exception( str(cami_type) + " CAMI type not supported")

  @staticmethod
  def pn_as_cami(pn_type):
  
    for ( cami_t , pn_t , method_name ) in cami_mapping.mapping:
      if type(pn_type) == pn_t:
        return cami_t

    raise Exception( str(type(pn_type)) + " pn type not supported")
