/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  checkctl.c
*
* Part of: PROD's program probe
*
* Authors: Keijo Heljanko
*
* Current version:
*     $Id: checkctl.c,v 3.1 1997/04/17 15:43:53 kepa Exp $
*
* Description:
*     This module implements the main loop of them CTL model
*     checker. See the module temporal.c for subroutines
*     for checking the temporal formulas.
*
* Version history:
*     $Log: checkctl.c,v $
 * Revision 3.1  1997/04/17  15:43:53  kepa
 * *** empty log message ***
 *
***************************************************************************/


/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "baltree.h"
#include "ptrbl.h"
#include "bdd.h"
#include "gphutil.h"
#include "graph.h"
#include "mark.h"
#include "expr.h"
#include "formula.h"
#include "dassert.h"
#include "bitarr.h"
#include "stackpro.h"
#include "temporal.h"
#include "checkctl.h"
#include "prepctl.h"

static unsigned long *cad = NULL;
static unsigned long *caw = NULL;
/***************************************************************************
*
* Function:  ModelCheckCTL
* Purpose:   An interface from the probe code to the CTL model
*            checker. Also does the atomic subformula evaluation.
* Arguments: formula - The probe ctl_formula to be checked
*            graph   - The probe graph we are checking.
* Returns:   TRUE on success and FALSE on failure.
*            Also the results are printed out here.
* See also:
*
****************************************************************************
*/
FE3(BOOL, ModelCheckCTL,
    FORMULA *, formula,
    GRAPH *, graph,
    NODE, state)
{
    char *s;
    BOOL result = TRUE;
    LIST *formulas;
    unsigned long num_states;
    STACK *path;
    GRAPH_SUCC g;
    unsigned long node;
    
    /* First read in the arrows files to memory for speed.
       We must later add some checking to ensure that the
       files are small enough before loading them into memory.

       In that case a slower more disk-resident algorithm must
       be used in model-checking. */

    if(cad == NULL) {
	s = FileName(graph->name, "cad");
	cad = (unsigned long *)GetFile(s);
	if(cad == NULL) {
	    Deallocate((char *)s);
	    return FALSE;
	}
	Deallocate((char *)s);
    }

    if(caw == NULL) {
	s = FileName(graph->name, "caw");
	caw = (unsigned long *)GetFile(s);
	if(caw == NULL) {
	    Deallocate((char *)s);
	    return FALSE;
	}
	Deallocate((char *)s);
    }
    
    /* Initialize the CTL formulas. */
    
    formulas = LstCreate(&CTLFormulaTools);
    num_states = GphNdCount(graph);
    
    (void) InitVisitFormulas(formula, formulas, num_states);
    (void) RemoveNegations(formulas);

    /* Find and evaluate all atomic subformulas */

    (void)fprintf(_out, "Evaluating atomic subformulas.\n");
    (void)fflush(_out);

    (void) EvalAtomicFormulas(graph, formulas);
    
    (void)fprintf(_out, "Evaluating CTL subformulas.\n");
    (void)fflush(_out);

    /* Do the real model checking */

    path = CreateStack();
    g.cad = cad;
    g.caw = caw;
    
    result = CheckCTL(formulas,
		      state,
		      &g,
		      num_states,
		      path);

    /* Print results */

    if(result == FALSE) {
	s = "false";
    } else {
	s = "true";
    }

    (void)fprintf(_out, "Result: %s\n", s);
    (void)fflush(_out);

    if(!EmptyStack(path)) {
    
	(void)fprintf(_out, "Path:\n");

	while(!EmptyStack(path)) {

	    node = PopStack(path);

	    (void)fprintf(_out, "%lu", node);

	    if(!EmptyStack(path)) {
		(void)fprintf(_out, " [> ");
	    }
	    
	}
	(void)fprintf(_out, "\n");
	(void)fflush(_out);

    }
    
    /* Deallocate everything */

    LstDestroy(formulas, TRUE);

#if 0
    Dehallocate((char *) caw);
    Dehallocate((char *) cad);
#endif

    DeleteStack(path);
    
    return result;
}


/***************************************************************************
*
* Function:  RemoveNegations
* Purpose:   Removes negations from the end of the subformula list
*            to enable counterexample and witness processing.
*
* Arguments: formulas   - a list of CTL formulas
*
* Returns:   
* See also:
*
****************************************************************************
*/
FE1(BOOL, RemoveNegations,
    LIST *, formulas)
{
    /* int size; */
    int negations;
    BOOL odd;
    CTL_FORMULA *node;

    /* size=0; */
    negations=0;
    odd=FALSE;

    LSTFOR(formulas, node, CTL_FORMULA) {
	/* size++; */
	if(ctl_formula_type(node) == NOT_FORMULA){
	    negations++;
	} else {
	    negations=0;
	}
    }

    if((negations%2) != 0) {
	odd=TRUE;
    }

    while(negations--){
	(void) LstDelLast(formulas, TRUE);		
    }

    ((CTL_FORMULA *)LstLast(formulas))->top_level = TRUE;
    if(odd == TRUE) {
	((CTL_FORMULA *)LstLast(formulas))->negated = TRUE;
    }

    return TRUE;
}


/***************************************************************************
*
* Function:  InitVisitFormulas
* Purpose:   Initialize the true CTL formulas from the parsed
*            ones and make a list with subformulas before formulas
*            that use them. The top-level formula will be the last in this
*            list. Also intialize lots of thing for the model checker.
* Arguments: formula    - a parsed formula
*            formulas   - a list of CTL formulas
*            num_states - the number of states in the graph
* Returns:   
* See also:
*
****************************************************************************
*/
FE3(CTL_FORMULA *, InitVisitFormulas,
    FORMULA *, formula,
    LIST *, formulas,
    unsigned long, num_states)
{
    CTL_FORMULA *cf;

    cf = NULL;

    /* This is NOT intelligent, fix this. */

    
    cf = (CTL_FORMULA *)Allocate(sizeof(CTL_FORMULA));

    cf->formulatype = EVALUATED_FORMULA;
    cf->left_subformula = NULL;
    cf->right_subformula = NULL;
    cf->result = NULL;
    cf->top_level = FALSE;
    cf->negated = FALSE;
    cf->p = NULL;
    
    switch(formula->operator) {

      case FO_CTL_FALSE:
      case FO_FALSE:
      case FO_CTL_TRUE:
      case FO_TRUE:

	cf->formulatype = EVALUATED_FORMULA;
	cf->result = CreateBitArray(num_states);
	if(formula->operator == FO_CTL_TRUE ||
	   formula->operator == FO_TRUE) {
	    BitArrayNot(cf->result, cf->result);
	}
	break;

      case FO_CTL_NOT:

	cf->formulatype = NOT_FORMULA;
	cf->right_subformula =
	    InitVisitFormulas(formula->uf.of.sub1f, formulas, num_states);
	break;

      case FO_CTL_EG:
      case FO_CTL_AG:

	cf->formulatype = NOT_FORMULA;
	cf->right_subformula = (CTL_FORMULA *)Allocate(sizeof(CTL_FORMULA));

        if(formula->operator == FO_CTL_EG) {
	    cf->right_subformula->formulatype = AU_FORMULA;
	} else {
	    cf->right_subformula->formulatype = EU_FORMULA;
	}

	cf->right_subformula->left_subformula = NULL;
	cf->right_subformula->right_subformula = NULL;
	cf->right_subformula->result = NULL;
	cf->right_subformula->top_level = FALSE;
	cf->right_subformula->negated = FALSE;
	cf->right_subformula->p = NULL;

	/* Create a dummy subformula which is TRUE in all states */
	cf->right_subformula->left_subformula = (CTL_FORMULA *)Allocate(sizeof(CTL_FORMULA));
	cf->right_subformula->left_subformula->formulatype = EVALUATED_FORMULA;
	cf->right_subformula->left_subformula->left_subformula = NULL;
	cf->right_subformula->left_subformula->right_subformula = NULL;
	cf->right_subformula->left_subformula->result = CreateBitArray(num_states);
	cf->right_subformula->left_subformula->top_level = FALSE;
	cf->right_subformula->left_subformula->negated = FALSE;
	cf->right_subformula->left_subformula->p = NULL;
	BitArrayNot(cf->right_subformula->left_subformula->result, cf->right_subformula->left_subformula->result);
	(void) LstInsLast(formulas, (char *)cf->right_subformula->left_subformula);

	cf->right_subformula->right_subformula = (CTL_FORMULA *)Allocate(sizeof(CTL_FORMULA));
	cf->right_subformula->right_subformula->formulatype = NOT_FORMULA;
	cf->right_subformula->right_subformula->left_subformula = NULL;
	cf->right_subformula->right_subformula->right_subformula = NULL;
	cf->right_subformula->right_subformula->result = NULL;
	cf->right_subformula->right_subformula->top_level = FALSE;
	cf->right_subformula->right_subformula->negated = FALSE;
	cf->right_subformula->right_subformula->p = NULL;

        cf->right_subformula->right_subformula->right_subformula =
	    InitVisitFormulas(formula->uf.of.sub1f, formulas, num_states);
	(void) LstInsLast(formulas, (char *)
			  cf->right_subformula->right_subformula);
	(void) LstInsLast(formulas, (char *)cf->right_subformula);
	break;

      case FO_CTL_EX:
      case FO_CTL_AX:

	if(formula->operator == FO_CTL_EX) {
	    cf->formulatype = EX_FORMULA;
	} else {
	    cf->formulatype = AX_FORMULA;
	}
	cf->left_subformula =
	    InitVisitFormulas(formula->uf.of.sub1f, formulas, num_states);
	cf->right_subformula = (CTL_FORMULA *)Allocate(sizeof(CTL_FORMULA));

	cf->right_subformula->formulatype = EVALUATED_FORMULA;
	cf->right_subformula->left_subformula = NULL;
	cf->right_subformula->right_subformula = NULL;
	cf->right_subformula->result = CreateBitArray(num_states);
	cf->right_subformula->top_level = FALSE;
	cf->right_subformula->negated = FALSE;
	cf->right_subformula->p = NULL;

	(void) LstInsLast(formulas, (char *)cf->right_subformula);

	break;

      case FO_CTL_AND:
      case FO_CTL_OR:
	if(formula->operator == FO_CTL_AND) {
	    cf->formulatype = AND_FORMULA;
	} else {
	    cf->formulatype = OR_FORMULA;
	}
	cf->left_subformula =
	    InitVisitFormulas(formula->uf.of.sub1f, formulas, num_states);
	cf->right_subformula =
	    InitVisitFormulas(formula->uf.of.sub2f, formulas, num_states);
	break;

	
      case FO_CTL_EF:
      case FO_CTL_AF:
      case FO_CTL_EU:
      case FO_CTL_AU:
	if(formula->operator == FO_CTL_EU ||
           formula->operator == FO_CTL_EF) {
	    cf->formulatype = EU_FORMULA;
	} else {
	    cf->formulatype = AU_FORMULA;
	}
	if(formula->operator == FO_CTL_EU ||
           formula->operator == FO_CTL_AU) {
	    cf->left_subformula =
		InitVisitFormulas(formula->uf.of.sub1f, formulas, num_states);
 	    cf->right_subformula =
		InitVisitFormulas(formula->uf.of.sub2f, formulas, num_states);
        } else {
	    /* Create a dummy subformula which is TRUE in all states */
	    cf->left_subformula = (CTL_FORMULA *)Allocate(sizeof(CTL_FORMULA));

	    cf->left_subformula->formulatype = EVALUATED_FORMULA;
	    cf->left_subformula->left_subformula = NULL;
	    cf->left_subformula->right_subformula = NULL;
	    cf->left_subformula->result = CreateBitArray(num_states);
	    cf->left_subformula->top_level = FALSE;
	    cf->left_subformula->negated = FALSE;
	    cf->left_subformula->p = NULL;
	    BitArrayNot(cf->left_subformula->result, cf->left_subformula->result);
	    (void) LstInsLast(formulas, (char *)cf->left_subformula);

            cf->right_subformula =
		InitVisitFormulas(formula->uf.of.sub1f, formulas, num_states);
	}
	break;

      case FO_CTL_EXPR:
	cf->formulatype = ATOMIC_FORMULA;
	cf->result = CreateBitArray(num_states);
	cf->p = formula->uf.expression;
	break;

      default:
	(void)fprintf(_err, "\nUnknown CTL formula type %u.\n",
		      formula->operator);
	(void)fflush(_err);
	abort();
    }

    if(cf != NULL) {
	(void) LstInsLast(formulas, (char *)cf);
    }

    return cf;
    
}



/***************************************************************************
*
* Function:  CheckCTL
* Purpose:   Do CTL model checking.
* Arguments: 
* Returns:   
* See also:
*
****************************************************************************
*/
FE5(BOOL, CheckCTL,
    LIST *, formulas,
    NODE, s,
    GRAPH_SUCC *, g,
    unsigned long, num_states,
    STACK *, path)
{
    CTL_FORMULA *node;
    BOOL result = FALSE;

    LSTFOR(formulas, node, CTL_FORMULA) {

	result = CheckCTLFormula(node,
				 s,
				 g,
				 num_states,
				 is_toplevel(node),
				 path);

	if(is_toplevel(node)) {
	    if(is_negated(node)){
		if(result == FALSE) {
		    result = TRUE;
		} else {
		    result = FALSE;
		}
	    }
	}
    }

    return result;

}

/***************************************************************************
*
* Function:  CheckCTLFormula
* Purpose:   Check the truth value of the formula f when the
*            sub-formulas it needs are already calculated.
*            Subformula truth-values are overwritten, so save them
*            for if they are needed afterwards.
*
* Arguments: f -    The formula to be checked
*            s -    The state we are interested in
*
*            g    - Graph pointer for finding out successor nodes.
*            num_states - Number of states in the graph.
*
*            top_level - Is a counter_example/witness wanted.
*
*            path - A counter-example/witness path if wanted.
*
* Returns:   The truth value of the formula in s. If the top_level argument
*            is FALSE, also a counter-example/witness path is returned
*            if requested. Note: In this case the formula f is NOT evaluated
*            for all states in the graph!
* See also:
*
****************************************************************************
*/
FE6(BOOL, CheckCTLFormula,
    CTL_FORMULA *, f,
    NODE, s,
    GRAPH_SUCC *, g,
    unsigned long, num_states,
    BOOL, top_level,
    STACK *, path)
{
    switch(ctl_formula_type(f)) {

      case EVALUATED_FORMULA:
	break;
	
      case ATOMIC_FORMULA:

	/* All atomic subformulas should be evaluated already. */

	ASSERT(FALSE);
	break;
	
      case AND_FORMULA:

	(void) LabelAnd(s,
		 left_subformula_result(f),
		 right_subformula_result(f),
		 /* g,
		 num_states, */
		 top_level,
		 path);

	break;

      case OR_FORMULA:

	(void) LabelOr(s,
		left_subformula_result(f),
		right_subformula_result(f),
		/* g,
		num_states, */
		top_level,
		path);

	break;

      case NOT_FORMULA:

	(void) LabelNot(s,
		 right_subformula_result(f),
		 /* g,
		 num_states, */
		 top_level,
		 path);

	break;

      case EX_FORMULA:

	(void) CheckEX(s,
		left_subformula_result(f),
		right_subformula_result(f),
		g,
		num_states,
		top_level,
		path);

	break;

      case AX_FORMULA:

	(void) CheckAX(s,
		left_subformula_result(f),
		right_subformula_result(f),
		g,
		num_states,
		top_level,
		path);

	break;

      case EU_FORMULA:

	(void) CheckEU(s,
		left_subformula_result(f),
		right_subformula_result(f),
		g,
		num_states,
		top_level,
		path);

	break;

      case AU_FORMULA:

	(void) CheckAU(s,
		left_subformula_result(f),
		right_subformula_result(f),
		g,
		num_states,
		top_level,
		path);

	break;

      default:
	ASSERT(FALSE);
	break;
    }


    /* The rightmost subformula is now overwritten
       by result. */

    if(ctl_formula_type(f) != EVALUATED_FORMULA) {
	set_result(f);
	mark_evaluated(f);
    }

    return GetBit(formula_result(f), s);
}


/***************************************************************************
*
* Function:  LabelAnd
* Purpose:   Take the logical AND of the two subformulas.
* Arguments: s -  state we are interested in
*            f1 - leftmost subformula
*            f2 - rightmost subformula, overwritten by f on exit.
*            top_level - Is a counter_example/witness wanted.
*            path - A counter-example/witness path if wanted.
* Returns:   The truth of the formula in state s.
* See also:
*
****************************************************************************
*/
FE5(BOOL, LabelAnd,
    NODE, s,
    BITARRAY *, f1,
    BITARRAY *, f2,
    BOOL, top_level,
    STACK *, path)
{

    if(top_level == FALSE) {
	BitArrayAnd(f1, f2, f2);
    } else {

	ClearStack(path);
    
	/* If we are evaluating a top-level formula,
	   we only calculate the truth value of the
	   formula in the start state. */
	
	if(GetBit(f1,s) &&
	   GetBit(f2,s)) {
	    SetBit(f2, s);
	} else {
	    ClearBit(f2, s);
	}

	/* Maybe we should also do the following: */
	
	/*
	PushStack(path, s);
	*/
	
    }

    return GetBit(f2, s);
}

/***************************************************************************
*
* Function:  LabelOr
* Purpose:   Take the logical OR of the two subformulas.
* Arguments: s -  state we are interested in
*            f1 - leftmost subformula
*            f2 - rightmost subformula, overwritten by f on exit.
*            top_level - Is a counter_example/witness wanted.
*            path - A counter-example/witness path if wanted.
* Returns:   The truth of the formula in state s.
* See also:
*
****************************************************************************
*/
FE5(BOOL, LabelOr,
    NODE, s,
    BITARRAY *, f1,
    BITARRAY *, f2,
    BOOL, top_level,
    STACK *, path)
{

    if(top_level == FALSE) {
	BitArrayOr(f1, f2, f2);
    } else {

	ClearStack(path);
    
	/* If we are evaluating a top-level formula,
	   we only calculate the truth value of the
	   formula in the start state. */
	
	if(GetBit(f1,s) ||
	   GetBit(f2,s)) {
	    SetBit(f2, s);
	} else {
	    ClearBit(f2, s);
	}

	/* Maybe we should also do the following: */
	
	/*
	PushStack(path, s);
	*/
	
    }

    return GetBit(f2, s);
}


/***************************************************************************
*
* Function:  LabelNot
* Purpose:   Take the logical NOT of the subformula.
* Arguments: s -  state we are interested in
*            f2 - rightmost subformula, overwritten by f on exit.
*            top_level - Is a counter_example/witness wanted.
*            path - A counter-example/witness path if wanted.
* Returns:   The truth of the formula in state s.
* See also:
*
****************************************************************************
*/
FE4(BOOL, LabelNot,
    NODE, s,
    BITARRAY *, f2,
    BOOL, top_level,
    STACK *, path)
{

    if(top_level == FALSE) {
	BitArrayNot(f2, f2);
    } else {

	ClearStack(path);
    
	/* If we are evaluating a top-level formula,
	   we only calculate the truth value of the
	   formula in the start state. */
	
	if(!GetBit(f2,s)){
	    SetBit(f2, s);
	} else {
	    ClearBit(f2, s);
	}

	/* Maybe we should also do the following: */
	
	/*
	PushStack(path, s);
	*/
	
    }

    return GetBit(f2, s);
}

/***************************************************************************
*
* Function:  CTLFormulaCp
* Purpose:   A function to use in CTL formula list handling.
*            Not implemented yet.
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE0(CTL_FORMULA *, CTLFormulaCp)
{
    ASSERT(FALSE);
    return NULL;
}

/***************************************************************************
*
* Function:  CTLFormulaDeallocate
* Purpose:   A function to use in CTL formula list handling.
*            Not implemented yet.
* Arguments: f - the formula to be deallocated
* Returns:   
* See also:
*
****************************************************************************
*/
FE1(void, CTLFormulaDeallocate,
    CTL_FORMULA *, f)
{
    if(f->result != NULL) {
	DeleteBitArray(f->result);
	f->result = NULL;
    }

    /* Other fields are removed from memory elsewhere. */

}

#ifdef DEBUG

/***************************************************************************
*
* Function:  CTLFormulaDump
* Purpose:   A function to use in CTL formula list handling.
*            Not implemented yet.
* Arguments: f - the formula to be deallocated
*            depth - indentation depth
* Returns:   
* See also:
*
****************************************************************************
*/
FE2(void, CTLFormulaDump,
    CTL_FORMULA *, f,
    unsigned, depth)
{
    ASSERT(FALSE);
}

#endif

/* A collection of routines for ctl formulas to be used by
 * list routines of dynam.c and table routines of ptrbl.c. */
TOOLS CTLFormulaTools = {
  (CFUNC) CTLFormulaCp, CTLFormulaDeallocate
#ifdef DEBUG
, CTLFormulaDump
#endif
};
