/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  dynam.c
*
* Part of: PROD's programs araprod, "gphgen", probe, prpp, strong, and t
*
* Authors: Marko Rauhamaa and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This module contains general list routines and string handling.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include <stdio.h>
#include <string.h>
#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"

static LIST *listStack; /* a global stack of list cursors */

/***************************************************************************
*
* Function:  DynamInit
* Purpose:   This routine creates a global stack of list cursors,
*            initialized to be empty.
* Arguments:
* Returns:
* See also:  LstStore, LstRestore
*
****************************************************************************
*/
FE0(void, DynamInit)
{
    listStack = LstCreate((TOOLS *) NULL);
}

/***************************************************************************
*
* Function:  DumpIndent
* Purpose:   This routine is used for indenting when data structures
*            are displayed. It prints a sequence of dots to indicate
*            a margin.
* Arguments: depth - this many dots are printed, starting from the first
*                    column
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, DumpIndent,
    unsigned, depth)
{
    while (depth--)
	(void) fputc((int) '.', _err);
}

/***************************************************************************
*
* Function:  StrCp
* Purpose:   This routine makes a copy of a character string.
* Arguments: string - the string to be copied
* Returns:   the copy
* See also:  StrDump, StrPrefix (gphutil.c)
*
****************************************************************************
*/
FE1(char *, StrCp,
    char *, string)
{
    return StrCopy(string);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  StrDump
* Purpose:   This routine displays a character string.
* Arguments: string - the string
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  StrCp, StrPrefix (gphutil.c)
*
****************************************************************************
*/
FE2(void, StrDump,
    char *, string,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "STRING \"%s\"\n", string);
}
#endif

/* A collection of routines for character strings to be used by
 * list routines of dynam.c and table routines of ptrbl.c. */
TOOLS strTools = {
  (CFUNC) StrCp, Deallocate
#ifdef DEBUG
, StrDump
#endif
};

/***************************************************************************
*
* Function:  LstCatenate
* Purpose:   This routine concatenates two lists.
* Arguments: list1 - the first list (will contain the result)
*            list2 - the second list (will be made empty here)
* Returns:   the result of concatenating list2 after list1
* See also:  all routines having a name of the form Lst[A-Za-z]+
*
****************************************************************************
*/
FE2(LIST *, LstCatenate,
    LIST *, list1,
    LIST *, list2)
{
    if (!list1 || !list2)
	return(NULL);
    if (!list1->_first)
	{
	    list1->_first = list2->_first;
	    list1->_last = list2->_last;
	}
    else if (list2->_first)
	{
	    list1->_last->_next = list2->_first;
	    list1->_last = list2->_last;
	}
    list2->_first = list2->_this = list2->_last = NULL;
    return list1;
}

/***************************************************************************
*
* Function:  LstCopy
* Purpose:   This routine makes a copy of a list.
* Arguments: list - the list to be copied
*            trueCopy - elements are copied if and only if trueCopy is TRUE
* Returns:   the copy
* See also:  all routines having a name of the form Lst[A-Za-z]+
*
****************************************************************************
*/
FE2(LIST *, LstCopy,
    LIST *, list,
    BOOL, trueCopy)
{
    LIST *copy;
    VOID *node, *nodecopy;

    if (!(copy = LstCreate(list->_tools)))
	return(NULL);
    if (trueCopy)
	LSTFOR (list, node, void)
	    {
		if (!(nodecopy = (*(list->_tools->_copy))(node)))
		    {
			LstDestroy(copy, TRUE);
			return(NULL);
		    }
		if (!LstInsLast(copy, (char *) nodecopy))
		    {
			(*(list->_tools->_discard))(nodecopy);
			LstDestroy(copy, TRUE);
			return(NULL);
		    }
	    }
    else
	LSTFOR (list, node, void)
	    {
		if (!LstInsLast(copy, (char *) node))
		    {
			LstDestroy(copy,FALSE);
			return(NULL);
		    }
	    }
    return copy;
}

/***************************************************************************
*
* Function:  LstCreate
* Purpose:   This routine creates a list, initialized to be empty.
* Arguments: tools - a collection of routines needed for copying, deleting
*                    and displaying of an element in a list
* Returns:   the created list
* See also:  all routines having a name of the form Lst[A-Za-z]+
*
****************************************************************************
*/
FE1(LIST *, LstCreate,
    TOOLS *, tools)
{
    LIST *list;

    if (!(list = (LIST *) Allocate(sizeof(LIST))))
	return(NULL);
    list->_first = list->_this = list->_last = NULL;
    list->_tools = tools;
    return list;
}

/***************************************************************************
*
* Function:  LstDelete
* Purpose:   This routine deletes an element from a list.
* Arguments: list - the list
*            node - the element to be deleted
*            discard - the element is removed from the memory if and
*                      only if discard is TRUE
* Returns:   the element if it was found and discard is FALSE,
*            and NULL otherwise
* See also:  all routines having a name of the form Lst[A-Za-z]+
*
****************************************************************************
*/
FE3(VOID *, LstDelete,
    LIST *, list,
    char *, node,
    BOOL, discard)
{
    LISTNODE *l, *p;

    if (!list)
	return(NULL);
    for (p = NULL, l = list->_first;
	 l && (node != (char *) (l->_node));
	 p = l, l = l->_next)
	;
    if (!l)
	return(NULL);
    if (p)
	p->_next = l->_next;
    if (l == list->_first)
	list->_first = l->_next;
    if (l == list->_this)
	list->_this = p;
    if (l == list->_last)
	list->_last = p;
    Deallocate((char *) l);
    if (discard)
	{
	    (*(list->_tools->_discard))(node);
	    return(NULL);
	}
    else
	return(Pvoid(node));
}

/***************************************************************************
*
* Function:  LstDelFirst
* Purpose:   This is like LstDelete, except that the first element
*            in the list is deleted
* Arguments: list - the list
*            discard - as in LstDelete
* Returns:   as in LstDelete
* See also:  all routines having a name of the form Lst[A-Za-z]+
*
****************************************************************************
*/
FE2(VOID *, LstDelFirst,
    LIST *, list,
    BOOL, discard)
{
    VOID *node;
    LISTNODE *listnode;

    if (LstEmpty(list))
	return(NULL);
    listnode = list->_first;
    list->_first = listnode->_next;
    if (listnode == list->_last)
	list->_last = NULL;
    if (listnode == list->_this)
	list->_this = list->_first;
    node = listnode->_node;
    Deallocate((char *) listnode);
    if (discard)
	{
	    (*(list->_tools->_discard))(node);
	    return(NULL);
	}
    else
	return node;
}

/***************************************************************************
*
* Function:  LstDelLast
* Purpose:   This is like LstDelete, except that the last element
*            in the list is deleted.
* Arguments: list - the list
*            discard - as in LstDelete
* Returns:   as in LstDelete
* See also:  all routines having a name of the form Lst[A-Za-z]+
*
****************************************************************************
*/
FE2(VOID *, LstDelLast,
    LIST *, list,
    BOOL, discard)
{
    VOID *node;
    LISTNODE *listnode, *l, *p;

    if (LstEmpty(list))
	return(NULL);
    listnode = list->_last;
    for (p = NULL, l = list->_first;
	 l != list->_last;
	 p = l, l = l->_next)
	;
    list->_last = p;
    if (p)
	p->_next = NULL;
    if (listnode == list->_first)
	list->_first = NULL;
    if (listnode == list->_this)
	list->_this = p;
    node = listnode->_node;
    Deallocate((char *) listnode);
    if (discard)
	{
	    (*(list->_tools->_discard))(node);
	    return(NULL);
	}
    else
	return node;
}

/***************************************************************************
*
* Function:  LstDelNext
* Purpose:   This is like LstDelete, except that the next element with
*            respect to the "current" cursor is deleted.
* Arguments: list - the list
*            discard - as in LstDelete
* Returns:   as in LstDelete
* See also:  all routines having a name of the form Lst[A-Za-z]+
*
****************************************************************************
*/
FE2(VOID *, LstDelNext,
    LIST *, list,
    BOOL, discard)
{
    VOID *node;
    LISTNODE *listnode;
    
    if (!list)
	return(NULL);
    if (list->_this)
	{
	    if (!list->_this->_next)
		return(NULL);
	    if (list->_last == list->_this->_next)
		list->_last = list->_this;
	    listnode = list->_this->_next->_next;
	    node = list->_this->_next->_node;
	    Deallocate((char *) (list->_this->_next));
	    list->_this->_next = listnode;
	    if (discard)
		{
		    (*(list->_tools->_discard))(node);
		    return(NULL);
		}
	    else
		return node;
	}
    else
	return LstDelFirst(list,discard);
}

/***************************************************************************
*
* Function:  LstDestroy
* Purpose:   This routine removes a list from the memory.
* Arguments: list - the list
*            discard - elements are removed from the memory if and only if
*                      discard is TRUE
* Returns:
* See also:  all routines having a name of the form Lst[A-Za-z]+
*
****************************************************************************
*/
FE2(void, LstDestroy,
    LIST *, list,
    BOOL, discard)
{
    LISTNODE *listnode, *next;

    listnode = list->_first;
    while (listnode)
	{
	    next = listnode->_next;
	    if (discard)
		(*(list->_tools->_discard))(listnode->_node);
	    Deallocate((char *) listnode);
	    listnode = next;
	}
    Deallocate((char *) list);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  LstDump
* Purpose:   This routine displays a list.
* Arguments: list - the list
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  all routines having a name of the form Lst[A-Za-z]+
*
****************************************************************************
*/
FE2(void, LstDump,
    LIST *, list,
    unsigned, depth)
{
    VOID *node;

    DumpIndent(depth);
    (void) fprintf(_err, "LIST\n");
    LstStore(list);
    for (node = LstBegin(list); node; node = LstStep(list))
	(*(list->_tools->_dump))(node,depth + 1);
    (void) LstRestore(list);
    DumpIndent(depth);
    (void) fprintf(_err, "End of LIST\n");
}
#endif

/***************************************************************************
*
* Function:  LstInsFirst
* Purpose:   This routine inserts an element to the head of a list.
* Arguments: list - the list
*            node - the element
* Returns:   the element
* See also:  Allocate (alloc.c), all routines having a name of the form
*            Lst[A-Za-z]+
*
****************************************************************************
*/
FE2(VOID *, LstInsFirst,
    LIST *, list,
    char *, node)
{
    LISTNODE *listnode;
    
    if (!(listnode = (LISTNODE *) Allocate(sizeof(LISTNODE))))
	return(NULL);
    listnode->_node = Pvoid(node);
    listnode->_next = list->_first;
    list->_first = listnode;
    if (!list->_last)
	list->_last = list->_first;
    list->_this = listnode;
    return(listnode->_node);
}

/***************************************************************************
*
* Function:  LstInsLast
* Purpose:   This routine inserts an element to the tail of a list.
* Arguments: list - the list
*            node - the element
* Returns:   the element
* See also:  Allocate (alloc.c), all routines having a name of the form
*            Lst[A-Za-z]+
*
****************************************************************************
*/
FE2(VOID *, LstInsLast,
    LIST *, list,
    char *, node)
{
    LISTNODE *listnode;

    if (!(listnode = (LISTNODE *) Allocate(sizeof(LISTNODE))))
	return(NULL);
    listnode->_node = Pvoid(node);
    listnode->_next = NULL;
    if (list->_last)
	list->_last = list->_last->_next = listnode;
    else
	list->_last = list->_first = listnode;
    list->_this = listnode;
    return(listnode->_node);
}

/***************************************************************************
*
* Function:  LstInsNext
* Purpose:   This routine inserts an element immediately after the
*            "current" cursor in a list. The "current" cursor is
*            set to point to the new basic unit.
* Arguments: list - the list
*            node - the element
* Returns:   the element
* See also:  Allocate (alloc.c), all routines having a name of the form
*            Lst[A-Za-z]+
*
****************************************************************************
*/
FE2(VOID *, LstInsNext,
    LIST *, list,
    char *, node)
{
    LISTNODE *listnode;

    if (!(listnode = (LISTNODE *) Allocate(sizeof(LISTNODE))))
	return(NULL);
    listnode->_node = Pvoid(node);
    if (list->_this)
	{
	    listnode->_next = list->_this->_next;
	    list->_this->_next = listnode;
	}
    else
	{
	    listnode->_next = list->_first;
	    list->_first = listnode;
	}
    if (list->_this == list->_last)
	list->_last = listnode;
    list->_this = listnode;
    return(listnode->_node);
}

/***************************************************************************
*
* Function:  LstNth
* Purpose:   This routine returns the nth element in a list
* Arguments: list - the list
*            n - as said above
* Returns:   the nth element if such exists, and NULL otherwise
* See also:  all routines having a name of the form Lst[A-Za-z]+
*
****************************************************************************
*/
FE2(VOID *, LstNth,
    LIST *, list,
    unsigned, n)
{
    LISTNODE *listnode;
    unsigned i;

    if (!n)
	return(NULL);
    for (i = 1, listnode = list->_first;
	 i < n && listnode;
	 i++, listnode = listnode->_next)
	;
    return listnode ? listnode->_node : NULL;
}

/***************************************************************************
*
* Function:  LstRestore
* Purpose:   This routine sets the "current" cursor of a list equal
*            to a cursor that is popped from a global stack.
* Arguments: list - the list
* Returns:   the element pointed by the cursor if such element exists,
*            and NULL otherwise
* See also:  DynamInit, all routines having a name of the form Lst[A-Za-z]+
*
****************************************************************************
*/
FE1(VOID *, LstRestore,
    LIST *, list)
{
    return (list->_this = (LISTNODE *) LstDelFirst(listStack, FALSE))
	? list->_this->_node : NULL;
}

/***************************************************************************
*
* Function:  LstSize
* Purpose:   This routine computes the number of elements in a list.
* Arguments: list - the list
* Returns:   the number of elements
* See also:  all routines having a name of the form Lst[A-Za-z]+
*
****************************************************************************
*/
FE1(unsigned, LstSize,
    LIST *, list)
{
    unsigned n;
    LISTNODE *listnode;
  
    if (!list)
	return 0;
    for (listnode = list->_first, n = 0;
	 listnode;
	 listnode = listnode->_next, n++)
	;
    return n;
}

/***************************************************************************
*
* Function:  LstStore
* Purpose:   This routine pushes the "current" cursor of a list onto a
*            global stack.
* Arguments: list - the list
* Returns:
* See also:  DynamInit, all routines having a name of the form Lst[A-Za-z]+
*
****************************************************************************
*/
FE1(void, LstStore,
    LIST *, list)
{
    (void) LstInsFirst(listStack, (char *) (list->_this));
}

/***************************************************************************
*
* Function:  LstWhich
* Purpose:   This routine returns the order number of an element in a list.
* Arguments: list - the list
*            node - the element
* Returns:   the order number (>= 1) if the element is in the list,
*            and 0 otherwise
* See also:  all routines having a name of the form Lst[A-Za-z]+
*
****************************************************************************
*/
FE2(unsigned, LstWhich,
    LIST *, list,
    char *, node)
{
    LISTNODE *listnode;
    unsigned i;

    for (listnode = list->_first, i = 1;
	 listnode && (node != (char *) (listnode->_node));
	 listnode = listnode->_next, i++)
	;
    return listnode ? i : 0;
}

/***************************************************************************
*
* Function:  LstCp
* Purpose:   This instance of LstCopy is needed when copying must be
*            done by a function having one argument only.
* Arguments: list - the list to be copied
* Returns:   the copy
* See also:  all routines having a name of the form Lst[A-Za-z]+
*
****************************************************************************
*/
FE1(LIST *, LstCp,
    LIST *, list)
{
    return LstCopy(list,TRUE);
}

/***************************************************************************
*
* Function:  LstDc
* Purpose:   This instance of LstDestroy is needed when destroying
*            must be done by a function having one argument only.
* Arguments: list - the list
* Returns:
* See also:  all routines having a name of the form Lst[A-Za-z]+
*
****************************************************************************
*/
FE1(void, LstDc,
    LIST *, list)
{
    LstDestroy(list, TRUE);
}

/* A collection of routines for the LIST type to be used by
 * list routines of dynam.c and table routines of ptrbl.c. */
TOOLS lstTools = {
  (CFUNC) LstCp, LstDc
#ifdef DEBUG
, LstDump
#endif
};
