/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  expr.c
*
* Part of: PROD's programs probe and prpp
*
* Authors: Marko Rauhamaa and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This module is concentrated on the handling of integer and
*     marking expressions.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include <stdio.h>
#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "gphutil.h"
#include "mark.h"
#include "encode.h"
#include "expr.h"

int errorFlag = NO_ERROR;         /* indicates different kinds of syntactic
                                   * or semantic errors */
BOOL continueEvaluation = TRUE;   /* In probe, this is FALSE when
                                   * an execution of a command, including
                                   * all evaluations, is to be interrupted. */
TUPLE *fogTuple;                  /* a context tuple in a submarking
                                   * calculation */

/***************************************************************************
*
* Function:  MexCons
* Purpose:   This routine constructs a marking expression.
* Arguments: operator - the outermost operator of the expression
*            mep1 - the first operand if such exists
*            mep2 - the second operand if such exists
*            place - a number of a place of a net if operator is MK_PLACE
* Returns:   the expression
* See also:
*
****************************************************************************
*/
FE4(MARKEXPR *, MexCons,
    unsigned, operator,
    MARKEXPR *, mep1,
    MARKEXPR *, mep2,
    unsigned, place)
{
    MARKEXPR *markexpr;

    markexpr = MexAlloc();
    markexpr->operator = operator;
    switch (MexKind(markexpr)) 
	{
	case MK_CONSTANT:
	    markexpr->um.marking = (MARKING *) mep1;
	    break;
	case MK_PLACE:
	    markexpr->um.place = place;
	    break;
	case MK_TRANGE:
	    markexpr->um.trange = (TRANGE *) mep1;
	    break;
	case MK_EXPRMARK:
	    markexpr->um.em.esub = (EXPRESSION *) mep1;
	    markexpr->um.em.msub = mep2;
	    break;
	case MK_MARKMARK:
	    markexpr->um.mm.sub1m = mep1;
	    markexpr->um.mm.sub2m = mep2;
	    break;
	}
    return(markexpr);
}

/***************************************************************************
*
* Function:  MexAux
* Purpose:   This routine is an interface to MexCons and is needed because
*            the construction of expressions in ltl.y or rules.y is not
*            immediately stopped after a finding of a syntax error.
* Arguments: operator - as in MexCons
*            mep1 - as in MexCons
*            mep2 - as in MexCons
*            place - as in MexCons
* Returns:   NULL in error cases, and otherwise the expression
* See also:  MexCons
*
****************************************************************************
*/
FE4(MARKEXPR *, MexAux,
    unsigned, operator,
    MARKEXPR *, mep1,
    MARKEXPR *, mep2,
    unsigned, place)
{
    if (errorFlag == NO_ERROR)
	return(MexCons(operator, mep1, mep2, place));
    switch (mexOpTable[operator].kind) 
	{
	case MK_CONSTANT:
	    if (mep1)
		MarDiscard((MARKING *) mep1);
	    break;
	case MK_PLACE:
	    break;
	case MK_TRANGE:
	    if (mep1)
		TrgDiscard((TRANGE *) mep1);
	    break;
	case MK_EXPRMARK:
	    if (mep1)
		XprDiscard((EXPRESSION *) mep1);
	    if (mep2)
		MexDiscard(mep2);
	    break;
	case MK_MARKMARK:
	    if (mep1)
		MexDiscard(mep1);
	    if (mep2)
		MexDiscard(mep2);
	    break;
	}
    return(NULL);
}

/***************************************************************************
*
* Function:  MexCopy
* Purpose:   This routine makes a copy of a marking expression.
* Arguments: markexpr - the expression to be copied
* Returns:   the copy
* See also:  MexForce
*
****************************************************************************
*/
FE1(MARKEXPR *, MexCopy,
    MARKEXPR *, markexpr)
{
    MARKEXPR *copy;

    copy = MexAlloc();
    copy->operator = markexpr->operator;
    switch (MexKind(markexpr)) 
	{
	case MK_CONSTANT:
	    copy->um.marking = MarCopy(markexpr->um.marking);
	    break;
	case MK_PLACE:
	    copy->um.place = markexpr->um.place;
	    break;
	case MK_TRANGE:
	    copy->um.trange = TrgCopy(markexpr->um.trange);
	    break;
	case MK_EXPRMARK:
	    copy->um.em.esub = XprCopy(markexpr->um.em.esub);
	    copy->um.em.msub = MexCopy(markexpr->um.em.msub);
	    break;
	case MK_MARKMARK:
	    copy->um.mm.sub1m = MexCopy(markexpr->um.mm.sub1m);
	    copy->um.mm.sub2m = MexCopy(markexpr->um.mm.sub2m);
	    break;
	}
    return(copy);
}

/***************************************************************************
*
* Function:  MexForce
* Purpose:   This routine replaces the contents of a marking expression
*            by the contents of another marking expression.
* Arguments: markexpr1 - the target
*            markexpr2 - the source (removed from the memory here)
* Returns:
* See also:  MexCopy
*
****************************************************************************
*/
FE2(void, MexForce,
    MARKEXPR *, mep1,
    MARKEXPR *, mep2)
{
    mep1->operator = mep2->operator;
    switch (MexKind(mep2)) 
	{
	case MK_CONSTANT:
	    mep1->um.marking = mep2->um.marking;
	    break;
	case MK_PLACE:
	    mep1->um.place = mep2->um.place;
	    break;
	case MK_TRANGE:
	    mep1->um.trange = mep2->um.trange;
	    break;
	case MK_EXPRMARK:
	    mep1->um.em.esub = mep2->um.em.esub;
	    mep1->um.em.msub = mep2->um.em.msub;
	    break;
	case MK_MARKMARK:
	    mep1->um.mm.sub1m = mep2->um.mm.sub1m;
	    mep1->um.mm.sub2m = mep2->um.mm.sub2m;
	    break;
	}
    Deallocate((char *) mep2);
}

/***************************************************************************
*
* Function:  MexDiscard
* Purpose:   This routine removes a marking expression from the memory.
* Arguments: markexpr - the expression
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, MexDiscard,
    MARKEXPR *, markexpr)
{
    switch (MexKind(markexpr)) 
	{
	case MK_CONSTANT:
	    MarDiscard(markexpr->um.marking);
	    break;
	case MK_PLACE:
	    break;
	case MK_TRANGE:
	    TrgDiscard(markexpr->um.trange);
	    break;
	case MK_EXPRMARK:
	    XprDiscard(markexpr->um.em.esub);
	    MexDiscard(markexpr->um.em.msub);
	    break;
	case MK_MARKMARK:
	    MexDiscard(markexpr->um.mm.sub1m);
	    MexDiscard(markexpr->um.mm.sub2m);
	    break;
	}
    Deallocate((char *) markexpr);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  MexDump
* Purpose:   This routine displays a marking expression.
* Arguments: markexpr - the expression
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, MexDump,
    MARKEXPR *, markexpr,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "MARKEXPR %u\n", markexpr->operator);
    (void) fflush(_err);
    switch (MexKind(markexpr)) 
	{
	case MK_CONSTANT:
	    MarDump(markexpr->um.marking, depth + 1);
	    break;
	case MK_PLACE:
	    DumpIndent(depth + 1);
	    (void) fprintf(_err, " place = %u\n", markexpr->um.place);
	    (void) fflush(_err);
	    break;
	case MK_TRANGE:
	    TrgDump(markexpr->um.trange, depth + 1);
	    break;
	case MK_EXPRMARK:
	    XprDump(markexpr->um.em.esub, depth + 1);
	    MexDump(markexpr->um.em.msub, depth + 1);
	    break;
	case MK_MARKMARK:
	    MexDump(markexpr->um.mm.sub1m, depth + 1);
	    MexDump(markexpr->um.mm.sub2m, depth + 1);
	    break;
	}
}
#endif

/***************************************************************************
*
* Function:  MexEConstant
* Purpose:   This routine evaluates a marking expression that points to
*            a constant marking.
* Arguments: markexpr - the expression
* Returns:   a copy of the constant marking
* See also:  all routines having a name of the form MexE[A-Za-z]+
*
****************************************************************************
*/
FS1(MARKING *, MexEConstant,
    MARKEXPR *, markexpr)
{
    return(MarCopy(markexpr->um.marking));
}

/***************************************************************************
*
* Function:  MexETrange
* Purpose:   This routine evaluates a marking expression that points
*            to a range list.
* Arguments: markexpr - the expression
* Returns:   a marking corresponding to the range list
* See also:  all routines having a name of the form MexE[A-Za-z]+
*
****************************************************************************
*/
FS1(MARKING *, MexETrange,
    MARKEXPR *, markexpr)
{
    return(TrgMarking(markexpr->um.trange));
}

/***************************************************************************
*
* Function:  MexETimes
* Purpose:   This routine evaluates a marking expression that represents
*            a product.
* Arguments: markexpr - the expression 
* Returns:   the product
* See also:  all routines having a name of the form MexE[A-Za-z]+
*
****************************************************************************
*/
FS1(MARKING *, MexETimes,
    MARKEXPR *, markexpr)
{
    register MARKING *marking;
    register DOMAIN value;

    value = XprEval(markexpr->um.em.esub);
    if (value == (DOMAIN) 0L)
	return(MarCreate());
    marking = MexEval(markexpr->um.em.msub);
    return(MarProduct(marking, (unsigned long) value));
}

/***************************************************************************
*
* Function:  MexECartes
* Purpose:   This routine evaluates a marking expression that represents
*            a Cartesian product.
* Arguments: markexpr - the expression
* Returns:   the Cartesian product
* See also:  all routines having a name of the form MexE[A-Za-z]+
*
****************************************************************************
*/
FS1(MARKING *, MexECartes,
    MARKEXPR *, markexpr)
{
    register MARKING *marking1, *marking2, *marking3;

    marking1 = MexEval(markexpr->um.mm.sub1m);
    if (MarEmpty(marking1))
	return(marking1);
    marking2 = MexEval(markexpr->um.mm.sub2m);
    if (MarEmpty(marking2))
        {
            MarDiscard(marking1);
            return(marking2);
        }
    marking3 = MarCartes(marking1, marking2);
    MarDiscard(marking1);
    MarDiscard(marking2);
    return(marking3);
}

/***************************************************************************
*
* Function:  MexESum
* Purpose:   This routine evaluates a marking expression that represents
*            a sum.
* Arguments: markexpr - the expression
* Returns:   the sum
* See also:  all routines having a name of the form MexE[A-Za-z]+
*
****************************************************************************
*/
FS1(MARKING *, MexESum,
    MARKEXPR *, markexpr)
{
    register MARKING *marking1, *marking2;

    marking1 = MexEval(markexpr->um.mm.sub1m);
    marking2 = MexEval(markexpr->um.mm.sub2m);
    (void) MarAdd(marking1, marking2);
    MarDiscard(marking2);
    return(marking1);
}

/***************************************************************************
*
* Function:  MexEDiff
* Purpose:   This routine evaluates a marking expression that represents
*            a complementation.
* Arguments: markexpr - the expression
* Returns:   the value of the complementation operation
* See also:  all routines having a name of the form MexE[A-Za-z]+
*
****************************************************************************
*/
FS1(MARKING *, MexEDiff,
    MARKEXPR *, markexpr)
{
    register MARKING *marking1, *marking2;

    marking1 = MexEval(markexpr->um.mm.sub1m);
    if (MarEmpty(marking1))
	return(marking1);
    marking2 = MexEval(markexpr->um.mm.sub2m);
    (void) MarSubtract(marking1, marking2);
    MarDiscard(marking2);
    return(marking1);
}

/***************************************************************************
*
* Function:  MexESelect
* Purpose:   This routine evaluates a marking expression that represents
*            a submarking satisfying a condition.
* Arguments: markexpr - the expression
* Returns:   the submarking
* See also:  all routines having a name of the form MexE[A-Za-z]+
*
****************************************************************************
*/
FS1(MARKING *, MexESelect,
    MARKEXPR *, markexpr)
{
    register MARKING *marking;
    register MULTUP *multup;
    register VOID *cursor;
    TUPLE *savTuple;

    marking = MexEval(markexpr->um.em.msub);
    if (MarEmpty(marking))
	return(marking);
    savTuple = FogGTuple();
    cursor = NULL;
    MARFOR (marking, multup)
	{
	    FogSTuple(MultupGTuple(multup));
	    if (XprEval(markexpr->um.em.esub) == (DOMAIN) 0L)
		{
		    MarSet(marking, cursor);
		    MarDelNext(marking);
		}
	    else
		cursor = MarGet(marking);
	}
    FogSTuple(savTuple);
    return(marking);
}

/***************************************************************************
*
* Function:  MexEInters
* Purpose:   This routine evaluates a marking expression that represents
*            an intersection.
* Arguments: markexpr - the expression
* Returns:   the intersection
* See also:  all routines having a name of the form MexE[A-Za-z]+
*
****************************************************************************
*/
FS1(MARKING *, MexEInters,
    MARKEXPR *, markexpr)
{
    register MARKING *marking1, *marking2;

    marking1 = MexEval(markexpr->um.mm.sub1m);
    if (MarEmpty(marking1))
	return(marking1);
    marking2 = MexEval(markexpr->um.mm.sub2m);
    (void) MarIntersect(marking2, marking1);
    MarDiscard(marking1);
    return(marking2);
}

/***************************************************************************
*
* Function:  MexRConstant
* Purpose:   This routine reduces a marking expression that points to
*            a constant marking.
* Arguments: markexpr - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            marking, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form MexR[A-Za-z]+
*
****************************************************************************
*/
FS0(unsigned, MexRConstant)
{
    return(RE_CONSTANT);
}

/***************************************************************************
*
* Function:  MexRPLace
* Purpose:   This routine reduces a marking expression that represents
*            a marking of a place.
* Arguments: markexpr - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            marking, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form MexR[A-Za-z]+
*
****************************************************************************
*/
FE0(unsigned, MexRPlace)
{
    return(RE_LOCAL);
}

/***************************************************************************
*
* Function:  MexRTimes
* Purpose:   This routine reduces a marking expression that represents
*            a product.
* Arguments: markexpr - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            marking, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form MexR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, MexRTimes,
    MARKEXPR *, markexpr)
{
    unsigned rlevel1, rlevel2;
    MARKING *marking;
    unsigned long coeff;

    rlevel1 = XprReduce(markexpr->um.em.esub);
    if (rlevel1 == RE_CONSTANT)
	{
	    coeff = (unsigned long) XprEval(markexpr->um.em.esub);
	    if (coeff == 0L)
		{
		    XprDiscard(markexpr->um.em.esub);
		    MexDiscard(markexpr->um.em.msub);
		    markexpr->operator = MO_CONSTANT;
		    markexpr->um.marking = MarCreate();
		    return(RE_CONSTANT);
		}
	    rlevel2 = MexReduce(markexpr->um.em.msub);
	    if (rlevel2 == RE_CONSTANT) 
		{
		    marking = MexEval(markexpr);
		    XprDiscard(markexpr->um.em.esub);
		    MexDiscard(markexpr->um.em.msub);
		    markexpr->operator = MO_CONSTANT;
		    markexpr->um.marking = marking;
		}
	    else if (coeff == 1L)
		{
		    XprDiscard(markexpr->um.em.esub);
		    MexForce(markexpr, markexpr->um.em.msub);
		}
	    return(rlevel2);
	}
    rlevel2 = MexReduce(markexpr->um.em.msub);
    if (rlevel2 == RE_CONSTANT)
	{
	    marking = MexEval(markexpr->um.em.msub);
	    if (MarEmpty(marking))
		{
		    XprDiscard(markexpr->um.em.esub);
		    MexDiscard(markexpr->um.em.msub);
		    markexpr->operator = MO_CONSTANT;
		    markexpr->um.marking = marking;
		    return(RE_CONSTANT);
		}
	    MarDiscard(marking);
	}
    return((rlevel1 > rlevel2) ? rlevel1 : rlevel2);
}

/***************************************************************************
*
* Function:  MexRSum
* Purpose:   This routine reduces a marking expression that represents
*            a sum.
* Arguments: markexpr - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            marking, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form MexR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, MexRSum,
    MARKEXPR *, markexpr)
{
    unsigned rlevel1, rlevel2;
    MARKING *marking;

    rlevel1 = MexReduce(markexpr->um.mm.sub1m);
    rlevel2 = MexReduce(markexpr->um.mm.sub2m);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (rlevel2 == RE_CONSTANT)
		{
		    marking = MexEval(markexpr);
		    MexDiscard(markexpr->um.mm.sub1m);
		    MexDiscard(markexpr->um.mm.sub2m);
		    markexpr->operator = MO_CONSTANT;
		    markexpr->um.marking = marking;
		    return(RE_CONSTANT);
		}
	    marking = MexEval(markexpr->um.mm.sub1m);
	    if (MarEmpty(marking))
		{
		    MexDiscard(markexpr->um.mm.sub1m);
		    MexForce(markexpr, markexpr->um.mm.sub2m);
		}
	    MarDiscard(marking);
	    return(rlevel2);
	}
    if (rlevel2 == RE_CONSTANT) 
	{
	    marking = MexEval(markexpr->um.mm.sub2m);
	    if (MarEmpty(marking))
		{
		    MexDiscard(markexpr->um.mm.sub2m);
		    MexForce(markexpr, markexpr->um.mm.sub1m);
		}
	    MarDiscard(marking);
	}
    return((rlevel1 > rlevel2) ? rlevel1 : rlevel2);
}

/***************************************************************************
*
* Function:  MexRDiff
* Purpose:   This routine reduces a marking expression that represents
*            a complementation.
* Arguments: markexpr - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            marking, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form MexR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, MexRDiff,
    MARKEXPR *, markexpr)
{
    unsigned rlevel1, rlevel2;
    MARKING *marking;

    rlevel1 = MexReduce(markexpr->um.mm.sub1m);
    rlevel2 = MexReduce(markexpr->um.mm.sub2m);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (rlevel2 == RE_CONSTANT)
		{
		    marking = MexEval(markexpr);
		    MexDiscard(markexpr->um.mm.sub1m);
		    MexDiscard(markexpr->um.mm.sub2m);
		    markexpr->operator = MO_CONSTANT;
		    markexpr->um.marking = marking;
		    return(RE_CONSTANT);
		}
	    marking = MexEval(markexpr->um.mm.sub1m);
	    if (MarEmpty(marking))
		{
		    MexDiscard(markexpr->um.mm.sub1m);
		    MexDiscard(markexpr->um.mm.sub2m);
		    markexpr->operator = MO_CONSTANT;
		    markexpr->um.marking = marking;
		    return(RE_CONSTANT);
		}
	    MarDiscard(marking);
	    return(rlevel2);
	}
    if (rlevel2 == RE_CONSTANT) 
	{
	    marking = MexEval(markexpr->um.mm.sub2m);
	    if (MarEmpty(marking))
		{
		    MexDiscard(markexpr->um.mm.sub2m);
		    MexForce(markexpr, markexpr->um.mm.sub1m);
		}
	    MarDiscard(marking);
	}
    return((rlevel1 > rlevel2) ? rlevel1 : rlevel2);
}

/***************************************************************************
*
* Function:  MexRSelect
* Purpose:   This routine reduces a marking expression that represents
*            a submarking satisfying a condition.
* Arguments: markexpr - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            marking, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form MexR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, MexRSelect,
    MARKEXPR *, markexpr)
{
    unsigned rlevel1, rlevel2;
    MARKING *marking;

    rlevel1 = XprReduce(markexpr->um.em.esub);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (XprEval(markexpr->um.em.esub) == (DOMAIN) 0L)
		{
		    XprDiscard(markexpr->um.em.esub);
		    MexDiscard(markexpr->um.em.msub);
		    markexpr->operator = MO_CONSTANT;
		    markexpr->um.marking = MarCreate();
		    return(RE_CONSTANT);
		}
	    XprDiscard(markexpr->um.em.esub);
	    rlevel2 = MexReduce(markexpr->um.em.msub);
	    MexForce(markexpr, markexpr->um.em.msub);
	    return(rlevel2);
	}
    rlevel2 = MexReduce(markexpr->um.em.msub);
    if (rlevel2 == RE_CONSTANT)
	{
	    if (rlevel1 == RE_GLOBAL)
		{
		    marking = MexEval(markexpr);
		    XprDiscard(markexpr->um.em.esub);
		    MexDiscard(markexpr->um.em.msub);
		    markexpr->operator = MO_CONSTANT;
		    markexpr->um.marking = marking;
		    return(RE_CONSTANT);
		}
	    marking = MexEval(markexpr->um.em.msub);
	    if (MarEmpty(marking))
		{
		    XprDiscard(markexpr->um.em.esub);
		    MexDiscard(markexpr->um.em.msub);
		    markexpr->operator = MO_CONSTANT;
		    markexpr->um.marking = marking;
		    return(RE_CONSTANT);
		}
	    MarDiscard(marking);
	}
    return((rlevel1 > rlevel2) ? rlevel1 : rlevel2);
}

/***************************************************************************
*
* Function:  MexRInters
* Purpose:   This routine reduces a marking expression that represents
*            an intersection.
* Arguments: markexpr - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            marking, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form MexR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, MexRInters,
    MARKEXPR *, markexpr)
{
    unsigned rlevel1, rlevel2;
    MARKING *marking;

    rlevel1 = MexReduce(markexpr->um.mm.sub1m);
    rlevel2 = MexReduce(markexpr->um.mm.sub2m);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (rlevel2 == RE_CONSTANT)
		{
		    marking = MexEval(markexpr);
		    MexDiscard(markexpr->um.mm.sub1m);
		    MexDiscard(markexpr->um.mm.sub2m);
		    markexpr->operator = MO_CONSTANT;
		    markexpr->um.marking = marking;
		    return(RE_CONSTANT);
		}
	    marking = MexEval(markexpr->um.mm.sub1m);
	    if (MarEmpty(marking))
		{
		    MexDiscard(markexpr->um.mm.sub1m);
		    MexDiscard(markexpr->um.mm.sub2m);
		    markexpr->operator = MO_CONSTANT;
		    markexpr->um.marking = marking;
		    return(RE_CONSTANT);
		}
	    MarDiscard(marking);
	    return(rlevel2);
	}
    if (rlevel2 == RE_CONSTANT) 
	{
	    marking = MexEval(markexpr->um.mm.sub2m);
	    if (MarEmpty(marking))
		{
		    MexDiscard(markexpr->um.mm.sub1m);
		    MexDiscard(markexpr->um.mm.sub2m);
		    markexpr->operator = MO_CONSTANT;
		    markexpr->um.marking = marking;
		    return(RE_CONSTANT);
		}
	    MarDiscard(marking);
	}
    return((rlevel1 > rlevel2) ? rlevel1 : rlevel2);
}

/***************************************************************************
*
* Function:  XprCons
* Purpose:   This routine constructs an integer expression.
* Arguments: operator - the outermost operator of the expression
*            ep1 - the first operand if such exists
*            ep2 - the second operand if such exists
*            ep3 - the third operand if such exists
*            value - an integer number if operator is EK_CONSTANT
* Returns:   the expression
* See also:
*
****************************************************************************
*/
FE5(EXPRESSION *, XprCons,
    unsigned, operator,
    EXPRESSION *, ep1,
    EXPRESSION *, ep2,
    EXPRESSION *, ep3,
    DOMAIN, value)
{
    EXPRESSION *expression;

    expression = XprAlloc();
    expression->operator = operator;
    switch (XprKind(expression)) 
	{
	case EK_CONSTANT:
	    expression->ue.value = value;
	    break;
	case EK_CARD:
	    expression->ue.me.msub1 = (MARKEXPR *) ep1;
	    break;
	case EK_MCOMP:
	    expression->ue.me.msub1 = (MARKEXPR *) ep1;
	    expression->ue.me.msub2 = (MARKEXPR *) ep2;
	    break;
	case EK_NULLARY:
	    break;
	case EK_UNARY:
	    expression->ue.oe.sub1e = ep1;
	    break;
	case EK_BINARY:
	    expression->ue.oe.sub1e = ep1;
	    expression->ue.oe.sub2e = ep2;
	    break;
	case EK_TRINARY:
	    expression->ue.oe.sub1e = ep1;
	    expression->ue.oe.sub2e = ep2;
	    expression->ue.oe.sub3e = ep3;
	    break;
	}
    return(expression);
}

/***************************************************************************
*
* Function:  XprAux
* Purpose:   This routine is an interface to XprCons and is needed because
*            the construction of expressions in ltl.y or rules.y is not
*            immediately stopped after a finding of a syntax error.
* Arguments: operator - as in XprCons
*            ep1 - as in XprCons
*            ep2 - as in XprCons
*            ep3 - as in XprCons
*            value - as in XprCons
* Returns:   NULL in error cases, and the expression otherwise
* See also:  XprCons
*
****************************************************************************
*/
FE5(EXPRESSION *, XprAux,
    unsigned, operator,
    EXPRESSION *, ep1,
    EXPRESSION *, ep2,
    EXPRESSION *, ep3,
    DOMAIN, value)
{
    if (errorFlag == NO_ERROR)
	return(XprCons(operator, ep1, ep2, ep3, value));
    switch (xprOpTable[operator].kind) 
	{
	case EK_CONSTANT:
	    break;
	case EK_CARD:
	    if (ep1)
		MexDiscard((MARKEXPR *) ep1);
	    break;
	case EK_MCOMP:
	    if (ep1)
		MexDiscard((MARKEXPR *) ep1);
	    if (ep2)
		MexDiscard((MARKEXPR *) ep2);
	    break;
	case EK_NULLARY:
	    break;
	case EK_UNARY:
	    if (ep1)
		XprDiscard(ep1);
	    break;
	case EK_BINARY:
	    if (ep1)
		XprDiscard(ep1);
	    if (ep2)
		XprDiscard(ep2);
	    break;
	case EK_TRINARY:
	    if (ep1)
		XprDiscard(ep1);
	    if (ep2)
		XprDiscard(ep2);
	    if (ep3)
		XprDiscard(ep3);
	    break;
	}
    return(NULL);
}

/***************************************************************************
*
* Function:  XprCopy
* Purpose:   This routine makes a copy of an integer expression.
* Arguments: expression - the expression to be copied
* Returns:   the copy
* See also:  XprForce
*
****************************************************************************
*/
FE1(EXPRESSION *, XprCopy,
    EXPRESSION *, expression)
{
    EXPRESSION *copy;

    copy = XprAlloc();
    copy->operator = expression->operator;
    switch (XprKind(expression))
	{
	case EK_CONSTANT:
	    copy->ue.value = expression->ue.value;
	    break;
	case EK_CARD:
	    copy->ue.me.msub1 = MexCopy(expression->ue.me.msub1);
	    break;
	case EK_MCOMP:
	    copy->ue.me.msub1 = MexCopy(expression->ue.me.msub1);
	    copy->ue.me.msub2 = MexCopy(expression->ue.me.msub2);
	    break;
	case EK_NULLARY:
	    break;
	case EK_UNARY:
	    copy->ue.oe.sub1e = XprCopy(expression->ue.oe.sub1e);
	    break;
	case EK_BINARY:
	    copy->ue.oe.sub1e = XprCopy(expression->ue.oe.sub1e);
	    copy->ue.oe.sub2e = XprCopy(expression->ue.oe.sub2e);
	    break;
	case EK_TRINARY:
	    copy->ue.oe.sub1e = XprCopy(expression->ue.oe.sub1e);
	    copy->ue.oe.sub2e = XprCopy(expression->ue.oe.sub2e);
	    copy->ue.oe.sub3e = XprCopy(expression->ue.oe.sub3e);
	    break;
	}
    return(copy);
}

/***************************************************************************
*
* Function:  XprForce
* Purpose:   This routine replaces the contents of an integer expression
*            by the contents of another integer expression.
* Arguments: expression1 - the target
*            expression2 - the source (removed from the memory here)
* Returns:
* See also:  XprCopy
*
****************************************************************************
*/
FE2(void, XprForce,
    EXPRESSION *, ep1,
    EXPRESSION *, ep2)
{
    ep1->operator = ep2->operator;
    switch (XprKind(ep2))
	{
	case EK_CONSTANT:
	    ep1->ue.value = ep2->ue.value;
	    break;
	case EK_CARD:
	    ep1->ue.me.msub1 = ep2->ue.me.msub1;
	    break;
	case EK_MCOMP:
	    ep1->ue.me.msub1 = ep2->ue.me.msub1;
	    ep1->ue.me.msub2 = ep2->ue.me.msub2;
	    break;
	case EK_NULLARY:
	    break;
	case EK_UNARY:
	    ep1->ue.oe.sub1e = ep2->ue.oe.sub1e;
	    break;
	case EK_BINARY:
	    ep1->ue.oe.sub1e = ep2->ue.oe.sub1e;
	    ep1->ue.oe.sub2e = ep2->ue.oe.sub2e;
	    break;
	case EK_TRINARY:
	    ep1->ue.oe.sub1e = ep2->ue.oe.sub1e;
	    ep1->ue.oe.sub2e = ep2->ue.oe.sub2e;
	    ep1->ue.oe.sub3e = ep2->ue.oe.sub3e;
	    break;
	}
    Deallocate((char *) ep2);
}

/***************************************************************************
*
* Function:  XprDiscard
* Purpose:   This routine removes an integer expression from the memory.
* Arguments: expression - the expression
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, XprDiscard,
    EXPRESSION *, expression)
{
    switch (XprKind(expression)) 
	{
	case EK_CONSTANT:
	    break;
	case EK_CARD:
	    MexDiscard(expression->ue.me.msub1);
	    break;
	case EK_MCOMP:
	    MexDiscard(expression->ue.me.msub1);
	    MexDiscard(expression->ue.me.msub2);
	    break;
	case EK_NULLARY:
	    break;
	case EK_UNARY:
	    XprDiscard(expression->ue.oe.sub1e);
	    break;
	case EK_BINARY:
	    XprDiscard(expression->ue.oe.sub1e);
	    XprDiscard(expression->ue.oe.sub2e);
	    break;
	case EK_TRINARY:
	    XprDiscard(expression->ue.oe.sub1e);
	    XprDiscard(expression->ue.oe.sub2e);
	    XprDiscard(expression->ue.oe.sub3e);
	    break;
	}
    Deallocate((char *) expression);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  XprDump
* Purpose:   This routine displays an integer expression.
* Arguments: expression - the expression
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, XprDump,
    EXPRESSION *, expression,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "EXPRESSION %u\n", expression->operator);
    (void) fflush(_err);
    switch (XprKind(expression)) 
	{
	case EK_CONSTANT:
	    DumpIndent(depth + 1);
	    (void) fprintf(_err, "value ");
	    (void) fprintf(_err, DOMFORMAT, expression->ue.value);
	    (void) fprintf(_err, "\n");
	    (void) fflush(_err);
	    break;
	case EK_CARD:
	    MexDump(expression->ue.me.msub1, depth + 1);
	    break;
	case EK_MCOMP:
	    MexDump(expression->ue.me.msub1, depth + 1);
	    MexDump(expression->ue.me.msub2, depth + 1);
	    break;
	case EK_NULLARY:
	    break;
	case EK_UNARY:
	    XprDump(expression->ue.oe.sub1e, depth + 1);
	    break;
	case EK_BINARY:
	    XprDump(expression->ue.oe.sub1e, depth + 1);
	    XprDump(expression->ue.oe.sub2e, depth + 1);
	    break;
	case EK_TRINARY:
	    XprDump(expression->ue.oe.sub1e, depth + 1);
	    XprDump(expression->ue.oe.sub2e, depth + 1);
	    XprDump(expression->ue.oe.sub3e, depth + 1);
	    break;
	}
}
#endif

/***************************************************************************
*
* Function:  XprEDom
* Purpose:   This routine evaluates an integer expression that points
*            to a constant.
* Arguments: expression - the expression
* Returns:   the constant
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprEDom,
    EXPRESSION *, expression)
{
    return(expression->ue.value);
}

/***************************************************************************
*
* Function:  XprECard
* Purpose:   This routine evaluates an integer expression that represents
*            the cardinality of a marking.
* Arguments: expression - the expression
* Returns:   the cardinality
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprECard,
    EXPRESSION *, expression)
{
    unsigned long card;
    MARKING *marking;
    MULTUP *multup;

    marking = MexEval(expression->ue.me.msub1);
    card = 0L;
    MARFOR (marking, multup)
	card += MultupGCount(multup);
    MarDiscard(marking);
    return((DOMAIN) card);
}

/***************************************************************************
*
* Function:  XprENeg
* Purpose:   This routine evaluates an integer expression that represents
*            the opposite number of a number.
* Arguments: expression - the expression
* Returns:   the opposite number
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprENeg,
    EXPRESSION *, expression)
{
    register DOMAIN value;

    value = XprEval(expression->ue.oe.sub1e);
    /* Since unary minus is marginal, we can handle it in a pedantic way. */
    return ((value == (DOMAIN) MINDOMAIN) ? value : ~(value - (DOMAIN) 1L));
}

/***************************************************************************
*
* Function:  XprENot
* Purpose:   This routine evaluates an integer expression that represents
*            a boolean negation.
* Arguments: expression - the expression
* Returns:   the result of the negation
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprENot,
    EXPRESSION *, expression)
{
    return(!XprEval(expression->ue.oe.sub1e));
}

/***************************************************************************
*
* Function:  XprEBnot
* Purpose:   This routine evaluates an integer expression that represents
*            a bitwise complementation.
* Arguments: expression - the expression
* Returns:   the bitwise complement
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprEBnot,
    EXPRESSION *, expression)
{
    return(~XprEval(expression->ue.oe.sub1e));
}

/***************************************************************************
*
* Function:  XprETimes
* Purpose:   This routine evaluates an integer expression that represents
*            a product.
* Arguments: expression - the expression
* Returns:   the product
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprETimes,
    EXPRESSION *, expression)
{
    DOMAIN a;

    if ((a = XprEval(expression->ue.oe.sub1e)) == (DOMAIN) 0L)
	return((DOMAIN) 0L);
    return(a * XprEval(expression->ue.oe.sub2e));
}

/***************************************************************************
*
* Function:  XprEMinus
* Purpose:   This routine evaluates an integer expression that represents
*            a subtraction.
* Arguments: expression - the expression
* Returns:   the result of the subtraction
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprEMinus,
    EXPRESSION *, expression)
{
    return(XprEval(expression->ue.oe.sub1e) - XprEval(expression->ue.oe.sub2e));
}

/***************************************************************************
*
* Function:  XprEPlus
* Purpose:   This routine evaluates an integer expression that represents
*            a sum.
* Arguments: expression - the expression
* Returns:   the sum
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprEPlus,
    EXPRESSION *, expression)
{
    return(XprEval(expression->ue.oe.sub1e) + XprEval(expression->ue.oe.sub2e));
}

/***************************************************************************
*
* Function:  XprELshift
* Purpose:   This routine evaluates an integer expression that represents
*            a shifting of bits to the left.
* Arguments: expression - the expression
* Returns:   the result of the shifting
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprELshift,
    EXPRESSION *, expression)
{
    DOMAIN a;

    if ((a = XprEval(expression->ue.oe.sub1e)) == (DOMAIN) 0L)
	return((DOMAIN) 0L);
    return(a << XprEval(expression->ue.oe.sub2e));
}

/***************************************************************************
*
* Function:  XprERshift
* Purpose:   This routine evaluates an integer expression that represents
*            a shifting of bits to the right.
* Arguments: expression - the expression
* Returns:   the result of the shifting
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprERshift,
    EXPRESSION *, expression)
{
    DOMAIN a;

    if ((a = XprEval(expression->ue.oe.sub1e)) == (DOMAIN) 0L)
	return((DOMAIN) 0L);
    return(a >> XprEval(expression->ue.oe.sub2e));
}

/***************************************************************************
*
* Function:  XprELss
* Purpose:   This routine evaluates an integer expression that represents
*            a "less than" comparison of numbers.
* Arguments: expression - the expression
* Returns:   the boolean value of the comparison
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprELss,
    EXPRESSION *, expression)
{
    return(XprEval(expression->ue.oe.sub1e) < XprEval(expression->ue.oe.sub2e));
}

/***************************************************************************
*
* Function:  XprELeq
* Purpose:   This routine evaluates an integer expression that represents
*            a "less than or equal to" comparison of numbers.
* Arguments: expression - the expression
* Returns:   the boolean value of the comparison
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprELeq,
    EXPRESSION *, expression)
{
    return(XprEval(expression->ue.oe.sub1e) <= XprEval(expression->ue.oe.sub2e));
}

/***************************************************************************
*
* Function:  XprEEqu
* Purpose:   This routine evaluates an integer expression that represents
*            an equality comparison of numbers.
* Arguments: expression - the expression
* Returns:   the boolean value of the comparison
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprEEqu,
    EXPRESSION *, expression)
{
    return(XprEval(expression->ue.oe.sub1e) == XprEval(expression->ue.oe.sub2e));
}

/***************************************************************************
*
* Function:  XprENeq
* Purpose:   This routine evaluates an integer expression that represents
*            an inequality comparison of numbers.
* Arguments: expression - the expression
* Returns:   the boolean value of the comparison
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprENeq,
    EXPRESSION *, expression)
{
    return(XprEval(expression->ue.oe.sub1e) != XprEval(expression->ue.oe.sub2e));
}

/***************************************************************************
*
* Function:  XprEBand
* Purpose:   This routine evaluates an integer expression that represents
*            a bitwise "and" operation.
* Arguments: expression - the expression
* Returns:   the value of the operation
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprEBand,
    EXPRESSION *, expression)
{
    DOMAIN a;

    if ((a = XprEval(expression->ue.oe.sub1e)) == (DOMAIN) 0L)
	return((DOMAIN) 0L);
    return(a & XprEval(expression->ue.oe.sub2e));
}

/***************************************************************************
*
* Function:  XprEBxor
* Purpose:   This routine evaluates an integer expression that represents
*            a bitwise "exclusive or" operation.
* Arguments: expression - the expression
* Returns:   the value of the operation
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprEBxor,
    EXPRESSION *, expression)
{
    return(XprEval(expression->ue.oe.sub1e) ^ XprEval(expression->ue.oe.sub2e));
}

/***************************************************************************
*
* Function:  XprEBor
* Purpose:   This routine evaluates an integer expression that represents
*            a bitwise "or" operation.
* Arguments: expression - the expression
* Returns:   the value of the operation
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprEBor,
    EXPRESSION *, expression)
{
    DOMAIN a;

    if ((a = XprEval(expression->ue.oe.sub1e)) == ~((DOMAIN) 0L))
	return(~((DOMAIN) 0L));
    return(a | XprEval(expression->ue.oe.sub2e));
}

/***************************************************************************
*
* Function:  XprEAnd
* Purpose:   This routine evaluates an integer expression that represents
*            a short-circuit boolean "and" operation.
* Arguments: expression - the expression
* Returns:   the value of the operation
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprEAnd,
    EXPRESSION *, expression)
{
    return(XprEval(expression->ue.oe.sub1e) && XprEval(expression->ue.oe.sub2e));
}

/***************************************************************************
*
* Function:  XprEOr
* Purpose:   This routine evaluates an integer expression that represents
*            a short-circuit boolean "or" operation.
* Arguments: expression - the expression
* Returns:   the value of the operation
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprEOr,
    EXPRESSION *, expression)
{
    return(XprEval(expression->ue.oe.sub1e) || XprEval(expression->ue.oe.sub2e));
}

/***************************************************************************
*
* Function:  XprEIf
* Purpose:   This routine evaluates an integer expression that represents
*            an "if-then-else" operation.
* Arguments: expression - the expression
* Returns:   the value of the operation
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprEIf,
    EXPRESSION *, expression)
{
    return(XprEval(expression->ue.oe.sub1e)
	   ? XprEval(expression->ue.oe.sub2e) : XprEval(expression->ue.oe.sub3e));
}

/***************************************************************************
*
* Function:  XprEMls
* Purpose:   This routine evaluates an integer expression that represents
*            a "less than" comparison of markings.
* Arguments: expression - the expression
* Returns:   the boolean value of the comparison
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprEMls,
    EXPRESSION *, expression)
{
    register MARKING *marking1, *marking2;
    register DOMAIN value;

    marking2 = MexEval(expression->ue.me.msub2);
    if (MarEmpty(marking2))
	value = (DOMAIN) 0L;
    else
	{
	    marking1 = MexEval(expression->ue.me.msub1);
	    value = (DOMAIN) MarLess(marking1, marking2);
	    MarDiscard(marking1);
	}
    MarDiscard(marking2);
    return(value);
}

/***************************************************************************
*
* Function:  XprEMlq
* Purpose:   This routine evaluates an integer expression that represents
*            a "less than or equal to" comparison of markings.
* Arguments: expression - the expression
* Returns:   the boolean value of the comparison
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprEMlq,
    EXPRESSION *, expression)
{
    register MARKING *marking1, *marking2;
    register DOMAIN value;

    marking1 = MexEval(expression->ue.me.msub1);
    if (MarEmpty(marking1))
	value = 1L;
    else
	{
	    marking2 = MexEval(expression->ue.me.msub2);
	    value = (DOMAIN) MarLequal(marking1, marking2);
	    MarDiscard(marking2);
	}
    MarDiscard(marking1);
    return(value);
}

/***************************************************************************
*
* Function:  XprEMeq
* Purpose:   This routine evaluates an integer expression that represents
*            an equality comparison of markings.
* Arguments: expression - the expression
* Returns:   the boolean value of the comparison
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprEMeq,
    EXPRESSION *, expression)
{
    register MARKING *marking1, *marking2;
    register DOMAIN value;

    marking1 = MexEval(expression->ue.me.msub1);
    marking2 = MexEval(expression->ue.me.msub2);
    value = (DOMAIN) MarEqual(marking1, marking2);
    MarDiscard(marking1);
    MarDiscard(marking2);
    return(value);
}

/***************************************************************************
*
* Function:  XprEMnq
* Purpose:   This routine evaluates an integer expression that represents
*            an inequality comparison of markings.
* Arguments: expression - the expression
* Returns:   the boolean value of the comparison
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprEMnq,
    EXPRESSION *, expression)
{
    return(!XprEMeq(expression));
}

/***************************************************************************
*
* Function:  XprRField
* Purpose:   This routine reduces an integer expression that represents
*            a field of a known tuple.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form XprR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, XprRField,
    EXPRESSION *, expression)
{
    unsigned rlevel;

    rlevel = XprReduce(expression->ue.oe.sub1e);
    return((rlevel > RE_GLOBAL) ? rlevel : RE_GLOBAL);
}

/***************************************************************************
*
* Function:  XprRCard
* Purpose:   This routine reduces an integer expression that represents
*            the cardinality of a marking.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form XprR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, XprRCard,
    EXPRESSION *, expression)
{
    unsigned rlevel;
    DOMAIN card;

    rlevel = MexReduce(expression->ue.me.msub1);
    if (rlevel == RE_CONSTANT) 
	{
	    card = XprEval(expression);
	    MexDiscard(expression->ue.me.msub1);
	    expression->operator = EO_DOM;
	    expression->ue.value = card;
	}
    return(rlevel);
}

/***************************************************************************
*
* Function:  XprRNeg
* Purpose:   This routine reduces an integer expression that represents
*            the opposite number of a number.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form XprR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, XprRNeg,
    EXPRESSION *, expression)
{
    unsigned rlevel;
    DOMAIN value;

    rlevel = XprReduce(expression->ue.oe.sub1e);
    if (rlevel == RE_CONSTANT) 
	{
	    value = XprEval(expression);
	    XprDiscard(expression->ue.oe.sub1e);
	    expression->operator = EO_DOM;
	    expression->ue.value = value;
	}
    return(rlevel);
}

/***************************************************************************
*
* Function:  XprRTimes
* Purpose:   This routine reduces an integer expression that represents
*            a product.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form XprR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, XprRTimes,
    EXPRESSION *, expression)
{
    unsigned rlevel1, rlevel2;
    DOMAIN value;

    rlevel1 = XprReduce(expression->ue.oe.sub1e);
    rlevel2 = XprReduce(expression->ue.oe.sub2e);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (rlevel2 == RE_CONSTANT)
		{
		    value = XprEval(expression);
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = value;
		    return(RE_CONSTANT);
		}
	    value = XprEval(expression->ue.oe.sub1e);
	    if (value == (DOMAIN) -1L)
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    expression->ue.oe.sub1e = expression->ue.oe.sub2e;
		    expression->operator = EO_NEG;
		    return(rlevel2);
		}
	    if (value == (DOMAIN) 0L)
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = (DOMAIN) 0L;
		    return(RE_CONSTANT);
		}
	    if (value == (DOMAIN) 1L)
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprForce(expression, expression->ue.oe.sub2e);
		    return(rlevel2);
		}
	    return(rlevel2);
	}
    if (rlevel2 == RE_CONSTANT)
	{
	    value = XprEval(expression->ue.oe.sub2e);
	    if (value == (DOMAIN) -1L)
		{
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_NEG;
		    return(rlevel1);
		}
	    if (value == (DOMAIN) 0L)
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = (DOMAIN) 0L;
		    return(RE_CONSTANT);
		}
	    if (value == (DOMAIN) 1L)
		{
		    XprDiscard(expression->ue.oe.sub2e);
		    XprForce(expression, expression->ue.oe.sub1e);
		    return(rlevel1);
		}
	    return(rlevel1);
	}
    return((rlevel2 > rlevel1) ? rlevel2 : rlevel1);
}

/***************************************************************************
*
* Function:  XprRMinus
* Purpose:   This routine reduces an integer expression that represents
*            a subtraction.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form XprR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, XprRMinus,
    EXPRESSION *, expression)
{
    unsigned rlevel1, rlevel2;
    DOMAIN value;

    rlevel1 = XprReduce(expression->ue.oe.sub1e);
    rlevel2 = XprReduce(expression->ue.oe.sub2e);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (rlevel2 == RE_CONSTANT)
		{
		    value = XprEval(expression);
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = value;
		    return(RE_CONSTANT);
		}
	    if (!XprEval(expression->ue.oe.sub1e)) 
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    expression->ue.oe.sub1e = expression->ue.oe.sub2e;
		    expression->operator = EO_NEG;
		}
	    return(rlevel2);
	}
    if (rlevel2 == RE_CONSTANT) 
	{
	    if (!XprEval(expression->ue.oe.sub2e))
		{
		    XprDiscard(expression->ue.oe.sub2e);
		    XprForce(expression, expression->ue.oe.sub1e);
		}
	    return(rlevel1);
	}
    return((rlevel2 > rlevel1) ? rlevel2 : rlevel1);
}

/***************************************************************************
*
* Function:  XprRPlus
* Purpose:   This routine reduces an integer expression that represents
*            a sum.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form XprR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, XprRPlus,
    EXPRESSION *, expression)
{
    unsigned rlevel1, rlevel2;
    DOMAIN value;

    rlevel1 = XprReduce(expression->ue.oe.sub1e);
    rlevel2 = XprReduce(expression->ue.oe.sub2e);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (rlevel2 == RE_CONSTANT)
		{
		    value = XprEval(expression);
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = value;
		    return(RE_CONSTANT);
		}
	    if (!XprEval(expression->ue.oe.sub1e)) 
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprForce(expression, expression->ue.oe.sub2e);
		}
	    return(rlevel2);
	}
    if (rlevel2 == RE_CONSTANT) 
	{
	    if (!XprEval(expression->ue.oe.sub2e)) 
		{
		    XprDiscard(expression->ue.oe.sub2e);
		    XprForce(expression, expression->ue.oe.sub1e);
		}
	    return(rlevel1);
	}
    return((rlevel2 > rlevel1) ? rlevel2 : rlevel1);
}

/***************************************************************************
*
* Function:  XprRLshift
* Purpose:   This routine reduces an integer expression that represents
*            a shifting of bits to the left.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form XprR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, XprRLshift,
    EXPRESSION *, expression)
{
    unsigned rlevel1, rlevel2;
    DOMAIN value;

    rlevel1 = XprReduce(expression->ue.oe.sub1e);
    rlevel2 = XprReduce(expression->ue.oe.sub2e);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (rlevel2 == RE_CONSTANT)
		{
		    value = XprEval(expression);
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = value;
		    return(RE_CONSTANT);
		}
	    else if (!XprEval(expression->ue.oe.sub1e))
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->ue.oe.sub1e = expression->ue.oe.sub2e;
		    expression->operator = EO_DOM;
		    expression->ue.value = (DOMAIN) 0L;
		    return(RE_CONSTANT);
		}
	    else 
		return(rlevel2);
	}
    else if (rlevel2 == RE_CONSTANT) 
	{
	    if (!XprEval(expression->ue.oe.sub2e)) 
		{
		    XprDiscard(expression->ue.oe.sub2e);
		    XprForce(expression, expression->ue.oe.sub1e);
		}
	    return(rlevel1);
	}
    else
	return(rlevel2 > rlevel1 ? rlevel2 : rlevel1);
}

/***************************************************************************
*
* Function:  XprRLss
* Purpose:   This routine reduces an integer expression that represents
*            a "less than" comparison of numbers.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form XprR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, XprRLss,
    EXPRESSION *, expression)
{
    unsigned rlevel1, rlevel2;
    DOMAIN value;

    rlevel1 = XprReduce(expression->ue.oe.sub1e);
    rlevel2 = XprReduce(expression->ue.oe.sub2e);
    if ((rlevel1 == RE_CONSTANT) && (rlevel2 == RE_CONSTANT)) 
	{
	    value = XprEval(expression);
	    XprDiscard(expression->ue.oe.sub1e);
	    XprDiscard(expression->ue.oe.sub2e);
	    expression->operator = EO_DOM;
	    expression->ue.value = value;
	    return(RE_CONSTANT);
	}
    else 
	return(rlevel2 > rlevel1 ? rlevel2 : rlevel1);
}

/***************************************************************************
*
* Function:  XprRBand
* Purpose:   This routine reduces an integer expression that represents
*            a bitwise "and" operation.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form XprR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, XprRBand,
    EXPRESSION *, expression)
{
    unsigned rlevel1, rlevel2;
    DOMAIN value;

    rlevel1 = XprReduce(expression->ue.oe.sub1e);
    rlevel2 = XprReduce(expression->ue.oe.sub2e);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (rlevel2 == RE_CONSTANT)
		{
		    value = XprEval(expression);
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = value;
		    return(RE_CONSTANT);
		}
	    else
		{
		    value = XprEval(expression->ue.oe.sub1e);
		    if (value == (DOMAIN) 0L)
			{
			    XprDiscard(expression->ue.oe.sub1e);
			    XprDiscard(expression->ue.oe.sub2e);
			    expression->operator = EO_DOM;
			    expression->ue.value = (DOMAIN) 0L;
			    return(RE_CONSTANT);
			}
		    else if (value == ~((DOMAIN) 0L))
			{
			    XprDiscard(expression->ue.oe.sub1e);
			    XprForce(expression, expression->ue.oe.sub2e);
			    return(rlevel2);
			}
		    else
			return(rlevel2);
		}
	}
    else if (rlevel2 == RE_CONSTANT)
	{
	    value = XprEval(expression->ue.oe.sub2e);
	    if (value == (DOMAIN) 0L)
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = (DOMAIN) 0L;
		    return(RE_CONSTANT);
		}
	    else if (value == ~((DOMAIN) 0L))
		{
		    XprDiscard(expression->ue.oe.sub2e);
		    XprForce(expression, expression->ue.oe.sub1e);
		    return(rlevel1);
		}
	    else
		return(rlevel1);
	}
    else
	return(rlevel2 > rlevel1 ? rlevel2 : rlevel1);
}

/***************************************************************************
*
* Function:  XprRBxor
* Purpose:   This routine reduces an integer expression that represents
*            a bitwise "exclusive or" operation.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form XprR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, XprRBxor,
    EXPRESSION *, expression)
{
    unsigned rlevel1, rlevel2;
    DOMAIN value;

    rlevel1 = XprReduce(expression->ue.oe.sub1e);
    rlevel2 = XprReduce(expression->ue.oe.sub2e);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (rlevel2 == RE_CONSTANT) 
		{
		    value = XprEval(expression);
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = value;
		    return(RE_CONSTANT);
		}
	    else
		{
		    value = XprEval(expression->ue.oe.sub1e);
		    if (value == (DOMAIN) 0L)
			{
			    XprDiscard(expression->ue.oe.sub1e);
			    XprDiscard(expression->ue.oe.sub2e);
			    expression->operator = EO_DOM;
			    expression->ue.value = (DOMAIN) 0L;
			    return(RE_CONSTANT);
			}
		    else if (value == ~((DOMAIN) 0L))
			{
			    XprDiscard(expression->ue.oe.sub1e);
			    expression->ue.oe.sub1e = expression->ue.oe.sub2e;
			    expression->operator = EO_BNOT;
			    return(rlevel2);
			}
		    else
			return(rlevel2);
		}
	}
    else if (rlevel2 == RE_CONSTANT)
	{
	    value = XprEval(expression->ue.oe.sub2e);
	    if (value == (DOMAIN) 0L)
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = (DOMAIN) 0L;
		    return(RE_CONSTANT);
		}
	    else if (value == ~((DOMAIN) 0L))
		{
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_BNOT;
		    return(rlevel1);
		}
	    else
		return(rlevel1);
	}
    else 
	return(rlevel2 > rlevel1 ? rlevel2 : rlevel1);
}

/***************************************************************************
*
* Function:  XprRBOr
* Purpose:   This routine reduces an integer expression that represents
*            a bitwise "or" operation.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form XprR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, XprRBor,
    EXPRESSION *, expression)
{
    unsigned rlevel1, rlevel2;
    DOMAIN value;

    rlevel1 = XprReduce(expression->ue.oe.sub1e);
    rlevel2 = XprReduce(expression->ue.oe.sub2e);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (rlevel2 == RE_CONSTANT) 
		{
		    value = XprEval(expression);
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = value;
		    return(RE_CONSTANT);
		}
	    else
		{
		    value = XprEval(expression->ue.oe.sub1e);
		    if (value == (DOMAIN) 0L)
			{
			    XprDiscard(expression->ue.oe.sub1e);
			    XprForce(expression, expression->ue.oe.sub2e);
			    return(rlevel2);
			}
		    else if (value == ~((DOMAIN) 0L))
			{
			    XprDiscard(expression->ue.oe.sub1e);
			    XprDiscard(expression->ue.oe.sub2e);
			    expression->operator = EO_DOM;
			    expression->ue.value = ~((DOMAIN) 0L);
			    return(RE_CONSTANT);
			}
		    else
			return(rlevel2);
		}
	}
    else if (rlevel2 == RE_CONSTANT)
	{
	    value = XprEval(expression->ue.oe.sub2e);
	    if (value == (DOMAIN) 0L)
		{
		    XprDiscard(expression->ue.oe.sub2e);
		    XprForce(expression, expression->ue.oe.sub1e);
		    return(rlevel1);
		}
	    else if (value == ~((DOMAIN) 0L))
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = ~((DOMAIN) 0L);
		    return(RE_CONSTANT);
		}
	    else
		return(rlevel1);
	}
    else
	return(rlevel2 > rlevel1 ? rlevel2 : rlevel1);
}

/***************************************************************************
*
* Function:  XprRAnd
* Purpose:   This routine reduces an integer expression that represents
*            a short-circuit boolean "and" operation.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form XprR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, XprRAnd,
    EXPRESSION *, expression)
{
    unsigned rlevel1, rlevel2;
    DOMAIN value;

    rlevel1 = XprReduce(expression->ue.oe.sub1e);
    rlevel2 = XprReduce(expression->ue.oe.sub2e);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (rlevel2 == RE_CONSTANT)
		{
		    value = XprEval(expression);
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = value;
		    return(RE_CONSTANT);
		}
	    else if (XprEval(expression->ue.oe.sub1e)) 
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprForce(expression, expression->ue.oe.sub2e);
		    return(rlevel2);
		}
	    else 
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = (DOMAIN) 0L;
		    return(RE_CONSTANT);
		}
	}
    else if (rlevel2 == RE_CONSTANT)
	{
	    if (XprEval(expression->ue.oe.sub2e)) 
		{
		    XprDiscard(expression->ue.oe.sub2e);
		    XprForce(expression, expression->ue.oe.sub1e);
		    return(rlevel1);
		}
	    else 
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = (DOMAIN) 0L;
		    return(RE_CONSTANT);
		}
	}
    else
	return(rlevel2 > rlevel1 ? rlevel2 : rlevel1);
}

/***************************************************************************
*
* Function:  XprROr
* Purpose:   This routine reduces an integer expression that represents
*            a short-circuit boolean "or" operation.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form XprR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, XprROr,
    EXPRESSION *, expression)
{
    unsigned rlevel1, rlevel2;
    DOMAIN value;

    rlevel1 = XprReduce(expression->ue.oe.sub1e);
    rlevel2 = XprReduce(expression->ue.oe.sub2e);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (rlevel2 == RE_CONSTANT) 
		{
		    value = XprEval(expression);
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = value;
		    return(RE_CONSTANT);
		}
	    else if (XprEval(expression->ue.oe.sub1e)) 
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = (DOMAIN) 1L;
		    return(RE_CONSTANT);
		}
	    else 
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprForce(expression, expression->ue.oe.sub2e);
		    return(rlevel2);
		}
	}
    else if (rlevel2 == RE_CONSTANT)
	{
	    if (XprEval(expression->ue.oe.sub2e)) 
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = (DOMAIN) 1L;
		    return(RE_CONSTANT);
		}
	    else 
		{
		    XprDiscard(expression->ue.oe.sub2e);
		    XprForce(expression, expression->ue.oe.sub1e);
		    return(rlevel1);
		}
	}
    else
	return(rlevel2 > rlevel1 ? rlevel2 : rlevel1);
}

/***************************************************************************
*
* Function:  XprRIf
* Purpose:   This routine reduces an integer expression that represents
*            an "if-then-else" operation.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form XprR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, XprRIf,
    EXPRESSION *, expression)
{
    unsigned rlevel1, rlevel2, rlevel3;
    DOMAIN value;

    rlevel1 = XprReduce(expression->ue.oe.sub1e);
    rlevel2 = XprReduce(expression->ue.oe.sub2e);
    rlevel3 = XprReduce(expression->ue.oe.sub3e);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (XprEval(expression->ue.oe.sub1e))
		{
		    if (rlevel2 == RE_CONSTANT) 
			{
			    value = XprEval(expression);
			    XprDiscard(expression->ue.oe.sub1e);
			    XprDiscard(expression->ue.oe.sub2e);
			    XprDiscard(expression->ue.oe.sub3e);
			    expression->operator = EO_DOM;
			    expression->ue.value = value;
			    return(RE_CONSTANT);
			}
		    else 
			{
			    XprDiscard(expression->ue.oe.sub1e);
			    XprDiscard(expression->ue.oe.sub3e);
			    XprForce(expression, expression->ue.oe.sub2e);
			    return(rlevel2);
			}
		}
	    else if (rlevel3 == RE_CONSTANT) 
		{
		    value = XprEval(expression);
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    XprDiscard(expression->ue.oe.sub3e);
		    expression->operator = EO_DOM;
		    expression->ue.value = value;
		    return(RE_CONSTANT);
		}
	    else 
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    XprForce(expression, expression->ue.oe.sub3e);
		    return(rlevel3);
		}
	}
    else 
	return(rlevel2 > rlevel1 ? rlevel3 > rlevel2 ?
	       rlevel3 : rlevel2 : rlevel1);
}

/***************************************************************************
*
* Function:  XprRMls
* Purpose:   This routine reduces an integer expression that represents
*            a "less than" comparison of markings.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form XprR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, XprRMls,
    EXPRESSION *, expression)
{
    unsigned rlevel1, rlevel2;
    DOMAIN value;

    rlevel1 = MexReduce(expression->ue.me.msub1);
    rlevel2 = MexReduce(expression->ue.me.msub2);
    if ((rlevel1 == RE_CONSTANT) && (rlevel2 == RE_CONSTANT)) 
	{
	    value = XprEval(expression);
	    MexDiscard(expression->ue.me.msub1);
	    MexDiscard(expression->ue.me.msub2);
	    expression->operator = EO_DOM;
	    expression->ue.value = value;
	    return(RE_CONSTANT);
	}
    return((rlevel2 > rlevel1) ? rlevel2 : rlevel1);
}

/***************************************************************************
*
* Function:  XprRMlq
* Purpose:   This routine reduces an integer expression that represents
*            a "less than or equal to" comparison of markings.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  all routines having a name of the form XprR[A-Za-z]+
*
****************************************************************************
*/
FS1(unsigned, XprRMlq,
    EXPRESSION *, expression)
{
    unsigned rlevel1, rlevel2;
    DOMAIN value;
    MARKING *marking;

    rlevel1 = MexReduce(expression->ue.me.msub1);
    rlevel2 = MexReduce(expression->ue.me.msub2);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (rlevel2 == RE_CONSTANT)
		{
		    value = XprEval(expression);
		    MexDiscard(expression->ue.me.msub1);
		    MexDiscard(expression->ue.me.msub2);
		    expression->operator = EO_DOM;
		    expression->ue.value = value;
		    return(RE_CONSTANT);
		}
	    marking = MexEval(expression->ue.me.msub1);
	    if (MarEmpty(marking))
		{
		    MarDiscard(marking);
		    MexDiscard(expression->ue.me.msub1);
		    MexDiscard(expression->ue.me.msub2);
		    expression->operator = EO_DOM;
		    expression->ue.value = 1L;
		    return(RE_CONSTANT);
		}
	    MarDiscard(marking);
	}
    return((rlevel2 > rlevel1) ? rlevel2 : rlevel1);
}

/***************************************************************************
*
* Function:  XrgCons
* Purpose:   This routine constructs a range expression.
* Arguments: lower - the start of the range
*            upper - the end of the range
* Returns:   the expression
* See also:
*
****************************************************************************
*/
FE2(XRANGE *, XrgCons,
    EXPRESSION *, lower,
    EXPRESSION *, upper)
{
    XRANGE *xrange;

    xrange = XrgAlloc();
    xrange->lower = lower;
    xrange->upper = upper;
    return(xrange);
}

/***************************************************************************
*
* Function:  XrgAux
* Purpose:   This routine is an interface to XrgCons and is needed because
*            the construction of expressions in ltl.y or rules.y is not
*            immediately stopped after a finding of a syntax error.
* Arguments: lower - as in XrgCons
*            upper - as in XrgCons
* Returns:   NULL in error cases, and otherwise the expression
* See also:  XrgCons
*
****************************************************************************
*/
FE2(XRANGE *, XrgAux,
    EXPRESSION *, lower,
    EXPRESSION *, upper)
{
    if (errorFlag == NO_ERROR)
	return(XrgCons(lower, upper));
    if (lower)
	XprDiscard(lower);
    if (upper)
	XprDiscard(upper);
    return(NULL);
}

/***************************************************************************
*
* Function:  XrgCopy
* Purpose:   This routine makes a copy of a range expression.
* Arguments: xrange - the expression to be copied
* Returns:   the copy
* See also:
*
****************************************************************************
*/
FE1(XRANGE *, XrgCopy,
    XRANGE *, xrange)
{
    XRANGE *copy;

    copy = XrgAlloc();
    copy->lower = XprCopy(xrange->lower);
    copy->upper = XprCopy(xrange->upper);
    return(copy);
}

/***************************************************************************
*
* Function:  XrgDiscard
* Purpose:   This routine removes a range expression from the memory.
* Arguments: xrange - the expression
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, XrgDiscard,
    XRANGE *, xrange)
{
    XprDiscard(xrange->lower);
    XprDiscard(xrange->upper);
    Deallocate((char *) xrange);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  XrgDump
* Purpose:   This routine displays a range expression.
* Arguments: xrange - the range expression
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, XrgDump,
    XRANGE *, xrange,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "XRANGE\n");
    (void) fflush(_err);
    XprDump(xrange->lower, depth + 1);
    XprDump(xrange->upper, depth + 1);
}
#endif

TOOLS xrgTools =
{ (CFUNC) XrgCopy, XrgDiscard
#ifdef DEBUG
, XrgDump
#endif
 };

#ifdef DEBUG
/***************************************************************************
*
* Function:  TrgDump
* Purpose:   This routine displays a range expression list
* Arguments: trange - the list
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  XrgDump
*
****************************************************************************
*/
FE2(void, TrgDump,
    TRANGE *, trange,
    unsigned, depth)
{
    XRANGE *xrange;
    DumpIndent(depth);
    (void) fprintf(_err, "TRANGE\n");
    (void) fflush(_err);
    TRGFOR (trange, xrange) 
	XrgDump(xrange, depth + 1);
    DumpIndent(depth);
    (void) fprintf(_err, "End of TRANGE\n");
    (void) fflush(_err);
}
#endif

/***************************************************************************
*
* Function:  TrgMake
* Purpose:   This routine is used in constructing the marking that
*            corresponds to a range expression list.
* Arguments: tuple - the current tuple
*            lower - the least tuple
*            upper - the greatest tuple
*            index - the index of the current field
* Returns:   the marking consisting of tuples constructed so far
* See also:  TrgMarking
*
****************************************************************************
*/
FS4(MARKING *, TrgMake,
    TUPLE *, tuple,
    TUPLE *, lower,
    TUPLE *, upper,
    unsigned, index)
{
    MARKING *marking;

    if ((errorFlag != NO_ERROR) || !continueEvaluation)
	return(MarCreate());
    marking = MarCreate();
    if (index == TupGArity(tuple)) 
	{
	    MarInsLast(marking, MultupCons((unsigned long) 1L,
					   TupCopy(tuple)));
	    return(marking);
	}
    TupSField(tuple, index, TupGField(lower, index));
    (void) MarCatenate(marking, TrgMake(tuple, lower, upper, index + 1));
    while ((errorFlag == NO_ERROR) && continueEvaluation
	   && (TupGField(tuple, index) < TupGField(upper, index)))
	{
	    TupIncField(tuple, index);
	    (void) MarCatenate(marking, TrgMake(tuple, lower, upper,
						index + 1));
	}
    return(marking);
}

/***************************************************************************
*
* Function:  TrgMarking
* Purpose:   This routine constructs a marking that corresponds to
*            a range expression list.
* Arguments: trange - the list
* Returns:   the marking
* See also:  TrgMake
*
****************************************************************************
*/
FE1(MARKING *, TrgMarking,
    TRANGE *, trange)
{
    XRANGE *xrange;
    TUPLE *tuple, *lower, *upper;
    unsigned i, trgArity;
    MARKING *marking;

    trgArity = TrgArity(trange);
    lower = TupCons(trgArity);
    upper = TupCons(trgArity);
    TRGNUM (trange, xrange, i)
	{
	    TupSField(lower, i, (DOMAIN) XprEval(XrgGLower(xrange)));
	    if ((errorFlag != NO_ERROR) || !continueEvaluation)
		{
		    TupDiscard(lower);
		    TupDiscard(upper);
		    return(MarCreate());
		}
	    TupSField(upper, i, (DOMAIN) XprEval(XrgGUpper(xrange)));
	    if ((errorFlag != NO_ERROR) || (!continueEvaluation)
		|| (TupGField(lower, i) > TupGField(upper, i)))
		{
		    TupDiscard(lower);
		    TupDiscard(upper);
		    return(MarCreate());
		}
	}
    tuple = TupCons(trgArity);
    marking = TrgMake(tuple, lower, upper, 0);
    TupDiscard(tuple);
    TupDiscard(lower);
    TupDiscard(upper);
    if ((errorFlag != NO_ERROR) || !continueEvaluation)
	{
	    if (marking)
		MarDiscard(marking);
	    return(MarCreate());
	}
    return(marking);
}

#define MexRCartes MexRInters
#define XprRDom MexRConstant
#define XprRNot XprRNeg
#define XprRBnot XprRNeg
#define XprRRshift XprRLshift
#define XprRLeq XprRLss
#define XprREqu XprRLss
#define XprRNeq XprRLss
#define XprRMeq XprRMls
#define XprRMnq XprRMls

MEXOPENTRY mexOpTable[SZ_MO] = 
{
    { MK_CONSTANT, MexEConstant, MexRConstant },
    { MK_PLACE,    (MARF) NULL,  MexRPlace },
    { MK_TRANGE,   MexETrange,   (UNSF) NULL },
    { MK_EXPRMARK, MexETimes,    MexRTimes },
    { MK_MARKMARK, MexECartes,   MexRCartes },
    { MK_MARKMARK, MexESum,      MexRSum },
    { MK_MARKMARK, MexEDiff,     MexRDiff },
    { MK_EXPRMARK, MexESelect,   MexRSelect },
    { MK_MARKMARK, MexEInters,   MexRInters }
};

XPROPENTRY xprOpTable[SZ_EO] = 
{
    { EK_CONSTANT, XprEDom,      XprRDom },
    { EK_UNARY,    (DOMF) NULL,  XprRField },
    { EK_CARD,     XprECard,     XprRCard },
    { EK_UNARY,    XprENeg,      XprRNeg },
    { EK_UNARY,    XprENot,      XprRNot },
    { EK_UNARY,    XprEBnot,     XprRBnot },
    { EK_BINARY,   XprETimes,    XprRTimes },
    { EK_BINARY,   (DOMF) NULL,  (UNSF) NULL },
    { EK_BINARY,   (DOMF) NULL,  (UNSF) NULL },
    { EK_BINARY,   XprEMinus,    XprRMinus },
    { EK_BINARY,   XprEPlus,     XprRPlus },
    { EK_BINARY,   XprELshift,   XprRLshift },
    { EK_BINARY,   XprERshift,   XprRRshift },
    { EK_BINARY,   XprELss,      XprRLss },
    { EK_BINARY,   XprELeq,      XprRLeq },
    { EK_BINARY,   XprEEqu,      XprREqu },
    { EK_BINARY,   XprENeq,      XprRNeq },
    { EK_BINARY,   XprEBand,     XprRBand },
    { EK_BINARY,   XprEBxor,     XprRBxor },
    { EK_BINARY,   XprEBor,      XprRBor },
    { EK_BINARY,   XprEAnd,      XprRAnd },
    { EK_BINARY,   XprEOr,       XprROr },
    { EK_TRINARY,  XprEIf,       XprRIf },
    { EK_MCOMP,    XprEMls,      XprRMls },
    { EK_MCOMP,    XprEMlq,      XprRMlq },
    { EK_MCOMP,    XprEMeq,      XprRMeq },
    { EK_MCOMP,    XprEMnq,      XprRMnq },
    { EK_CONSTANT, (DOMF) NULL,  (UNSF) NULL },
    { EK_NULLARY,  (DOMF) NULL,  (UNSF) NULL },
    { EK_NULLARY,  (DOMF) NULL,  (UNSF) NULL }
};
