/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  formula.c
*
* Part of: PROD's program probe
*
* Authors: Keijo Heljanko, Marko Rauhamaa and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This module is concentrated on the handling and especially
*     the evaluation of the formulas of the graph query language.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include <stdio.h>
#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "graph.h"
#include "mark.h"
#include "encode.h"
#include "expr.h"
#include "formula.h"
#include "mbufs.h"
#include "macptree.h"
#include "comline.h"
#include "probe.h"

GRAPH *fogGraph;
unsigned long fogWorld;
unsigned fogTransition;
DOMAIN *fogInstance;
unsigned fogPriority;
BOOL fogFaithful;

/***************************************************************************
*
* Function:  MexEPlace
* Purpose:   This routine evaluates a marking of a place.
* Arguments: markexpr - a marking expression pointing to the place
* Returns:   the marking
* See also:  MexEPlace (prpexpr.c)
*
****************************************************************************
*/
FS1(MARKING *, MexEPlace,
    MARKEXPR *, markexpr)
{
    STATE *state;
    MARKING *marking;

    state = StaAlloc();
    BStaDecode(GphNode(FogGGraph(), FogGWorld()), state);
    marking = MarCopy(StaGMarking(state, markexpr->um.place));
    StaDiscard(state);
    return(marking);
}

/***************************************************************************
*
* Function:  MexRTrange
* Purpose:   This routine reduces a marking expression that points to
*            a range list.
* Arguments: markexpr - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            marking, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  MexRTrange (prpexpr.c)
*
****************************************************************************
*/
FS1(unsigned, MexRTrange,
    MARKEXPR *, markexpr)
{
    unsigned rlevel1, rlevel2, rlevel3;
    XRANGE *xrange;
    MARKING *marking;

    rlevel1 = RE_CONSTANT;
    TRGFOR (markexpr->um.trange, xrange) 
	{
	    rlevel2 = XprReduce(xrange->lower);
	    if (rlevel2 > rlevel1) 
		rlevel1 = rlevel2;
	    rlevel3 = XprReduce(xrange->upper);
	    if ((rlevel2 == RE_CONSTANT) && (rlevel3 == RE_CONSTANT) &&
		(XprEval(xrange->lower) > XprEval(xrange->upper)))
		{
		    TrgDiscard(markexpr->um.trange);
		    markexpr->operator = MO_CONSTANT;
		    markexpr->um.marking = MarCreate();
		    return(RE_CONSTANT);
		}
	    if (rlevel3 > rlevel1) 
		rlevel1 = rlevel3;
	}
    if (rlevel1 == RE_CONSTANT) 
	{
	    marking = MexEval(markexpr);
	    TrgDiscard(markexpr->um.trange);
	    markexpr->operator = MO_CONSTANT;
	    markexpr->um.marking = marking;
	}
    return(rlevel1);
}

/***************************************************************************
*
* Function:  XprEField
* Purpose:   This routine evaluates an integer expression that represents
*            a field of a known tuple.
* Arguments: expression - the expression
* Returns:   the field
* See also:  XprEField (prpexpr.c)
*
****************************************************************************
*/
FS1(DOMAIN, XprEField,
    EXPRESSION *, expression)
{
    unsigned field;

    field = (unsigned) XprEval(expression->ue.oe.sub1e);
    if ((!FogGTuple()) || (TupGArity(FogGTuple()) <= field))
	{
	    if (errorFlag == NO_ERROR)
		{
		    (void) fprintf(_err, "Illegal field reference: %u\n",
				   field);
		    (void) fflush(_err);
		    if (FogGTuple())
			TupDump(FogGTuple(), (unsigned) 0);
		    errorFlag = XPR_ERROR;
		}
	    return((DOMAIN) 0L);
	}
    return(TupGField(FogGTuple(), field));
}

/***************************************************************************
*
* Function:  XprEDiv
* Purpose:   This routine evaluates an integer expression that represents
*            the quotient of a division.
* Arguments: expression - the expression 
* Returns:   the quotient
* See also:  XprEDiv (prpexpr.c)
*
****************************************************************************
*/
FS1(DOMAIN, XprEDiv,
    EXPRESSION *, expression)
{
    register DOMAIN divisor;

    divisor = XprEval(expression->ue.oe.sub2e);
    if (divisor)
	return(XprEval(expression->ue.oe.sub1e) / divisor);
    if (errorFlag == NO_ERROR)
	{
	    (void) fprintf(_err, "Division by zero\n");
	    (void) fflush(_err);
	    errorFlag = XPR_ERROR;
	}
    return((DOMAIN) 0L);
}

/***************************************************************************
*
* Function:  XprERem
* Purpose:   This routine evaluates an integer expression that represents
*            the remainder of a division.
* Arguments: expression - the expression
* Returns:   the remainder
* See also:  XprERem (prpexpr.c)
*
****************************************************************************
*/
FS1(DOMAIN, XprERem,
    EXPRESSION *, expression)
{
    register DOMAIN divisor;

    divisor = XprEval(expression->ue.oe.sub2e);
    if (divisor)
	return(XprEval(expression->ue.oe.sub1e) % divisor);
    if (errorFlag == NO_ERROR)
	{
	    (void) fprintf(_err, "Division by zero\n");
	    (void) fflush(_err);
	    errorFlag = XPR_ERROR;
	}
    return((DOMAIN) 0L);
}

/***************************************************************************
*
* Function:  XprETravar
* Purpose:   This routine evaluates an integer expression that represents
*            a variable of a transition.
* Arguments: expression - the expression
* Returns:   the value of the variable
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS1(DOMAIN, XprETravar,
    EXPRESSION *, expression)
{
    return(FogGInstance((unsigned) expression->ue.value));
}

/***************************************************************************
*
* Function:  XprEPreClass
* Purpose:   This routine evaluates an integer expression that represents
*            a precedence class of a transition instance
* Arguments: expression - the expression
* Returns:   the class
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS0(DOMAIN, XprEPreClass)
{
    return((DOMAIN) FogGPriority());
}

/***************************************************************************
*
* Function:  XprEFaithful
* Purpose:   This routine evaluates an integer expression that represents
*            a flag indicating if a sequence of transition instances
*            actually leads to the marking pointed by the corresponding
*            arrow.
* Arguments: expression - the expression
* Returns:   the flag
* See also:  all routines having a name of the form XprE[A-Za-z]+
*
****************************************************************************
*/
FS0(DOMAIN, XprEFaithful)
{
    return(FogGFaithful() ? ((DOMAIN) 1L) : (DOMAIN) 0L);
}

/***************************************************************************
*
* Function:  XprRDiv
* Purpose:   This routine reduces an integer expression that represents
*            the quotient of a division.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  XprRDiv (prpexpr.c)
*
****************************************************************************
*/
FS1(unsigned, XprRDiv,
    EXPRESSION *, expression)
{
    unsigned rlevel1, rlevel2;
    DOMAIN value;

    rlevel1 = XprReduce(expression->ue.oe.sub1e);
    rlevel2 = XprReduce(expression->ue.oe.sub2e);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (rlevel2 == RE_CONSTANT)
		{
		    value = XprEval(expression);
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = value;
		    return(RE_CONSTANT);
		}
	    if (!XprEval(expression->ue.oe.sub1e))
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = (DOMAIN) 0L;
		    return(RE_CONSTANT);
		}
	    return(rlevel2);
	}
    if (rlevel2 == RE_CONSTANT)
	{
	    value = XprEval(expression->ue.oe.sub2e);
	    if (value == (DOMAIN) 0L)
		{
		    if (errorFlag == NO_ERROR)
			{
			    (void) fprintf(_err, "Division by zero\n");
			    (void) fflush(_err);
			    errorFlag = XPR_ERROR;
			}
		    return(rlevel1);
		}
	    if (value == (DOMAIN) -1L)
		{
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_NEG;
		    return(rlevel1);
		}
	    if (value == (DOMAIN) 1L)
		{
		    XprDiscard(expression->ue.oe.sub2e);
		    XprForce(expression, expression->ue.oe.sub1e);
		    return(rlevel1);
		}
	    return(rlevel1);
	}
    return((rlevel2 > rlevel1) ? rlevel2 : rlevel1);
}

/***************************************************************************
*
* Function:  XprRRem
* Purpose:   This routine reduces an integer expression that represents
*            the remainder of a division.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  XprRRem (prpexpr.c)
*
****************************************************************************
*/
FS1(unsigned, XprRRem,
    EXPRESSION *, expression)
{
    unsigned rlevel1, rlevel2;
    DOMAIN value;

    rlevel1 = XprReduce(expression->ue.oe.sub1e);
    rlevel2 = XprReduce(expression->ue.oe.sub2e);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (rlevel2 == RE_CONSTANT) 
		{
		    value = XprEval(expression);
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = value;
		    return(RE_CONSTANT);
		}
	    if (!XprEval(expression->ue.oe.sub1e)) 
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = (DOMAIN) 0L;
		    return(RE_CONSTANT);
		}
	    return(rlevel2);
	}
    if (rlevel2 == RE_CONSTANT)
	{
	    if ((XprEval(expression->ue.oe.sub2e) == (DOMAIN) 0L) &&
		(errorFlag == NO_ERROR))
		{
		    (void) fprintf(_err, "Division by zero\n");
		    (void) fflush(_err);
		    errorFlag = XPR_ERROR;
		}
	    return(rlevel1);
	}
    return((rlevel2 > rlevel1) ? rlevel2 : rlevel1);
}

/***************************************************************************
*
* Function:  FclCons
* Purpose:   This routine constructs a firing class expression.
* Arguments: operator - the outermost operator of the expression
*            fcl1 - the first operand if such exists
*            fcl2 - the second operand if such exists
*            transition - a number of a transition if operator is FC_ATOM
* Returns:   the expression
* See also:
*
****************************************************************************
*/
FE4(FCLASS *, FclCons,
    unsigned, operator,
    FCLASS *, fcl1,
    FCLASS *, fcl2,
    unsigned, transition)
{
    FCLASS *fclass;

    fclass = FclAlloc();
    fclass->operator = operator;
    switch (operator) 
	{
	case FC_ATOM:
	    fclass->uc.tc.expression = (EXPRESSION *) fcl1;
	    fclass->uc.tc.transition = transition;
	    break;
	case FC_NOT:
	    fclass->uc.sc.sub1c = fcl1;
	    break;
	case FC_OR:
	case FC_AND:
	    fclass->uc.sc.sub1c = fcl1;
	    fclass->uc.sc.sub2c = fcl2;
	    break;
	}
    return(fclass);
}

/***************************************************************************
*
* Function:  FclAux
* Purpose:   This routine is an interface to FclCons and is needed because
*            the construction of expressions in rules.y is not
*            immediately stopped after a finding of a syntax error.
* Arguments: operator - as in FclCons
*            fcl1 - as in FclCons
*            fcl2 - as in FclCons
*            transition - as in FclCons
* Returns:   NULL in error cases, and otherwise the expression
* See also:  FclCons
*
****************************************************************************
*/
FE4(FCLASS *, FclAux,
    unsigned, operator,
    FCLASS *, fcl1,
    FCLASS *, fcl2,
    unsigned, transition)
{
    if (errorFlag == NO_ERROR)
	return(FclCons(operator, fcl1, fcl2, transition));
    switch (operator) 
	{
	case FC_ATOM:
	    if (fcl1)
		XprDiscard((EXPRESSION *) fcl1);
	    break;
	case FC_NOT:
	    if (fcl1)
		FclDiscard(fcl1);
	    break;
	case FC_OR:
	case FC_AND:
	    if (fcl1)
		FclDiscard(fcl1);
	    if (fcl2)
		FclDiscard(fcl2);
	    break;
	}
    return(NULL);
}

/***************************************************************************
*
* Function:  FclCopy
* Purpose:   This routine makes a copy of a firing class expression.
* Arguments: fclass - the expression
* Returns:   the copy
* See also:
*
****************************************************************************
*/
FE1(FCLASS *, FclCopy,
    FCLASS *, fclass)
{
    FCLASS *copy;

    copy = FclAlloc();
    copy->operator = fclass->operator;
    switch (fclass->operator) 
	{
	case FC_ATOM:
	    copy->uc.tc.transition = fclass->uc.tc.transition;
	    copy->uc.tc.expression = XprCopy(fclass->uc.tc.expression);
	    break;
	case FC_NOT:
	    copy->uc.sc.sub1c = FclCopy(fclass->uc.sc.sub1c);
	    break;
	case FC_OR:
	case FC_AND:
	    copy->uc.sc.sub1c = FclCopy(fclass->uc.sc.sub1c);
	    copy->uc.sc.sub2c = FclCopy(fclass->uc.sc.sub2c);
	    break;
	}
    return(copy);
}

/***************************************************************************
*
* Function:  FclDiscard
* Purpose:   This routine removes a firing class expression from
*            the memory.
* Arguments: fclass - the expression
* Returns:
* See also:
*
****************************************************************************
*/
FE1(void, FclDiscard,
    FCLASS *, fclass)
{
    switch (fclass->operator) 
	{
	case FC_ATOM:
	    XprDiscard(fclass->uc.tc.expression);
	    break;
	case FC_NOT:
	    FclDiscard(fclass->uc.sc.sub1c);
	    break;
	case FC_OR:
	case FC_AND:
	    FclDiscard(fclass->uc.sc.sub1c);
	    FclDiscard(fclass->uc.sc.sub2c);
	    break;
	}
    Deallocate((char *) fclass);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  FclDump
* Purpose:   This routine displays a firing class expression.
* Arguments: fclass - the expression
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, FclDump,
    FCLASS *, fclass,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "FCLASS %u\n", fclass->operator);
    (void) fflush(_err);
    switch (fclass->operator) 
	{
	case FC_ATOM:
	    DumpIndent(depth + 1);
	    (void) fprintf(_err, "transition = %u\n", fclass->uc.tc.transition);
	    (void) fflush(_err);
	    XprDump(fclass->uc.tc.expression, depth + 1);
	    break;
	case FC_NOT:
	    FclDump(fclass->uc.sc.sub1c, depth + 1);
	    break;
	case FC_OR:
	case FC_AND:
	    FclDump(fclass->uc.sc.sub1c, depth + 1);
	    FclDump(fclass->uc.sc.sub2c, depth + 1);
	    break;
	}
}
#endif

/***************************************************************************
*
* Function:  FclInClass
* Purpose:   This routine checks if a chosen transition instance is
*            in the set represented by a firing class expression.
* Arguments: fclass - the expression
* Returns:   TRUE iff the instance is in the set
* See also:
*
****************************************************************************
*/
FE1(BOOL, FclInClass,
    FCLASS *, fclass)
{
    switch (fclass->operator) 
	{
	case FC_ATOM:
	    return((FogGTransition() == fclass->uc.tc.transition) &&
		   (XprEval(fclass->uc.tc.expression) != (DOMAIN) 0L));
	case FC_NOT:
	    return(!FclInClass(fclass->uc.sc.sub1c));
	case FC_OR:
	    return(FclInClass(fclass->uc.sc.sub1c) ||
		   FclInClass(fclass->uc.sc.sub2c));
	case FC_AND:
	    return(FclInClass(fclass->uc.sc.sub1c) &&
		   FclInClass(fclass->uc.sc.sub2c));
	}
    return(FALSE);
}

/***************************************************************************
*
* Function:  ForCons
* Purpose:   This routine constructs a formula.
* Arguments: operator - the outermost operator of the formuka
*            fp1 - the first operand if such exists
*            fp2 - the second operand if such exists
*            fp3 - the third operand if such exists
* Returns:   the formula
* See also:  ForSnoc (prpexpr.c)
*
****************************************************************************
*/
FE4(FORMULA *, ForCons,
    unsigned, operator,
    FORMULA *, fp1,
    FORMULA *, fp2,
    FORMULA *, fp3)
{
    FORMULA *formula;
    register EXPRESSION *ep;

    formula = ForAlloc();
    formula->operator = operator;
    switch (ForKind(formula)) 
	{
	case FK_SET:
	    formula->uf.setOper = (SETOPER *) fp1;
	    break;
	case FK_FIRE:
	    formula->uf.ff.fclass = (FCLASS *) fp1;
	    formula->uf.ff.sub = fp2;
	    break;
	case FK_EXPRESSION:
	    ep = (EXPRESSION *) fp1;
	    if ((formula->operator == FO_EXPR) &&
		(ep->operator == EO_DOM))
		{
		    if (ep->ue.value == (DOMAIN) 0L)
			formula->operator = FO_FALSE;
		    else
			formula->operator = FO_TRUE;
		    XprDiscard((EXPRESSION *) fp1);
		}
	    else
		formula->uf.expression = ep;
	    break;
	case FK_NULLARY:
	    break;
	case FK_UNARY:
	    formula->uf.of.sub1f = fp1;
	    break;
	case FK_BINARY:
	    formula->uf.of.sub1f = fp1;
	    formula->uf.of.sub2f = fp2;
	    break;
	case FK_TRINARY:
	    formula->uf.of.sub1f = fp1;
	    formula->uf.of.sub2f = fp2;
	    formula->uf.of.sub3f = fp3;
	    break;
	}
    return(formula);
}

/***************************************************************************
*
* Function:  ForAux
* Purpose:   This routine is an interface to ForCons and is needed because
*            the construction of formulas in rules.y is not immediately
*            stopped after a finding of a syntax error.
* Arguments: operator - as in ForCons
*            fp1 - as in ForCons
*            fp2 - as in ForCons
*            fp3 - as in ForCons
* Returns:   NULL in error cases, and otherwise the formula
* See also:  ForCons, ForXua (prpexpr.c)
*
****************************************************************************
*/
FE4(FORMULA *, ForAux,
    unsigned, operator,
    FORMULA *, fp1,
    FORMULA *, fp2,
    FORMULA *, fp3)
{
    if (errorFlag == NO_ERROR)
	return(ForCons(operator, fp1, fp2, fp3));
    switch (forOpTable[operator]._kind) 
	{
	case FK_SET:
	    if (fp1)
		SopDiscard((SETOPER *) fp1);
	    break;
	case FK_FIRE:
	    if (fp1)
		FclDiscard((FCLASS *) fp1);
	    if (fp2)
		ForDiscard(fp2);
	    break;
	case FK_EXPRESSION:
	    if (fp1)
		XprDiscard((EXPRESSION *) fp1);
	    break;
	case FK_NULLARY:
	    break;
	case FK_UNARY:
	    if (fp1)
		ForDiscard(fp1);
	    break;
	case FK_BINARY:
	    if (fp1)
		ForDiscard(fp1);
	    if (fp2)
		ForDiscard(fp2);
	    break;
	case FK_TRINARY:
	    if (fp1)
		ForDiscard(fp1);
	    if (fp2)
		ForDiscard(fp2);
	    if (fp3)
		ForDiscard(fp3);
	    break;
	}
    return(NULL);
}

/***************************************************************************
*
* Function:  ForCopy
* Purpose:   This routine makes a copy of a formula.
* Arguments: formula - the formula to be copied
* Returns:   the copy
* See also:  ForYpoc (prpexpr.c)
*
****************************************************************************
*/
FE1(FORMULA *, ForCopy,
    FORMULA *, formula)
{
    FORMULA *copy;

    copy = ForAlloc();
    copy->operator = formula->operator;
    switch (ForKind(formula)) 
	{
	case FK_SET:
	    copy->uf.setOper = SopCopy(formula->uf.setOper);
	    break;
	case FK_FIRE:
	    copy->uf.ff.fclass = FclCopy(formula->uf.ff.fclass);
	    copy->uf.ff.sub = ForCopy(formula->uf.ff.sub);
	    break;
	case FK_EXPRESSION:
	    copy->uf.expression = XprCopy(formula->uf.expression);
	    break;
	case FK_NULLARY:
	    break;
	case FK_TRINARY:
	    copy->uf.of.sub3f = ForCopy(formula->uf.of.sub3f);
	    /* intentional abuse of switch */ 
	case FK_BINARY:
	    copy->uf.of.sub2f = ForCopy(formula->uf.of.sub2f);
	    /* intentional abuse of switch */ 
	case FK_UNARY:
	    copy->uf.of.sub1f = ForCopy(formula->uf.of.sub1f);
	    break;
	}
    return(copy);
}

/***************************************************************************
*
* Function:  ForDiscard
* Purpose:   This routine removes a formula from the memory.
* Arguments: formula - the formula
* Returns:
* See also:  ForDracsid (prpexpr.c)
*
****************************************************************************
*/
FE1(void, ForDiscard,
    FORMULA *, formula)
{
    switch (ForKind(formula)) 
	{
	case FK_SET:
	    SopDiscard(formula->uf.setOper);
	    break;
	case FK_FIRE:
	    FclDiscard(formula->uf.ff.fclass);
	    ForDiscard(formula->uf.ff.sub);
	    break;
	case FK_EXPRESSION:
	    XprDiscard(formula->uf.expression);
	    break;
	case FK_NULLARY:
	    break;
	case FK_UNARY:
	    ForDiscard(formula->uf.of.sub1f);
	    break;
	case FK_BINARY:
	    ForDiscard(formula->uf.of.sub1f);
	    ForDiscard(formula->uf.of.sub2f);
	    break;
	case FK_TRINARY:
	    ForDiscard(formula->uf.of.sub1f);
	    ForDiscard(formula->uf.of.sub2f);
	    ForDiscard(formula->uf.of.sub3f);
	    break;
	}
    Deallocate((char *) formula);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  ForDump
* Purpose:   This routine displays a formula.
* Arguments: formula - the formula
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  ForPmud (prpexpr.c)
*
****************************************************************************
*/
FE2(void, ForDump,
    FORMULA *, formula,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "FORMULA: operator %u\n", formula->operator);
    (void) fflush(_err);
    switch (ForKind(formula)) 
	{
	case FK_SET:
	    SopDump(formula->uf.setOper, depth + 1);
	    break;
	case FK_FIRE:
	    FclDump(formula->uf.ff.fclass, depth + 1);
	    ForDump(formula->uf.ff.sub, depth + 1);
	    break;
	case FK_EXPRESSION:
	    XprDump(formula->uf.expression, depth + 1);
	    break;
	case FK_NULLARY:
	    break;
	case FK_UNARY:
	    ForDump(formula->uf.of.sub1f, depth + 1);
	    break;
	case FK_BINARY:
	    ForDump(formula->uf.of.sub1f, depth + 1);
	    ForDump(formula->uf.of.sub2f, depth + 1);
	    break;
	case FK_TRINARY:
	    ForDump(formula->uf.of.sub1f, depth + 1);
	    ForDump(formula->uf.of.sub2f, depth + 1);
	    ForDump(formula->uf.of.sub3f, depth + 1);
	    break;
	}
    DumpIndent(depth);
    (void) fprintf(_err, "End of FORMULA\n");
    (void) fflush(_err);
}
#endif

/***************************************************************************
*
* Function:  ForIFalse
* Purpose:   This routine initializes a constant false formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS0(void, ForIFalse)
{
}

/***************************************************************************
*
* Function:  ForEFalse
* Purpose:   This routine evaluates a constant false formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:   NULL
* See also:
*
****************************************************************************
*/
FS0(PATH *, ForEFalse)
{
    return(NULL);
}

/***************************************************************************
*
* Function:  ForITrue
* Purpose:   This routine initializes a constant true formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS1(void, ForITrue,
    FORMULA *, formula)
{
    formula->mf.visited = FALSE;
}

/***************************************************************************
*
* Function:  ForETrue
* Purpose:   This routine evaluates a constant true formula.
* Arguments: formula - the formula
*            world - address of evaluation node in gph-file
* Returns:   the evaluation node "path" or NULL
* See also:
*
****************************************************************************
*/
FE2(PATH *, ForETrue,
    FORMULA *, formula,
    unsigned long, world)
{
    return((formula->mf.visited = !formula->mf.visited)
	   ? PthCreate(world) : NULL);
}

/***************************************************************************
*
* Function:  ForINot
* Purpose:   This routine initializes a negation formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForINot,
    FORMULA *, formula,
    GRAPH *, graph)
{
    formula->mf.visited = FALSE;
    ForInit(formula->uf.of.sub1f, graph);
}

/***************************************************************************
*
* Function:  ForCNot
* Purpose:   This routine clears a negation formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:
* See also:
*
****************************************************************************
*/
FS3(void, ForCNot,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    formula->mf.visited = FALSE;
    ForClear(formula->uf.of.sub1f, graph, world);
}

/***************************************************************************
*
* Function:  ForENot
* Purpose:   This routine evaluates a negation formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:   the evaluation node "path" or NULL
* See also:
*
****************************************************************************
*/
FS3(PATH *, ForENot,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    PATH *path;

    if (formula->mf.visited) 
	{
	    formula->mf.visited = FALSE;
	    return(NULL);
	}
    if ((path = ForEval(formula->uf.of.sub1f, graph, world)) )
	{
	    PthDiscard(path);
	    ForClear(formula->uf.of.sub1f, graph, world);
	    return(NULL);
	}
    else 
	{
	    formula->mf.visited = TRUE;
	    return(PthCreate(world));
	}
}

/***************************************************************************
*
* Function:  ForQNot
* Purpose:   This routine closes a negation formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForQNot,
    FORMULA *, formula,
    GRAPH *, graph)
{
    ForClose(formula->uf.of.sub1f, graph);
}

/***************************************************************************
*
* Function:  ForINode
* Purpose:   This routine initializes a global enumeration formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForINode,
    FORMULA *, formula,
    GRAPH *, graph)
{
    formula->mf.nf.state = formula->mf.nf.wdCount = GphNdCount(graph);
    ForInit(formula->uf.of.sub1f, graph);
}

/***************************************************************************
*
* Function:  ForCNode
* Purpose:   This routine clears a global enumeration formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForCNode,
    FORMULA *, formula,
    GRAPH *, graph)
{
    ForClear(formula->uf.of.sub1f, graph,
	     GphId(graph, formula->mf.nf.state));
    formula->mf.nf.state = formula->mf.nf.wdCount;
}

/***************************************************************************
*
* Function:  ForENode
* Purpose:   This routine evaluates a global enumeration formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:   a path from a matching node or NULL
* See also:
*
****************************************************************************
*/
FS2(PATH *, ForENode,
    FORMULA *, formula,
    GRAPH *, graph)
{
    PATH *path;

    if (formula->mf.nf.state == formula->mf.nf.wdCount) 
	{
	    formula->mf.nf.state = 0L;
	    formula->mf.nf.wdCount = GphNdCount(graph);
	    if (formula->mf.nf.wdCount == 0L)
		return(NULL);
	}
    while (continueEvaluation && (errorFlag == NO_ERROR)) 
	{
	    if ((path = ForEval(formula->uf.of.sub1f, graph,
				GphId(graph, formula->mf.nf.state))))
		return(path);
	    if (++formula->mf.nf.state == formula->mf.nf.wdCount) 
		return(NULL);
	}
    return(NULL);
}

/***************************************************************************
*
* Function:  ForQNode
* Purpose:   This routine closes a global enumeration formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForQNode,
    FORMULA *, formula,
    GRAPH *, graph)
{
    ForClose(formula->uf.of.sub1f, graph);
}

/***************************************************************************
*
* Function:  ForIStep
* Purpose:   This routine initializes a step formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForIStep,
    FORMULA *, formula,
    GRAPH *, graph)
{
    formula->mf.af.awCount = 0L;
    ForInit(formula->uf.of.sub1f, graph);
}

/***************************************************************************
*
* Function:  ForCStep
* Purpose:   This routine clears a step formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:
* See also:
*
****************************************************************************
*/
FS3(void, ForCStep,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    formula->mf.af.awCount = 0L;
    ForClear(formula->uf.of.sub1f, graph, world);
}

/***************************************************************************
*
* Function:  ForEStep
* Purpose:   This routine evaluates a step formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:   a single-edge path from the evaluation node or NULL
* See also:
*
****************************************************************************
*/
FS3(PATH *, ForEStep,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    PATH *path;

    if (formula->mf.af.awCount == 0L) 
	{
	    formula->mf.af.awCount = GphAwCount(graph, world);  
	    if (formula->mf.af.awCount == (unsigned long) -1L)
		formula->mf.af.awCount = 0L;
	    if (formula->mf.af.awCount == 0L)
		return(NULL);
	    GphFlFirst(graph, world, &(formula->mf.af.awAddr),
		       &(formula->mf.af.succ));
	}
    while (!(path = ForEval(formula->uf.of.sub1f, graph, formula->mf.af.succ)) &&
	   continueEvaluation && (errorFlag == NO_ERROR))
	{
	    if (--formula->mf.af.awCount == 0L)
		return(NULL);
	    GphFlNext(graph, &(formula->mf.af.awAddr), &(formula->mf.af.succ));
	}
    if (continueEvaluation && (errorFlag == NO_ERROR))
	{
	    PthInsFirst(path, world, formula->mf.af.awAddr);
	    return(path);
	}
    if (path) 
	PthDiscard(path);
    return(NULL);
}

/***************************************************************************
*
* Function:  ForQStep
* Purpose:   This routine closes a step formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForQStep,
    FORMULA *, formula,
    GRAPH *, graph)
{
    ForClose(formula->uf.of.sub1f, graph);
}

/***************************************************************************
*
* Function:  ForIFire
* Purpose:   This routine initializes a fire formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForIFire,
    FORMULA *, formula,
    GRAPH *, graph)
{
    formula->mf.af.awCount = 0L;
    ForInit(formula->uf.ff.sub, graph);
}

/***************************************************************************
*
* Function:  ForCFire
* Purpose:   This routine clears a fire formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:
* See also:
*
****************************************************************************
*/
FS3(void, ForCFire,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    formula->mf.af.awCount = 0L;
    ForClear(formula->uf.ff.sub, graph, world);
}

/***************************************************************************
*
* Function:  ForEFire
* Purpose:   This routine evaluates a fire formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:   a single-edge path from the evaluation node or NULL
* See also:
*
****************************************************************************
*/
FS3(PATH *, ForEFire,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    PATH *path;
    DOMAIN *instance = NULL;
    DATATYPE *pointer;
    BOOL inClass = FALSE;

    FogSGraph(graph);
    if (formula->mf.af.awCount == 0L) 
	{
	    formula->mf.af.awCount = GphAwCount(graph, world);  
	    if (formula->mf.af.awCount == (unsigned long) -1L)
		formula->mf.af.awCount = 0L;
	    if (formula->mf.af.awCount == 0L)
		return(NULL);
	    GphFlFirst(graph, world, &(formula->mf.af.awAddr),
		       &(formula->mf.af.succ));
	    pointer = GphArrow(graph, formula->mf.af.awAddr);
	    while (TRUE)
		{
		    instance = BInsDecode(FogATransition(), FogAPriority(),
					  FogAFaithful(), &pointer);
		    FogSInstance(instance);
		    inClass = FclInClass(formula->uf.ff.fclass);
		    if (inClass || !pointer)
			break;
		    Deallocate((char *) instance);
		}
	    while (!inClass)
		{
		    Deallocate((char *) instance);
		    if (--formula->mf.af.awCount == 0L)
			return(NULL);
		    GphFlNext(graph, &(formula->mf.af.awAddr),
			      &(formula->mf.af.succ));
		    pointer = GphArrow(graph, formula->mf.af.awAddr);
		    while (TRUE)
			{
			    instance = BInsDecode(FogATransition(),
						  FogAPriority(),
						  FogAFaithful(), &pointer);
			    FogSInstance(instance);
			    inClass = FclInClass(formula->uf.ff.fclass);
			    if (inClass || !pointer)
				break;
			    Deallocate((char *) instance);
			}
		}
	}
    else
	instance = NULL;
    while ((!(path = ForEval(formula->uf.ff.sub, graph, formula->mf.af.succ))) &&
	   continueEvaluation && (errorFlag == NO_ERROR))
	do 
	    {
		if (instance)
		    Deallocate((char *) instance);
		if (--formula->mf.af.awCount == 0L)
		    return(NULL);
		GphFlNext(graph, &(formula->mf.af.awAddr), &(formula->mf.af.succ));
		pointer = GphArrow(graph, formula->mf.af.awAddr);
		while (TRUE)
		    {
			instance = BInsDecode(FogATransition(), FogAPriority(),
					      FogAFaithful(), &pointer);
			FogSInstance(instance);
			inClass = FclInClass(formula->uf.ff.fclass);
			if (inClass || !pointer)
			    break;
			Deallocate((char *) instance);
		    }
	    } while (!inClass);
    if (instance)
	Deallocate((char *) instance);
    if (continueEvaluation && (errorFlag == NO_ERROR)) 
	{
	    PthInsFirst(path, world, formula->mf.af.awAddr);
	    return(path);
	}
    if (path) 
	PthDiscard(path);
    return(NULL);
}

/***************************************************************************
*
* Function:  ForQFire
* Purpose:   This routine closes a fire formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForQFire,
    FORMULA *, formula,
    GRAPH *, graph)
{
    ForClose(formula->uf.ff.sub, graph);
}

/***************************************************************************
*
* Function:  ForIBPath
* Purpose:   <<<<<this routine initialzes a bpath formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForIBPath,
    FORMULA *, formula,
    GRAPH *, graph)
{
    formula->mf.bf.front = LstCreate(&pthTools);
    ForInit(formula->uf.of.sub1f, graph);
    ForInit(formula->uf.of.sub2f, graph);
    ForInit(formula->uf.of.sub3f, graph);
}

/***************************************************************************
*
* Function:  ForCBPath
* Purpose:   This routine clears a bpath formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:
* See also:
*
****************************************************************************
*/
FS3(void, ForCBPath,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    if (!LstEmpty(formula->mf.bf.front))
	{
	    LstDestroy(formula->mf.bf.front, TRUE);
	    formula->mf.bf.front = LstCreate(&pthTools);
	}
    ForClear(formula->uf.of.sub3f, graph, world);
}

/***************************************************************************
*
* Function:  ForEBPath
* Purpose:   This routine evaluates a bpath formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:   a path from the evaluation node or NULL
* See also:
*
****************************************************************************
*/
FS3(PATH *, ForEBPath,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    PATH *path, *next, *aux;
    unsigned long awCount, awAddr = 0L, succ = 0L;

    if (LstEmpty(formula->mf.bf.front)) 
	{
	    (void) LstInsLast(formula->mf.bf.front, (char *) PthCreate(world));
	    formula->mf.bf.phase = FP_NOLOOP;
	    awCount = GphAwCount(graph, world);
	    if (awCount == (unsigned long) -1L)
		awCount = 0L;
	    if ((awCount != 0L) &&
		(aux = ForEval(formula->uf.of.sub1f, graph, world))) 
		{
		    ForClear(formula->uf.of.sub1f, graph, world);
		    PthDiscard(aux);
		    path = PthCreate(world);
		    GphFlFirst(graph, world, &awAddr, &succ); 
		    PthInsLast(path, awAddr, succ);
		    (void) LstInsLast(formula->mf.bf.front, (char *) path);
		}
	}
    while (continueEvaluation && (errorFlag == NO_ERROR)) 
	{
	    path = (PATH *) LstFirst(formula->mf.bf.front);
	    switch (formula->mf.bf.phase) 
		{
		case FP_NOLOOP:
		    if ((next =
			 ForEval(formula->uf.of.sub3f, graph, PthEnd(path))))
			return(PthCatenate(PthCopy(path), next));
		    break;
		case FP_ISLOOP:
		    if ((aux =
			 ForEval(formula->uf.of.sub2f, graph, PthEnd(path))))
			{
			    ForClear(formula->uf.of.sub2f, graph, PthEnd(path));
			    PthDiscard(aux);
			    formula->mf.bf.phase = FP_LOOPDONE;
			    return(PthCopy(path));
			}
		    break;
		case FP_LOOPDONE:
		    break;
		}
	    if (PthLast(path)
		&&
		(FreeExpansion()
		 ?
		 (GphAwOrdinal(graph, StpGAwAddr(PthLast(path))) != 0L)
		 :
		 (GphAwOrdinal(graph, StpGAwAddr(PthLast(path))) + 1L
		  < GphAwCount(graph, StpGWorld(PthLast(path))))))
		GphFlNext(graph, StpAAwAddr(PthLast(path)), PthAEnd(path));
	    else 
		{
		    (void) LstDelFirst(formula->mf.bf.front, TRUE);
		    if (LstEmpty(formula->mf.bf.front)) 
			return(NULL);
		    path = (PATH *) LstFirst(formula->mf.bf.front);
		}
	    if (PthLoop(path)) 
		formula->mf.bf.phase = FP_ISLOOP;
	    else 
		{
		    formula->mf.bf.phase = FP_NOLOOP;
		    awCount = GphAwCount(graph, PthEnd(path));
		    if (awCount == (unsigned long) -1L)
			awCount = 0L;
		    if ((awCount != 0L) &&
			(aux = ForEval(formula->uf.of.sub1f,
				       graph, PthEnd(path))))
			{
			    ForClear(formula->uf.of.sub1f, graph, PthEnd(path));
			    PthDiscard(aux);
			    next = PthCopy(path);
			    GphFlFirst(graph, PthEnd(next), &awAddr, &succ);
			    PthInsLast(next, awAddr, succ);
			    (void) LstInsLast(formula->mf.bf.front,
					      (char *) next);
			}
		}
	}
    return(NULL);
}

/***************************************************************************
*
* Function:  ForQBPath
* Purpose:   This routine closes a bpath formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForQBPath,
    FORMULA *, formula,
    GRAPH *, graph)
{
    LstDestroy(formula->mf.bf.front, TRUE);
    ForClose(formula->uf.of.sub1f, graph);
    ForClose(formula->uf.of.sub2f, graph);
    ForClose(formula->uf.of.sub3f, graph);
}

/***************************************************************************
*
* Function:  ForIDPath
* Purpose:   This routine initializes a dpath formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForIDPath,
    FORMULA *, formula,
    GRAPH *, graph)
{
    formula->mf.df.path = (PATH *) NULL;
    formula->mf.df.inPath = BstCons(GphNdCount(graph));
    BstClear(formula->mf.df.inPath);
    ForInit(formula->uf.of.sub1f, graph);
    ForInit(formula->uf.of.sub2f, graph);
    ForInit(formula->uf.of.sub3f, graph);
}

/***************************************************************************
*
* Function:  ForCDPath
* Purpose:   This routine clears a dpath formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:
* See also:
*
****************************************************************************
*/
FS3(void, ForCDPath,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    if (formula->mf.df.path)
	PthDiscard(formula->mf.df.path);
    BstClear(formula->mf.df.inPath);
    ForClear(formula->uf.of.sub3f, graph, world);
}

/***************************************************************************
*
* Function:  ForEDPath
* Purpose:   This routine evaluates a dpath formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:   a path from the evaluation node or NULL
* See also:
*
****************************************************************************
*/
FS3(PATH *, ForEDPath,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    PATH *path, *next, *aux;
    unsigned long awCount, awAddr = 0L, succ = 0L;

    if (!formula->mf.df.path) 
	{
	    formula->mf.df.path = PthCreate(world);
	    formula->mf.df.phase = FP_FORWARD;
	}
    path = formula->mf.df.path;
    while (continueEvaluation && (errorFlag == NO_ERROR))
	{
	    switch(formula->mf.df.phase)
		{
		case FP_FORWARD:
		    if ((next =
			 ForEval(formula->uf.of.sub3f, graph, PthEnd(path))))
			return(PthCatenate(PthCopy(path), next));
		    awCount = GphAwCount(graph, PthEnd(path));
		    if (awCount == (unsigned long) -1L) 
			awCount = 0L;
		    if ((awCount != 0L) && 
			(aux = ForEval(formula->uf.of.sub1f,
				       graph, PthEnd(path))))
			{
			    ForClear(formula->uf.of.sub1f, graph, PthEnd(path));
			    PthDiscard(aux);
			    BstInsert(formula->mf.df.inPath,
				      GphOrdinal(graph, PthEnd(path)));
			    GphFlFirst(graph, PthEnd(path), &awAddr, &succ);
			    PthInsLast(path, awAddr, succ);
			} 
		    else
			formula->mf.df.phase = FP_BACKTRACK;
		    break;
		case FP_ISLOOP:
		    formula->mf.df.phase = FP_BACKTRACK;
		    if ((aux =
			 ForEval(formula->uf.of.sub2f, graph, PthEnd(path))))
			{
			    ForClear(formula->uf.of.sub2f, graph, PthEnd(path));
			    PthDiscard(aux);
			    return(PthCopy(path));
			}
		    break;
		case FP_BACKTRACK:
		    break;
		}
	    if (formula->mf.df.phase == FP_BACKTRACK)
		{
		    while (PthLast(path)
			   &&
			   (FreeExpansion()
			    ?
			    (GphAwOrdinal(graph, StpGAwAddr(PthLast(path)))
			     == 0L)
			    :
			    (GphAwOrdinal(graph, StpGAwAddr(PthLast(path)))
			     + 1L
			     >= GphAwCount(graph, StpGWorld(PthLast(path))))))
			{
			    PthDelLast(path);
			    BstDelete(formula->mf.df.inPath,
				      GphOrdinal(graph, PthEnd(path)));
			}
		    if (!PthLast(path)) 
			{
			    PthDiscard(path);
			    formula->mf.df.path = (PATH *) NULL;
			    return(NULL);
			}
		    GphFlNext(graph, StpAAwAddr(PthLast(path)), PthAEnd(path));
		}
	    if (BstIn(formula->mf.df.inPath, GphOrdinal(graph, PthEnd(path))))
		formula->mf.df.phase = FP_ISLOOP;
	    else 
		formula->mf.df.phase = FP_FORWARD;
	}
    return(NULL);
}

/***************************************************************************
*
* Function:  ForQDPath
* Purpose:   This routine closes a dpath formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForQDPath,
    FORMULA *, formula,
    GRAPH *, graph)
{
    if (formula->mf.df.path)
	PthDiscard(formula->mf.df.path);
    else
	formula->mf.df.path = (PATH *) NULL;
    BstDiscard(formula->mf.df.inPath);
    ForClose(formula->uf.of.sub1f, graph);
    ForClose(formula->uf.of.sub2f, graph);
    ForClose(formula->uf.of.sub3f, graph);
}

/***************************************************************************
*
* Function:  ForIBSpan
* Purpose:   This routine initializes a bspan formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForIBSpan,
    FORMULA *, formula,
    GRAPH *, graph)
{
    formula->mf.pf.path = (PATH *) NULL;
    ForInit(formula->uf.of.sub1f, graph);
    ForInit(formula->uf.of.sub2f, graph);
}

/***************************************************************************
*
* Function:  ForCBSpan
* Purpose:   This routine clears a bspan formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:
* See also:
*
****************************************************************************
*/
FS3(void, ForCBSpan,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    if (((PATH *) NULL) != formula->mf.pf.path)
	{
	    PthDiscard(formula->mf.pf.path);
	    formula->mf.pf.path = (PATH *) NULL;
	}
    ForClear(formula->uf.of.sub2f, graph, world);
}

FS3(BOOL, Along,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    PATH *path;

    path = ForEval(formula->uf.of.sub1f, graph, world);
    if (errorFlag != NO_ERROR)
	continueEvaluation = FALSE;
    if (((PATH *) NULL) == path)
	return(FALSE);
    PthDiscard(path);
    ForClear(formula->uf.of.sub1f, graph, world);
    return(TRUE);
}

FS3(PATH *, Admit,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    PATH *path;

    path = ForEval(formula->uf.of.sub2f, graph, world);
    if (errorFlag != NO_ERROR)
	continueEvaluation = FALSE;
    else if (continueEvaluation)
	return(path);
    if (((PATH *) NULL) != path)
	PthDiscard(path);
    return((PATH *) NULL);
}

/***************************************************************************
*
* Function:  ForEBSpan
* Purpose:   This routine evaluates a bspan formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:   a path from the evaluation node or NULL
* See also:
*
****************************************************************************
*/
FS3(PATH *, ForEBSpan,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    if ((errorFlag != NO_ERROR) || !continueEvaluation)
	{
	    if (((PATH *) NULL) != formula->mf.pf.path)
		{
		    PthDiscard(formula->mf.pf.path);
		    formula->mf.pf.path = (PATH *) NULL;
		}
	    return((PATH *) NULL);
	}
    if (((PATH *) NULL) == formula->mf.pf.path)
	formula->mf.pf.present = world;
    return(GphShortest(graph, &(formula->mf.pf.present),
		       &(formula->mf.pf.head), &(formula->mf.pf.tail),
		       &(formula->mf.pf.path), &continueEvaluation,
		       (VOID *) formula, Along, Admit));
}

/***************************************************************************
*
* Function:  ForQBSpan
* Purpose:   This routine closes a bspan formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForQBSpan,
    FORMULA *, formula,
    GRAPH *, graph)
{
    if (((PATH *) NULL) != formula->mf.pf.path)
	{
	    PthDiscard(formula->mf.pf.path);
	    formula->mf.pf.path = (PATH *) NULL;
	}
    ForClose(formula->uf.of.sub1f, graph);
    ForClose(formula->uf.of.sub2f, graph);
}

/***************************************************************************
*
* Function:  ForIDSpan
* Purpose:   This routine initializes a dspan formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForIDSpan,
    FORMULA *, formula,
    GRAPH *, graph)
{
    formula->mf.df.path = (PATH *) NULL;
    formula->mf.df.used = BstCons(GphNdCount(graph));
    BstClear(formula->mf.df.used);
    formula->mf.df.inPath = BstCons(GphNdCount(graph));
    BstClear(formula->mf.df.inPath);
    ForInit(formula->uf.of.sub1f, graph);
    ForInit(formula->uf.of.sub2f, graph);
    ForInit(formula->uf.of.sub3f, graph);
}

/***************************************************************************
*
* Function:  ForCDSpan
* Purpose:   This routine clears a dspan formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS3(void, ForCDSpan,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    if (formula->mf.df.path) 
	PthDiscard(formula->mf.df.path);
    BstClear(formula->mf.df.used);
    BstClear(formula->mf.df.inPath);
    ForClear(formula->uf.of.sub3f, graph, world);
}

/***************************************************************************
*
* Function:  ForEDSpan
* Purpose:   This routine evaluates a dspan formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:   a path from the evaluation node or NULL
* See also:
*
****************************************************************************
*/
FS3(PATH *, ForEDSpan,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    PATH *path, *next, *aux;
    unsigned long awCount, awAddr = 0L, succ = 0L;

    if (!formula->mf.df.path) 
	{
	    formula->mf.df.path = PthCreate(world);
	    formula->mf.df.phase = FP_FORWARD;
	}
    path = formula->mf.df.path;
    while (continueEvaluation && (errorFlag == NO_ERROR))
	{
	    switch(formula->mf.df.phase)
		{
		case FP_FORWARD:
		    if ((next =
			 ForEval(formula->uf.of.sub3f, graph, PthEnd(path))))
			return(PthCatenate(PthCopy(path), next));
		    BstInsert(formula->mf.df.used,
			      GphOrdinal(graph, PthEnd(path)));
		    awCount = GphAwCount(graph, PthEnd(path));
		    if (awCount == (unsigned long) -1L) 
			awCount = 0L;
		    if ((awCount != 0L) && 
			(aux = ForEval(formula->uf.of.sub1f,
				       graph, PthEnd(path))))
			{
			    ForClear(formula->uf.of.sub1f, graph, PthEnd(path));
			    PthDiscard(aux);
			    BstInsert(formula->mf.df.inPath,
				      GphOrdinal(graph, PthEnd(path)));
			    GphFlFirst(graph, PthEnd(path), &awAddr, &succ);
			    PthInsLast(path, awAddr, succ);
			} 
		    else
			formula->mf.df.phase = FP_BACKTRACK;
		    break;
		case FP_ISLOOP:
		    formula->mf.df.phase = FP_BACKTRACK;
		    if ((aux =
			 ForEval(formula->uf.of.sub2f, graph, PthEnd(path))))
			{
			    ForClear(formula->uf.of.sub2f, graph, PthEnd(path));
			    PthDiscard(aux);
			    return(PthCopy(path));
			}
		    break;
		case FP_BACKTRACK:
		    break;
		}
	    if (formula->mf.df.phase == FP_BACKTRACK)
		{
		    while (PthLast(path)
			   &&
			   (FreeExpansion()
			    ?
			    (GphAwOrdinal(graph, StpGAwAddr(PthLast(path)))
			     == 0L)
			    :
			    (GphAwOrdinal(graph, StpGAwAddr(PthLast(path)))
			     + 1L
			     >= GphAwCount(graph, StpGWorld(PthLast(path))))))
			{
			    PthDelLast(path);
			    BstDelete(formula->mf.df.inPath,
				      GphOrdinal(graph, PthEnd(path)));
			}
		    if (!PthLast(path)) 
			{
			    PthDiscard(path);
			    formula->mf.df.path = (PATH *) NULL;
			    BstClear(formula->mf.df.used);
			    return(NULL);
			}
		    GphFlNext(graph, StpAAwAddr(PthLast(path)), PthAEnd(path));
		}
	    if (BstIn(formula->mf.df.used, GphOrdinal(graph, PthEnd(path))))
		{
		    if (BstIn(formula->mf.df.inPath,
			      GphOrdinal(graph, PthEnd(path))))
			formula->mf.df.phase = FP_ISLOOP;
		    else
			formula->mf.df.phase = FP_BACKTRACK;
		}
	    else 
		formula->mf.df.phase = FP_FORWARD;
	}
    return(NULL);
}

/***************************************************************************
*
* Function:  ForQDSpan
* Purpose:   This routine closes a dspan formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForQDSpan,
    FORMULA *, formula,
    GRAPH *, graph)
{
    if (formula->mf.df.path)
	PthDiscard(formula->mf.df.path);
    else
	formula->mf.df.path = (PATH *) NULL;
    BstDiscard(formula->mf.df.used);
    BstDiscard(formula->mf.df.inPath);
    ForClose(formula->uf.of.sub1f, graph);
    ForClose(formula->uf.of.sub2f, graph);
    ForClose(formula->uf.of.sub3f, graph);
}

/***************************************************************************
*
* Function:  ForEAt
* Purpose:   This routine evaluates an "at" formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:   the evaluation node "path" or NULL
* See also:
*
****************************************************************************
*/
FS3(PATH *, ForEAt,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    unsigned long state;

    if (formula->mf.visited) 
	{
	    formula->mf.visited = FALSE;
	    return(NULL);
	}
    FogSGraph(graph);
    FogSWorld(world);
    state = (unsigned long) XprEval(formula->uf.expression);
    if ((errorFlag != NO_ERROR) || !continueEvaluation)
	return(NULL);
    if (state >= GphNdCount(graph))
	{
	    (void) fprintf(_err,
			   "Node %lu does not exist\n", state);
	    (void) fflush(_err);
	    errorFlag = FOR_ERROR;
	    return(NULL);
	}
    if (GphId(graph, state) == world) 
	{
	    formula->mf.visited = TRUE;
	    return(PthCreate(world));
	}
    return(NULL);
}

/***************************************************************************
*
* Function:  ForEInset
* Purpose:   This routine evaluates an "in-the-set" formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:   the evaluation node "path" or NULL
* See also:
*
****************************************************************************
*/
FS3(PATH *, ForEInSet,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    if (formula->mf.visited) 
	{
	    formula->mf.visited = FALSE;
	    return(NULL);
	}
    FogSGraph(graph);
    FogSWorld(world);
    if (SopIn(formula->uf.setOper, world))
	{
	    formula->mf.visited = TRUE;
	    return(PthCreate(world));
	}
    return(NULL);
}

/***************************************************************************
*
* Function:  ForELocal
* Purpose:   This routine evaluates an atomic formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:   the evaluation node "path" or NULL
* See also:
*
****************************************************************************
*/
FS3(PATH *, ForELocal,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    if (formula->mf.visited) 
	{
	    formula->mf.visited = FALSE;
	    return(NULL);
	}
    FogSGraph(graph);
    FogSWorld(world);
    if (XprEval(formula->uf.expression)) 
	{
	    formula->mf.visited = TRUE;
	    return(PthCreate(world));
	}
    return(NULL);
}

/***************************************************************************
*
* Function:  ForIAnd
* Purpose:   This routine initializes a conjunction formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForIAnd,
    FORMULA *, formula,
    GRAPH *, graph)
{
    formula->mf.visited = FALSE;
    ForInit(formula->uf.of.sub1f, graph);
    ForInit(formula->uf.of.sub2f, graph);
}

/***************************************************************************
*
* Function:  ForCAnd
* Purpose:   This routine clears a conjunction formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:
* See also:
*
****************************************************************************
*/
FS3(void, ForCAnd,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    formula->mf.visited = FALSE;
    ForClear(formula->uf.of.sub1f, graph, world);
    ForClear(formula->uf.of.sub2f, graph, world);
}

/***************************************************************************
*
* Function:  ForEAnd
* Purpose:   This routine evaluates a conjunction formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:   a path from the evaluation node or NULL
* See also:
*
****************************************************************************
*/
FS3(PATH *, ForEAnd,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    PATH *path;

    if (formula->mf.visited) 
	{
	    if ((path = ForEval(formula->uf.of.sub2f, graph, world)) )
		return(path);
	    formula->mf.visited = FALSE;
	    ForClear(formula->uf.of.sub2f, graph, world);
	    return(NULL);
	}
    if ((path = ForEval(formula->uf.of.sub1f, graph, world)) )
	{
	    PthDiscard(path);
	    ForClear(formula->uf.of.sub1f, graph, world);
	    if ((path = ForEval(formula->uf.of.sub2f, graph, world)) )
		{
		    formula->mf.visited = TRUE;
		    return(path);
		}
	}
    return(NULL);
}

/***************************************************************************
*
* Function:  ForQAnd
* Purpose:   This routine closes a conjunction formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForQAnd,
    FORMULA *, formula,
    GRAPH *, graph)
{
    ForClose(formula->uf.of.sub1f, graph);
    ForClose(formula->uf.of.sub2f, graph);
}

/***************************************************************************
*
* Function:  ForIOr
* Purpose:   This routine initializes a disjunction formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForIOr,
    FORMULA *, formula,
    GRAPH *, graph)
{
    formula->mf.sf.count = 0;
    formula->mf.sf.paths = BalTrCreate(&pthBalTools);
    ForInit(formula->uf.of.sub1f, graph);
    ForInit(formula->uf.of.sub2f, graph);
}

/***************************************************************************
*
* Function:  ForCOr
* Purpose:   This routine clears a disjunction formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:
* See also:
*
****************************************************************************
*/
FS3(void, ForCOr,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    formula->mf.sf.count = 0;
    if (BalTrGElCount(formula->mf.sf.paths) != 0L)
	{
	    BalTrDestroy(formula->mf.sf.paths, TRUE);
	    formula->mf.sf.paths = BalTrCreate(&pthBalTools);
	}
    ForClear(formula->uf.of.sub1f, graph, world);
    ForClear(formula->uf.of.sub2f, graph, world);
}

/***************************************************************************
*
* Function:  ForEOr
* Purpose:   This routine evaluates a disjucntion formula.
* Arguments: formula - the formula
*            graph - the graph
*            world - address of evaluation node in gph-file
* Returns:   a path from the evaluation node or NULL
* See also:
*
****************************************************************************
*/
FS3(PATH *, ForEOr,
    FORMULA *, formula,
    GRAPH *, graph,
    unsigned long, world)
{
    PATH *path, *aux;

    switch (formula->mf.sf.count) 
	{
	case 0:
	    if ((path = ForEval(formula->uf.of.sub1f, graph, world)))
		{
		    aux = PthCopy(path);
		    if (BalTrInsert(formula->mf.sf.paths, (char *) aux,
				    (char *) aux, TRUE))
			PthDiscard(aux);
		    return(path);
		}
	    ForClear(formula->uf.of.sub1f, graph, world);
	    formula->mf.sf.count = 1;
	    /* intentional abuse of switch */ 
	case 1:
	    while (continueEvaluation && (errorFlag == NO_ERROR) &&
		   (path = ForEval(formula->uf.of.sub2f, graph, world)))
		{
		    if (BalTrElement(formula->mf.sf.paths, (char *) path))
			PthDiscard(path);
		    else
			return(path);
		}
	    ForClear(formula->uf.of.sub2f, graph, world);
	    /* intentional abuse of switch */ 
	default:
	    formula->mf.sf.count = 0;
	    return(NULL);
	}
}

/***************************************************************************
*
* Function:  ForQOr
* Purpose:   This routine closes a disjunction formula.
* Arguments: formula - the formula
*            graph - the graph
* Returns:
* See also:
*
****************************************************************************
*/
FS2(void, ForQOr,
    FORMULA *, formula,
    GRAPH *, graph)
{
    BalTrDestroy(formula->mf.sf.paths, TRUE);
    ForClose(formula->uf.of.sub1f, graph);
    ForClose(formula->uf.of.sub2f, graph);
}

#define XprRTravar MexRPlace
#define XprRPreClass MexRPlace
#define XprRFaithful MexRPlace

/***************************************************************************
*
* Function:  SetExMex
* Purpose:   This routine sets the probe specific entries in
*            the expression evaluation and reduction function tables.
* Arguments:
* Returns:
* See also:  SetExMex (prpexpr.c)
*
****************************************************************************
*/
FE0(void, SetExMex)
{
    SetMexEval(MO_PLACE, MexEPlace);
    SetMexReduce(MO_TRANGE, MexRTrange);
    SetXprEval(EO_FIELD, XprEField);
    SetXprEval(EO_DIV, XprEDiv);
    SetXprEval(EO_REM, XprERem);
    SetXprEval(EO_TRAVAR, XprETravar);
    SetXprEval(EO_PRECLASS, XprEPreClass);
    SetXprEval(EO_FAITHFUL, XprEFaithful);
    SetXprReduce(EO_DIV, XprRDiv);
    SetXprReduce(EO_REM, XprRRem);
    SetXprReduce(EO_TRAVAR, XprRTravar);
    SetXprReduce(EO_PRECLASS, XprRPreClass);
    SetXprReduce(EO_FAITHFUL, XprRFaithful);
}

#define ForIAt     ForITrue
#define ForIInSet  ForITrue
#define ForILocal  ForITrue
#define ForCFalse  ForIFalse
#define ForCTrue   ForITrue
#define ForCAt     ForITrue
#define ForCInSet  ForITrue
#define ForCLocal  ForITrue
#define ForQFalse  ForIFalse
#define ForQTrue   ForIFalse
#define ForQAt     ForIFalse
#define ForQInSet  ForIFalse
#define ForQLocal  ForIFalse
#define ForICTLFalse ForIFalse
#define ForCCTLFalse ForCFalse
#define ForECTLFalse ForEFalse
#define ForQCTLFalse ForQFalse
#define ForICTLTrue ForIFalse
#define ForCCTLTrue ForCFalse
#define ForECTLTrue ForEFalse
#define ForQCTLTrue ForQFalse
#define ForICTLNot ForIFalse
#define ForCCTLNot ForCFalse
#define ForECTLNot ForEFalse
#define ForQCTLNot ForQFalse
#define ForICTLEF ForIFalse
#define ForCCTLEF ForCFalse
#define ForECTLEF ForEFalse
#define ForQCTLEF ForQFalse
#define ForICTLEG ForIFalse
#define ForCCTLEG ForCFalse
#define ForECTLEG ForEFalse
#define ForQCTLEG ForQFalse
#define ForICTLEU ForIFalse
#define ForCCTLEU ForCFalse
#define ForECTLEU ForEFalse
#define ForQCTLEU ForQFalse
#define ForICTLEX ForIFalse
#define ForCCTLEX ForCFalse
#define ForECTLEX ForEFalse
#define ForQCTLEX ForQFalse
#define ForICTLAF ForIFalse
#define ForCCTLAF ForCFalse
#define ForECTLAF ForEFalse
#define ForQCTLAF ForQFalse
#define ForICTLAG ForIFalse
#define ForCCTLAG ForCFalse
#define ForECTLAG ForEFalse
#define ForQCTLAG ForQFalse
#define ForICTLAU ForIFalse
#define ForCCTLAU ForCFalse
#define ForECTLAU ForEFalse
#define ForQCTLAU ForQFalse
#define ForICTLAX ForIFalse
#define ForCCTLAX ForCFalse
#define ForECTLAX ForEFalse
#define ForQCTLAX ForQFalse
#define ForICTLLocal ForIFalse
#define ForCCTLLocal ForCFalse
#define ForECTLLocal ForEFalse
#define ForQCTLLocal ForQFalse
#define ForICTLAnd ForIFalse
#define ForCCTLAnd ForCFalse
#define ForECTLAnd ForEFalse
#define ForQCTLAnd ForQFalse
#define ForICTLOr ForIFalse
#define ForCCTLOr ForCFalse
#define ForECTLOr ForEFalse
#define ForQCTLOr ForQFalse

FOROPENTRY forOpTable[SZ_FO] = 
{
    { FK_NULLARY,    ForIFalse,  ForCFalse,  ForEFalse,  ForQFalse },
    { FK_NULLARY,    ForITrue,   ForCTrue,   (PTHF) NULL, ForQTrue },
    { FK_UNARY,      ForINot,    ForCNot,    ForENot,    ForQNot },
    { FK_UNARY,      ForINode,   ForCNode,   ForENode,   ForQNode },
    { FK_UNARY,      ForIStep,   ForCStep,   ForEStep,   ForQStep },
    { FK_FIRE,       ForIFire,   ForCFire,   ForEFire,   ForQFire },
    { FK_TRINARY,    ForIBPath,  ForCBPath,  ForEBPath,  ForQBPath },
    { FK_TRINARY,    ForIDPath,  ForCDPath,  ForEDPath,  ForQDPath },
    { FK_BINARY,     ForIBSpan,  ForCBSpan,  ForEBSpan,  ForQBSpan },
    { FK_TRINARY,    ForIDSpan,  ForCDSpan,  ForEDSpan,  ForQDSpan },
    { FK_EXPRESSION, ForIAt,     ForCAt,     ForEAt,     ForQAt },
    { FK_SET,        ForIInSet,  ForCInSet,  ForEInSet,  ForQInSet },
    { FK_EXPRESSION, ForILocal,  ForCLocal,  ForELocal,  ForQLocal },
    { FK_BINARY,     ForIAnd,    ForCAnd,    ForEAnd,    ForQAnd },
    { FK_BINARY,     ForIOr,     ForCOr,     ForEOr,     ForQOr },
    { FK_NULLARY, ForICTLFalse, ForCCTLFalse, ForECTLFalse, ForQCTLFalse },
    { FK_NULLARY, ForICTLTrue, ForCCTLTrue, ForECTLTrue, ForQCTLTrue },
    { FK_UNARY, ForICTLNot, ForCCTLNot, ForECTLNot, ForQCTLNot },
    { FK_UNARY, ForICTLEF, ForCCTLEF, ForECTLEF, ForQCTLEF },
    { FK_UNARY, ForICTLEG, ForCCTLEG, ForECTLEG, ForQCTLEG },
    { FK_BINARY, ForICTLEU, ForCCTLEU, ForECTLEU, ForQCTLEU },
    { FK_UNARY, ForICTLEX, ForCCTLEX, ForECTLEX, ForQCTLEX },
    { FK_UNARY, ForICTLAF, ForCCTLAF, ForECTLAF, ForQCTLAF },
    { FK_UNARY, ForICTLAG, ForCCTLAG, ForECTLAG, ForQCTLAG },
    { FK_BINARY, ForICTLAU, ForCCTLAU, ForECTLAU, ForQCTLAU },
    { FK_UNARY, ForICTLAX, ForCCTLAX, ForECTLAX, ForQCTLAX },
    { FK_EXPRESSION, ForICTLLocal, ForCCTLLocal, ForECTLLocal, ForQCTLLocal },
    { FK_BINARY, ForICTLAnd, ForCCTLAnd, ForECTLAnd, ForQCTLAnd },
    { FK_BINARY, ForICTLOr, ForCCTLOr, ForECTLOr, ForQCTLOr }
};
