/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  gphutil.c
*
* Part of: PROD's programs araprod, "gphgen", probe, prpp, and strong
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     The routines in this module were designed to help the construction
*     of a reachability graph of a net, but they can well be used for
*     other purposes, too.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include <stdio.h>
#include <string.h>
#ifdef UNIX
#include <sys/types.h>
#else
#ifdef TNIL
#include <sys/types.h>
#else
#include <sys\types.h>
#endif
#endif
#include <stdlib.h>
#include <malloc.h>
#ifdef UNIX
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifndef S_IRWXU
#define S_IRWXU 0000700
#define S_IRUSR 0000400
#define S_IWUSR 0000200
#define S_IXUSR 0000100
#define S_IRWXG 0000070
#define S_IRGRP 0000040
#define S_IWGRP 0000020
#define S_IXGRP 0000010
#define S_IRWXO 0000007
#define S_IROTH 0000004
#define S_IWOTH 0000002
#define S_IXOTH 0000001
#endif
#endif
#ifndef SEEK_SET
#define SEEK_SET 0
#define SEEK_CUR 1
#define SEEK_END 2
#endif
#ifndef UNIX
#define Hmalloc(n) malloc((size_t) n)
#define Hfree(p) free(p)
#else
#ifdef HALLOC
#define Hmalloc(n) _halloc((long) (n), (size_t) 1)
#define Hfree(p) _hfree(p)
#else
#define Hmalloc(n) malloc((size_t) n)
#define Hfree(p) free(p)
#endif
#endif
#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "error.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "gphutil.h"

/***************************************************************************
*
* Function:  MapToMem
* Purpose:   This routine creates a table to be used instead of a binary
*            file.
* Arguments: name - the name of the file
*            bFile - the file
*            len - the size of the table
*            actLen - the current size of the file
* Returns:   a pointer to a structure needed for the table
* See also:  MapFlush, Unmap, MapSkAbs, MapReLng, MapWrLng, MapReUns,
*            MapWrUns, MapReData, MapWrData
*
****************************************************************************
*/
FE4(MMFILE *, MapToMem,
    char *, name,
    BINARYFILE, bFile,
    unsigned long, len,
    unsigned long, actLen)
{
    MMFILE *mmFile;

    if (len < actLen)
	{
	    (void) fprintf(_err, "\n%s is used directly\n", name);
	    (void) fflush(_err);
	    Deallocate((char *) name);
	    return(NULL);
	}
    mmFile = (MMFILE *) Allocate(sizeof(MMFILE));
    mmFile->len = len;
    mmFile->offset = 0;
    /* Do not replace this Hmalloc by Hallocate since size can be
     * arbitrarily great, and an unsuccesful Hallocate would abort
     * the program. */
    if (!(mmFile->start = (char *) Hmalloc(len)))
	{
	    (void) fprintf(_err, "\n%s is used directly\n", name);
	    (void) fflush(_err);
	    ClearErrno();
	    Deallocate((char *) name);
	    Deallocate((char *) mmFile);
	    return(NULL);
	}
    SeekAbs(bFile, (unsigned long) 0L);
    ReadChr(mmFile->start, actLen, bFile);
    mmFile->name = name;
    mmFile->bFile = bFile;
    mmFile->actLen = actLen;
    return(mmFile);
}

/***************************************************************************
*
* Function:  MapFlush
* Purpose:   This routine updates a binary file that has earlier been
*            replaced by a table.
* Arguments: mmFile - a pointer to a structure needed for the table
*                     and also pointing to the file
* Returns:
* See also:  MapToMem, Unmap, MapSkAbs, MapReLng, MapWrLng, MapReUns,
*            MapWrUns, MapReData, MapWrData
*
****************************************************************************
*/
FE1(void, MapFlush,
    MMFILE *, mmFile)
{
    SeekAbs(mmFile->bFile, (unsigned long) 0L);
    WriteChr(mmFile->start, mmFile->actLen, mmFile->bFile);
}

/***************************************************************************
*
* Function:  Unmap
* Purpose:   This routine deletes a table that was used to replace a binary
*            file.
* Arguments: mmFile - a pointer to a structure needed for the table
*                     and also pointing to the file
*            wr - if this is TRUE, the file is updated before the deletion
*                 of the table
* Returns:
* See also:  MapToMem, MapFlush, MapSkAbs, MapReLng, MapWrLng, MapReUns,
*            MapWrUns, MapReData, MapWrData
*
****************************************************************************
*/
FE2(void, Unmap,
    MMFILE *, mmFile,
    BOOL, wr)
{
    if (wr)
	MapFlush(mmFile);
    /* Do not replace this Hfree by Dehallocate since the area
     * has not been allocated by Hallocate. */
    Hfree(mmFile->start);
    Deallocate((char *) mmFile->name);
    Deallocate((char *) mmFile);
}

/***************************************************************************
*
* Function:  CreateUpdate
* Purpose:   This routine creates an empty file for writing and reading.
* Arguments: name - the name of the file
* Returns:   the file
* See also:  CreateWrite, OpenRead, OpenUpdate, OwnClose, OwnFlush,
*            SeekAbs, SeekEnd, ReadBytes, WriteBytes
*
****************************************************************************
*/
FE1(BINARYFILE, CreateUpdate,
    char *, name)
{
    register BINARYFILE file;

    file = (BINARYFILE) 
#ifdef UNIX
	open(name, (int) (O_RDWR | O_CREAT | O_TRUNC), (int)
	     (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH));
#else
    fopen(name, "w+b");
#endif
    if (NotBinaryFile(file))
	{
	    (void) fprintf(_err, 
			   "\nCannot open empty file %s for reading and writing\n",
			   name);
	    ErrnoOutput(_err);
	    ClearErrno();
	}
    return(file);
}

/***************************************************************************
*
* Function:  CreateWrite
* Purpose:   This routine creates an empty file for writing.
* Arguments: name - the name of the file
* Returns:   the file
* See also:  CreateUpdate, OpenRead, OpenUpdate, OwnClose, OwnFlush,
*            SeekAbs, SeekEnd, ReadBytes, WriteBytes
*
****************************************************************************
*/
FE1(BINARYFILE, CreateWrite,
    char *, name)
{
    register BINARYFILE file;

    file = (BINARYFILE)
#ifdef UNIX
	open(name, (int) (O_WRONLY | O_CREAT | O_TRUNC), (int)
	     (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH));
#else
    fopen(name, "wb");
#endif
    if (NotBinaryFile(file))
	{
	    (void) fprintf(_err, "\nCannot open empty file %s for writing\n",
			   name);
	    ErrnoOutput(_err);
	    ClearErrno();
	}
    return(file);
}

/***************************************************************************
*
* Function:  OpenRead
* Purpose:   This routine opens an existing file for reading.
* Arguments: name - the name of the file
* Returns:   the file
* See also:  CreateUpdate, CreateWrite, OpenUpdate, OwnClose, OwnFlush,
*            SeekAbs, SeekEnd, ReadBytes, WriteBytes
*
****************************************************************************
*/
FE2(BINARYFILE, OpenRead,
    char *, name,
    BOOL, strict)
{
    register BINARYFILE file;

    file = (BINARYFILE)
#ifdef UNIX
	open(name, (int) O_RDONLY, (int) 0);
#else
    fopen(name, "rb");
#endif
    if (NotBinaryFile(file))
	{
	    if (strict)
		{
		    (void) fprintf(_err, "\nCannot open file %s for reading\n",
				   name);
		    ErrnoOutput(_err);
		}
	    ClearErrno();
	}
    return(file);
}

/***************************************************************************
*
* Function:  OpenUpdate
* Purpose:   This routine opens an existing file for writing and reading.
* Arguments: name - the name of the file
* Returns:   the file
* See also:  CreateUpdate, CreateWrite, OpenRead, OwnClose, OwnFlush,
*            SeekAbs, SeekEnd, ReadBytes, WriteBytes
*
****************************************************************************
*/
FE1(BINARYFILE, OpenUpdate,
    char *, name)
{
    register BINARYFILE file;

    file = (BINARYFILE)
#ifdef UNIX
	open(name, (int) O_RDWR, (int) 0);
#else
    fopen(name, "r+b");
#endif
    if (NotBinaryFile(file))
	{
	    (void) fprintf(_err, "\nCannot open file %s for updating\n",
			   name);
	    ErrnoOutput(_err);
	    ClearErrno();
	}
    return(file);
}

/***************************************************************************
*
* Function:  OwnClose
* Purpose:   This routine closes a binary file.
* Arguments: file - the file
* Returns:
* See also:  CreateUpdate, CreateWrite, OpenRead, OpenUpdate, OwnFlush,
*            SeekAbs, SeekEnd, ReadBytes, WriteBytes
*
****************************************************************************
*/
FE1(void, OwnClose,
    BINARYFILE, file)
{
    register int status;

    if ((status =
#ifdef UNIX
	 close((int) file)
#else
	 fclose((FILE *) file)
#endif
	 ))
	{
	    (void) fprintf(_err, 
#ifdef UNIX
			   "\nOwnClose: close(%d) returned %d\n",
			   (int) file,
#else
			   "\nOwnClose: fclose(0x%lx) returned %d\n",
			   (unsigned long) file,
#endif
			   status);
	    ErrnoOutput(_err);
	    ClearErrno();
	}
}

#ifndef UNIX
/***************************************************************************
*
* Function:  OwnFlush
* Purpose:   This routine updates a binary file if the buffering write
*            operations of the standard input/output library have been
*            used for the file.
* Arguments: file - the file
* Returns:
* See also:  CreateUpdate, CreateWrite, OpenRead, OpenUpdate, OwnClose,
*            SeekAbs, SeekEnd, ReadBytes, WriteBytes
*
****************************************************************************
*/
FE1(void, OwnFlush,
    BINARYFILE, file)
{
    register int status;

    if ((status = fflush((FILE *) file)))
	{
	    (void) fprintf(_err, 
			   "\nOwnFlush: fflush(0x%lx) returned %d\n",
			   (unsigned long) file, status);
	    ErrnoOutput(_err);
	    ClearErrno();
	}
}
#endif

/***************************************************************************
*
* Function:  SeekAbs
* Purpose:   This routine moves a file cursor to a new position.
* Arguments: file - the file
*            pos - the new position
* Returns:
* See also:  CreateUpdate, CreateWrite, OpenRead, OpenUpdate, OwnClose,
*            OwnFlush, SeekEnd, ReadBytes, WriteBytes
*
****************************************************************************
*/
FE2(void, SeekAbs,
    BINARYFILE, file,
    unsigned long, pos)
{
#ifndef UNIX
    register int status;
#endif

    if (
#ifdef UNIX
	lseek((int) file, (off_t) pos, (int) SEEK_SET) == (off_t) -1
#else
	(status = fseek((FILE *) file, (long) pos, (int) SEEK_SET))
#endif
	)
	{
 	    (void) fprintf(_out, "\n");
	    (void) fflush(_out);
	    (void) fprintf(_err,
#ifdef UNIX
			   "\nSeekAbs: lseek(%d, %ld, %d) returned -1\n",
			   (int) file, (long) pos, (int) SEEK_SET
#else
			   "\nSeekAbs: fseek(0x%lx, %ld, %d) returned %d\n",
			   (unsigned long) file, (long) pos,
			   (int) SEEK_SET, status
#endif
			   );
	    ErrnoOutput(_err);
	    abort();
	}
}

/***************************************************************************
*
* Function:  SeekEnd
* Purpose:   This routine moves a file cursor to the end of the file
*            and returns the new position. A RECEPTION OF A SIGNAL
*            DURING A WRITING OPERATION CAN MAKE THIS FUNCTION BEHAVE
*            UNEXPECTEDLY. (SeekAbs is more stable.)
* Arguments: file - the file
* Returns:   the new position, i.e. the current size of the file
* See also:  CreateUpdate, CreateWrite, OpenRead, OpenUpdate, OwnClose,
*            OwnFlush, SeekAbs, ReadBytes, WriteBytes
*
****************************************************************************
*/
FE1(unsigned long, SeekEnd,
    BINARYFILE, file)
{
    register unsigned long pos;
#ifndef UNIX
    register int status;
#endif

    if (
#ifdef UNIX
	(pos = (unsigned long) lseek((int) file, (off_t) 0L, (int) SEEK_END))
	== (unsigned long) -1L
#else
	(status = fseek((FILE *) file, (long) 0L, (int) SEEK_END))
#endif
	)
	{
 	    (void) fprintf(_out, "\n");
	    (void) fflush(_out);
	    (void) fprintf(_err,
#ifdef UNIX
			   "\nSeekAbs: lseek(%d, 0L, %d) returned -1\n",
			   (int) file, (int) SEEK_END
#else
			   "\nSeekAbs: fseek(0x%lx, 0L, %d) returned %d\n",
			   (unsigned long) file, (int) SEEK_END, status
#endif
			   );
	    ErrnoOutput(_err);
	    abort();
	}
#ifndef UNIX
    if ((pos = (unsigned long) ftell((FILE *) file)) == (unsigned long) -1L)
	{
 	    (void) fprintf(_out, "\n");
	    (void) fflush(_out);
	    (void) fprintf(_err,
			   "\nSeekAbs: ftell(0x%lx) returned -1\n",
			   (unsigned long) file);
	    ErrnoOutput(_err);
	    abort();
	}
#endif
    return(pos);
}

/***************************************************************************
*
* Function:  ReadBytes
* Purpose:   This routine reads a sequence of bytes from a binary file.
* Arguments: buf - the read stuff is put here
*            cnt - the number of bytes to be read
*            file - the file
* Returns:
* See also:  CreateUpdate, CreateWrite, OpenRead, OpenUpdate, OwnClose,
*            OwnFlush, SeekAbs, SeekEnd, WriteBytes
*
****************************************************************************
*/
FE3(void, ReadBytes,
    char *, buf,
    unsigned long, cnt,
    BINARYFILE, file)
{
    register char *p;
    register unsigned long readcnt, rest, totLeft;

    for (p = buf, totLeft = cnt; 0L != totLeft;)
	{
	    if (((unsigned long) BUFSIZ) < totLeft)
		rest = (unsigned long) BUFSIZ;
	    else
		rest = totLeft;
	    totLeft -= rest;
	    do
		{
		    readcnt = (unsigned long)
#ifdef UNIX
			read((int) file, p, (size_t) rest);
#else
		    fread(p, (size_t) 1, (size_t) rest, (FILE *) file);
#endif
		    if ((0L == readcnt) || (readcnt > rest))
			{
			    (void) fprintf(_out, "\n");
			    (void) fflush(_out);
			    (void) fprintf(_err, 
#ifdef UNIX
					   "\nReadBytes: read(%d, 0x%lx, %lu) returned %lu\n",
					   (int) file, (unsigned long) p, rest,
#else
					   "\nReadIt: fread(0x%lx, 1, %lu, 0x%lx) returned %lu\n",
					   (unsigned long) p, rest,
					   (unsigned long) file,
#endif
					   readcnt);
			    ErrnoOutput(_err);
			    abort();
			}
		    p += readcnt;
		    rest -= readcnt;
		} while (0L != rest);
	}
}

/***************************************************************************
*
* Function:  WriteBytes
* Purpose:   This routine writes a sequence of bytes to a binary file.
* Arguments: buf - the stuff to be written is taken from here
*            cnt - the number of bytes to be written
*            file - the file
* Returns:
* See also:  CreateUpdate, CreateWrite, OpenRead, OpenUpdate, OwnClose,
*            OwnFlush, SeekAbs, SeekEnd, ReadBytes
*
****************************************************************************
*/
FE3(void, WriteBytes,
    char *, buf,
    unsigned long, cnt,
    BINARYFILE, file)
{
    register char *p;
    register unsigned long writecnt, rest, totLeft;

    for (p = buf, totLeft = cnt; 0L != totLeft;)
	{
	    if (((unsigned long) BUFSIZ) < totLeft)
		rest = (unsigned long) BUFSIZ;
	    else
		rest = totLeft;
	    totLeft -= rest;
	    do
		{
		    writecnt = (unsigned long)
#ifdef UNIX
			write((int) file, p, (size_t) rest);
#else
		    fwrite(p, (size_t) 1, (size_t) rest, (FILE *) file);
#endif
		    if ((0L == writecnt) || (writecnt > rest))
			{
			    (void) fprintf(_out, "\n");
			    (void) fflush(_out);
			    (void) fprintf(_err, 
#ifdef UNIX
					   "\nWriteBytes: write(%d, 0x%lx, %lu) returned %lu\n",
					   (int) file, (unsigned long) p, rest,
#else
					   "\nWriteIt: fwrite(0x%lx, 1, %lu, 0x%lx) returned %lu\n",
					   (unsigned long) p, rest,
					   (unsigned long) file,
#endif
					   writecnt);
			    ErrnoOutput(_err);
			    abort();
			}
		    p += writecnt;
		    rest -= writecnt;
		} while (0L != rest);
	}
}

/***************************************************************************
*
* Function:  MapSkAbs
* Purpose:   This routine is like SeekAbs except that a table is used
*            instead of a file.
* Arguments: mmFile - a pointer to a structure needed for the table
*            pos - the new position of the cursor corresponding to a file
*                  cursor
* Returns:
* See also:  MapToMem, MapFlush, Unmap, MapReLng, MapWrLng, MapReUns,
*            MapWrUns, MapReData, MapWrData, SeekAbs
*
****************************************************************************
*/
FE2(void, MapSkAbs,
    MMFILE *, mmFile,
    unsigned long, pos)
{
    mmFile->offset = pos;
}

/***************************************************************************
*
* Function:  MapReBytes
* Purpose:   This routine is like ReadBytes except that a table is used
*            instead of a file.
* Arguments: buf - the read stuff is put here
*            cnt - the number of bytess to be read
*            mmFile - a pointer to a structure needed for the table
* Returns:
* See also:  MapToMem, MapFlush, Unmap, MapSkAbs, MapWrBytes, ReadBytes
*
****************************************************************************
*/
FE3(void, MapReBytes,
    char *, buf,
    unsigned long, cnt,
    MMFILE *, mmFile)
{
    register char *p, *q;
    register unsigned long i;

    p = buf;
    q = mmFile->start + mmFile->offset;
    i = cnt;
    mmFile->offset += i;
    if (mmFile->actLen < mmFile->offset)
	{
	    (void) fprintf(_out, "\n");
	    (void) fflush(_out);
	    (void) fprintf(_err,
			   "\nMapReBytes: actLen (%lu) < offset (%lu)\n",
			   mmFile->actLen, mmFile->offset);
	    (void) fflush(_err);
	    abort();
	}
    while (i--)
	*p++ = *q++;
}

/***************************************************************************
*
* Function:  MapWrBytes
* Purpose:   This routine is like WriteBytes except that a table is used
*            instead of a file.
* Arguments: buf - the stuff to be written is taken from here
*            cnt - the number of bytes to be written
*            mmFile - a pointer to a structure needed for the table
* Returns:
* See also:  MapToMem, MapFlush, Unmap, MapSkAbs, MapReBytes, WriteBytes
*
****************************************************************************
*/
FE3(void, MapWrBytes,
    char *, buf,
    unsigned long, cnt,
    MMFILE **, mmFile)
{
    register char *p, *q;
    register unsigned long i;

    p = (char *) buf;
    q = (*mmFile)->start + (*mmFile)->offset;
    i = cnt;
    (*mmFile)->offset += i;
    if ((*mmFile)->len < (*mmFile)->offset)
	{
	    (void) fprintf(_err, "\n%s is used directly\n", (*mmFile)->name);
	    (void) fflush(_err);
	    MapFlush(*mmFile);
	    SeekAbs((*mmFile)->bFile, (unsigned long) ((*mmFile)->offset - i));
	    WriteBytes(buf, cnt, (*mmFile)->bFile);
	    Unmap(*mmFile, FALSE);
	    *mmFile = NULL;
	    return;
	}
    while (i--)
	*q++ = *p++;
    if ((*mmFile)->actLen < (*mmFile)->offset)
	(*mmFile)->actLen = (*mmFile)->offset;
}

/***************************************************************************
*
* Function:  LngCons
* Purpose:   This routine reserves space for an element of unsigned long
*            and initializes the space.
* Arguments: value - the intended value
* Returns:   a pointer to the element
* See also:  LngCopy, LngDump, LngCompare, LngPrefix
*
****************************************************************************
*/
FE1(unsigned long *, LngCons,
    unsigned long, value)
{
    unsigned long *lng;

    lng = LngAlloc();
    *lng = value;
    return(lng);
}

/***************************************************************************
*
* Function:  LngCopy
* Purpose:   This routine reserves space for an element of unsigned long
*            and initializes the space according to another element.
* Arguments: lng - a pointer to the other element
* Returns:   a pointer to the new element
* See also:  LngCons, LngDump, LngCompare, LngPrefix
*
****************************************************************************
*/
FE1(unsigned long *, LngCopy,
    unsigned long *, lng)
{
    unsigned long *copy;

    copy = LngAlloc();
    *copy = *lng;
    return(copy);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  LngDump
* Purpose:   This routine displays an element of unsigned long.
* Arguments: lng - a pointer to the element
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  LngCons, LngCopy, LngCompare, LngPrefix
*
****************************************************************************
*/
FE2(void, LngDump,
    unsigned long *, lng,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "unsigned long: %lu\n", *lng);
    (void) fflush(_err);
}
#endif

/* A collection of routines for elements of unsigned long to be used by
 * list routines of dynam.c and table routines of ptrbl.c. */
TOOLS lngTools = 
{ (CFUNC) LngCopy, Deallocate
#ifdef DEBUG
, LngDump
#endif
 };

/***************************************************************************
*
* Function:  LngCompare
* Purpose:   This routine compares two elements of unsigned long
*            to each other.
* Arguments: lng1 - a pointer to the first element
*            lng2 - a pointer to the second element
* Returns:   -1 if the first is less than the second,
*             1 if the first is greater than the second, and
*             0 otherwise
* See also:  LngCons, LngCopy, LngDump, LngPrefix
*
****************************************************************************
*/
FE2(int, LngCompare,
    unsigned long *, lng1,
    unsigned long *, lng2)
{
    return((*lng1 == *lng2) ? 0 : ((*lng1 < *lng2) ? -1 : 1));
}

/***************************************************************************
*
* Function:  LngPrefix
* Purpose:   This routine checks whether an element of unsigned long
*            has a given prefix in the sense that all bits except the
*            least significant bit should be equal in the element and
*            in the prefix. (Though this definition is unconventional,
*            it has the advantage that the result of the check is
*            independent of the order of the arguments of this routine.)
* Arguments: lng - a pointer to the element
*            prefix - a pointer to the possible prefix
* Returns:   TRUE if the element has the prefix, and FALSE otherwise
* See also:  LngCons, LngCopy, LngDump, LngCompare
*
****************************************************************************
*/
FE2(BOOL, LngPrefix,
    unsigned long *, lng,
    unsigned long *, prefix)
{
    return((*lng >> 1) == (*prefix >> 1));
}

/* A collection of routines for elements of unsigned long to be used by
 * tree routines of baltree.c. */
BALTOOLS lngBalTools =
{ (CFUNC) LngCopy, Deallocate
#ifdef DEBUG
, LngDump
#endif
, LngCompare, LngPrefix };

/***************************************************************************
*
* Function:  DomCons
* Purpose:   This routine reserves space for an element of DOMAIN
*            and initializes the space.
* Arguments: value - the intended value
* Returns:   a pointer to the element
* See also:  DomCopy, DomDump, DomCompare, DomPrefix
*
****************************************************************************
*/
FE1(DOMAIN *, DomCons,
    DOMAIN, value)
{
    DOMAIN *dom;

    dom = DomAlloc();
    *dom = value;
    return(dom);
}

/***************************************************************************
*
* Function:  DomCopy
* Purpose:   This routine reserves space for an element of DOMAIN
*            and initializes the space according to another element.
* Arguments: dom - a pointer to the other element
* Returns:   a pointer to the new element
* See also:  DomCons, DomDump, DomCompare, DomPrefix
*
****************************************************************************
*/
FE1(DOMAIN *, DomCopy,
    DOMAIN *, dom)
{
    DOMAIN *copy;

    copy = DomAlloc();
    *copy = *dom;
    return(copy);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  DomDump
* Purpose:   This routine displays an element of DOMAIN.
* Arguments: dom - a pointer to the element
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  DomCons, DomCopy, DomCompare, DomPrefix
*
****************************************************************************
*/
FE2(void, DomDump,
    DOMAIN *, dom,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "DOMAIN: ");
    (void) fprintf(_err, DOMFORMAT, *dom);
    (void) fprintf(_err, "\n");
    (void) fflush(_err);
}
#endif

/* A collection of routines for elements of DOMAIN to be used by
 * list routines of dynam.c and table routines of ptrbl.c. */
TOOLS domTools = 
{ (CFUNC) DomCopy, Deallocate
#ifdef DEBUG
, DomDump
#endif
 };

/***************************************************************************
*
* Function:  DomCompare
* Purpose:   This routine compares two elements of DOMAIN
*            to each other.
* Arguments: dom1 - a pointer to the first element
*            dom2 - a pointer to the second element
* Returns:   -1 if the first is less than the second,
*             1 if the first is greater than the second, and
*             0 otherwise
* See also:  DomCons, DomCopy, DomDump, DomPrefix
*
****************************************************************************
*/
FE2(int, DomCompare,
    DOMAIN *, dom1,
    DOMAIN *, dom2)
{
    return((*dom1 == *dom2) ? 0 : ((*dom1 < *dom2) ? -1 : 1));
}

/***************************************************************************
*
* Function:  DomPrefix
* Purpose:   This routine checks whether an element of DOMAIN
*            has a given prefix in the sense that all bits except the
*            least significant bit should be equal in the element and
*            in the prefix. (Though this definition is unconventional,
*            it has the advantage that the result of the check is
*            independent of the order of the arguments of this routine.)
* Arguments: dom - a pointer to the element
*            prefix - a pointer to the possible prefix
* Returns:   TRUE if the element has the prefix, and FALSE otherwise
* See also:  DomCons, DomCopy, DomDump, DomCompare
*
****************************************************************************
*/
FE2(BOOL, DomPrefix,
    DOMAIN *, dom,
    DOMAIN *, prefix)
{
    return((*dom >> 1) == (*prefix >> 1));
}

/* A collection of routines for elements of DOMAIN to be used by
 * tree routines of baltree.c. */
BALTOOLS domBalTools =
{ (CFUNC) DomCopy, Deallocate
#ifdef DEBUG
, DomDump
#endif
, DomCompare, DomPrefix };

/***************************************************************************
*
* Function:  DataCopy
* Purpose:   This routine makes a copy of a string of DATATYPE.
* Arguments: t - the string to be copied
* Returns:   the copy
* See also:  DataDump, DataLen, DataCompare, DataPrefix
*
****************************************************************************
*/
FE1(DATATYPE *, DataCopy,
    DATATYPE *, t)
{
    DATATYPE *copy;
    register DATATYPE *s;

    s = t;
    while (*s++)
	;
    s = copy = (DATATYPE *) AllocArray((unsigned) (s - t), sizeof(DATATYPE));
    while ((*s++ = *t++))
	;
    return(copy);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  DataDump
* Purpose:   This routine displays a string of DATATYPE.
* Arguments: t - the string
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  DataCopy, DataLen, DataCompare, DataPrefix
*
****************************************************************************
*/
FE2(void, DataDump,
    DATATYPE *, t,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "DATA");
    while (*t)
	(void) fprintf(_err, " %lu", (unsigned long) (*t++));
    (void) fprintf(_err, "\n");
    (void) fflush(_err);
}
#endif

/* A collection of routines for strings of DATATYPE to be used by
 * list routines of dynam.c. */
TOOLS dataTools =
{ (CFUNC) DataCopy, Deallocate
#ifdef DEBUG
, DataDump
#endif
 };

/***************************************************************************
*
* Function:  DataLen
* Purpose:   This routine computes the length of a string of DATATYPE.
* Arguments: t - the string
* Returns:   the length
* See also:  DataCopy, DataDump, DataCompare, DataPrefix
*
****************************************************************************
*/
FE1(unsigned, DataLen,
    DATATYPE *, t)
{
    register DATATYPE *s;

    s = t;
    while (*s++)
	;
    return(((unsigned) (s - t)) - 1);
}

/***************************************************************************
*
* Function:  DataCompare
* Purpose:   This routine compares two strings of DATATYPE lexicographically
*            to each other. (April the 13th, 1992: strcmp returned 0
*            although the strings differed from each other before any
*            0 byte. Marko Rauhamaa found out that for any i, if the
*            (4 * i + 1)th byte was 128, then the byte was mistaken for 0.)
* Arguments: s - the first string
*            t - the second string
* Returns:   -1 if the first is less than the second,
*             1 if the first is greater than the second, and
*             0 otherwise
* See also:  DataCopy, DataDump, DataLen, DataPrefix
*
****************************************************************************
*/
FE2(int, DataCompare,
    DATATYPE *, s,
    DATATYPE *, t)
{
    for (; *s && (*s == *t); s++, t++)
	;
    return((*s == *t) ? 0 : ((*s < *t) ? -1 : 1));
}

/***************************************************************************
*
* Function:  DataPrefix
* Purpose:   This routine checks whether a string of DATATYPE has a given
*            prefix.
* Arguments: s - the string
*            p - the possible prefix
* Returns:   TRUE if the string has the prefix, and 0 otherwise
* See also:  DataCopy, DataDump, DataLen, DataCompare, StrPrefix
*
****************************************************************************
*/
FE2(BOOL, DataPrefix,
    DATATYPE *, s,
    DATATYPE *, p)
{
    for (; *p && (*s++ == *p); p++)
	;
    return(!*p);
}

/* A collection of routines for strings of DATATYPE to be used by
 * tree routines of baltree.c. */
BALTOOLS dataBalTools =
{ (CFUNC) DataCopy, Deallocate
#ifdef DEBUG
, DataDump
#endif
, DataCompare, DataPrefix };

/***************************************************************************
*
* Function:  StrPrefix
* Purpose:   This routine checks whether a character string has a given
*            prefix.
* Arguments: s - the string
*            p - the possible prefix
* Returns:   TRUE if the string has the prefix, and 0 otherwise
* See also:  DataPrefix, StrCp (dynam.c), StrDump (dynam.c)
*
****************************************************************************
*/
FE2(BOOL, StrPrefix,
    char *, str,
    char *, prefix)
{
    for (; *prefix && (*str++ == *prefix); prefix++)
	;
    return(!*prefix);
}

/* A collection of routines for character strings to be used by
 * tree routines of baltree.c. */
BALTOOLS strBalTools =
{ (CFUNC) StrCp, Deallocate
#ifdef DEBUG
, StrDump
#endif
, strcmp, StrPrefix };

/***************************************************************************
*
* Function:  RefCons
* Purpose:   This routine constructs an element of REF. The REF type is
*            used when there is an unordered table and we want to know
*            the position of a key.
* Arguments: key - the key
*            slot - the position
* Returns:   the constructed element
* See also:  RefCopy, RefDump, RefKeyCmp, RefPrefix
*
****************************************************************************
*/
FE2(REF *, RefCons,
    char *, key,
    unsigned long, slot)
{
    REF *ref;

    ref = RefAlloc();
    ref->key = key;
    ref->slot = slot;
    return(ref);
}

/***************************************************************************
*
* Function:  RefCopy
* Purpose:   This routine makes a copy of an element of REF.
* Arguments: ref - the element to be copied
* Returns:   the copy
* See also:  RefCons, RefDump, RefKeyCmp, RefPrefix
*
****************************************************************************
*/
FE1(REF *, RefCopy,
    REF *, ref)
{
    REF *copy;

    copy = RefAlloc();
    copy->key = ref->key;
    copy->slot = ref->slot;
    return(copy);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  RefDump
* Purpose:   This routine displays an element of REF.
* Arguments: ref - the element
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  RefCons, RefCopy, RefKeyCmp, RefPrefix
*
****************************************************************************
*/
FE2(void, RefDump,
    REF *, ref,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "REF slot %lu key %s\n", ref->slot, ref->key);
    (void) fflush(_err);
}
#endif

/***************************************************************************
*
* Function:  RefKeyCmp
* Purpose:   This routine compares the key of an element of REF to another
*            key.
* Arguments: ref - the element
*            key - a key
* Returns:   something negative if the key of the element is
*                                  less than the other key,
*            something positive if the key of the element is
*                                  greater than the other key, and
*            0 otherwise
* See also:  RefCons, RefCopy, RefDump, RefPrefix
*
****************************************************************************
*/
FE2(int, RefKeyCmp,
    REF *, ref,
    char *, key)
{
    return(strcmp(ref->key, key));
}

/***************************************************************************
*
* Function:  RefPrefix
* Purpose:   This routine checks whether the key of an element of REF
*            has a given prefix.
* Arguments: ref - the element
*            prefix - the possible prefix
* Returns:   TRUE if the key has the prefix, and 0 otherwise
* See also:  RefCons, RefCopy, RefDump, RefPrefix
*
****************************************************************************
*/
FE2(BOOL, RefPrefix,
    REF *, ref,
    char *, prefix)
{
    return(StrPrefix(ref->key, prefix));
}

/* A collection of routines for elements of REF to be used by
 * tree routines of baltree.c. */
BALTOOLS refBalTools =
{ (CFUNC) RefCopy, Deallocate
#ifdef DEBUG
, RefDump
#endif
, RefKeyCmp, RefPrefix };

/***************************************************************************
*
* Function:  DrefCons
* Purpose:   This routine constructs an element of DREF. The DREF type is
*            used when there is an unordered table and we want to know
*            the position of a key. DREF differs from the REF type in
*            the type of the key and in the intended usage. When an
*            element of DREF is deleted, the key is deleted, whereas
*            the deletion of an element of REF does not delete the key.
* Arguments: key - the key
*            slot - the position
* Returns:   the constructed element
* See also:  DrefCopy, DrefDiscard, DrefDump, DrefKeyCmp, DrefPrefix
*
****************************************************************************
*/
FE2(DREF *, DrefCons,
    DATATYPE *, key,
    unsigned long, slot)
{
    DREF *dref;

    dref = DrefAlloc();
    dref->key = key;
    dref->slot = slot;
    return(dref);
}

/***************************************************************************
*
* Function:  DrefCopy
* Purpose:   This routine makes a copy of an element of DREF.
* Arguments: dref - the element to be copied
* Returns:   the copy
* See also:  DrefCons, DrefDiscard, DrefDump, DrefKeyCmp, DrefPrefix
*
****************************************************************************
*/
FE1(DREF *, DrefCopy,
    DREF *, dref)
{
    DREF *copy;

    copy = DrefAlloc();
    copy->key = DataCopy(dref->key);
    copy->slot = dref->slot;
    return(copy);
}

/***************************************************************************
*
* Function:  DrefDiscard
* Purpose:   This routine removes an element of DREF from the memory.
* Arguments: dref - the element
* Returns:
* See also:  DrefCons, DrefCopy, DrefDump, DrefKeyCmp, DrefPrefix
*
****************************************************************************
*/
FE1(void, DrefDiscard,
    DREF *, dref)
{
    Deallocate((char *) (dref->key));
    Deallocate((char *) dref);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  DrefDump
* Purpose:   This routine displays an element of DREF.
* Arguments: dref - the element
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  DrefCons, DrefCopy, DrefDiscard, DrefKeyCmp, DrefPrefix
*
****************************************************************************
*/
FE2(void, DrefDump,
    DREF *, dref,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "DREF slot %lu\n", dref->slot);
    DumpIndent(depth);
    (void) fprintf(_err, "DREF key\n");
    (void) fflush(_err);
    DataDump(dref->key, depth + 1);
}
#endif

/***************************************************************************
*
* Function:  DrefKeyCmp
* Purpose:   This routine compares the key of an element of DREF to another
*            key.
* Arguments: dref - the element
*            key - a key
* Returns:   -1 if the key of the element is less than the other key,
*             1 if the key of the element is greater than the other key, and
*             0 otherwise
* See also:  DrefCons, DrefCopy, DrefDiscard, DrefDump, DrefPrefix
*
****************************************************************************
*/
FE2(int, DrefKeyCmp,
    DREF *, dref,
    DATATYPE *, key)
{
    return(DataCompare(dref->key, key));
}

/***************************************************************************
*
* Function:  DrefPrefix
* Purpose:   This routine checks whether the key of an element of DREF
*            has a given prefix.
* Arguments: dref - the element
*            prefix - the possible prefix
* Returns:   TRUE if the key has the prefix, and 0 otherwise
* See also:  DrefCons, DrefCopy, DrefDiscard, DrefDump, DrefPrefix
*
****************************************************************************
*/
FE2(BOOL, DrefPrefix,
    DREF *, dref,
    DATATYPE *, prefix)
{
    return(DataPrefix(dref->key, prefix));
}

/* A collection of routines for elements of DREF to be used by
 * tree routines of baltree.c. */
BALTOOLS drefBalTools =
{ (CFUNC) DrefCopy, DrefDiscard
#ifdef DEBUG
, DrefDump
#endif
, DrefKeyCmp, DrefPrefix };

/***************************************************************************
*
* Function:  StrLstCompare
* Purpose:   This routine compares two lists of character strings to each
*            other, in a two-fold lexicographic way.
* Arguments: list1 - the first list
*            list2 - the second list
* Returns:   something negative if the first is less than the second,
*            something positive if the first is greater than the second, and
*            0 otherwise
* See also:  StrLstPrefix
*
****************************************************************************
*/
FE2(int, StrLstCompare,
    LIST *, list1,
    LIST *, list2)
{
    char *str1, *str2;
    int cmpResult;

    if (list1 == list2)
	return(0);
    str1 = (char *) LstBegin(list1);
    str2 = (char *) LstBegin(list2);
    while (str1 && str2)
	{
	    if ((cmpResult = strcmp(str1, str2)))
		return(cmpResult);
	    str1 = (char *) LstStep(list1);
	    str2 = (char *) LstStep(list2);
	}
    return(str1 ? 1 : (str2 ? -1 : 0));
}

/***************************************************************************
*
* Function:  StrLstPrefix
* Purpose:   This routine checks whether a list of character strings has
*            a given prefix. As in StrLstCompare, a two-fold lexicographic
*            approach is used.
* Arguments: list - the list
*            prefix - the possible prefix
* Returns:   TRUE if the list has the prefix, and FALSE otherwise
* See also:  StrLstCompare
*
****************************************************************************
*/
FE2(BOOL, StrLstPrefix,
    LIST *, list,
    LIST *, prefix)
{
    char *str, *key;
    int cmpResult;

    if (list == prefix)
	return(TRUE);
    str = (char *) LstBegin(list);
    key = (char *) LstBegin(prefix);
    while (str && key)
	{
	    if ((cmpResult = strcmp(str, key)))
		return((cmpResult > 0) && (!LstNext(prefix)) &&
		       StrPrefix(str, key));
	    str = (char *) LstStep(list);
	    key = (char *) LstStep(prefix);
	}
    return(!key);
}

static BOOL lfLast;          /* From this, a routine manipulating
                              * a leaf segment of a segmented bit set table
                              * knows whether or not the leaf is the last
                              * leaf. */
static unsigned lfLastSize;  /* The size of the last leaf segment. */

/***************************************************************************
*
* Function:  LeafInit
* Purpose:   This routine creates a leaf segment of a segmented
*            bit set table.
* Arguments: leaf - a reference to the leaf segment
* Returns:
* See also:  LeafClear, HLeafInit (graph.c)
*
****************************************************************************
*/
FS1(BOOL, LeafInit,
    DATATYPE **, leaf)
{
    if (lfLast)
	{
	    *leaf = (DATATYPE *)
		AllocArray(lfLastSize, sizeof(DATATYPE));
	    lfLast = FALSE;
	}
    else
	*leaf = (DATATYPE *)
	    AllocArray(LEAFSIZE, sizeof(DATATYPE));
    return(FALSE);
}

/***************************************************************************
*
* Function:  LeafClear
* Purpose:   This routine clears the contents of a leaf segment
*            of a segmented bit set table.
* Arguments: leaf - the leaf segment
* Returns:
* See also:  LeafInit, HLeafClear (graph.c)
*
****************************************************************************
*/
FS1(BOOL, LeafClear,
    DATATYPE *, leaf)
{
    register unsigned i;

    if (lfLast)
	{
	    i = lfLastSize;
	    lfLast = FALSE;
	}
    else
	i = LEAFSIZE;
    while (i--) 
	leaf[i] = (DATATYPE) 0;
    return(FALSE);
}

/***************************************************************************
*
* Function:  BstCons
* Purpose:   This routine creates a segmented bit set table, needed
*            when there is a fixed set and we want to present some of its
*            subsets efficiently with respect to space and search time.
* Arguments: count - the number of bit positions in the table
* Returns:   the created table
* See also:  LeafInit, BstDiscard, BstDump, BstClear, BstIn, BstInsert,
*            BstDelete, HashCreate (graph.c)
*
****************************************************************************
*/
FE1(BITSET *, BstCons,
    unsigned long, count)
{
    BITSET *bitset;
    
    bitset = BstAlloc();
    BstSCount(bitset, count);
    BstSSet(bitset, PtrBlCreate(BstGNLeaves(bitset), &strTools));
    lfLastSize = BstGLastSize(bitset);
    lfLast = TRUE;
    (void) PtrBlBkdrApply(BstGSet(bitset), LeafInit);
    return(bitset);
}

/***************************************************************************
*
* Function:  BstDiscard
* Purpose:   This routine removes a segmented bit set table from the memory.
* Arguments: bitset - the table
* Returns:
* See also:  BstCons, BstDump, BstClear, BstIn, BstInsert, BstDelete,
*            HashDiscard (graph.c)
*
****************************************************************************
*/
FE1(void, BstDiscard,
    BITSET *, bitset)
{
    PtrBlDestroy(BstGSet(bitset), TRUE);
    Deallocate((char *) bitset);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  BstDump
* Purpose:   This routine displays a segmented bit set table.
* Arguments: bitset - the table
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  BstCons, BstDiscard, BstClear, BstIn, BstInsert, BstDelete
*
****************************************************************************
*/
FE2(void, BstDump,
    BITSET *, bitset,
    unsigned, depth)
{
    unsigned long n;
    unsigned k;

    DumpIndent(depth);
    (void) fprintf(_err, "BITSET count %lu\n", BstGCount(bitset));
    (void) fflush(_err);
    k = 0;
    for (n = 0; n < BstGCount(bitset); n++)
	if (BstIn(bitset, n)) 
	    {
		if (k == 19) 
		    {
			k = 0;
			(void) fprintf(_err, "\n");
			(void) fflush(_err);
			DumpIndent(depth + 1);
		    }
		else
		    k++;
		(void) fprintf(_err, " %lu", n);
		(void) fflush(_err);
	    }
    (void) fprintf(_err, "\nEnd of BITSET\n");
    (void) fflush(_err);
}
#endif

/***************************************************************************
*
* Function:  BstClear
* Purpose:   This routine clears the contents of a segmented bit set table.
* Arguments: bitset - the table
* Returns:
* See also:  LeafClear, BstCons, BstDiscard, BstDump, BstIn, BstInsert,
*            BstDelete, HashClear (graph.c)
*
****************************************************************************
*/
FE1(void, BstClear,
    BITSET *, bitset)
{
    lfLastSize = BstGLastSize(bitset);
    lfLast = TRUE;
    (void) PtrBlBackApply(BstGSet(bitset), LeafClear);
}

/***************************************************************************
*
* Function:  BstIn
* Purpose:   This routine checks whether a bit is in a bit set.
* Arguments: bitset - a segmented bit set table
*            number - the number of the bit to be checked
* Returns:   TRUE if the bit is in the set, and FALSE otherwise
* See also:  BstCons, BstDiscard, BstDump, BstClear, BstInsert, BstDelete,
*            HashGet (graph.c)
*
****************************************************************************
*/
FE2(BOOL, BstIn,
    BITSET *, bitset,
    unsigned long, number)
{
    DATATYPE *leaf;
    register unsigned long a;

    a = number / (BITSPERBYTE * sizeof(DATATYPE));
    leaf = (DATATYPE *) PtrBlGet(BstGSet(bitset), a >> LEAFBITS);
    return((1 & (unsigned)
	    (leaf[(unsigned) (a & LEAFMASK)]
	     >> (number %
		 (BITSPERBYTE * sizeof(DATATYPE)))))
	   != 0);
}

/***************************************************************************
*
* Function:  BstInsert
* Purpose:   This routine inserts a bit into a bit set.
* Arguments: bitset - a segmented bit set table
*            number - the number of the bit to be inserted
* Returns:
* See also:  BstCons, BstDiscard, BstDump, BstClear, BstIn, BstDelete,
*            HashPut (graph.c)
*
****************************************************************************
*/
FE2(void, BstInsert,
    BITSET *, bitset,
    unsigned long, number)
{
    DATATYPE *leaf;
    register unsigned long a;

    a = number / (BITSPERBYTE * sizeof(DATATYPE));
    leaf = (DATATYPE *) PtrBlGet(BstGSet(bitset), a >> LEAFBITS);
    leaf[(unsigned) (a & LEAFMASK)] 
	|= (((unsigned long) 1L) << (number %
				     (BITSPERBYTE * sizeof(DATATYPE))));
}

/***************************************************************************
*
* Function:  BstDelete
* Purpose:   This routine removes a bit from a bit set.
* Arguments: bitset - a segmented bit set table
*            number - the number of the bit to be removed
* Returns:
* See also:  BstCons, BstDiscard, BstDump, BstClear, BstIn, BstInsert
*
****************************************************************************
*/
FE2(void, BstDelete,
    BITSET *, bitset,
    unsigned long, number)
{
    DATATYPE *leaf;
    register unsigned long a;

    a = number / (BITSPERBYTE * sizeof(DATATYPE));
    leaf = (DATATYPE *) PtrBlGet(BstGSet(bitset), a >> LEAFBITS);
    leaf[(unsigned) (a & LEAFMASK)] 
	&= ~(((unsigned long) 1L) << (number %
				      (BITSPERBYTE * sizeof(DATATYPE))));
}

/***************************************************************************
*
* Function:  MonCons
* Purpose:   This routine creates an element of MON. The MON type is used
*            in on-the-fly verification. The elements of MON are called
*            monitor states. A monitor state consists of a state number and
*            a flag indicating what kind of a monitor state is in question.
* Arguments: state - a state number
* Returns:   the constructed element
* See also:  MonDump, MonKeyCmp
*
****************************************************************************
*/
FE1(MON *, MonCons,
    DOMAIN, state)
{
    MON *mon;

    mon = MonAlloc();
    MonSState(mon, state);
    MonSFlag(mon, (unsigned) 0);
    return(mon);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  MonCons
* Purpose:   This routine displays a monitor state.
* Arguments: mon - the monitor state
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  MonCons, MonKeyCmp
*
****************************************************************************
*/
FE2(void, MonDump,
    MON *, mon,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "MON state ");
    (void) fprintf(_err, DOMFORMAT, MonGState(mon));
    (void) fprintf(_err, " flag %u\n", MonGFlag(mon));
    (void) fflush(_err);
}
#endif

/***************************************************************************
*
* Function:  MonKeyCmp
* Purpose:   This routine compares the state number of a monitor state
*            to a key.
* Arguments: mon - the monitor state
*            key - the key key
* Returns:   -1 if the state number is less than the key,
*             1 if the state number is greater than the key, and
*             0 otherwise
* See also:  MonCons, MonDump
*
****************************************************************************
*/
FE2(int, MonKeyCmp,
    MON *, mon,
    DOMAIN *, key)
{
    return((MonGState(mon) < *key) ? -1 :
	   ((MonGState(mon) > *key) ? 1 : 0));
}

/* A collection of routines for monitor states to be used by
 * tree routines of baltree.c. */
BALTOOLS monBalTools =
{ (CFUNC) NULL, Deallocate
#ifdef DEBUG
, MonDump
#endif
, MonKeyCmp, (FOOL) NULL };

BALTREE *monitors;  /* a tree of monitor states */

/***************************************************************************
*
* Function:  StpCons
* Purpose:   This routine constructs a step, consisting of a node and one
*            of its immediate successor arrows in a reachability graph
*            of a net.
* Arguments: world - the address of the node
*            awAddr - the address of the arrow
* Returns:   the constructed step
* See also:  StpCopy, StpDump
*
****************************************************************************
*/
FE2(STEP *, StpCons,
    unsigned long, world,
    unsigned long, awAddr)
{
    register STEP *step;

    step = StpAlloc();
    step->world = world;
    step->awAddr = awAddr;
    return(step);
}

/***************************************************************************
*
* Function:  StpCopy
* Purpose:   This routine makes a copy of a step.
* Arguments: step - the step to be copied
* Returns:   the copy
* See also:  StpCons, StpDump
*
****************************************************************************
*/
FE1(STEP *, StpCopy,
    STEP *, step)
{
    register STEP *copy;

    copy = StpAlloc();
    copy->world = step->world;
    copy->awAddr = step->awAddr;
    return(copy);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  StpDump
* Purpose:   This routine displays a step.
* Arguments: step - the step
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  StpCons, StpCopy
*
****************************************************************************
*/
FE2(void, StpDump,
    STEP *, step,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "STEP: %lu %lu\n", step->world, step->awAddr);
    (void) fflush(_err);
}
#endif

/* A collection of routines for steps to be used by
 * list routines of dynam.c and table routines of ptrbl.c. */
TOOLS stpTools =
{ (CFUNC) StpCopy, Deallocate
#ifdef DEBUG
, StpDump
#endif
 };

/***************************************************************************
*
* Function:  PthCreate
* Purpose:   This routine constructs a path, corresponding to a path in a
*            reachability graph of a net and consisting of a list of
*            consecutive steps and an end node. The list is initialized
*            to be empty.
* Arguments: world - the end node
* Returns:   the constructed path
* See also:  PthCopy, PthDracsid, PthDump, PthInsFirst, PthInsLast,
*            PthDelLast, PthCatenate, PthCompare, PthSuffix
*
****************************************************************************
*/
FE1(PATH *, PthCreate,
    unsigned long, world)
{
    PATH *path;

    path = (PATH *) Allocate(sizeof(PATH));
    path->steps = LstCreate(&stpTools);
    path->end = world;
    return(path);
}

/***************************************************************************
*
* Function:  PthCopy
* Purpose:   This routine makes a copy of a path.
* Arguments: path - the path to be copied
* Returns:   the copy
* See also:  PthCreate, PthDracsid, PthDump, PthInsFirst, PthInsLast,
*            PthDelLast, PthCatenate, PthLoop, PthCompare, PthSuffix
*
****************************************************************************
*/
FE1(PATH *, PthCopy,
    PATH *, path)
{
    PATH *copy;

    copy = (PATH *) Allocate(sizeof(PATH));
    copy->steps = LstCopy(path->steps, TRUE);
    copy->end = path->end;
    return(copy);
}

/***************************************************************************
*
* Function:  PthDracsid
* Purpose:   This routines removes a path from the memory.
* Arguments: path - the path
* Returns:
* See also:  PthCreate, PthCopy, PthDump, PthInsFirst, PthInsLast,
*            PthDelLast, PthCatenate, PthLoop, PthCompare, PthSuffix
*
****************************************************************************
*/
FE1(void, PthDracsid,
    PATH *, path)
{
    LstDestroy(path->steps, TRUE);
    Deallocate((char *) path);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  PthDump
* Purpose:   This routine displays a path.
* Arguments: path - the path
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  PthCreate, PthCopy, PthDracsid, PthInsFirst, PthInsLast,
*            PthDelLast, PthCatenate, PthLoop, PthCompare, PthSuffix
*
****************************************************************************
*/
FE2(void, PthDump,
    PATH *, path,
    unsigned, depth)
{
    STEP *step;

    DumpIndent(depth);
    (void) fprintf(_err, "PATH\n");
    LSTFOR (path->steps, step, STEP) 
	StpDump(step, depth + 1);
    DumpIndent(depth);
    (void) fprintf(_err, "ENDING TO %lu\n", path->end);
    (void) fflush(_err);
}
#endif

/***************************************************************************
*
* Function:  PthInsFirst
* Purpose:   This routine adds a node and an arrow to the beginning
*            of a path.
* Arguments: path - the path
*            world - the address of the node (the source of the arrow)
*            awAddr - the address of the arrow
* Returns:
* See also:  PthCreate, PthCopy, PthDracsid, PthDump, PthInsLast,
*            PthDelLast, PthCatenate, PthLoop, PthCompare, PthSuffix
* See also:
*
****************************************************************************
*/
FE3(void, PthInsFirst,
    PATH *, path,
    unsigned long, world,
    unsigned long, awAddr)
{
    (void) LstInsLast(path->steps, (char *) StpCons(world, awAddr));
}

/***************************************************************************
*
* Function:  PthInsLast
* Purpose:   This routine adds an arrow and a node to the end of a path
* Arguments: path - the path
*            awAddr - the address of the arrow
*            world - the address of the node (the target of the arrow)
* Returns:
* See also:  PthCreate, PthCopy, PthDracsid, PthDump, PthInsFirst,
*            PthDelLast, PthCatenate, PthLoop, PthCompare, PthSuffix
*
****************************************************************************
*/
FE3(void, PthInsLast,
    PATH *, path,
    unsigned long, awAddr,
    unsigned long, world)
{
    (void) LstInsFirst(path->steps, (char *) StpCons(path->end, awAddr));
    path->end = world;
}

/***************************************************************************
*
* Function:  PthDelLast
* Purpose:   This routine removes the last arrow and node from a path.
* Arguments: path - the path
* Returns:
* See also:  PthCreate, PthCopy, PthDracsid, PthDump, PthInsFirst,
*            PthInsLast, PthCatenate, PthLoop, PthCompare, PthSuffix
*
****************************************************************************
*/
FE1(void, PthDelLast,
    PATH *, path)
{
    path->end = ((STEP *) LstFirst(path->steps))->world;
    (void) LstDelFirst(path->steps, TRUE);
}

/***************************************************************************
*
* Function:  PthCatenate
* Purpose:   This routine concatenates two paths.
* Arguments: path1 - the first path (will be deleted here)
*            path2 - the second path (will contain the result)
* Returns:   the result of concatenating path2 after path1
* See also:  PthCreate, PthCopy, PthDracsid, PthDump, PthInsFirst,
*            PthInsLast, PthDelLast, PthLoop, PthCompare, PthSuffix
*
****************************************************************************
*/
FE2(PATH *, PthCatenate,
    PATH *, path1,
    PATH *, path2)
{
    (void) LstCatenate(path2->steps, path1->steps);
    PthDracsid(path1);
    return(path2);
}

/***************************************************************************
*
* Function:  PthLoop
* Purpose:   This routine checks whether a path ends in a loop.
* Arguments: path - the path
* Returns:   TRUE if the path ends in a loop, and FALSE otherwise
* See also:  PthCreate, PthCopy, PthDracsid, PthDump, PthInsFirst,
*            PthInsLast, PthDelLast, PthCatenate, PthCompare, PthSuffix
*
****************************************************************************
*/
FE1(BOOL, PthLoop,
    PATH *, path)
{
    register STEP *step;

    LSTFOR (path->steps, step, STEP)
	if (step->world == path->end) 
	    return(TRUE);
    return(FALSE);
}

/* A collection of routines for elements of PATH to be used by
 * list routines of dynam.c and table routines of ptrbl.c. */
TOOLS pthTools =
{ (CFUNC) PthCopy, PthDracsid
#ifdef DEBUG
, PthDump
#endif
 };

/***************************************************************************
*
* Function:  PthCompare
* Purpose:   This routine compares two paths to each other, in a two-fold
*            lexicographic way.
* Arguments: path1 - the first path
*            path2 - the second path
* Returns:   -1 if the first is less than the second,
*             1 if the first is greater than the second, and
*             0 otherwise
* See also:  PthCreate, PthCopy, PthDracsid, PthDump, PthInsFirst,
*            PthInsLast, PthDelLast, PthCatenate, PthLoop, PthSuffix
*
****************************************************************************
*/
FE2(int, PthCompare,
    PATH *, path1,
    PATH *, path2)
{
    STEP *step1, *step2;

    if (path1 == path2)
	return(0);
    if (path1->end < path2->end) 
	return(-1);
    else if (path1->end > path2->end)
	return(1);
    else if (path1->steps == path2->steps)
	return(0);
    step1 = (STEP *) LstBegin(path1->steps);
    step2 = (STEP *) LstBegin(path2->steps);
    while (step1 && step2 && (step1->world == step2->world) &&
	   (step1->awAddr == step2->awAddr)) 
	{
	    step1 = (STEP *) LstStep(path1->steps);
	    step2 = (STEP *) LstStep(path2->steps);
	}
    if (step1 && step2)
	{
	    if (step1->world == step2->world)
		return((step1->awAddr > step2->awAddr) ? 1 : -1);
	    else
		return((step1->world > step2->world) ? 1 : -1);
	}
    else
	return(step1 ? 1 : (step2 ? -1 : 0));
}

/***************************************************************************
*
* Function:  PthSuffix
* Purpose:   This routine checks whether a path has a given suffix.
* Arguments: path - the path
*            suffix - the possible suffix
* Returns:   TRUE if the path has the suffix, and FALSE otherwise
* See also:  PthCreate, PthCopy, PthDracsid, PthDump, PthInsFirst,
*            PthInsLast, PthDelLast, PthCatenate, PthLoop, PthCompare
*
****************************************************************************
*/
FE2(BOOL, PthSuffix,
    PATH *, path,
    PATH *, suffix)
{
    STEP *step1, *step2;

    if (path == suffix)
	return(TRUE);
    if (path->end != suffix->end) 
	return(FALSE);
    else if (path->steps == suffix->steps)
	return(TRUE);
    step1 = (STEP *) LstBegin(path->steps);
    step2 = (STEP *) LstBegin(suffix->steps);
    while (step1 && step2 && (step1->world == step2->world) &&
	   (step1->awAddr == step2->awAddr)) 
	{
	    step1 = (STEP *) LstStep(path->steps);
	    step2 = (STEP *) LstStep(suffix->steps);
	}
    return (!step2);
}

/* A collection of routines for paths to be used by
 * tree routines of baltree.c. */
BALTOOLS pthBalTools =
{ (CFUNC) PthCopy, PthDracsid
#ifdef DEBUG
, PthDump
#endif
, PthCompare, PthSuffix };
