/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  graph.c
*
* Part of: PROD's programs araprod, "gphgen", probe, and strong
*
* Authors: Marko Rauhamaa and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     The routines in this module are needed for constructing and
*     inspecting a reachability graph of a net.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include <stdio.h>
#include <string.h>
#if (defined(UNIX) || defined(TNIL))
#include <unistd.h>
#endif
#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "error.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "mark.h"
#include "encode.h"
#include "graph.h"
#include "ptnet.h"
#include "araprod.h"
#include "cmdnm.h"

/* A segmented hash table is used in the generation of the reachability
 * graph of a net. The data of the hash table is kept in the leaf segments
 * of the segmented table. */
#define HLEAFBITS 11
#define HLEAFSIZE (((unsigned) 1) << HLEAFBITS)
#define HLEAFMASK (HLEAFSIZE - 1)

static BOOL hlLast;          /* From this, a routine manipulating
                              * a leaf segment of a segmented hash table
                              * knows whether or not the leaf is the last
                              * leaf. */
static unsigned hlLastSize;  /* The size of the last leaf segment. */

/***************************************************************************
*
* Function:  HLeafInit
* Purpose:   This routine creates a leaf segment of a segmented
*            hash table.
* Arguments: hleaf - a reference to the leaf segment
* Returns:
* See also:  HLeafClear, HLeafRead, HLeafWrite, HashCreate
*
****************************************************************************
*/
FS1(BOOL, HLeafInit,
    unsigned long **, hleaf)
{
    if (hlLast)
	{
	    *hleaf = (unsigned long *)
		AllocArray(hlLastSize, sizeof(unsigned long));
	    hlLast = FALSE;
	}
    else
	*hleaf = (unsigned long *)
	    AllocArray(HLEAFSIZE, sizeof(unsigned long));
    return(FALSE);
}

/***************************************************************************
*
* Function:  HLeafClear
* Purpose:   This routine clears the contents of a leaf segment
*            of a segmented hash table.
* Arguments: hleaf - the leaf segment
* Returns:
* See also:  HLeafInit, HLeafRead, HLeafWrite, HashClear
*
****************************************************************************
*/
FS1(BOOL, HLeafClear,
    unsigned long *, hleaf)
{
    register unsigned size;
    
    if (hlLast)
	{
	    size = hlLastSize;
	    hlLast = FALSE;
	}
    else
	size = HLEAFSIZE;
    while (size) 
	hleaf[--size] = (unsigned long) -1L;
    return(FALSE);
}

static BINARYFILE hashFile; /* A binary file used for storing
                             * the contents of a hash table. */

/***************************************************************************
*
* Function:  HLeafRead
* Purpose:   This routine reads data from a binary file and fills
*            a leaf segment of a segmented hash table with the data.
* Arguments: hleaf - the leaf segment
* Returns:
* See also:  HLeafInit, HLeafClear, HLeafWrite, HashRead
*
****************************************************************************
*/
FS1(BOOL, HLeafRead,
    unsigned long *, hleaf)
{
    if (hlLast)
	{
	    ReadLng(hleaf, hlLastSize, hashFile);
	    hlLast = FALSE;
	}
    else
	ReadLng(hleaf, HLEAFSIZE, hashFile);
    return(FALSE);
}

/***************************************************************************
*
* Function:  HLeafWrite
* Purpose:   This routine writes the contents of a leaf segment of
*            a segmented hash table to a binary file.
* Arguments: hleaf - the leaf segment
* Returns:
* See also:  HLeafInit, HLeafClear, HLeafRead, HashWrite
*
****************************************************************************
*/
FS1(BOOL, HLeafWrite,
    unsigned long *, hleaf)
{
    if (hlLast)
	{
	    WriteLng(hleaf, hlLastSize, hashFile);
	    hlLast = FALSE;
	}
    else
	WriteLng(hleaf, HLEAFSIZE, hashFile);
    return(FALSE);
}

/***************************************************************************
*
* Function:  HashCreate
* Purpose:   This routine creates a segmented hash table.
* Arguments: size - the number of data positions in the table.
* Returns:   the created table
* See also:  HLeafInit, HashClear, HashGet, HashPut, HashDiscard,
*            HashRead, HashWrite
*
****************************************************************************
*/
FE1(HASH *, HashCreate,
    unsigned long, size)
{
    HASH *hash;
    
    hash = PtrBlCreate(((size - 1L) >> HLEAFBITS) + 1L, &strTools);
    hlLastSize = (unsigned) (((size - 1L) & HLEAFMASK) + 1L);
    hlLast = TRUE;
    (void) PtrBlBkdrApply(hash, HLeafInit);
    return(hash);
}

/***************************************************************************
*
* Function:  HashClear
* Purpose:   This routine clears the contents of a segmented hash table.
* Arguments: hash - the table
*            size - the number of data positions in the table
* Returns:
* See also:  HLeafClear, HashCreate, HashGet, HashPut, HashDiscard,
*            HashRead, HashWrite
*
****************************************************************************
*/
FE2(void, HashClear,
    HASH *, hash,
    unsigned long, size)
{
    hlLastSize = (unsigned) (((size - 1L) & HLEAFMASK) + 1L);
    hlLast = TRUE;
    (void) PtrBlBackApply(hash, HLeafClear);
}

/***************************************************************************
*
* Function:  HashRead
* Purpose:   This routine reads data from a binary file and fills
*            a segmented hash table with the data.
* Arguments: hash - the table
*            size - the number of data positions in the table
*            file - the file
* Returns:
* See also:  HLeafRead, HashCreate, HashClear, HashGet, HashPut,
*            HashDiscard, HashWrite
*
****************************************************************************
*/
FE3(void, HashRead,
    HASH *, hash,
    unsigned long, size,
    BINARYFILE, file)
{
    hashFile = file;
    hlLastSize = (unsigned) (((size - 1L) & HLEAFMASK) + 1L);
    hlLast = TRUE;
    (void) PtrBlBackApply(hash, HLeafRead);
}

/***************************************************************************
*
* Function:  HashWrite
* Purpose:   This routine writes the contents of a segmented hash table
*            to a binary file.
* Arguments: hash - the table
*            size - the number of data positions in the table
*            file - the file
* Returns:
* See also:  HLeafWrite, HashCreate, HashClear, HashGet, HashPut,
*            HashDiscard, HashRead
*
****************************************************************************
*/
FE3(void, HashWrite,
    HASH *, hash,
    unsigned long, size,
    BINARYFILE, file)
{
    hashFile = file;
    hlLastSize = (unsigned) (((size - 1L) & HLEAFMASK) + 1L);
    hlLast = TRUE;
    (void) PtrBlBackApply(hash, HLeafWrite);
}

/***************************************************************************
*
* Function:  PrimeBelow
* Purpose:   This routine computes the greatest prime number less than
*            or egual to b.
* Arguments: b - as explained above
* Returns:   the prime number explained above
* See also:
*
****************************************************************************
*/
FE1(unsigned long, PrimeBelow,
    unsigned long, b)
{
    register unsigned long p, x, dx, sqx;
    
    if (b == 2L)
        return(2L);
    p = (b - 1L) | 1L;
    while (1)
	{
	    /* sqx == 9+8*n*(2+(n+1))/2 == (3+2*n)*(3+2*n) == x*x */
	    for (x = 3L, dx = 8L, sqx = 9L;
		 (sqx < p) && (p % x);
		 x += 2L, dx += 8L, sqx += dx)
		;
	    if (sqx > p)
		break;
	    p -= 2L;
	}
    return(p);
}

/***************************************************************************
*
* Function:  Hash
* Purpose:   This routine computes a hash value for a string. The value
*            is greater than or equal to 0L and less than or equal to
*            prime - 1L.
* Arguments: s - the string
*            prime - a prime number
*            coeff - the greatest prime number less than or equal to
*                    prime >> 1
* Returns:   the computed hash value
* See also:
*
****************************************************************************
*/
FE3(unsigned long, Hash,
    DATATYPE *, s,
    unsigned long, prime,
    unsigned long, coeff)
{
    register unsigned long l;
    
    for (l = 0L; *s; l = l * coeff ^ (*s++))
	;
    return(l % prime);
}

/***************************************************************************
*
* Function:  GphCreate
* Purpose:   This routine creates a reachability graph for a net.
*            The created graph is empty, but nodes and arrows can
*            be inserted into the graph by GphPutNode and GphPutArrow.
*            A hash table is created to be used for checking whether
*            the graph has a node corresponding to a computed marking
*            of the net. The construction of the graph uses the file
*            system. If one wants to use the virtual memory
*            instead, one can call GphMapToMem.
* Arguments: name - the name of the graph
*            hash_bits - the size of the hash table is the greatest
*                        prime number less than (or equal to !!)
*                        power(2, hash_bits)
*            bufsize - an initial size for a buffer to be used
*                      for an encoded marking or transition instance
*                      (later, the routines using the buffer will
*                       automatically expand the buffer if necessary)
*            seed - a seed number for a random number generator
*                   which may be used by the stubborn set method
*            createFlag - the last actual parameter; the following
*                         are contained in it
*            unfold - TRUE if and only if the net is unfolded
*                     before the construction of the graph
*            symm - TRUE if and only if the symmetry method is used
*            iMkSymm - TRUE if and only if the symmetry method
*                      is used with the condition that no symmetry
*                      maps the initial marking to any other marking
*            stub - TRUE if and only if the stubborn set method is
*                   used
*            delAlg - TRUE if and only if the deletion algorithm
*                     of the stubborn set method is used
*            cffd - TRUE if and only if the CFFD-preserving version
*                   of the stubborn set method is used
*            defVis - TRUE if and only if transition instances
*                     are visible by default
*            sleep - TRUE if and only if the sleep set method
*                    is used
*            bdds - TRUE if and only if BDD's are used
*            dfs - TRUE if and only if the graph is constructed
*                  in a depth first order
*            freeExpansion - TRUE if and only if dfs is TRUE
*                            and the successor information of a node
*                            can be updated arbitrarily many times
*            inftr - TRUE if and only if on-the-fly verification is
*                    done with infinite path monitor states
*            rotSuff - TRUE if and only if the graph is an auxiliary
*                      graph (then the suffix of a file name
*                      is obtained from the default suffix by
*                      a rotation to left)
* Returns:   the graph
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE5(GRAPH *, GphCreate,
    char *, name,
    unsigned, hash_bits,
    unsigned, bufsize,
    UNS32, seed,
    unsigned long, createFlag)
{
    BOOL unfold;
    BOOL symm;
    BOOL iMkSymm;
    BOOL stub;
    BOOL delAlg;
    BOOL cffd;
    BOOL defVis;
    BOOL s_l_e_e_p;
    BOOL bdds;
    BOOL dfs;
    BOOL freeExpansion;
    BOOL inftr;
    BOOL rotSuff;
    unsigned long mask;
    GRAPH *graph;
    char *s;

    mask = ((unsigned long) 1L) << 12;
    unfold = (0L != (createFlag & mask));
    mask >>= 1;
    symm = (0L != (createFlag & mask));
    mask >>= 1;
    iMkSymm = (0L != (createFlag & mask));
    mask >>= 1;
    stub = (0L != (createFlag & mask));
    mask >>= 1;
    delAlg = (0L != (createFlag & mask));
    mask >>= 1;
    cffd = (0L != (createFlag & mask));
    mask >>= 1;
    defVis = (0L != (createFlag & mask));
    mask >>= 1;
    s_l_e_e_p = (0L != (createFlag & mask));
    mask >>= 1;
    bdds = (0L != (createFlag & mask));
    mask >>= 1;
    dfs = (0L != (createFlag & mask));
    mask >>= 1;
    freeExpansion = (0L != (createFlag & mask));
    mask >>= 1;
    inftr = (0L != (createFlag & mask));
    mask >>= 1;
    rotSuff = (0L != (createFlag & mask));
    if (mask != (unsigned long) 1L)
	{
	    (void) fprintf(_out, "\n");
	    (void) fflush(_out);
	    (void) fprintf(_err, "\nInternal mask error in GphCreate\n");
	    (void) fflush(_err);
	    abort();
	}
    graph = (GRAPH *) Allocate(sizeof(GRAPH));
    graph->reason = GPH_GENERATE;
    graph->mmnodes = NULL;
    graph->mmarrows = NULL;
    graph->mmaddrs = NULL;
    graph->mmstack = NULL;
    s = FileName(name, rotSuff ? "phg" : "gph");
    if (NotBinaryFile(graph->nodes = CreateUpdate(s)))
	{
	    Deallocate((char *) s);
	    Deallocate((char *) graph);
	    return(NULL);
	}
    Deallocate((char *) s);
    s = FileName(name, rotSuff ? "wsa" : "aws");
    if (NotBinaryFile(graph->arrows = CreateUpdate(s)))
	{
	    OwnClose(graph->nodes);
	    Deallocate((char *) s);
	    Deallocate((char *) graph);
	    return(NULL);
	}
    Deallocate((char *) s);
    s = FileName(name, rotSuff ? "dra" : "adr");
    if (NotBinaryFile(graph->addrs = CreateUpdate(s)))
	{
	    OwnClose(graph->nodes);
	    OwnClose(graph->arrows);
	    Deallocate((char *) s);
	    Deallocate((char *) graph);
	    return(NULL);
	}
    Deallocate((char *) s);
    if (s_l_e_e_p)
	{
	    s = FileName(name, rotSuff ? "tks" : "stk");
	    if (NotBinaryFile(graph->stack = CreateUpdate(s)))
		{
		    OwnClose(graph->nodes);
		    OwnClose(graph->arrows);
		    OwnClose(graph->addrs);
		    Deallocate((char *) s);
		    Deallocate((char *) graph);
		    return(NULL);
		}
	    Deallocate((char *) s);
	}
    graph->glob = (unsigned long *)
        AllocArray(NUM_GRAPH_GLOBALS, sizeof(unsigned long));
    graph->glob[FLAG_POS] = 0L;
    graph->glob[NODE_COUNT_POS] = 0L; 
    graph->glob[ARROW_COUNT_POS] = 0L;
    graph->glob[SGC_COUNT_POS] = 0L;
    graph->glob[TERMCOMP_COUNT_POS] = 0L;
    graph->glob[LATEST_TERMCOMP_POS] = (unsigned long) -1L;
    graph->glob[TERMNODE_COUNT_POS] = 0L;
    graph->glob[LATEST_TERMNODE_POS] = (unsigned long) -1L;
    graph->glob[GEN_COUNT_POS] = 0L;
    graph->glob[EOF_NODES_POS] = (unsigned long)
	(NUM_GRAPH_GLOBALS * sizeof(unsigned long));
    graph->glob[EOF_ARROWS_POS] = 0L;
    graph->glob[LAST_ARROW_POS] = 0L;
    if (dfs)
	{
	    _sdfs(graph);
	    graph->glob[FIRST_UNHANDLED_POS] = (unsigned long) -1L;
	}
    else
	graph->glob[FIRST_UNHANDLED_POS] = graph->glob[EOF_NODES_POS];
    graph->glob[HASH_BITS_POS] = (unsigned long) hash_bits;
    graph->glob[HASH_SIZE_POS] = PrimeBelow(((unsigned long) 1L) << hash_bits);
    graph->glob[HASH_COEFF_POS] = PrimeBelow(graph->glob[HASH_SIZE_POS] >> 1);
    graph->glob[BUF_SIZE_POS] = (unsigned long) bufsize;
    graph->glob[MAX_DATALEN_POS] = 0L;
    graph->glob[SEED_POS] = (unsigned long) seed;
    graph->glob[TOPMOST_POS] = (unsigned long) -1L;
    graph->glob[EOF_STACK_POS] = 0L;
    graph->bdd = NULL;
    graph->balTree = NULL;
    if (rotSuff)
	_srotsuff(graph);
    if (dfs)
	{
	    graph->next_lower_pos = FOLLOWERS_POS + 1;
	    if (inftr)
		{
		    _sinftr(graph);
		    graph->check_pos = graph->next_lower_pos + 1;
		    graph->sgc_val_pos = graph->check_pos + 1;
		}
	    else
		graph->sgc_val_pos = graph->next_lower_pos + 1;
	}
    else
	{
	    graph->next_lower_pos = (unsigned) -1;
	    graph->sgc_val_pos = FOLLOWERS_POS + 1;
	}
    graph->predecessors_pos = graph->sgc_val_pos + 1;
    graph->left_child_pos = graph->predecessors_pos + 1;
    if (bdds)
	{
	    _sbdd(graph);
	    graph->datalen_pos = graph->left_child_pos + 1;
	}
    else
	{
	    graph->right_child_pos = graph->left_child_pos + 1;
	    graph->datalen_pos = graph->right_child_pos + 1;
	}
    graph->num_node_fields = graph->datalen_pos + 1;
    if (unfold)
	_sunfold(graph);
    if (symm)
	{
	    _ssymm(graph);
	    if (iMkSymm)
		_simksymm(graph);
	}
    if (stub)
	{
	    _sstub(graph);
	    if (delAlg)
		_sdelalg(graph);
	    if (cffd)
		_scffd(graph);
	}
    if (defVis)
	_sdefvis(graph);
    if (dfs && freeExpansion)
	{
	    _sfrexpansion(graph);
	    graph->aw_datalen_pos = AW_NEXT_POS + 1;
	}
    else
	graph->aw_datalen_pos = TARGET_POS + 1;
    if (s_l_e_e_p)
	_s_s_l_e_e_p(graph);
    graph->num_arrow_fields = graph->aw_datalen_pos + 1;
    WriteLng(graph->glob, (unsigned) NUM_GRAPH_GLOBALS, graph->nodes);
    if (!bdds)
	{
	    graph->hash = HashCreate(graph->glob[HASH_SIZE_POS]);
	    HashClear(graph->hash, graph->glob[HASH_SIZE_POS]);
	}
    graph->latest_node = (unsigned long) -1L;
    graph->latest_rec = (unsigned long *)
        AllocArray(graph->num_node_fields, sizeof(unsigned long));
    graph->latest_contents = (DATATYPE *)
	AllocArray((unsigned) (graph->glob[BUF_SIZE_POS]), sizeof(DATATYPE));
    graph->arrow_rec = (unsigned long *)
	AllocArray(graph->num_arrow_fields, sizeof(unsigned long));
    graph->sgc_rec = (unsigned long *)
	AllocArray(NUM_SGC_FIELDS, sizeof(unsigned long));
    graph->name = StrCopy(name);
    if (!rotSuff)
        {
            s = FileName(name, "cad");
            (void) unlink(s);
            Deallocate((char *) s);
            s = FileName(name, "caw");
            (void) unlink(s);
            Deallocate((char *) s);
        }
    return(graph);
}
/***************************************************************************
*
* Function:  GphGet
* Purpose:   This routine returns an untouched node to be processed
*            in the generation of a reachability graph of a net.
*            However, in the case of the sleep set method, this routine
*            returns the topmost node in the search stack.
* Arguments: graph - the reachability graph
*            buffer - an encoded marking corresponding to the node
*                     is read into this buffer
*            size - the size of the buffer, changed if necessary
*            sleep1 - a reference to a sleep set associated with the node
*                     (important only if the sleep set method is used)
*            sleep2 - a reference to another sleep set associated with
*                     the node (important only if the sleep set method
*                     is used)
* Returns:   the node explained above if it exists, and -1L otherwise
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE5(unsigned long, GphGet,
    GRAPH *, graph,
    DATATYPE **, buffer,
    unsigned *, size,
    BALTREE **, sleep1,
    BALTREE **, sleep2)
{
    unsigned long node_addr;
    unsigned len = 0;
    unsigned long a;
    register DATATYPE *pointer;
    
    graph->latest_node = (unsigned long) -1L;
    if (_s_l_e_e_p(graph))
	{
	    if (graph->glob[TOPMOST_POS] == (unsigned long) -1L)
		return((unsigned long) -1L);
	    StkSkAbs(graph, graph->glob[TOPMOST_POS]);
	    StkReUns(&len, 1, graph);
	    if (len >= (unsigned) (graph->glob[BUF_SIZE_POS]))
		{
		    graph->glob[BUF_SIZE_POS] =
			(unsigned long) ((len >> 1) + len + 1);
		    Deallocate((char *) (graph->latest_contents));
		    graph->latest_contents = (DATATYPE *)
			AllocArray((unsigned) (graph->glob[BUF_SIZE_POS]),
				   sizeof(DATATYPE));
		}
	    StkReData(graph->latest_contents, len, graph);
	    pointer = graph->latest_contents;
	    GetValue(unsigned long, a, pointer);
	    graph->glob[EOF_STACK_POS] = graph->glob[TOPMOST_POS];
	    graph->glob[TOPMOST_POS] = a;
	    GetValue(unsigned long, node_addr, pointer);
	    *sleep1 = BalTrCreate(&lngBalTools);
	    while (*pointer)
		{
		    GetValue(unsigned long, a, pointer);
		    (void) BalTrInsert(*sleep1, (char *) LngCons(a),
				       (char *) (&a), TRUE);
		}
	    NdsSkAbs(graph,
		     node_addr +
		     graph->next_lower_pos * sizeof(unsigned long));
	    NdsReLng(graph->latest_rec + graph->next_lower_pos,
		     graph->datalen_pos - graph->next_lower_pos + 1, graph);
	    len = (unsigned) (graph->latest_rec[graph->datalen_pos]);
	    if (len >= *size)
		{
		    *size = (len >> 1) + len + 1;
		    Deallocate((char *) *buffer);
		    *buffer = (DATATYPE *) AllocArray(*size, sizeof(DATATYPE));
		}
	    NdsReData(*buffer, len, graph);
	    if (graph->latest_rec[graph->next_lower_pos]
		>= graph->glob[EOF_ARROWS_POS])
		{
		    if (graph->latest_rec[graph->next_lower_pos]
			== (unsigned long) -1L)
			*sleep2 = NULL;
		    else
			*sleep2 = BalTrCreate(&lngBalTools);
		    return(node_addr);
		}
	    AwsSkAbs(graph, graph->latest_rec[graph->next_lower_pos]);
	    AwsReUns(&len, 1, graph);
	    if (len >= (unsigned) (graph->glob[BUF_SIZE_POS]))
		{
		    graph->glob[BUF_SIZE_POS] =
			(unsigned long) ((len >> 1) + len + 1);
		    Deallocate((char *) (graph->latest_contents));
		    graph->latest_contents = (DATATYPE *)
			AllocArray((unsigned) (graph->glob[BUF_SIZE_POS]),
				   sizeof(DATATYPE));
		}
	    AwsReData(graph->latest_contents, len, graph);
	    pointer = graph->latest_contents;
	    *sleep2 = BalTrCreate(&lngBalTools);
	    while (*pointer)
		{
		    GetValue(unsigned long, a, pointer);
		    (void) BalTrInsert(*sleep2, (char *) LngCons(a),
				       (char *) (&a), TRUE);
		}
	    return(node_addr);
	}
    do
	{
	    node_addr = graph->glob[FIRST_UNHANDLED_POS];
	    if (node_addr >= graph->glob[EOF_NODES_POS])
		return((unsigned long) -1L);
	    NdsSkAbs(graph,
		     node_addr + SUCCESSOR_COUNT_POS * sizeof(unsigned long));
	    NdsReLng(graph->latest_rec + SUCCESSOR_COUNT_POS,
		     graph->datalen_pos - SUCCESSOR_COUNT_POS + 1, graph);
	    if (_dfs(graph))
		graph->glob[FIRST_UNHANDLED_POS] =
		    graph->latest_rec[graph->next_lower_pos];
	    else
		graph->glob[FIRST_UNHANDLED_POS] +=
		    graph->latest_rec[graph->datalen_pos] * sizeof(DATATYPE)
			+ graph->num_node_fields * sizeof(unsigned long);
	} while ((graph->latest_rec[SUCCESSOR_COUNT_POS]
		  != (unsigned long) -1L)
		 &&
		 ((!_inftr(graph))
		  ||
		  (graph->latest_rec[graph->check_pos] & 1L)));
    a = graph->latest_rec[graph->datalen_pos];
    if (*size <= (unsigned) a)
	{
	    *size = (((unsigned) a) >> 1) + ((unsigned) a) + 1;
	    Deallocate((char *) *buffer);
	    *buffer = (DATATYPE *) AllocArray(*size, sizeof(DATATYPE));
	}
    NdsReData(*buffer, a, graph);
    return(node_addr);
}

/***************************************************************************
*
* Function:  GphPutNode
* Purpose:   This routine inserts a node into a reachability graph of
*            a net if the corresponding marking is new.
* Arguments: graph - the reachability graph
*            node - an encoded marking corresponding to the node
*            len - the length of the node string
*            back - returns the address of the latest incoming arrow
*                   if such arrow exists, and (unsigned long) -1L otherwise
*                   (exception: a NULL pointer returns nothing)
* Returns:   the address of a node which has the marking and is in
*            in the graph
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE4(unsigned long, GphPutNode,
    GRAPH *, graph,
    DATATYPE *, node,
    unsigned, len,
    unsigned long *, back)
{
    unsigned long hash_value;
    unsigned long node_addr;
    int cmp_result;
    DREF *dref;
    
    if (((unsigned long *) NULL) != back)
	*back = (unsigned long) -1L;
    if (_bdd(graph))
	{
	    node_addr = (unsigned long) -1L;
	    if (graph->balTree)
		{
		    if ((dref = (DREF *)
			 BalTrElement(graph->balTree, (char *) node)))
			node_addr = DrefGSlot(dref);
		}
	    if (graph->bdd)
		{
		    if (!BddFind(graph->bdd, node, &node_addr))
			node_addr = (unsigned long) -1L;
		}
	    if (((unsigned long) -1L) != node_addr)
		{
		    if (((unsigned long *) NULL) != back)
			{
			    NdsSkAbs(graph,
				     node_addr + graph->predecessors_pos *
				     sizeof(unsigned long));
			    NdsReLng(back, 1, graph);
			    if (*back != graph->glob[EOF_ARROWS_POS])
				{
				    NdsSkAbs(graph,
					     node_addr +
					     graph->predecessors_pos *
					     sizeof(unsigned long));
				    NdsWrLng(graph->glob + EOF_ARROWS_POS,
					     1, graph);
				}
			    if (*back >= graph->glob[EOF_ARROWS_POS])
				*back = (unsigned long) -1L;
			}
		    return(node_addr);
		}
	    node_addr = graph->glob[EOF_NODES_POS];
	    if (!graph->balTree)
		graph->balTree = BalTrCreate(&drefBalTools);
	    (void) BalTrInsert(graph->balTree, (char *)
			       DrefCons(DataCopy(node), node_addr),
			       (char *) node, TRUE);
	    if (BalTrGElCount(graph->balTree) == (unsigned long) 64)
		{
		    BalTrBdd(graph->balTree, &(graph->bdd));
		    graph->balTree = NULL;
		}
	}
    else
	{
	    hash_value = Hash(node, graph->glob[HASH_SIZE_POS],
			      graph->glob[HASH_COEFF_POS]);
	    node_addr = HashGet(graph->hash, hash_value);
	    if (node_addr == (unsigned long) -1L)
		{
		    node_addr = graph->glob[EOF_NODES_POS];
		    HashPut(graph->hash, hash_value, node_addr);
		}
	    else
		{
		    while (1)
			{
			    if (((unsigned long *) NULL) != back)
				{
				    NdsSkAbs(graph,
					     node_addr +
					     graph->predecessors_pos *
					     sizeof(unsigned long));
				    NdsReLng(graph->latest_rec +
					     graph->predecessors_pos,
					     4, graph);
				}
			    else
				{
				    NdsSkAbs(graph, 
					     node_addr + 
					     graph->left_child_pos *
					     sizeof(unsigned long));
				    NdsReLng(graph->latest_rec +
					     graph->left_child_pos,
					     3, graph);
				}
			    graph->latest_node = (unsigned long) -1L;
			    NdsReData(graph->latest_contents,
				      graph->latest_rec[graph->datalen_pos],
				      graph);
			    if (!(cmp_result =
				  DataCompare(node, graph->latest_contents)))
				{
				    if (((unsigned long *) NULL) != back)
					{
					    *back =
						graph->latest_rec[graph->predecessors_pos];
					    if (*back !=
						graph->glob[EOF_ARROWS_POS])
						{
						    NdsSkAbs(graph,
							     node_addr +
							     graph->predecessors_pos *
							     sizeof(unsigned long));
						    NdsWrLng(graph->glob +
							     EOF_ARROWS_POS,
							     1, graph);
						}
					    if (*back >=
						graph->glob[EOF_ARROWS_POS])
						*back = (unsigned long) -1L;
					}
				    return(node_addr);
				}
			    if (cmp_result < 0)
				{
				    if (graph->latest_rec[graph->left_child_pos]
					== (unsigned long) -1L)
					{
					    NdsSkAbs(graph,
						     node_addr +
						     graph->left_child_pos *
						     sizeof(unsigned long));
					    break;
					}
				    node_addr =
					graph->latest_rec[graph->left_child_pos];
				}
			    else
				{
				    if (graph->latest_rec[graph->right_child_pos]
					== (unsigned long) -1L)
					{
					    NdsSkAbs(graph,
						     node_addr +
						     graph->right_child_pos *
						     sizeof(unsigned long));
					    break;
					}
				    node_addr =
					graph->latest_rec[graph->right_child_pos];
				}
			}
		    node_addr = graph->glob[EOF_NODES_POS];
		    NdsWrLng(&node_addr, 1, graph);
		}
	}
    AdrSkAbs(graph, graph->glob[NODE_COUNT_POS] * sizeof(unsigned long));
    AdrWrLng(&node_addr, 1, graph);
    graph->latest_node = (unsigned long) -1L;
    graph->latest_rec[ORDINAL_POS] = graph->glob[NODE_COUNT_POS];
    graph->latest_rec[SUCCESSOR_COUNT_POS] = (unsigned long) -1L;
    if (_s_l_e_e_p(graph))
	graph->latest_rec[graph->next_lower_pos] = (unsigned long) -1L;
    else if (_dfs(graph))
	{
	    if (_inftr(graph))
		graph->latest_rec[graph->check_pos] = 0L;
	    graph->latest_rec[graph->next_lower_pos] =
		graph->glob[FIRST_UNHANDLED_POS];
	    graph->glob[FIRST_UNHANDLED_POS] = node_addr;
	}
    graph->latest_rec[FOLLOWERS_POS] = (unsigned long) -1L;
    graph->latest_rec[graph->sgc_val_pos] = 0L;
    graph->latest_rec[graph->predecessors_pos] =
	(graph->latest_rec[ORDINAL_POS] != 0L)
	    ? graph->glob[EOF_ARROWS_POS] : (unsigned long) -1L;
    graph->latest_rec[graph->left_child_pos] = (unsigned long) -1L;
    if (!_bdd(graph))
	graph->latest_rec[graph->right_child_pos] = (unsigned long) -1L;
    graph->latest_rec[graph->datalen_pos] = (unsigned long) len;
    NdsSkAbs(graph, node_addr);
    NdsWrLng(graph->latest_rec, graph->num_node_fields, graph);
    NdsWrData(node, len, graph);
    (graph->glob[NODE_COUNT_POS])++;
    graph->glob[EOF_NODES_POS] +=
	len * sizeof(DATATYPE) +
	    graph->num_node_fields * sizeof(unsigned long);
    if (len >= (unsigned) (graph->glob[BUF_SIZE_POS]))
	{
	    graph->glob[BUF_SIZE_POS] =
		(unsigned long) ((len >> 1) + len + 1);
	    Deallocate((char *) (graph->latest_contents));
	    graph->latest_contents = (DATATYPE *)
		AllocArray((unsigned) (graph->glob[BUF_SIZE_POS]),
			   sizeof(DATATYPE));
	}
    if (len > (unsigned) (graph->glob[MAX_DATALEN_POS]))
	graph->glob[MAX_DATALEN_POS] = (unsigned long) len;
    return(node_addr);
}

/***************************************************************************
*
* Function:  GphFndNode
* Purpose:   This routine searches for a marking in a reachability graph
*            of a net.
* Arguments: graph - the reachability graph
*            node - an encoded marking
*            back - returns the address of the latest incoming arrow
*                   if such arrow exists, and (unsigned long) -1L otherwise
*                   (exception: a NULL pointer returns nothing)
*                   If back is not NULL, the predecessor record
*                   of a found node is updated.
* Returns:   the address of a node which has the marking and is in
*            in the graph, or -1L if such node does not exist
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE3(unsigned long, GphFndNode,
    GRAPH *, graph,
    DATATYPE *, node,
    unsigned long *, back)
{
    unsigned long hash_value;
    unsigned long node_addr;
    int cmp_result;
    DREF *dref;
    
    if (((unsigned long *) NULL) != back)
	*back = (unsigned long) -1L;
    if (_bdd(graph))
	{
	    node_addr = (unsigned long) -1L;
	    if (graph->balTree)
		{
		    if ((dref = (DREF *)
			 BalTrElement(graph->balTree, (char *) node)))
			node_addr = DrefGSlot(dref);
		}
	    if (graph->bdd)
		{
		    if (!BddFind(graph->bdd, node, &node_addr))
			node_addr = (unsigned long) -1L;
		}
	    if ((((unsigned long) -1L) != node_addr) &&
		(((unsigned long *) NULL) != back))
		{
		    NdsSkAbs(graph,
			     node_addr + graph->predecessors_pos *
			     sizeof(unsigned long));
		    NdsReLng(back, 1, graph);
		    if (*back != graph->glob[EOF_ARROWS_POS])
			{
			    NdsSkAbs(graph,
				     node_addr + graph->predecessors_pos *
				     sizeof(unsigned long));
			    NdsWrLng(graph->glob + EOF_ARROWS_POS, 1, graph);
			}
		    if (*back >= graph->glob[EOF_ARROWS_POS])
			*back = (unsigned long) -1L;
		}
	    return(node_addr);
	}
    hash_value = Hash(node, graph->glob[HASH_SIZE_POS],
		      graph->glob[HASH_COEFF_POS]);
    node_addr = HashGet(graph->hash, hash_value);
    while (node_addr != (unsigned long) -1L)
	{
	    if (((unsigned long *) NULL) != back)
		{
		    NdsSkAbs(graph,
			     node_addr + graph->predecessors_pos *
			     sizeof(unsigned long));
		    NdsReLng(graph->latest_rec + graph->predecessors_pos,
			     4, graph);
		}
	    else
		{
		    NdsSkAbs(graph,
			     node_addr +  graph->left_child_pos *
			     sizeof(unsigned long));
		    NdsReLng(graph->latest_rec + graph->left_child_pos,
			     3, graph);
		}
	    graph->latest_node = (unsigned long) -1L;
	    NdsReData(graph->latest_contents,
		      graph->latest_rec[graph->datalen_pos],
		      graph);
	    if (!(cmp_result = DataCompare(node, graph->latest_contents)))
		{
		    if (((unsigned long *) NULL) != back)
			{
			    *back =
				graph->latest_rec[graph->predecessors_pos];
			    if (*back != graph->glob[EOF_ARROWS_POS])
				{
				    NdsSkAbs(graph,
					     node_addr +
					     graph->predecessors_pos *
					     sizeof(unsigned long));
				    NdsWrLng(graph->glob + EOF_ARROWS_POS,
					     1, graph);
				}
			    if (*back >= graph->glob[EOF_ARROWS_POS])
				*back = (unsigned long) -1L;
			}
		    break;
		}
	    if (cmp_result < 0)
		node_addr = graph->latest_rec[graph->left_child_pos];
	    else
		node_addr = graph->latest_rec[graph->right_child_pos];
	}
    return(node_addr);
}

/***************************************************************************
*
* Function:  GphErase
* Purpose:   This routine removes nodes from a reachability graph
*            of a net, starting from a given node. However, it is
*            not allowed to remove any node that has been handled by
*            GphUpdate or GphAddTermNode.
* Arguments: graph - the reachability graph
*            start_addr - the address of the first node to be removed
*            buffer - during the removal of a node, an encoding of the
*                     marking of the node is kept in this buffer
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE3(void, GphErase,
    GRAPH *, graph,
    unsigned long, start_addr,
    DATATYPE *, buffer)
{
    unsigned long hash_value;
    unsigned long del_addr;
    unsigned long node_addr;
    unsigned long prev_addr;
    unsigned long left;
    unsigned long right;
    unsigned long len;
    unsigned child;
    
    if (_bdd(graph) || (start_addr >= graph->glob[EOF_NODES_POS]))
	{
	    (void) fprintf(_out, "\n");
	    (void) fflush(_out);
	    (void) fprintf(_err, "\nIllegal GphErase call:\n");
	    if (_bdd(graph))
		(void) fprintf(_err, "BDD-based graph generation\n");
	    else
		(void) fprintf(_err, "start_addr %lu >= end %lu\n",
			       start_addr, graph->glob[EOF_NODES_POS]);
	    (void) fflush(_err);
	    abort();
	}
    graph->latest_node = (unsigned long) -1L;
    del_addr = start_addr;
    do
	{
	    NdsSkAbs(graph, del_addr +
		     SUCCESSOR_COUNT_POS * sizeof(unsigned long));
	    NdsReLng(graph->latest_rec + SUCCESSOR_COUNT_POS,
		     graph->datalen_pos - SUCCESSOR_COUNT_POS + 1,
		     graph);
	    if (graph->latest_rec[SUCCESSOR_COUNT_POS] != (unsigned long) -1L)
		{
		    (void) fprintf(_out, "\n");
		    (void) fflush(_out);
		    (void) fprintf(_err,
				   "\nIllegal GphErase call:\nAn attempt to delete a handled node\n");
		    (void) fflush(_err);
		    abort();
		}
	    if (_dfs(graph) && (!_s_l_e_e_p(graph)) && (del_addr == start_addr))
		graph->glob[FIRST_UNHANDLED_POS] =
		    graph->latest_rec[graph->next_lower_pos];
	    left = graph->latest_rec[graph->left_child_pos];
	    right = graph->latest_rec[graph->right_child_pos];
	    len = graph->latest_rec[graph->datalen_pos];
	    NdsReData(buffer,
		      graph->latest_rec[graph->datalen_pos],
		      graph);
	    hash_value = Hash(buffer, graph->glob[HASH_SIZE_POS],
			      graph->glob[HASH_COEFF_POS]);
	    node_addr = HashGet(graph->hash, hash_value);
	    if (node_addr == del_addr)
		HashPut(graph->hash, hash_value,
			(right == (unsigned long) -1L) ? left : right);
	    else
		{
		    do
			{
			    NdsSkAbs(graph, 
				     node_addr +
				     graph->left_child_pos
				     * sizeof(unsigned long));
			    NdsReLng(graph->latest_rec + graph->left_child_pos,
				     3, graph);
			    NdsReData(graph->latest_contents,
				      graph->latest_rec[graph->datalen_pos],
				      graph);
			    prev_addr = node_addr;
			    child = (DataCompare(buffer,
						 graph->latest_contents) < 0)
				? (graph->left_child_pos)
				    : (graph->right_child_pos);
			    node_addr = graph->latest_rec[child];
			} while (node_addr != del_addr);
		    NdsSkAbs(graph, 
			     prev_addr + child * sizeof(unsigned long));
		    NdsWrLng((right == (unsigned long) -1L)
			     ? (&left) : (&right), 1, graph);
		}
	    if ((left != (unsigned long) -1L) &&
		(right != (unsigned long) -1L))
		{
		    prev_addr = right;
		    do
			{
			    NdsSkAbs(graph, prev_addr +
				     graph->left_child_pos
				     * sizeof(unsigned long));
			    NdsReLng(&node_addr, 1, graph);
			    if (node_addr != (unsigned long) -1L)
				prev_addr = node_addr;
			} while (node_addr != (unsigned long) -1L);
		    NdsSkAbs(graph, prev_addr +
			     graph->left_child_pos * sizeof(unsigned long));
		    NdsWrLng(&left, 1, graph);
		}
	    (graph->glob[NODE_COUNT_POS])--;
	    del_addr += len + graph->num_node_fields * sizeof(unsigned long);
	} while (del_addr != graph->glob[EOF_NODES_POS]);
    graph->glob[EOF_NODES_POS] = start_addr;
}

/***************************************************************************
*
* Function:  GphPutArrow
* Purpose:   This routine inserts an arrow into a reachability
*            graph of a net.
* Arguments: graph - the reachability graph
*            arrow - an encoded transition instance corresponding to
*                    the arrow
*            len - the length of the arrow string
*            ordinal - the order number of the arrow for its source node
*            source_node_addr - the address of the source node of the arrow
*            target_node_addr - the address of the target node of the arrow
*            aw_next_addr - the address of the most recently inserted
*                           arrow of the source node (-1L if such arrow
*                           does not exist) (this parameter is used
*                           only if the freeExpansion parameter of
*                           GphCreate was TRUE)
*            aw_back_addr - the address of the latest incoming arrow
*                           of the target node if such arrow exists,
*                           and (unsigned long) -1L otherwise
*            invisible - TRUE if the instance is invisible,
*                        and FALSE otherwise
* Returns:   the address of the inserted arrow
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE9(unsigned long, GphPutArrow,
    GRAPH *, graph,
    DATATYPE *, arrow,
    unsigned, len,
    unsigned long, ordinal,
    unsigned long, source_node_addr,
    unsigned long, target_node_addr,
    unsigned long, aw_next_addr,
    unsigned long, aw_back_addr,
    BOOL, invisible)
{
    if (target_node_addr >= graph->glob[EOF_NODES_POS])
	{
	    (void) fprintf(_out, "\n");
	    (void) fflush(_out);
	    (void) fprintf(_err,
			   "\nIllegal node address for GphPutArrow: %lu\n",
			   target_node_addr);
	    (void) fflush(_err);
	    abort();
	}
    graph->arrow_rec[AW_ORDINAL_POS] = (ordinal << 2) | (invisible ? 2L : 1L);
    graph->arrow_rec[AW_BACK_POS] = aw_back_addr;
    graph->arrow_rec[SOURCE_POS] = source_node_addr;
    graph->arrow_rec[TARGET_POS] = target_node_addr;
    if (_frexpansion(graph))
	graph->arrow_rec[AW_NEXT_POS] = aw_next_addr;
    graph->arrow_rec[graph->aw_datalen_pos] = (unsigned long) len;
    AwsSkAbs(graph, graph->glob[EOF_ARROWS_POS]);
    AwsWrLng(graph->arrow_rec, graph->num_arrow_fields, graph);
    AwsWrData(arrow, len, graph);
    (graph->glob[ARROW_COUNT_POS])++;
    if (len >= (unsigned) (graph->glob[BUF_SIZE_POS]))
	{
	    graph->glob[BUF_SIZE_POS] =
		(unsigned long) ((len >> 1) + len + 1);
	    Deallocate((char *) (graph->latest_contents));
	    graph->latest_contents = (DATATYPE *)
		AllocArray((unsigned) (graph->glob[BUF_SIZE_POS]),
			   sizeof(DATATYPE));
	}
    if (len > (unsigned) graph->glob[MAX_DATALEN_POS])
	graph->glob[MAX_DATALEN_POS] = (unsigned long) len;
    graph->glob[LAST_ARROW_POS] = graph->glob[EOF_ARROWS_POS];
    graph->glob[EOF_ARROWS_POS] +=
	len * sizeof(DATATYPE) +
	    graph->num_arrow_fields * sizeof(unsigned long);
    return(graph->glob[LAST_ARROW_POS]);
}

static unsigned gphLen;      /* The length of a string of encoded
                              * information. */
static DATATYPE *gphPointer; /* A pointer to a position in a string of
                              * encoded information. */

/***************************************************************************
*
* Function:  LenInc
* Purpose:   This routine calculates how many bytes are needed to
*            encode *a. The result is added to gphLen.
* Arguments: a - a pointer to a number
* Returns:
* See also:  ValuePut
*
****************************************************************************
*/
FS1(BOOL, LenInc,
    unsigned long *, a)
{
    register unsigned long b;

    b = *a;
    IncLen(b, gphLen);
    return(FALSE);
}

/***************************************************************************
*
* Function:  ValuePut
* Purpose:   This routine encodes *a. The result is stored in addresses
*            pointed by gphPointer.
* Arguments: a - a pointer to a number
* Returns:
* See also:  LenInc
*
****************************************************************************
*/
FS1(BOOL, ValuePut,
    unsigned long *, a)
{
    register unsigned long b;

    b = *a;
    PutValue(b, gphPointer);
    return(FALSE);
}

/***************************************************************************
*
* Function:  GphUpdate
* Purpose:   This routine updates the successor information of a node
*            in a reachability graph of a net.
* Arguments: graph - the reachability graph
*            node_addr - the address of the node
*            arrow_count - the number of immediate successor arrows
*                          of the node
*            first_arrow_addr - the address of the first immediate
*                               successor arrow of the node
*            sleepSet - a sleep set associated with the node (used
*                       only if the sleep set method is used)
* Returns:
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE5(void, GphUpdate,
    GRAPH *, graph,
    unsigned long, node_addr,
    unsigned long, arrow_count,
    unsigned long, first_arrow_addr,
    BALTREE *, sleepSet)
{
    unsigned long *a;

    if (node_addr >= graph->glob[EOF_NODES_POS])
	{
	    (void) fprintf(_out, "\n");
	    (void) fflush(_out);
	    (void) fprintf(_err,
			   "\nIllegal node address for GphUpdate: %lu\n", 
			   node_addr);
	    (void) fflush(_err);
	    abort();
	}
    if ((arrow_count == (unsigned long) -1L) ||
	((!_s_l_e_e_p(graph)) && (arrow_count == 0L)))
	{
	    (void) fprintf(_out, "\n");
	    (void) fflush(_out);
	    (void) fprintf(_err,
			   "\nIllegal arrow count for GphUpdate: %lu\n",
			   arrow_count);
	    (void) fflush(_err);
	    abort();
	}
    if ((arrow_count != 0L) &&
	(first_arrow_addr >= graph->glob[EOF_ARROWS_POS]))
	{
	    (void) fprintf(_out, "\n");
	    (void) fflush(_out);
	    (void) fprintf(_err,
			   "\nIllegal arrow address for GphUpdate: %lu\n",
			   first_arrow_addr);
	    (void) fflush(_err);
	    abort();
	}
    graph->latest_node = (unsigned long) -1L;
    graph->latest_rec[SUCCESSOR_COUNT_POS] = arrow_count;  
    graph->latest_rec[FOLLOWERS_POS] = first_arrow_addr;
    a = graph->latest_rec + graph->next_lower_pos;
    if (_s_l_e_e_p(graph))
	{
	    if (sleepSet && (BalTrGElCount(sleepSet) != 0L))
		{
		    NdsSkAbs(graph, node_addr +
			     graph->next_lower_pos * sizeof(unsigned long));
		    NdsReLng(a, 1, graph);
		}
	    else
		*a = (unsigned long) -2L;
	}
    NdsSkAbs(graph, node_addr + SUCCESSOR_COUNT_POS * sizeof(unsigned long));
    if (_s_l_e_e_p(graph) && (*a > graph->glob[EOF_ARROWS_POS]))
	{
	    if (sleepSet && (BalTrGElCount(sleepSet) != 0L))
		*a = graph->glob[EOF_ARROWS_POS];
	    NdsWrLng(graph->latest_rec + SUCCESSOR_COUNT_POS,
		     graph->next_lower_pos
		     - SUCCESSOR_COUNT_POS + 1, graph);
	}
    else
	NdsWrLng(graph->latest_rec + SUCCESSOR_COUNT_POS, 2, graph);
    if (_s_l_e_e_p(graph) && sleepSet && (BalTrGElCount(sleepSet) != 0L))
	{
	    AwsSkAbs(graph, *a);
	    gphLen = 0;
	    (void) BalTrApply(sleepSet, LenInc);
	    if (gphLen >= (unsigned) (graph->glob[BUF_SIZE_POS]))
		{
		    graph->glob[BUF_SIZE_POS] =
			(unsigned long) ((gphLen >> 1) + gphLen + 1);
		    Deallocate((char *) (graph->latest_contents));
		    graph->latest_contents = (DATATYPE *)
			AllocArray((unsigned) (graph->glob[BUF_SIZE_POS]),
				   sizeof(DATATYPE));
		}
	    AwsWrUns(&gphLen, (unsigned) 1, graph);
	    gphPointer = graph->latest_contents;
	    (void) BalTrApply(sleepSet, ValuePut);
	    AwsWrData(graph->latest_contents, gphLen, graph);
	    graph->glob[EOF_ARROWS_POS] +=
		sizeof(unsigned) + gphLen * sizeof(DATATYPE);
	}
}

/***************************************************************************
*
* Function:  GphAddTermNode
* Purpose:   This routine writes down that the given node in
*            a reachability graph of a net is terminal, i.e.
*            no transition is enabled at the corresponding marking or
*            some reject state occurs in the marking.
* Arguments: graph - the reachability graph
*            node_addr - 
* Returns:
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE2(void, GphAddTermNode,
    GRAPH *, graph,
    unsigned long, node_addr)
{
    if (node_addr >= graph->glob[EOF_NODES_POS])
	{
	    (void) fprintf(_out, "\n");
	    (void) fflush(_out);
	    (void) fprintf(_err,
			   "\nIllegal node address for GphAddTermNode: %lu\n", 
			   node_addr);
	    (void) fflush(_err);
	    abort();
	}
    graph->latest_node = (unsigned long) -1L;
    graph->latest_rec[SUCCESSOR_COUNT_POS] = 0L;
    graph->latest_rec[FOLLOWERS_POS] =
	graph->glob[LATEST_TERMNODE_POS];
    NdsSkAbs(graph,
	     node_addr + SUCCESSOR_COUNT_POS * sizeof(unsigned long));
    if (_s_l_e_e_p(graph))
	{
	    graph->latest_rec[graph->next_lower_pos]
		= (unsigned long) -2L;
	    NdsWrLng(graph->latest_rec + SUCCESSOR_COUNT_POS,
		     graph->next_lower_pos
		     - SUCCESSOR_COUNT_POS + 1, graph);
	}
    else
	NdsWrLng(graph->latest_rec + SUCCESSOR_COUNT_POS, 2, graph);
    graph->glob[LATEST_TERMNODE_POS] = node_addr;
    (graph->glob[TERMNODE_COUNT_POS])++;
}

/***************************************************************************
*
* Function:  GphPush
* Purpose:   This routine pushes a node together with a sleep set
*            onto a search stack. This routine is called only
*            if the sleep set method is used.
* Arguments: graph - the reachability graph
*            node_addr - the address of the node
*            sleepSet - a sleep set associated with the node
* Returns:
* See also:  all routines having a name of the form Gph[A-Za-z]+
*
****************************************************************************
*/
FE3(void, GphPush,
    GRAPH *, graph,
    unsigned long, node_addr,
    BALTREE *, sleepSet)
{
    register unsigned long a;
    
    if (!_s_l_e_e_p(graph))
	{
	    (void) fprintf(_out, "\n");
	    (void) fflush(_out);
	    (void) fprintf(_err, "\nIllegal GphPush call\n");
	    (void) fflush(_err);
	    abort();
	}
    if (node_addr >= graph->glob[EOF_NODES_POS])
	{
	    (void) fprintf(_out, "\n");
	    (void) fflush(_out);
	    (void) fprintf(_err,
			   "\nIllegal node address for GphPush: %lu\n", 
			   node_addr);
	    (void) fflush(_err);
	    abort();
	}
    StkSkAbs(graph, graph->glob[EOF_STACK_POS]);
    gphLen = 0;
    a = graph->glob[TOPMOST_POS];
    IncLen(a, gphLen);
    a = node_addr;
    IncLen(a, gphLen);
    if (sleepSet && (BalTrGElCount(sleepSet) != 0L))
	(void) BalTrApply(sleepSet, LenInc);
    if (gphLen >= (unsigned) (graph->glob[BUF_SIZE_POS]))
	{
	    graph->glob[BUF_SIZE_POS] =
		(unsigned long) ((gphLen >> 1) + gphLen + 1);
	    Deallocate((char *) (graph->latest_contents));
	    graph->latest_contents = (DATATYPE *)
		AllocArray((unsigned) (graph->glob[BUF_SIZE_POS]),
			   sizeof(DATATYPE));
	}
    StkWrUns(&gphLen, (unsigned) 1, graph);
    gphPointer = graph->latest_contents;
    a = graph->glob[TOPMOST_POS];
    PutValue(a, gphPointer);
    a = node_addr;
    PutValue(a, gphPointer);
    if (sleepSet && (BalTrGElCount(sleepSet) != 0L))
	(void) BalTrApply(sleepSet, ValuePut);
    StkWrData(graph->latest_contents, gphLen, graph);
    graph->glob[TOPMOST_POS] = graph->glob[EOF_STACK_POS];
    graph->glob[EOF_STACK_POS] +=
	sizeof(unsigned) + gphLen * sizeof(DATATYPE);
}
