/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  inst.c
*
* Part of: PROD's program "gphgen"
*
* Authors: Marko Rauhamaa and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This module contains routines needed when a reachability graph
*     of a net is constructed without unfolding the net.
*     The following are needed for understanding: Valmari's CAV '93 paper
*     and the article of Courcoubetis, Vardi, Wolper & Yannakakis in
*     Formal Methods in System Design 1 (1992) 2/3, pp. 275-288.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include <stdio.h>
#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "graph.h"
#include "mark.h"
#include "encode.h"
#include "inst.h"
#include "ptnet.h"

/***************************************************************************
*
* Function:  PeeCons
* Purpose:   This routine constructs a peer. A peer is needed for saving
*            an enabled transition instance for a later precedence check.
*            Peers are not needed if there is only the default independent
*            precedence class 0.
* Arguments: trIndex - the transition
*            instance - the values of the variables of the transition
*            buffer - may contain an encoding of the instance
* Returns:   the constructed peer
* See also:  PeeCopy, PeeDiscard, PeeDump
*
****************************************************************************
*/
FE4(PEER *, PeeCons,
    unsigned, trIndex,
    DOMAIN *, instance,
    DATATYPE *, buffer,
    unsigned, visible)
{
    PEER *peer;
    register unsigned i;

    peer = PeeAlloc();
    peer->flag = 0;
    if (visible)
	PeeSVisible(peer);
    peer->trIndex = trIndex;
    i = IngGProfile(trIndex);
    if (Buchi())
	i += 2;
    peer->instance = (DOMAIN *) AllocArray(i, sizeof(DOMAIN));
    while (i--)
	peer->instance[i] = instance[i];
    if (buffer)
	peer->buffer = DataCopy(buffer);
    else
	peer->buffer = NULL;
    return(peer);
}

/***************************************************************************
*
* Function:  PeeCopy
* Purpose:   This routine makes a copy of a peer.
* Arguments: peer - the peer to be copied
* Returns:   the copy
* See also:  PeeCons, PeeDiscard, PeeDump
*
****************************************************************************
*/
FE1(PEER *, PeeCopy,
    PEER *, peer)
{
    PEER *copy;
    register unsigned i;

    copy = PeeAlloc();
    copy->flag = peer->flag;
    copy->trIndex = peer->trIndex;
    i = IngGProfile(peer->trIndex);
    copy->instance = (DOMAIN *) AllocArray(i, sizeof(DOMAIN));
    while (i--)
	copy->instance[i] = peer->instance[i];
    if (peer->buffer)
	copy->buffer = DataCopy(peer->buffer);
    else
	copy->buffer = NULL;
    return(copy);
}

/***************************************************************************
*
* Function:  PeeDiscard
* Purpose:   This routine removes a peer from the memory.
* Arguments: peer - the peer
* Returns:
* See also:  PeeCons, PeeCopy, PeeDump
*
****************************************************************************
*/
FE1(void, PeeDiscard,
    PEER *, peer)
{
    Deallocate((char *) (peer->instance));
    if (peer->buffer)
	Deallocate((char *) (peer->buffer));
    Deallocate((char *) peer);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  PeeDump
* Purpose:   This routine displays a peer.
* Arguments: peer - the peer
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  PeeCons, PeeCopy, PeeDiscard
*
****************************************************************************
*/
FE2(void, PeeDump,
    PEER *, peer,
    unsigned, depth)
{
    unsigned i, dimension;

    DumpIndent(depth);
    (void) fprintf(_err, "PEER trIndex %u flag %u\n",
		   peer->trIndex, peer->flag);
    DumpIndent(depth + 1);
    (void) fprintf(_err, "instance");
    dimension = IngGProfile(peer->trIndex);
    for (i = 0; i < dimension; i++)
	{
	    (void) fprintf(_err, " ");
	    (void) fprintf(_err, DOMFORMAT, peer->instance[i]);
	}
    (void) fprintf(_err, "\n");
    DumpIndent(depth + 1);
    (void) fprintf(_err, "buffer\n");
    if (peer->buffer)
	DataDump(peer->buffer, depth + 2);
    (void) fflush(_err);
}
#endif

/* A collection of routines for peers to be used by
 * list routines of dynam.c and table routines of ptrbl.c. */
TOOLS peeTools =
{ (CFUNC) PeeCopy, PeeDiscard
#ifdef DEBUG
, PeeDump
#endif
 };

static MON *doMon;            /* the monitor state of a marking */
static MON *bigMon;           /* the monitor state of the marking at which
                               * transitions are being fired */
static unsigned doU;          /* a short-term auxiliary variable */
static unsigned long doCount; /* the number of those immediate successor
                               * arrows of a node that have not been
                               * considered yet */

/***************************************************************************
*
* Function:  IllTrace
* Purpose:   This on-the-fly verification routine handles a possible case
*            where an error of the form specified by the reject states of
*            the tester is found.
* Arguments:
* Returns:
* See also:  IllStabFail, IllDivTrace, IllInfTrace, DoOnTheFly, DoInfinite,
*            PtC1 (ptnet.c), ExOnTheFly (graph.c)
*
****************************************************************************
*/
FE0(void, IllTrace)
{
    doMon = (MON *)
	BalTrElement(monitors, (char *)
		     MultupAField(MarFirst(StaGMarking(theFolState,
						       MonPlInd())),
			       (unsigned) 0));
    if ((!doMon) || !MonReject(doMon))
	return;
    if (theAwAddr != (unsigned long) -1L)
	{
	    GphUpdate(theGraph, PthEnd(nowPath), theAwCount, theAwFirst, 
		      (BALTREE *) NULL);
	    GphIncGenCount(theGraph);
	    if (theDots)
		DotOutput();
	}
    if (Divergence())
	BalTrDestroy(theNodeTree, FALSE);
    if (Ara())
	LTSEnd();
    (void) fprintf(_out, "\nReject state reached\n");
    if (PthLast(nowPath))
	{
	    LstDestroy(PthGSteps(nowPath), TRUE);
	    PthSSteps(nowPath, LstCreate(&stpTools));
	}
    PthSEnd(nowPath, theNdAddr);
    ExOnTheFly(theGraph, nowPath, FALSE);
}

/***************************************************************************
*
* Function:  IllStabFail
* Purpose:   This on-the-fly verification routine handles a case where
*            an error of the form specified by the deadlock monitor
*            states of the tester is found.
* Arguments:
* Returns:
* See also:  IllTrace, IllDivTrace, IllInfTrace, DoOnTheFly, DoInfinite,
*            PtC2 (ptnet.c), ExOnTheFly (graph.c)
*
****************************************************************************
*/
FS0(void, IllStabFail)
{
    if (Divergence())
	BalTrDestroy(theNodeTree, FALSE);
    if (Ara())
	LTSEnd();
    (void) fprintf(_out, "\nDeadlock reached\n");
    if (PthLast(nowPath))
	{
	    LstDestroy(PthGSteps(nowPath), TRUE);
	    PthSSteps(nowPath, LstCreate(&stpTools));
	}
    ExOnTheFly(theGraph, nowPath, FALSE);
}

/***************************************************************************
*
* Function:  IllDivTrace
* Purpose:   This on-the-fly verification routine handles a possible case
*            where an error of the form specified by the livelock monitor
*            states of the tester is found.
* Arguments:
* Returns:
* See also:  IllTrace, IllStabFail, IllInfTrace, DoOnTheFly, DoInfinite,
*            PtC3 (ptnet.c), ExOnTheFly (graph.c)
*
****************************************************************************
*/
FE1(void, IllDivTrace,
    BOOL, isExpanding)
{
    if (isExpanding && (theNdAddr != PthEnd(thePath)) &&
	!BalTrElement(theNodeTree, (char *) &theNdAddr))
	return;
    if (isExpanding)
	{
	    GphUpdate(theGraph, PthEnd(thePath), theAwCount, theAwFirst,
		      (BALTREE *) NULL);
	    GphIncGenCount(theGraph);
	    if (theDots)
		DotOutput();
	}
    BalTrDestroy(theNodeTree, FALSE);
    PthInsLast(thePath, theAwAddr, theNdAddr);
    if (Ara())
	LTSEnd();
    (void) fprintf(_out, "\nLivelock reached\n");
    ExOnTheFly(theGraph, thePath, TRUE);
}

/***************************************************************************
*
* Function:  IllInfTrace
* Purpose:   This on-the-fly verification routine handles a possible case
*            where an error of the form specified by the infinite path
*            monitor states of the tester is found.
* Arguments:
* Returns:
* See also:  IllTrace, IllStabFail, IllDivTrace, DoOnTheFly, DoInfinite,
*            PtC4 (ptnet.c), ExOnTheFly (graph.c)
*
****************************************************************************
*/
FE1(void, IllInfTrace,
    BOOL, isExpanding)
{
    if (Buchi() ? (theNdAddr != PthEnd(the2Path))
	: (theNdAddr != StpGWorld(PthLast(the2Path))))
	return;
    if (isExpanding)
	{
	    GphUpdate(theGraph, PthEnd(the3Path), theAwCount, theAwFirst,
		      (BALTREE *) NULL);
	    GphIncGenCount(theGraph);
	    if (theDots)
		DotOutput();
	}
    PthDiscard(the2Path);
    if (Divergence())
	BalTrDestroy(theNodeTree, FALSE);
    PthInsLast(the3Path, theAwAddr, theNdAddr);
    if (Ara())
	LTSEnd();
    (void) fprintf(_out, "\nIllegal infinite path found\n");
    ExOnTheFly(theGraph, the3Path, TRUE);
}

/***************************************************************************
*
* Function:  DoOnTheFly
* Purpose:   This on-the-fly verification routine searches for errors
*            of the form specified by the livelock monitor states,
*            reject states and deadlock monitor states of the tester.
* Arguments:
* Returns:
* See also:  IllTrace, IllStabFail, IllDivTrace, IllInfTrace, DoInfinite,
*            PtOnTheFly (ptnet.c)
*
****************************************************************************
*/
FE0(void, DoOnTheFly)
{
    nowPath = thePath;
    ClInfNow();
    bigMon = (MON *)
	BalTrElement(monitors, (char *)
		     MultupAField(MarFirst(StaGMarking(theState, MonPlInd())),
			       (unsigned) 0));
    if (bigMon && MonLivelock(bigMon))
	SetDivNow();
    else
	ClDivNow();
    while (TRUE)
	{
	    if ((!PthLast(thePath))
		||
		((Infinite()
		  ?
		  GphFresh(theGraph, PthEnd(thePath), (unsigned long) 1L)
		  :
		  (GphAwCount(theGraph, PthEnd(thePath)) ==
		   (unsigned long) -1L))
		 &&
		 GphAwInvisible(theGraph,
				StpGAwAddr(PthLast(thePath)))))
		{
		    if (Infinite())
			GphConsume(theGraph, PthEnd(thePath),
				   (unsigned long) 1L);
		    if (Infinite() &&
			((theAwCount =
			  GphAwCount(theGraph, PthEnd(thePath)))
			 != (unsigned long) -1L))
			{
			    if (!PthLast(thePath))
				for (doU = 0;
				     doU < MagGNofPlaces(); doU++)
				    MarDiscard(StaGMarking(theState,
							   doU));
			    if ((theAwCount != 0L) && DivNow())
				{
				    GphFlFirst(theGraph,
					       PthEnd(thePath),
					       &theAwFirst,
					       &theNdFirst);
				    theAwAddr = theAwFirst;
				    theNdAddr = theNdFirst;
				    doCount = theAwCount;
				    while (TRUE)
					{
					    if (((theNdAddr ==
						  PthEnd(thePath))
						 ||
						 BalTrElement(theNodeTree,
							      (char *)
							      &theNdAddr))
						&&
						GphAwInvisible(theGraph,
							       theAwAddr))
						IllDivTrace(FALSE);
					    if (--doCount == 0L)
						break;
					    GphFlNext(theGraph,
						      &theAwAddr,
						      &theNdAddr);
					}
				}
			}
		    else
			{
			    theAwCount = 0L;
			    theAwFirst = (unsigned long) -1L;
			    if (PthLast(thePath))
				StaDecode(GphNode(theGraph,
						  PthEnd(thePath)),
					  theState);
			    _Followers();
			    for (doU = 0; doU < MagGNofPlaces(); doU++)
				MarDiscard(StaGMarking(theState, doU));
			    if (theAwCount != 0L)
				GphUpdate(theGraph, PthEnd(thePath),
					  theAwCount, theAwFirst, 
					  (BALTREE *) NULL);
			    else
				GphAddTermNode(theGraph,
					       PthEnd(thePath));
			    GphIncGenCount(theGraph);
			    if (theDots)
				DotOutput();
			    if ((theAwCount == 0L) &&
				bigMon && MonDeadlock(bigMon))
				IllStabFail();
			}
		    if ((theAwCount != 0L) && DivNow())
			{
			    PthInsLast(thePath,
				       theAwFirst, theNdFirst);
			    (void)
				BalTrInsert(theNodeTree,
					    (char *)
					    StpAWorld(PthLast(thePath)),
					    (char *)
					    StpAWorld(PthLast(thePath)),
					    TRUE);
			    continue;
			}
		}
	    if (!PthLast(thePath))
		return;
	    if (FreeExpansion()
		?
		(GphAwOrdinal(theGraph, StpGAwAddr(PthLast(thePath))) != 0L)
		:
		(GphAwOrdinal(theGraph, StpGAwAddr(PthLast(thePath))) + 1L
		 < GphAwCount(theGraph, StpGWorld(PthLast(thePath)))))
		GphFlNext(theGraph, StpAAwAddr(PthLast(thePath)),
			  PthAEnd(thePath));
	    else
		{
		    (void) BalTrDelete(theNodeTree, (char *) NULL,
				       (char *) StpAWorld(PthLast(thePath)),
				       FALSE);
		    PthDelLast(thePath);
		    if (!PthLast(thePath))
			return;
		}
	}
}


/***************************************************************************
*
* Function:  DoInside
* Purpose:   This is an auxiliary routine for DoInfinite.
* Arguments:
* Returns:
* See also:  DoInfinite
*
****************************************************************************
*/
FS0(void, DoInside)
{
    nowPath = the3Path;
    StaDecode(GphNode(theGraph, StpGWorld(PthLast(the2Path))), theState);
    doMon = (MON *)
	BalTrElement(monitors, (char *)
		     MultupAField(MarFirst(StaGMarking(theState, MonPlInd())),
			       (unsigned) 0));
    for (doU = 0; doU < MagGNofPlaces(); doU++)
	MarDiscard(StaGMarking(theState, doU));
    if ((!doMon) || !MonInfinite(doMon))
	return;
    if (GphAwInvisible(theGraph, StpGAwAddr(PthLast(the2Path))))
	return;
    PthSEnd(the3Path, StpGWorld(PthLast(the2Path)));
    theAwAddr = StpGAwAddr(PthLast(the2Path));
    theNdAddr = PthEnd(the2Path);
    IllInfTrace(FALSE);
    PthInsLast(the3Path, StpGAwAddr(PthLast(the2Path)), PthEnd(the2Path));
    while (TRUE)
	{
	    if (GphFresh(theGraph, PthEnd(the3Path), (unsigned long) 4L))
		{
		    GphConsume(theGraph, PthEnd(the3Path), (unsigned long) 4L);
		    if ((theAwCount = GphAwCount(theGraph, PthEnd(the3Path)))
			!= (unsigned long) -1L)
			{
			    if (theAwCount != 0L)
				{
				    GphFlFirst(theGraph, PthEnd(the3Path),
					       &theAwFirst, &theNdFirst);
				    theAwAddr = theAwFirst;
				    theNdAddr = theNdFirst;
				    doCount = theAwCount;
				    while (TRUE)
					{
					    IllInfTrace(FALSE);
					    if (--doCount == 0L)
						break;
					    GphFlNext(theGraph, &theAwAddr,
						      &theNdAddr);
					}
				}
			}
		    else
			{
			    theAwCount = 0L;
			    theAwFirst = (unsigned long) -1L;
			    StaDecode(GphNode(theGraph, PthEnd(the3Path)),
				      theState);
			    _Followers();
			    if (StabFail())
				doMon = (MON *)
				    BalTrElement(monitors, (char *)
						 MultupAField(MarFirst(StaGMarking(theState,
										   MonPlInd())),
							      (unsigned) 0));
			    for (doU = 0; doU < MagGNofPlaces(); doU++)
				MarDiscard(StaGMarking(theState, doU));
			    if (theAwCount != 0L)
				GphUpdate(theGraph, PthEnd(the3Path),
					  theAwCount, theAwFirst,
					  (BALTREE *) NULL);
			    else
				GphAddTermNode(theGraph, PthEnd(the3Path));
			    GphIncGenCount(theGraph);
			    if (theDots)
				DotOutput();
			    if (StabFail() && (theAwCount == 0L) &&
				doMon && MonDeadlock(doMon))
				IllStabFail();
			}
		    if (theAwCount != 0L)
			{
			    PthInsLast(the3Path, theAwFirst, theNdFirst);
			    continue;
			}
		}
	    if (!PthLast(the3Path))
		return;
	    if (FreeExpansion()
		?
		(GphAwOrdinal(theGraph, StpGAwAddr(PthLast(the3Path))) != 0L)
		:
		(GphAwOrdinal(theGraph, StpGAwAddr(PthLast(the3Path))) + 1L
		 < GphAwCount(theGraph, StpGWorld(PthLast(the3Path)))))
		GphFlNext(theGraph, StpAAwAddr(PthLast(the3Path)),
			  PthAEnd(the3Path));
	    else
		{
		    PthDelLast(the3Path);
		    if (!PthLast(the3Path))
			return;
		}
	}
}

/***************************************************************************
*
* Function:  DoInfinite
* Purpose:   This on-the-fly verification routine searches for errors
*            of the form specified by the infinite path monitor states,
*            reject states and deadlock monitor states of the tester.
* Arguments:
* Returns:
* See also:  IllTrace, IllStabFail, IllDivTrace, IllInfTrace, DoOnTheFly,
*            DoInside, PtInfinite (ptnet.c)
*
****************************************************************************
*/
FE0(void, DoInfinite)
{
    nowPath = the2Path;
    ClDivNow();
    ClInfNow();
    if (GphFresh(theGraph, PthEnd(the2Path), (unsigned long) 2L))
	{
	    GphConsume(theGraph, PthEnd(the2Path), (unsigned long) 2L);
	    if ((theAwCount = GphAwCount(theGraph, PthEnd(the2Path)))
		!= (unsigned long) -1L)
		{
		    if (theAwCount != 0L)
			GphFlFirst(theGraph, PthEnd(the2Path),
				   &theAwFirst, &theNdFirst);
		}
	    else
		{
		    theAwCount = 0L;
		    theAwFirst = (unsigned long) -1L;
		    StaDecode(GphNode(theGraph, PthEnd(the2Path)), theState);
		    _Followers();
		    if (StabFail())
			doMon = (MON *)
			    BalTrElement(monitors, (char *)
					 MultupAField(MarFirst(StaGMarking(theState,
									MonPlInd())),
					       (unsigned) 0));
		    for (doU = 0; doU < MagGNofPlaces(); doU++)
			MarDiscard(StaGMarking(theState, doU));
		    if (theAwCount != 0L)
			GphUpdate(theGraph, PthEnd(the2Path),
				  theAwCount, theAwFirst, (BALTREE *) NULL);
		    else
			GphAddTermNode(theGraph, PthEnd(the2Path));
		    GphIncGenCount(theGraph);
		    if (theDots)
			DotOutput();
		    if (StabFail() && (theAwCount == 0L) &&
			doMon && MonDeadlock(doMon))
			IllStabFail();
		}
	    if (theAwCount != 0L)
		{
		    PthInsLast(the2Path, theAwFirst, theNdFirst);
		    return;
		}
	    else if (!PthLast(the2Path))
		{
		    PthSEnd(the2Path, (unsigned long) -1L);
		    return;
		}
	}
    else
	{
	    SetInfNow();
	    DoInside();
	    ClInfNow();
	}
    if (FreeExpansion()
	?
	(GphAwOrdinal(theGraph, StpGAwAddr(PthLast(the2Path))) != 0L)
	:
	(GphAwOrdinal(theGraph, StpGAwAddr(PthLast(the2Path))) + 1L
	 < GphAwCount(theGraph, StpGWorld(PthLast(the2Path)))))
	{
	    GphFlNext(theGraph, StpAAwAddr(PthLast(the2Path)),
		      PthAEnd(the2Path));
	    return;
	}
    PthDelLast(the2Path);
    if (!PthLast(the2Path))
	PthSEnd(the2Path, (unsigned long) -1L);
}

/***************************************************************************
*
* Function:  DoInner
* Purpose:   This is an auxiliary routine for DoBuchi.
* Arguments:
* Returns:
* See also:  DoBuchi
*
****************************************************************************
*/
FE0(void, DoInner)
{
    nowPath = the3Path;
    StaDecode(GphNode(theGraph, PthEnd(the2Path)), theState);
    doMon = (MON *)
	BalTrElement(monitors, (char *)
		     MultupAField(MarFirst(StaGMarking(theState, MonPlInd())),
			       (unsigned) 0));
    for (doU = 0; doU < MagGNofPlaces(); doU++)
	MarDiscard(StaGMarking(theState, doU));
    if ((!doMon) || !MonInfinite(doMon))
	return;
    PthSEnd(the3Path, PthEnd(the2Path));
    while (TRUE)
	{
	    if (GphFresh(theGraph, PthEnd(the3Path), (unsigned long) 4L))
		{
		    GphConsume(theGraph, PthEnd(the3Path), (unsigned long) 4L);
		    if ((theAwCount = GphAwCount(theGraph, PthEnd(the3Path)))
			!= (unsigned long) -1L)
			{
			    if (theAwCount != 0L)
				{
				    GphFlFirst(theGraph, PthEnd(the3Path),
					       &theAwFirst, &theNdFirst);
				    theAwAddr = theAwFirst;
				    theNdAddr = theNdFirst;
				    ptM = theAwCount;
				    while (TRUE)
					{
					    IllInfTrace(FALSE);
					    if (--ptM == 0L)
						break;
					    GphFlNext(theGraph, &theAwAddr,
						      &theNdAddr);
					}
				}
			}
		    else
			{
			    theAwCount = 0L;
			    theAwFirst = (unsigned long) -1L;
			    StaDecode(GphNode(theGraph, PthEnd(the3Path)),
				      theState);
			    _Followers();
			    for (doU = 0; doU < MagGNofPlaces(); doU++)
				MarDiscard(StaGMarking(theState, doU));
			    if (theAwCount != 0L)
				GphUpdate(theGraph, PthEnd(the3Path),
					  theAwCount, theAwFirst,
					  (BALTREE *) NULL);
			    else
				GphAddTermNode(theGraph, PthEnd(the3Path));
			    GphIncGenCount(theGraph);
			    if (theDots)
				DotOutput();
			}
		    if (theAwCount != 0L)
			{
			    PthInsLast(the3Path, theAwFirst, theNdFirst);
			    continue;
			}
		}
	    if (!PthLast(the3Path))
		return;
	    if (FreeExpansion()
		?
		(GphAwOrdinal(theGraph, StpGAwAddr(PthLast(the3Path))) != 0L)
		:
		(GphAwOrdinal(theGraph, StpGAwAddr(PthLast(the3Path))) + 1L
		 < GphAwCount(theGraph, StpGWorld(PthLast(the3Path)))))
		GphFlNext(theGraph, StpAAwAddr(PthLast(the3Path)),
			  PthAEnd(the3Path));
	    else
		{
		    PthDelLast(the3Path);
		    if (!PthLast(the3Path))
			return;
		}
	}
}

/***************************************************************************
*
* Function:  DoBuchi
* Purpose:   This on-the-fly verification routine checks if the so far
*            computed intersection of the reachability graph and the
*            Buchi automaton (the latter corresponding to the negation of
*            the formula given by the user) accepts any word.
* Arguments:
* Returns:
* See also:  DoInner, PtBuchi
*
****************************************************************************
*/
FE0(void, DoBuchi)
{
    nowPath = the2Path;
    ClDivNow();
    ClInfNow();
    if (GphFresh(theGraph, PthEnd(the2Path), (unsigned long) 2L))
	{
	    GphConsume(theGraph, PthEnd(the2Path), (unsigned long) 2L);
	    if ((theAwCount = GphAwCount(theGraph, PthEnd(the2Path)))
		!= (unsigned long) -1L)
		{
		    if (theAwCount != 0L)
			GphFlFirst(theGraph, PthEnd(the2Path),
				   &theAwFirst, &theNdFirst);
		}
	    else
		{
		    theAwCount = 0L;
		    theAwFirst = (unsigned long) -1L;
		    StaDecode(GphNode(theGraph, PthEnd(the2Path)), theState);
		    _Followers();
		    for (doU = 0; doU < MagGNofPlaces(); doU++)
			MarDiscard(StaGMarking(theState, doU));
		    if (theAwCount != 0L)
			GphUpdate(theGraph, PthEnd(the2Path),
				  theAwCount, theAwFirst, (BALTREE *) NULL);
		    else
			GphAddTermNode(theGraph, PthEnd(the2Path));
		    GphIncGenCount(theGraph);
		    if (theDots)
			DotOutput();
		}
	    if (theAwCount != 0L)
		{
		    PthInsLast(the2Path, theAwFirst, theNdFirst);
		    if (GphFresh(theGraph, PthEnd(the2Path),
				 (unsigned long) 2L))
			return;
		}
	}
    while (TRUE)
	{
	    if (PthLast(the2Path)
		&&
		(FreeExpansion()
		 ?
		 (GphAwOrdinal(theGraph, StpGAwAddr(PthLast(the2Path))) != 0L)
		 :
		 (GphAwOrdinal(theGraph, StpGAwAddr(PthLast(the2Path))) + 1L
		  < GphAwCount(theGraph, StpGWorld(PthLast(the2Path))))))
		{
		    GphFlNext(theGraph, StpAAwAddr(PthLast(the2Path)),
			      PthAEnd(the2Path));
		    if (GphFresh(theGraph, PthEnd(the2Path),
				 (unsigned long) 2L))
			return;
		}
	    else
		{
		    SetInfNow();
		    DoInner();
		    ClInfNow();
		    nowPath = the2Path;
		    if (PthLast(the2Path))
			PthDelLast(the2Path);
		    else
			PthSEnd(the2Path, (unsigned long) -1L);
		    return;
		}
	}
}

/***************************************************************************
*
* Function:  AraStep
* Purpose:   This routine writes an arc to an ARA LTS file. The arc
*            corresponds to an arrow in a reachability graph.
* Arguments: visible - if this is TRUE, the action of the arc is visible
* Returns:
* See also:  LTSBegin (ptelem.c), PtAraOutput (ptelem.c), LTSEnd (ptelem.c),
*            TrnFire (ptnet.c), InsOutput (encode.c)
*
****************************************************************************
*/
FE1(void, AraStep,
    unsigned, visible)
{
    DREF *dref;

    (void) fprintf(AraArcFile(), "%lu %lu ",
		   GphOrdinal(theGraph, PthEnd(nowPath)),
		   GphOrdinal(theGraph, theNdAddr));
    if (visible)
	{
	    dref = (DREF *) BalTrElement(AraTree(), (char *) EngBuffer());
	    if (!dref)
		{
		    dref = DrefCons(DataCopy(EngBuffer()),
				    BalTrGElCount(AraTree()) + 1L);
		    (void) BalTrInsert(AraTree(), (char *) dref,
				       (char *) EngBuffer(), TRUE);
		    InsOutput(DrefGSlot(dref), EngBuffer(), AraFile(), FALSE);
		}
	    (void) fprintf(AraArcFile(), "%lu\n", DrefGSlot(dref));
	}
    else
	(void) fprintf(AraArcFile(), "0\n");
}
