/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  net.c
*
* Part of: PROD's program prpp
*
* Authors: Marko Rauhamaa and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This module contains routines for the manipulation of high level
*     net elements.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "gphutil.h"
#include "mark.h"
#include "expr.h"
#include "net.h"

#include "util.h"

/* All these Cons, Copy, Discard, Dump, and Tools are trivial and
 * similar to those explained in detail in modules such as gphutil.c. */

FE6(NET *, NetCons,
    LIST *, places,
    LIST *, transes,
    LIST *, precs,
    unsigned, nofClasses,
    PLACE *, monPl,
    LIST *, atoms)
{
    NET *net;
    
    net = NetAlloc();
    net->_places = places;
    net->_transes = transes;
    net->_precs = precs;
    net->_nofClasses = nofClasses;
    net->_monPlace = monPl;
    net->_atoms = atoms;
    return(net);
}

FE1(void, NetDiscard,
    NET *, net)
{
    if (net->_atoms)
        LstDestroy(net->_atoms, TRUE);
    Deallocate((char *) net);
}

#ifdef DEBUG
FE2(void, NetDump,
    NET *, net,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "NET\n");
    DumpIndent(depth);
    (void) fprintf(_err, "Places\n");
    LstDump(net->_places, depth + 1);
    DumpIndent(depth);
    (void) fprintf(_err, "Transes\n");
    LstDump(net->_transes, depth + 1);
    DumpIndent(depth);
    (void) fprintf(_err, "Precs\n");
    LstDump(net->_precs, depth + 1);
    DumpIndent(depth);
    (void) fprintf(_err, "Number of precedence classes: %u\n", 
                   net->_nofClasses);
    if (net->_monPlace)
	{
	    DumpIndent(depth);
	    (void) fprintf(_err, "Monitor:\n");
            PlaDump(net->_monPlace, depth + 1);
	}
    if (net->_atoms)
	{
	    DumpIndent(depth);
	    (void) fprintf(_err, "Atoms:\n");
            LstDump(net->_atoms, depth + 1);
	}
    DumpIndent(depth);
    (void) fprintf(_err, "End of NET\n");
    (void) fflush(_err);
}
#endif

FE4(PLACE *, PlaCons,
    char *, name,
    LIST *, low,
    LIST *, high,
    MARKING *, marking)
{
    PLACE *place;

    place = PlaAlloc();
    place->_name = name;
    place->_arity = (unsigned) -1;
    place->_range._low = low;
    place->_range._high = high;
    place->_marking = marking;
#ifdef TRAIN
    PlaCDown(place);
    PlaCUp(place);
    PlaCCapacity(place);
    PlaCComplement(place);
#endif
    return(place);
}

FE1(PLACE *, PlaCopy,
    PLACE *, place)
{
    PLACE *copy;

    copy = (PLACE *) Allocate(sizeof(PLACE));
    copy->_name = StrCopy(place->_name);
    copy->_arity = place->_arity;
    copy->_range._low =
        place->_range._low ? LstCopy(place->_range._low, TRUE) : NULL;
    copy->_range._high =
        place->_range._high ? LstCopy(place->_range._high, TRUE) : NULL;
    copy->_marking = MarCopy(place->_marking);
#ifdef TRAIN
    PlaCDown(copy);
    PlaIDown(copy, PlaGDown(place));
    PlaCUp(copy);
    PlaIUp(copy, PlaGUp(place));
    PlaSCapacity(copy, PlaGCapacity(place));
    PlaCComplement(copy);
#endif
    return(copy);
}

FE1(void, PlaDiscard,
    PLACE *, place)
{
    Deallocate((char *) (place->_name));
    if (place->_range._low)
        LstDestroy(place->_range._low, TRUE);
    if (place->_range._high)
        LstDestroy(place->_range._high, TRUE);
    MarDiscard(place->_marking);
    Deallocate((char *) place);
}

#ifdef DEBUG
FE2(void, PlaDump,
    PLACE *, place,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "PLACE %s\n", place->_name);
    DumpIndent(depth + 1);
    (void) fprintf(_err, "Arity %u\n", place->_arity);
    if (place->_range._low)
	{
	    DumpIndent(depth + 1);
	    (void) fprintf(_err, "Lower bound\n");
            LstDump(place->_range._low, depth + 2);
	}
    if (place->_range._high)
	{
	    DumpIndent(depth + 1);
	    (void) fprintf(_err, "Higher bound\n");
            LstDump(place->_range._high, depth + 2);
	}
#ifdef TRAIN
    DumpIndent(depth + 1);
    (void) fprintf(_err, "Capacity %lu, down %lu, up %lu\n",
		   PlaGCapacity(place), PlaGDown(place), PlaGUp(place));
    if (PlaGComplement(place))
	{
	    DumpIndent(depth + 1);
	    (void) fprintf(_err, "Complement place %s\n",
			   PlaGName(PlaGComplement(place)));
	}
#endif
    DumpIndent(depth + 1);
    (void) fprintf(_err, "Marking\n");
    MarDump(place->_marking, depth + 2);
}
#endif

TOOLS plaTools =
{ (CFUNC) PlaCopy, PlaDiscard
#ifdef DEBUG
, PlaDump
#endif
 };

FE1(VARIABLE *, VarCons,
    char *, name)
{
    VARIABLE *variable;
    
    variable = (VARIABLE *) Allocate(sizeof(VARIABLE));
    variable->_name = name;
    variable->_label = NULL;
    variable->_low = (DOMAIN) MINDOMAIN;
    variable->_high = (DOMAIN) MAXDOMAIN;
    variable->_auxLow = (DOMAIN) MINDOMAIN;
    variable->_auxHigh = (DOMAIN) MAXDOMAIN;
    variable->_flag = 0;
    variable->_coeffs = NULL;
    return(variable);
}

FE1(VARIABLE *, VarCopy,
    VARIABLE *, variable)
{
    VARIABLE *copy;
    
    copy = (VARIABLE *) Allocate(sizeof(VARIABLE));
    if (variable->_name)
        copy->_name = StrCopy(variable->_name);
    else
        copy->_name = NULL;
    if (variable->_label)
        copy->_label = StrCopy(variable->_label);
    else
        copy->_label = NULL;
    copy->_low = variable->_low;
    copy->_high = variable->_high;
    copy->_auxLow = variable->_auxLow;
    copy->_auxHigh = variable->_auxHigh;
    copy->_flag = variable->_flag;
    if (variable->_coeffs)
        copy->_coeffs = LstCopy(variable->_coeffs, FALSE);
    else
        copy->_coeffs = NULL;
    return(copy);
}

FE1(void, VarDiscard,
    VARIABLE *, variable)
{
    if (variable->_name)
        Deallocate((char *) (variable->_name));
    if (variable->_label)
        Deallocate((char *) (variable->_label));
    if (variable->_coeffs)
        LstDestroy(variable->_coeffs, FALSE);
    Deallocate((char *) variable);
}

#ifdef DEBUG
FE2(void, VarDump,
    VARIABLE *, variable,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "VARIABLE%s%s%s%s flag %u\n",
                   variable->_name ? " name " : "",
                   variable->_name ? variable->_name : "",
                   variable->_label ? " label " : "",
                   variable->_label ? variable->_label : "", variable->_flag);
    DumpIndent(depth + 1);
    (void) fprintf(_err, "low ");
    (void) fprintf(_err, DOMFORMAT, variable->_low);
    (void) fprintf(_err, " high ");
    (void) fprintf(_err, DOMFORMAT, variable->_high);
    (void) fprintf(_err, " auxLow ");
    (void) fprintf(_err, DOMFORMAT, variable->_auxLow);
    (void) fprintf(_err, " auxHigh ");
    (void) fprintf(_err, DOMFORMAT, variable->_auxHigh);
    (void) fprintf(_err, "\n");
    DumpIndent(depth + 1);
    (void) fprintf(_err, "coeffs\n");
    if (variable->_coeffs)
        LstDump(variable->_coeffs, depth + 2);
    (void) fflush(_err);
}
#endif

TOOLS varTools =
{ (CFUNC) VarCopy, VarDiscard
#ifdef DEBUG
, VarDump
#endif
 };

FE1(TRANSITION *, TraCons,
    char *, name)
{
    TRANSITION *transition;
    
    transition = TraAlloc();
    transition->_name = name;
    transition->_kind = TK_NORM;
    transition->_fdWidth = 0;
    transition->_inputs = LstCreate(&arcTools);
    transition->_outputs = LstCreate(&arcTools);
    transition->_arcs = LstCreate(&arcTools);
    transition->_vars = LstCreate(&varTools);
    transition->_computation = NULL;
    transition->_gates = LstCreate(&symTools);
    transition->_range = placeLimits ? NULL : LstCreate(&dataTools);
    transition->_blocks = LstCreate(&lngTools);
    return(transition);
}

FE1(void, TraDiscard,
    TRANSITION *, transition)
{
    Deallocate((char *) (transition->_name));
    if (transition->_computation)
        LstDestroy(transition->_computation, TRUE);
    LstDestroy(transition->_inputs, TRUE);
    LstDestroy(transition->_outputs, TRUE);
    LstDestroy(transition->_arcs, FALSE);
    LstDestroy(transition->_vars, TRUE);
    LstDestroy(transition->_gates, TRUE);
    if (transition->_range)
        LstDestroy(transition->_range, TRUE);
    LstDestroy(transition->_blocks, FALSE);
    Deallocate((char *) transition);
}

#ifdef DEBUG
FE2(void, TraDump,
    TRANSITION *, transition,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "TRANSITION %s kind %u fdWidth %u\n",
                   transition->_name, transition->_kind, transition->_fdWidth);
    DumpIndent(depth + 1);
    (void) fprintf(_err, "Input arcs\n");
    LstDump(transition->_inputs, depth + 2);
    DumpIndent(depth + 1);
    (void) fprintf(_err, "Output arcs\n");
    LstDump(transition->_outputs, depth + 2);
    DumpIndent(depth + 1);
    (void) fprintf(_err, "Arcs\n");
    LstDump(transition->_arcs, depth + 2);
    DumpIndent(depth + 1);
    (void) fprintf(_err, "Vars\n");
    LstDump(transition->_vars, depth + 2);
    if (transition->_computation)
	{
	    DumpIndent(depth + 1);
	    (void) fprintf(_err, "Computation\n");
            LstDump(transition->_computation, depth + 2);
	}
    DumpIndent(depth + 1);
    (void) fprintf(_err, "Gates\n");
    LstDump(transition->_gates, depth + 2);
    if (transition->_range)
	{
	    DumpIndent(depth + 1);
            (void) fprintf(_err, "Range\n");
            LstDump(transition->_range, depth + 2);
	}
    DumpIndent(depth + 1);
    (void) fprintf(_err, "Blocks\n");
    LstDump(transition->_blocks, depth + 2);
}
#endif

/***************************************************************************
*
* Function:  TraAttInput
* Purpose:   This routine associates a new input arc with a transition.
* Arguments: transition - the transition
*            place - the source of the arc
*            symsum - the arc expression
* Returns:
* See also:  TraAttOutput
*
****************************************************************************
*/
FE3(ARC *, TraAttInput,
    TRANSITION *, transition,
    PLACE *, place,
    LIST *, symsum)
{
    ARC *arc;

    arc = ArcCons(place, symsum);
    (void) LstInsLast(TraGInputs(transition), (char *) arc);
    return(arc);
}

/***************************************************************************
*
* Function:  TraAttOutput
* Purpose:   This routine associates a new output arc with a transition.
* Arguments: transition - the transition
*            place - the target of the arc
*            symsum - the arc expression
* Returns:
* See also:  TraAttInput
*
****************************************************************************
*/
FE3(ARC *, TraAttOutput,
    TRANSITION *, transition,
    PLACE *, place,
    LIST *, symsum)
{
    ARC *arc;

    arc = ArcCons(place, symsum);
    (void) LstInsLast(TraGOutputs(transition), (char *) arc);
    return(arc);
}

TOOLS traTools =
{ (CFUNC) NULL, TraDiscard
#ifdef DEBUG
, TraDump
#endif
 };

FE1(SYMMULTUP *, SytCons,
    unsigned, arity)
{
    SYMMULTUP *symmultup;
    unsigned i;

    symmultup = SytAlloc();
    symmultup->_cffSym = NULL;
    symmultup->_multi = 0L;
    symmultup->_arity = arity;
    if (symmultup->_arity)
	{
            symmultup->_symtup = (SYMBOL **)
                AllocArray(symmultup->_arity, sizeof(SYMBOL *));
            for (i = 0; i < symmultup->_arity; i++)
                symmultup->_symtup[i] = NULL;
	}
    else
        symmultup->_symtup = NULL;
    symmultup->_label = NULL;
    symmultup->_number = 0;
    return(symmultup);
}

FE1(SYMMULTUP *, SytCopy,
    SYMMULTUP *, symmultup)
{
    SYMMULTUP *copy;
    unsigned i;

    copy = SytAlloc();
    if (symmultup->_cffSym)
        copy->_cffSym = SymCopy(symmultup->_cffSym);
    else
        copy->_cffSym = NULL;
    copy->_multi = symmultup->_multi;
    copy->_arity = symmultup->_arity;
    if (symmultup->_arity)
	{
            copy->_symtup = (SYMBOL **)
                AllocArray(symmultup->_arity, sizeof(SYMBOL *));
            for (i = 0; i < symmultup->_arity; i++)
		{
                    if (symmultup->_symtup[i])
                        copy->_symtup[i] = SymCopy(symmultup->_symtup[i]);
		    else
                        copy->_symtup[i] = NULL;
		}
	}
    else
        copy->_symtup = NULL;
    if (symmultup->_label)
        copy->_label = StrCopy(symmultup->_label);
    else
        copy->_label = NULL;
    copy->_number = symmultup->_number;
    return(copy);
}

FE1(void, SytDiscard,
    SYMMULTUP *, symmultup)
{
    unsigned i;

    if (symmultup->_cffSym)
        SymDiscard(symmultup->_cffSym);
    for (i = 0; i < symmultup->_arity; i++)
	{
            if (symmultup->_symtup[i])
                SymDiscard(symmultup->_symtup[i]);
	}
    if (symmultup->_arity)
        Deallocate((char *) (symmultup->_symtup));
    if (symmultup->_label)
        Deallocate((char *) (symmultup->_label));
    Deallocate((char *) symmultup);
}

#ifdef DEBUG
FE2(void, SytDump,
    SYMMULTUP *, symmultup,
    unsigned, depth)
{
    unsigned i;

    DumpIndent(depth);
    (void) fprintf(_err, "SYMMULTUP%s%s number %u multi %lu arity %u\n",
                   symmultup->_label ? " label " : "",
                   symmultup->_label ? symmultup->_label : "",
                   symmultup->_number, symmultup->_multi, symmultup->_arity);
    DumpIndent(depth + 1);
    (void) fprintf(_err, "cffSym\n");
    if (symmultup->_cffSym)
        SymDump(symmultup->_cffSym, depth + 2);
    for (i = 0; i < symmultup->_arity; i++)
	{
	    DumpIndent(depth + 1);
	    (void) fprintf(_err, "symtup %u\n", i);
            if (symmultup->_symtup[i])
                SymDump(symmultup->_symtup[i], depth + 2);
	}
    DumpIndent(depth);
    (void) fprintf(_err, "End of SYMMULTUP\n");
}
#endif

TOOLS sytTools =
{ (CFUNC) SytCopy, SytDiscard
#ifdef DEBUG
, SytDump
#endif
 };

FE2(SYMBOL *, SymCons,
    unsigned, kind,
    char *, any)
{
    SYMBOL *symbol;
    unsigned long a = 0L;
    DOMAIN value;

    symbol = SymAlloc();
    symbol->_flag = 0;
    if (kind == SK_CONSTANT)
	{
	    SymSIsConstant(symbol);
            symbol->_uy._constant = any;
	    if ((!any) || (!*any) || (!isdigit((UCHAR) *any)) || (*any == '0'))
		return(symbol);
	    for (any++; *any; any++)
		if (!isdigit((UCHAR) *any))
		    return(symbol);
            (void) sscanf(symbol->_uy._constant, "%lu", &a);
	    value = (DOMAIN) a;
	    if (value
#if MINDOMAIN
		< (DOMAIN) 0L
#else
		> (DOMAIN) ~value
#endif
		)
		{
                    Deallocate((char *) (symbol->_uy._constant));
		    (void) sprintf(numStr, DOMXFORMAT, (UDOMAIN) value);
                    symbol->_uy._constant = StrCopy(numStr);
		}
	}
    else if (kind == SK_VARIABLE)
	{
	    SymSIsVariable(symbol);
            symbol->_uy._var = (VARIABLE *) Pvoid(any);
	}
    else
	{
	    SymSIsOperation(symbol);
            symbol->_uy._operation = (LIST *) Pvoid(any);
	}
    return(symbol);
}

FE1(SYMBOL *, SymCopy,
    SYMBOL *, symbol)
{
    SYMBOL *copy;

    copy = SymAlloc();
    copy->_flag = symbol->_flag;
    if (SymIsConstant(symbol))
        copy->_uy._constant = StrCopy(symbol->_uy._constant);
    else if (SymIsVariable(symbol))
        copy->_uy._var = symbol->_uy._var;
    else
        copy->_uy._operation = LstCopy(symbol->_uy._operation, TRUE);
    return(copy);
}

FE1(void, SymDiscard,
    SYMBOL *, symbol)
{
    if (SymIsConstant(symbol))
        Deallocate((char *) (symbol->_uy._constant));
    else if (SymIsOperation(symbol))
        LstDestroy(symbol->_uy._operation, TRUE);
    Deallocate((char *) symbol);
}

#ifdef DEBUG
FE2(void, SymDump,
    SYMBOL *, symbol,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "SYMBOL flag %u ", symbol->_flag);
    if (SymIsConstant(symbol))
        (void) fprintf(_err, "constant %s\n", symbol->_uy._constant);
    else if (SymIsVariable(symbol))
	{
	    (void) fprintf(_err, "var\n");
            if (symbol->_uy._var)
                VarDump(symbol->_uy._var, depth + 1);
	}
    else
	{
	    (void) fprintf(_err, "operation\n");
            LstDump(symbol->_uy._operation, depth + 1);
	}
    (void) fflush(_err);
}
#endif

TOOLS symTools =
{ (CFUNC) SymCopy, SymDiscard
#ifdef DEBUG
, SymDump
#endif
 };

/***************************************************************************
*
* Function:  SymSatisfied
* Purpose:   This routine checks if an integer-valued C expression can be
*            evaluated. It can be evaluated if and only if all transition
*            variables in it have been set.
* Arguments: symbol - tje expression
* Returns:   TRUE if the expression can be evaluated, and FALSE otherwise
* See also:
*
****************************************************************************
*/
FE1(BOOL, SymSatisfied,
    SYMBOL *, symbol)
{
    SYMBOL *symbol2;

    if (SymIsConstant(symbol))
	return(TRUE);
    else if (SymIsVariable(symbol))
        return(VarUnified(symbol->_uy._var) != 0);
    else
	{
            LSTFOR (symbol->_uy._operation, symbol2, SYMBOL)
		{
		    if (!SymSatisfied(symbol2))
			return(FALSE);
		}
	    return(TRUE);
	}
}

/***************************************************************************
*
* Function:  PrintSymbol
* Purpose:   This routine displays an integer-valued C expression.
* Arguments: f - the output file
*            symbol - the expression
* Returns:
* See also:
*
****************************************************************************
*/
FE2(void, PrintSymbol,
    FILE *, f,
    SYMBOL *, symbol)
{
    SYMBOL *symbol2;

    if (SymIsConstant(symbol))
	(void) fprintf(f, "%s", SymGConstant(symbol));
    else if (SymIsVariable(symbol))
	(void) fprintf(f, "%s", VarGLabel(SymGVariable(symbol)));
    else
	{
	    LSTFOR (SymGOperation(symbol), symbol2, SYMBOL)
		PrintSymbol(f, symbol2);
	}
}

FE2(ARC *, ArcCons,
    PLACE *, place,
    LIST *, symsum)
{
    ARC *arc;
    
    arc = ArcAlloc();
    arc->_place = place;
    arc->_symsum = symsum;
    arc->_number = 0;
    arc->_flag = 0;
    return(arc);
}

FE1(ARC *, ArcCopy,
    ARC *, arc)
{
    ARC *copy;
    
    copy = ArcAlloc();
    copy->_place = arc->_place;
    copy->_symsum = LstCopy(arc->_symsum, TRUE);
    copy->_number = arc->_number;
    copy->_flag = arc->_flag;
    return(copy);
}

FE1(void, ArcDiscard,
    ARC *, arc)
{
    LstDestroy(arc->_symsum, TRUE);
    Deallocate((char *) arc);
}

#ifdef DEBUG
FE2(void, ArcDump,
    ARC *, arc,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "ARC number %u flag %u\n", arc->_number, arc->_flag);
    DumpIndent(depth);
    (void) fprintf(_err, "Place %s\n", PlaGName(arc->_place));
    DumpIndent(depth);
    (void) fprintf(_err, "Arc expression\n");
    LstDump(arc->_symsum, depth + 1);
}
#endif

TOOLS arcTools =
{ (CFUNC) ArcCopy, ArcDiscard
#ifdef DEBUG
, ArcDump
#endif
 };

FE1(PRECEDENCE *, PreCons,
    unsigned, value)
{
    PRECEDENCE *precedence;
    
    precedence = PreAlloc();
    precedence->_value = value;
    precedence->_flag = 0;
    precedence->_succ = NULL;
    precedence->_pred = NULL;
    return(precedence);
}

FE1(PRECEDENCE *, PreCopy,
    PRECEDENCE *, precedence)
{
    PRECEDENCE *copy;
    
    copy = PreAlloc();
    copy->_value = precedence->_value;
    copy->_flag = precedence->_flag;
    if (copy->_succ)
        copy->_succ = LstCopy(precedence->_succ, FALSE);
    else
        copy->_succ = NULL;
    if (copy->_pred)
        copy->_pred = LstCopy(precedence->_pred, FALSE);
    else
        copy->_pred = NULL;
    return(copy);
}

FE1(void, PreDiscard,
    PRECEDENCE *, precedence)
{
    if (precedence->_succ)
        LstDestroy(precedence->_succ, FALSE);
    if (precedence->_pred)
        LstDestroy(precedence->_pred, FALSE);
    Deallocate((char *) precedence);
}

#ifdef DEBUG
FE2(void, PreDump,
    PRECEDENCE *, precedence,
    unsigned, depth)
{
    PRECEDENCE *p;
    VOID *cursor;
    
    DumpIndent(depth);
    (void) fprintf(_err, "PRECEDENCE value %u flag %u\n", 
                   precedence->_value, precedence->_flag);
    DumpIndent(depth + 1);
    (void) fprintf(_err, "succ\n");
    if (precedence->_succ)
	{
            cursor = LstGet(precedence->_succ);
            LSTFOR(precedence->_succ, p, PRECEDENCE)
		{
		    DumpIndent(depth + 2);
                    (void) fprintf(_err, "PRECEDENCE %u\n", p->_value);
		}
            LstSet(precedence->_succ, cursor);
	}
    DumpIndent(depth + 1);
    (void) fprintf(_err, "pred\n");
    if (precedence->_pred)
	{
            cursor = LstGet(precedence->_pred);
            LSTFOR(precedence->_pred, p, PRECEDENCE)
		{
		    DumpIndent(depth + 2);
                    (void) fprintf(_err, "PRECEDENCE %u\n", p->_value);
		}
            LstSet(precedence->_pred, cursor);
	}
    DumpIndent(depth);
    (void) fprintf(_err, "End of PRECEDENCE\n");
    (void) fflush(_err);
}
#endif

TOOLS preTools =
{ (CFUNC) PreCopy, PreDiscard
#ifdef DEBUG
, PreDump
#endif
 };

/***************************************************************************
*
* Function:  DomStr
* Purpose:
* Arguments: value - 
* Returns:
* See also:
*
****************************************************************************
*/
FE1(char *, DomStr,
    DOMAIN, value)
{
    UDOMAIN uval;

    uval = (UDOMAIN) value;
    (void) sprintf(numStr,
                   (uval > (UDOMAIN) ~uval) ? DOMXFORMAT : UDOMFORMAT,
                   uval);
    return(numStr);
}

/***************************************************************************
*
* Function:  UnsStr
* Purpose:
* Arguments: value - 
* Returns:
* See also:
*
****************************************************************************
*/
FE1(char *, UnsStr,
    unsigned, value)
{
    (void) sprintf(numStr,
                   (value > (unsigned) ~value) ? "0x%x" : "%u",
                   value);
    return(numStr);
}

/***************************************************************************
*
* Function:  UlongStr
* Purpose:
* Arguments: value - 
* Returns:
* See also:
*
****************************************************************************
*/
FE1(char *, UlongStr,
    unsigned long, value)
{
    (void) sprintf(numStr,
                   (value > (unsigned long) ~value) ? "0x%lx" : "%lu",
                   value);
    return(numStr);
}
