/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  prec.c
*
* Part of: PROD's program prpp
*
* Authors: Johannes Helander and Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This module contains routines for the manipulation of net
*     precedences. Unfortunately, the used approach is more or less
*     impractical, even on the level of abstraction of the reference
*     manual. For example, if the user wants to realize the ordinary
*     order of numbers, he/she has no better alternative than
*     #prec class1 << class2 << class3 << class4 << class5 << class6
*     #prec class6 << class7 << class8 << class9 << class10 << class11
*     etc.
*     When these declarations are expanded, the resulting number of
*     pairs is quadratic with respect to the number of classes.
*     Instead of commenting this module in detail, we suggest
*     that someone designs something strictly better from scratch
*     or simply discards the whole stuff. (KV's opinion: discard
*     the whole stuff because the static priority method does nothing
*     better than the stubborn set method.)
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include <stdio.h>
#include <string.h>
#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ppio.h"
#include "token.h"
#include "place.h"
#include "prpp.h"
#include "lexer.h"
#include "macro.h"
#include "parse.h"
#include "util.h"
#include "ptrbl.h"
#include "baltree.h"
#include "gphutil.h"
#include "mark.h"
#include "expr.h"
#include "net.h"
#include "prec.h"
#include "pplst.h"

typedef struct SHF
{
    PRECEDENCE *precedence;
    LIST *succ;
} SHF;

#define ShfAlloc() ((SHF *) Allocate(sizeof(SHF)))
#define ShfGPrecedence(shf) ((shf)->precedence)
#define ShfGValue(shf) (((shf)->precedence)->_value)
#define ShfGSucc(shf) ((shf)->succ)
#define ShfGPSucc(shf) (((shf)->precedence)->_succ)
#define ShfGPPred(shf) (((shf)->precedence)->_pred)
#define ShfSSucc(shf) ((shf)->succ = LstCreate(&preTools))
#define PRC_NOSHIFT 0
#define PRC_LEFTSHIFT 1
#define PRC_RIGHTSHIFT 2

FS1(SHF *, ShfCons,
    PRECEDENCE *, precedence)
{
    SHF *shf;

    shf = ShfAlloc();
    shf->precedence = precedence;
    shf->succ = NULL;
    return(shf);
}

FS1(SHF *, ShfCopy,
    SHF *, shf)
{
    SHF *copy;

    copy = ShfAlloc();
    copy->precedence = shf->precedence;
    if (shf->succ)
	copy->succ = LstCopy(shf->succ, FALSE);
    else
	copy->succ = NULL;
    return(copy);
}

FS1(void, ShfDiscard,
    SHF *, shf)
{
    if (shf->succ)
	LstDestroy(shf->succ, FALSE);
    Deallocate((char *) shf);
}

#ifdef DEBUG
FS2(void, ShfDump,
    SHF *, shf,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "SHF\n");
    DumpIndent(depth + 1);
    (void) fprintf(_err, "precedence\n");
    PreDump(shf->precedence, depth + 2);
    DumpIndent(depth + 1);
    (void) fprintf(_err, "succ\n");
    if (shf->succ)
	LstDump(shf->succ, depth + 2);
    DumpIndent(depth);
    (void) fprintf(_err, "End of SHF\n");
    (void) fflush(_err);
}
#endif

TOOLS shfTools =
{ (CFUNC) ShfCopy, ShfDiscard
#ifdef DEBUG
, ShfDump
#endif
 };

FS3(void, add_shf,
    PRECEDENCE *, p,
    PRECEDENCE *, q,
    LIST *, list)
{
    SHF *shf;
    PRECEDENCE *r;
    VOID *cursor;

    cursor = NULL;
    LSTFOR (list, shf, SHF)
	{
	    if (ShfGValue(shf) >= PreGValue(p))
		break;
	    cursor = LstGet(list);
	}
    if (shf && (ShfGValue(shf) == PreGValue(p)))
	{
	    if (ShfGSucc(shf))
		{
		    cursor = NULL;
		    LSTFOR (ShfGSucc(shf), r, PRECEDENCE)
			{
			    if (PreGValue(r) >= PreGValue(q))
				break;
			    cursor = LstGet(ShfGSucc(shf));
			}
		    if (r && (PreGValue(r) == PreGValue(q)))
			return;
		    LstSet(ShfGSucc(shf), cursor);
		}
	}
    else
	{
	    LstSet(list, cursor);
	    shf = ShfCons(p);
	    (void) LstInsNext(list, (char *) shf);
	}
    if (ShfGSucc(shf))
	(void) LstInsNext(ShfGSucc(shf), (char *) q);
    else
	{
	    ShfSSucc(shf);
	    (void) LstInsLast(ShfGSucc(shf), (char *) q);
	}
}

FS3(void, add_precedence,
    unsigned, low,
    unsigned, high,
    int, mode)
{
    PRECEDENCE *plow, *phigh, *p;
    VOID *cursor;

    cursor = NULL;
    LSTFOR (PrecLst, plow, PRECEDENCE)
	{
	    if (PreGValue(plow) >= low)
		break;
	    cursor = LstGet(PrecLst);
	}
    if (plow && (PreGValue(plow) == low))
	{
	    if (PreGSucc(plow))
		{
		    cursor = NULL;
		    LSTFOR(PreGSucc(plow), phigh, PRECEDENCE)
			{
			    if (PreGValue(phigh) >= high)
				break;
			    cursor = LstGet(PreGSucc(plow));
			}
		    if (phigh && (PreGValue(phigh) == high))
			{
			    if (mode == PRC_LEFTSHIFT)
				add_shf(plow, phigh, LshLst);
			    else if (mode == PRC_RIGHTSHIFT)
				add_shf(phigh, plow, RshLst);
			    return;
			}
		    LstSet(PreGSucc(plow), cursor);
		}
	}
    else
	{
	    LstSet(PrecLst, cursor);
	    plow = PreCons(low);
	    (void) LstInsNext(PrecLst, (char *) plow);
	}
    if (low == high)
	{
	    phigh = plow;
	    if (PreGPred(phigh))
		{
		    cursor = NULL;
		    LSTFOR(PreGPred(phigh), p, PRECEDENCE)
			{
			    if (PreGValue(p) >= low)
				break;
			    cursor = LstGet(PreGPred(phigh));
			}
		    LstSet(PreGPred(phigh), cursor);
		}
	}
    else
	{
	    cursor = NULL;
	    LSTFOR (PrecLst, phigh, PRECEDENCE)
		{
		    if (PreGValue(phigh) >= high)
			break;
		    cursor = LstGet(PrecLst);
		}
	    if (phigh && (PreGValue(phigh) == high))
		{
		    if (PreGPred(phigh))
			{
			    cursor = NULL;
			    LSTFOR(PreGPred(phigh), p, PRECEDENCE)
				{
				    if (PreGValue(p) >= low)
					break;
				    cursor = LstGet(PreGPred(phigh));
				}
			    LstSet(PreGPred(phigh), cursor);
			}
		}
	    else
		{
		    LstSet(PrecLst, cursor);
		    phigh = PreCons(high);
		    (void) LstInsNext(PrecLst, (char *) phigh);
		}
	}
    if (PreGSucc(plow))
	(void) LstInsNext(PreGSucc(plow), (char *) phigh);
    else
	{
	    PreSSucc(plow);
	    (void) LstInsLast(PreGSucc(plow), (char *) phigh);
	}
    if (PreGPred(phigh))
	(void) LstInsNext(PreGPred(phigh), (char *) plow);
    else
	{
	    PreSPred(phigh);
	    (void) LstInsLast(PreGPred(phigh), (char *) plow);
	}
    if (mode == PRC_LEFTSHIFT)
	add_shf(plow, phigh, LshLst);
    else if (mode == PRC_RIGHTSHIFT)
	add_shf(phigh, plow, RshLst);
}

FE2(BOOL, IsPri,
    char *, name,
    unsigned *, priVal)
{
    REF *ref;
    
    ref = (REF *) BalTrElement(priTree, (char *) name);
    if (!ref)
	return(FALSE);
    *priVal = (unsigned) RefGSlot(ref);
    return(TRUE);
}

static unsigned maxPri = 1;

FE0(void, prec)
{
    PP_TOKEN *token;
    PP_TOKEN_TYPE type;
    unsigned x = 0;
    unsigned y = 0;
    REF *ref;

#if 0
    if (monName || verify)
	{
	    error("#tester or #verify makes #prec illegal");
	    token = get_token();
	    while (token && (token->type != NEWLINE))
		{
		    PpTokDiscard(token);
		    token = get_token();
		}
	    if (token)
		{
		    unget_token(token);
		    PpTokDiscard(token);
		}
	    return;
	}
#endif
    token = gTSkSpace();
    TOKEN_MATCH(token, NAME, "name in #prec");
    if (!IsPri(token->token, &x))
	{
	    (void) LstInsLast(PriLst, (char *) StrCopy(token->token));
	    ref = RefCons((char *) LstLast(PriLst), (unsigned long) maxPri);
	    (void) BalTrInsert(priTree, (char *) ref, (char *) RefGKey(ref),
			       TRUE);
	    x = maxPri;
	    maxPri++;
	}
    while (TRUE)
	{
	    PpTokDiscard(token);
	    token = gTSkSpace();
	    TOKEN_CHECK(token);
	    type = token->type;
	    if ((type != LESS_THAN) && (type != MORE_THAN) &&
		(type != SHIFT_LEFT) && (type != SHIFT_RIGHT))
		{
		    TOKEN_MATCH(token, NEVERMATCH, 
				"\"<\", \">\", \"<<\", or \">>\" in #prec");
		}
	    PpTokDiscard(token);
	    token = gTSkSpace();
	    TOKEN_MATCH(token, NAME, "name in #prec");
	    if (!IsPri(token->token, &y))
		{
		    (void) LstInsLast(PriLst, (char *) StrCopy(token->token));
		    ref = RefCons((char *) LstLast(PriLst),
				  (unsigned long) maxPri);
		    (void) BalTrInsert(priTree, (char *) ref,
				       (char *) RefGKey(ref), TRUE);
		    y = maxPri;
		    maxPri++;
		}
	    switch(type)
		{
		case LESS_THAN:
		    add_precedence(x, y, PRC_NOSHIFT);
		    break;
		case MORE_THAN:
		    add_precedence(y, x, PRC_NOSHIFT);
		    break;
		case SHIFT_LEFT:
		    add_precedence(x, y, PRC_LEFTSHIFT);
		    break;
		case SHIFT_RIGHT:
		    add_precedence(y, x, PRC_RIGHTSHIFT);
		    break;
		}
	    x = y;
	}
  completed: ;
  recover: ;
    if (token)
	PpTokDiscard(token);
}

FS1(void, forward_reach,
    PRECEDENCE *, p)
{
    PRECEDENCE *q;

    PreSet(p);
    if (!PreGSucc(p))
	return;
    LSTFOR (PreGSucc(p), q, PRECEDENCE)
	if (!PreMarked(q))
	    forward_reach(q);
}

FS1(void, backward_reach,
    PRECEDENCE *, p)
{
    PRECEDENCE *q;

    PreSet(p);
    if (!PreGPred(p))
	return;
    LSTFOR (PreGPred(p), q, PRECEDENCE)
	if (!PreMarked(q))
	    backward_reach(q);
}

FE0(void, shf_expansion)
{
   BOOL new;
   SHF *shf;
   PRECEDENCE *p, *q;
   VOID *cursor;

   new = TRUE;
   while (new)
       {
	   new = FALSE;
	   LSTFOR (LshLst, shf, SHF)
	       if (ShfGSucc(shf) && !LstEmpty(ShfGSucc(shf)))
		   {
		       LSTFOR (PrecLst, p, PRECEDENCE)
			   PreReset(p);
		       LSTFOR (ShfGSucc(shf), p, PRECEDENCE)
			   forward_reach(p);
		       LSTFOR (ShfGPSucc(shf), p, PRECEDENCE)
			   PreReset(p);
		       LSTFOR (PrecLst, p, PRECEDENCE)
			   if (PreMarked(p))
			       {
				   new = TRUE;
				   cursor = NULL;
				   LSTFOR (ShfGPSucc(shf), q, PRECEDENCE)
				       {
					   if (PreGValue(q) > PreGValue(p))
					       break;
					   cursor = LstGet(ShfGPSucc(shf));
				       }
				   LstSet(ShfGPSucc(shf), cursor);
				   (void) LstInsNext(ShfGPSucc(shf),
						     (char *) p);
				   if (PreGPred(p))
				       {
					   cursor = NULL;
					   LSTFOR (PreGPred(p), q, PRECEDENCE)
					       {
						   if (PreGValue(q) >
						       ShfGValue(shf))
						       break;
						   cursor =
						       LstGet(PreGPred(p));
					       }
					   LstSet(PreGPred(p), cursor);
					   (void) LstInsNext(PreGPred(p),
							     (char *)
							     ShfGPrecedence(shf));
				       }
				   else
				       {
					   PreSPred(p);
					   (void) LstInsLast(PreGPred(p),
							     (char *)
							     ShfGPrecedence(shf));
				       }
			       }
		   }
	   LSTFOR (RshLst, shf, SHF)
	       if (ShfGSucc(shf) && !LstEmpty(ShfGSucc(shf)))
		   {
		       LSTFOR (PrecLst, p, PRECEDENCE)
			   PreReset(p);
		       LSTFOR (ShfGSucc(shf), p, PRECEDENCE)
			   backward_reach(p);
		       LSTFOR (ShfGPPred(shf), p, PRECEDENCE)
			   PreReset(p);
		       LSTFOR (PrecLst, p, PRECEDENCE)
			   if (PreMarked(p))
			       {
				   new = TRUE;
				   cursor = NULL;
				   LSTFOR (ShfGPPred(shf), q, PRECEDENCE)
				       {
					   if (PreGValue(q) > PreGValue(p))
					       break;
					   cursor = LstGet(ShfGPPred(shf));
				       }
				   LstSet(ShfGPPred(shf), cursor);
				   (void) LstInsNext(ShfGPPred(shf),
						     (char *) p);
				   if (PreGSucc(p))
				       {
					   cursor = NULL;
					   LSTFOR (PreGSucc(p), q, PRECEDENCE)
					       {
						   if (PreGValue(q) >
						       ShfGValue(shf))
						       break;
						   cursor =
						       LstGet(PreGSucc(p));
					       }
					   LstSet(PreGSucc(p), cursor);
					   (void) LstInsNext(PreGSucc(p),
							     (char *)
							     ShfGPrecedence(shf));
				       }
				   else
				       {
					   PreSSucc(p);
					   (void) LstInsLast(PreGSucc(p),
							     (char *)
							     ShfGPrecedence(shf));
				       }
			       }
		   }
       }
}

FE0(void, pre_loops)
{
    PRECEDENCE *p, *q;
    VOID *cursor1, *cursor2;

    LSTFOR (PrecLst, p, PRECEDENCE)
	if (PreGPred(p))
	    {
		LSTFOR (PreGPred(p), q, PRECEDENCE)
		    if (PreGValue(q) >= PreGValue(p))
			break;
		if (q && (PreGValue(q) == PreGValue(p)))
		    PreSSelfLoop(p);
	    }
    LSTFOR (PrecLst, p, PRECEDENCE)
	if (PreGSelfLoop(p))
	    {
		cursor1 = LstGet(PrecLst);
		LSTFOR (PrecLst, q, PRECEDENCE)
		    PreReset(q);
		LSTFOR (PreGPred(p), q, PRECEDENCE)
		    {
			cursor2 = LstGet(PreGPred(p));
			backward_reach(q);
			LstSet(PreGPred(p), cursor2);
		    }
		LSTFOR (PrecLst, q, PRECEDENCE)
		    if (PreMarked(q) && !PreGSelfLoop(q))
			break;
		if (!q)
		    PreSEasy(p);
		LstSet(PrecLst, cursor1);
	    }
}
