/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  prpexpr.c
*
* Part of: PROD's program prpp
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This module contains routines for reading of formulas and
*     integer and marking expressions, for producing code that
*     evaluates non-constant expressions during the generation of
*     a reachability graph, and for the construction of a Buchi
*     automaton corresponding to the negation of a formula.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include <stdio.h>
#include <string.h>
#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ppio.h"
#include "token.h"
#include "output.h"
#include "place.h"
#include "prpp.h"
#include "lexer.h"
#include "macro.h"
#include "parse.h"
#include "util.h"
#include "ptrbl.h"
#include "baltree.h"
#include "gphutil.h"
#include "mark.h"
#include "encode.h"
#include "expr.h"
#include "net.h"
#include "prec.h"
#include "pplst.h"
#include "prpexpr.h"

LIST *prpExpList;
LIST *prpVerList;
FILE *pxFile;
unsigned long pxGotoNumber;
BOOL binaryMode;

#define F pxFile

/***************************************************************************
*
* Function:  MexDepth
* Purpose:   This routine computes the depth needed for _xprVal in order to
*            evaluate a given marking expression.
* Arguments: markexpr - the expression
* Returns:   the height
* See also:  XprDepth, all routines having a name of the form
*            MexO[A-Za-z]+ or XprO[A-Za-z]+
*
****************************************************************************
*/
FE1(unsigned, MexDepth,
    MARKEXPR *, markexpr)
{
    unsigned depth1, depth2;
    XRANGE *xrange;

    switch (MexKind(markexpr)) 
	{
	case MK_CONSTANT:
	case MK_PLACE:
	    return(0);
	case MK_TRANGE:
	    depth1 = 0;
	    TRGFOR (markexpr->um.trange, xrange) 
		{
		    depth2 = XprDepth(xrange->lower) + 1;
		    if (depth2 > depth1)
			depth1 = depth2;
		    depth2 = XprDepth(xrange->upper) + 1;
		    if (depth2 > depth1)
			depth1 = depth2;
		}
	    return(depth1);
	case MK_EXPRMARK:
	    depth1 = XprDepth(markexpr->um.em.esub) + 1;
	    depth2 = MexDepth(markexpr->um.em.msub) + 1;
	    return((depth1 > depth2) ? depth1 : depth2);
	case MK_MARKMARK:
	    depth1 = MexDepth(markexpr->um.mm.sub1m) + 1;
	    depth2 = MexDepth(markexpr->um.mm.sub2m) + 1;
	    return((depth1 > depth2) ? depth1 : depth2);
	}
    return(0);
}

/***************************************************************************
*
* Function:  MexEPlace
* Purpose:   This routine evaluates the initial marking of a place.
* Arguments: markexpr - a marking expression pointing to the place
* Returns:   the initial marking of the place
* See also:  MexEPlace (formula.c)
*
****************************************************************************
*/
FS1(MARKING *, MexEPlace,
    MARKEXPR *, markexpr)
{
    return(MarCopy(PlaGMarking((PLACE *)
			       LstNth(PlaceLst, markexpr->um.place + 1))));
}

/***************************************************************************
*
* Function:  MexRTrange
* Purpose:   This routine reduces a marking expression that points to
*            a range list. To avoid problems in code generation, the list
*            is evaluated only if some range is constantly empty or all
*            ranges are constant singletons.
* Arguments: markexpr - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            marking, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  MexRTrange (formula.c)
*
****************************************************************************
*/
FS1(unsigned, MexRTrange,
    MARKEXPR *, markexpr)
{
    unsigned rlevel1, rlevel2, rlevel3;
    XRANGE *xrange;
    MARKING *marking;
    BOOL equal;
    DOMAIN value1, value2;

    rlevel1 = RE_CONSTANT;
    equal = TRUE;
    TRGFOR (markexpr->um.trange, xrange) 
	{
	    rlevel2 = XprReduce(xrange->lower);
	    if (rlevel2 > rlevel1) 
		rlevel1 = rlevel2;
	    rlevel3 = XprReduce(xrange->upper);
	    if ((rlevel2 == RE_CONSTANT) && (rlevel3 == RE_CONSTANT))
		{
		    value1 = (DOMAIN) XprEval(xrange->lower);
		    value2 = (DOMAIN) XprEval(xrange->upper);
		    if (value1 > value2)
			{
			    TrgDiscard(markexpr->um.trange);
			    markexpr->operator = MO_CONSTANT;
			    markexpr->um.marking = MarCreate();
			    return(RE_CONSTANT);
			}
		    if (value1 < value2)
			equal = FALSE;
		}
	    if (rlevel3 > rlevel1) 
		rlevel1 = rlevel3;
	}
    if (equal && (rlevel1 == RE_CONSTANT))
	{
	    marking = MexEval(markexpr);
	    TrgDiscard(markexpr->um.trange);
	    markexpr->operator = MO_CONSTANT;
	    markexpr->um.marking = marking;
	    return(RE_CONSTANT);
	}
    return((rlevel1 == RE_CONSTANT) ? RE_LOCAL : rlevel1);
}

/***************************************************************************
*
* Function:  MexOConstant
* Purpose:   This routine produces code for evaluating a marking expression
*            that points to a constant marking.
* Arguments: markexpr - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form MexO[A-Za-z]+
*
****************************************************************************
*/
FS1(void, MexOConstant,
    MARKEXPR *, markexpr)
{
    MULTUP *multup;
    unsigned i;
    UDOMAIN uval;

    P(F, "    _xprMk = MarCreate();\n");
    MARFOR (markexpr->um.marking, multup)
	{
	    P(F, "    _tuple = TupCons((unsigned) %u);\n", MultupGArity(multup));
	    for (i = 0; i < MultupGArity(multup); i++)
		{
		    uval = (UDOMAIN) MultupGField(multup, i);
		    P(F, "    TupSField(_tuple, (unsigned) %u, (DOMAIN) ", i);
		    P(F, (uval > (UDOMAIN) ~uval) ? DOMXFORMAT : UDOMFORMAT,
		      uval);
		    P(F, ");\n");
		}
	    P(F, "    MarInsLast(_xprMk, MultupCons((unsigned long) %lu, _tuple));\n",
	      MultupGCount(multup));
	}
}

/***************************************************************************
*
* Function:  MexOPlace
* Purpose:   This routine produces code for evaluating a marking expression
*            that points to a place.
* Arguments: markexpr - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form MexO[A-Za-z]+
*
****************************************************************************
*/
FS1(void, MexOPlace,
    MARKEXPR *, markexpr)
{
    P(F, "    _xprMk = MarCopy(StaGMarking(_evalState, (unsigned) %u));\n",
      markexpr->um.place);
}

/***************************************************************************
*
* Function:  MexOTrange
* Purpose:   This routine produces code for evaluating a marking expression
*            that points to a range list.
* Arguments: markexpr - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form MexO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, MexOTrange,
    MARKEXPR *, markexpr,
    unsigned, depth)
{
    XRANGE *xrange;
    unsigned i, trgArity;
    unsigned long goNumber;

    goNumber = pxGotoNumber++;
    trgArity = TrgArity(markexpr->um.trange);
    P(F, "    _xprVal[%u].b.t0 = TupCons((unsigned) %u);\n    _xprVal[%u].b.t1 = TupCons((unsigned) %u);\n    _xprVal[%u].b.t2 = TupCons((unsigned) %u);\n    _xprVal[%u].b.mb = MarCreate();\n",
      depth, trgArity, depth, trgArity, depth, trgArity, depth);
    TRGNUM (markexpr->um.trange, xrange, i)
	{
	    XprOut(XrgGLower(xrange), depth + 1);
	    P(F, "    TupSField(_xprVal[%u].b.t0, (unsigned) %u, (DOMAIN) _xprDom);\n",
	      depth, i);
	    XprOut(XrgGUpper(xrange), depth + 1);
	    P(F, "    TupSField(_xprVal[%u].b.t1, (unsigned) %u, (DOMAIN) _xprDom);\n    if (TupGField(_xprVal[%u].b.t0, (unsigned) %u) >\n        TupGField(_xprVal[%u].b.t1, (unsigned) %u))\n        goto _q%lu;\n",
	      depth, i, depth, i, depth, i, goNumber);
	}
    for (i = 0; i < trgArity; i++)
	P(F, "    TupSField(_xprVal[%u].b.t2, (unsigned) %u,\n              TupGField(_xprVal[%u].b.t0, (unsigned) %u));\n_p%lu_%u: ;\n",
	  depth, i, depth, i, goNumber, i);
    P(F, "    MarInsLast(_xprVal[%u].b.mb,\n               MultupCons((unsigned long) 1L, TupCopy(_xprVal[%u].b.t2)));\n",
      depth, depth);
    while (i--)
	P(F, "    if (TupGField(_xprVal[%u].b.t2, (unsigned) %u) <\n        TupGField(_xprVal[%u].b.t1, (unsigned) %u))\n        {\n            TupIncField(_xprVal[%u].b.t2, (unsigned) %u);\n            goto _p%lu_%u;\n        }\n",
	  depth, i, depth, i, depth, i, goNumber, i);
    P(F, "_q%lu: ;\n    TupDiscard(_xprVal[%u].b.t0);\n    TupDiscard(_xprVal[%u].b.t1);\n    TupDiscard(_xprVal[%u].b.t2);\n    _xprMk = _xprVal[%u].b.mb;\n",
      goNumber, depth, depth, depth, depth);
}

/***************************************************************************
*
* Function:  MexOTimes
* Purpose:   This routine produces code for evaluating a marking expression
*            that represents a product.
* Arguments: markexpr - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form MexO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, MexOTimes,
    MARKEXPR *, markexpr,
    unsigned, depth)
{
    unsigned long goNumber;

    goNumber = pxGotoNumber++;
    XprOut(markexpr->um.em.esub, depth + 1);
    P(F, "    if (!_xprDom)\n        {\n            _xprMk = MarCreate();\n            goto _r%lu;\n        }\n    _xprVal[%u].d = _xprDom;\n",
      goNumber, depth);
    MexOut(markexpr->um.em.msub, depth + 1);
    P(F, "    _xprMk = MarProduct(_xprMk, (unsigned long) _xprVal[%u].d);\n_r%lu: ;\n",
      depth, goNumber);
}

/***************************************************************************
*
* Function:  MexOCartes
* Purpose:   This routine produces code for evaluating a marking expression
*            that represents a Cartesian product.
* Arguments: markexpr - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form MexO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, MexOCartes,
    MARKEXPR *, markexpr,
    unsigned, depth)
{
    unsigned long goNumber;

    goNumber = pxGotoNumber++;
    MexOut(markexpr->um.mm.sub1m, depth + 1);
    P(F, "    if (MarEmpty(_xprMk))\n        goto _r%lu;\n    _xprVal[%u].b.mb = _xprMk;\n",
      goNumber, depth);
    MexOut(markexpr->um.mm.sub2m, depth + 1);
    P(F, "    if (!MarEmpty(_xprMk))\n        (_tmpMk = MarCartes(_xprVal[%u].b.mb, _xprMk),\n         MarDiscard(_xprMk), _xprMk = _tmpMk);\n    MarDiscard(_xprVal[%u].b.mb);\n_r%lu: ;\n",
      depth, depth, goNumber);
}

/***************************************************************************
*
* Function:  MexOSum
* Purpose:   This routine produces code for evaluating a marking expression
*            that represents a sum.
* Arguments: markexpr - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form MexO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, MexOSum,
    MARKEXPR *, markexpr,
    unsigned, depth)
{
    MexOut(markexpr->um.mm.sub1m, depth + 1);
    P(F, "    _xprVal[%u].b.mb = _xprMk;\n", depth);
    MexOut(markexpr->um.mm.sub2m, depth + 1);
    P(F, "    (void) MarAdd(_xprMk, _xprVal[%u].b.mb);\n    MarDiscard(_xprVal[%u].b.mb);\n",
      depth, depth);
}

/***************************************************************************
*
* Function:  MexODiff
* Purpose:   This routine produces code for evaluating a marking expression
*            that represents a complementation.
* Arguments: markexpr - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form MexO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, MexODiff,
    MARKEXPR *, markexpr,
    unsigned, depth)
{
    unsigned long goNumber;

    goNumber = pxGotoNumber++;
    MexOut(markexpr->um.mm.sub1m, depth + 1);
    P(F, "    if (MarEmpty(_xprMk))\n        goto _r%lu;\n    _xprVal[%u].b.mb = _xprMk;\n",
      goNumber, depth);
    MexOut(markexpr->um.mm.sub2m, depth + 1);
    P(F, "    (void) MarSubtract(_xprVal[%u].b.mb, _xprMk);\n    MarDiscard(_xprMk);\n    _xprMk = _xprVal[%u].b.mb;\n_r%lu: ;\n",
      depth, depth, goNumber);
}

/***************************************************************************
*
* Function:  MexOSelect
* Purpose:   This routine produces code for evaluating a marking expression
*            that represents a submarking satisfying a condition.
* Arguments: markexpr - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form MexO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, MexOSelect,
    MARKEXPR *, markexpr,
    unsigned, depth)
{
    unsigned long goNumber;

    goNumber = pxGotoNumber++;
    MexOut(markexpr->um.em.msub, depth + 1);
    P(F, "    if (!MarBegin(_xprMk))\n        goto _c%lu;\n    _xprVal[%u].a.t = _xprTuple;\n    _xprVal[%u].a.ma = _xprMk;\n    _xprVal[%u].a.c = NULL;\n_b%lu: ;\n    _xprTuple = MultupGTuple(MarThis(_xprVal[%u].a.ma));\n",
      goNumber, depth, depth, depth, goNumber, depth);
    XprOut(markexpr->um.em.esub, depth + 1);
    P(F, "    if (!_xprDom)\n        {\n            MarSet(_xprVal[%u].a.ma, _xprVal[%u].a.c);\n            MarDelNext(_xprVal[%u].a.ma);\n        }\n    else\n        _xprVal[%u].a.c = MarGet(_xprVal[%u].a.ma);\n    if (MarStep(_xprVal[%u].a.ma))\n    goto _b%lu;\n    _xprTuple = _xprVal[%u].a.t;\n    _xprMk = _xprVal[%u].a.ma;\n_c%lu: ;\n",
      depth, depth, depth, depth, depth, depth, goNumber, depth, depth,
      goNumber);
}

/***************************************************************************
*
* Function:  MexOInters
* Purpose:   This routine produces code for evaluating a marking expression
*            that represents an intersection.
* Arguments: markexpr - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form MexO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, MexOInters,
    MARKEXPR *, markexpr,
    unsigned, depth)
{
    unsigned long goNumber;

    goNumber = pxGotoNumber++;
    MexOut(markexpr->um.mm.sub1m, depth + 1);
    P(F, "    if (MarEmpty(_xprMk))\n        goto _r%lu;\n    _xprVal[%u].b.mb = _xprMk;\n",
      goNumber, depth);
    MexOut(markexpr->um.mm.sub2m, depth + 1);
    P(F, "    if (!MarEmpty(_xprMk))\n        (void) MarIntersect(_xprMk, _xprVal[%u].b.mb);\n    MarDiscard(_xprVal[%u].b.mb);\n_r%lu: ;\n",
      depth, depth, goNumber);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  MexPConstant
* Purpose:   This routine displays a marking expression that points to
*            a constant marking.
* Arguments: markexpr - the expression
* Returns:
* See also:  all routines having a name of the form MexP[A-Za-z]+
*
****************************************************************************
*/
FS1(void, MexPConstant,
    MARKEXPR *, markexpr)
{
    MULTUP *multup;
    unsigned i;

    if (MarEmpty(markexpr->um.marking))
	{
	    P(F, "empty");
	    return;
	}
    if ((MarFirst(markexpr->um.marking) != MarLast(markexpr->um.marking)) ||
	(MultupGCount(MarFirst(markexpr->um.marking)) != 1L))
	P(F, "(");
    MARFOR (markexpr->um.marking, multup)
	{
	    if (multup != MarFirst(markexpr->um.marking))
		P(F, " + ");
	    if (MultupGCount(multup) != 1L)
		P(F, "%lu", MultupGCount(multup));
	    P(F, "<.");
	    for (i = 0; i < MultupGArity(multup); i++)
		{
		    if (i)
			P(F, ", ");
		    P(F, DOMFORMAT, MultupGField(multup, i));
		}
	    P(F, ".>");
	}
    if ((MarFirst(markexpr->um.marking) != MarLast(markexpr->um.marking)) ||
	(MultupGCount(MarFirst(markexpr->um.marking)) != 1L))
	P(F, ")");
}

/***************************************************************************
*
* Function:  MexPPlace
* Purpose:   This routine displays a marking expression that points to
*            a place.
* Arguments: markexpr - the expression
* Returns:
* See also:  all routines having a name of the form MexP[A-Za-z]+
* Purpose:
* Arguments: markexpr - 
* Returns:
* See also:
*
****************************************************************************
*/
FS1(void, MexPPlace,
    MARKEXPR *, markexpr)
{
    P(F, "%s", PlaGName((PLACE *) LstNth(PlaceLst, markexpr->um.place + 1)));
}

/***************************************************************************
*
* Function:  MexPTrange
* Purpose:   This routine displays a marking expression that points to
*            a range list.
* Arguments: markexpr - the expression
* Returns:
* See also:  all routines having a name of the form MexP[A-Za-z]+
*
****************************************************************************
*/
FS1(void, MexPTrange,
    MARKEXPR *, markexpr)
{
    XRANGE *xrange;

    P(F, "<.");
    TRGFOR (markexpr->um.trange, xrange)
	{
	    if (xrange != (XRANGE *) LstFirst(markexpr->um.trange))
		P(F, ", ");
	    XprPrint(XrgGLower(xrange));
	    P(F, "..");
	    XprPrint(XrgGUpper(xrange));
	}
    P(F, ".>");
}

/***************************************************************************
*
* Function:  MexPTimes
* Purpose:   This routine displays a marking expression that represents
*            a product.
* Arguments: markexpr - the expression
* Returns:
* See also:  all routines having a name of the form MexP[A-Za-z]+
*
****************************************************************************
*/
FS1(void, MexPTimes,
    MARKEXPR *, markexpr)
{
    P(F, "(");
    XprPrint(markexpr->um.em.esub);
    MexPrint(markexpr->um.em.msub);
    P(F, ")");
}

/***************************************************************************
*
* Function:  MexPBinary
* Purpose:   This routine displays a marking expression that represents
*            a binary operation on markings.
* Arguments: markexpr - the expression
* Returns:
* See also:  all routines having a name of the form MexP[A-Za-z]+
*
****************************************************************************
*/
FS2(void, MexPBinary,
    MARKEXPR *, markexpr,
    char *, op)
{
    P(F, "(");
    MexPrint(markexpr->um.mm.sub1m);
    P(F, " %s ", op);
    MexPrint(markexpr->um.mm.sub2m);
    P(F, ")");
}

/***************************************************************************
*
* Function:  MexPCartes
* Purpose:   This routine displays a marking expression that represents
*            a Cartesian product.
* Arguments: markexpr - the expression
* Returns:
* See also:  all routines having a name of the form MexP[A-Za-z]+
*
****************************************************************************
*/
FS1(void, MexPCartes,
    MARKEXPR *, markexpr)
{
    MexPBinary(markexpr, "*");
}

/***************************************************************************
*
* Function:  MexPSum
* Purpose:   This routine displays a marking expression that represents
*            a sum.
* Arguments: markexpr - the expression
* Returns:
* See also:  all routines having a name of the form MexP[A-Za-z]+
*
****************************************************************************
*/
FS1(void, MexPSum,
    MARKEXPR *, markexpr)
{
    MexPBinary(markexpr, "+");
}

/***************************************************************************
*
* Function:  MexPDiff
* Purpose:   This routine displays a marking expression that represents
*            a complementation.
* Arguments: markexpr - the expression
* Returns:
* See also:  all routines having a name of the form MexP[A-Za-z]+
*
****************************************************************************
*/
FS1(void, MexPDiff,
    MARKEXPR *, markexpr)
{
    MexPBinary(markexpr, "-");
}

/***************************************************************************
*
* Function:  MexPSelect
* Purpose:   This routine displays a marking expression that represents
*            a submarking satisfying a condition.
* Arguments: markexpr - the expression
* Returns:
* See also:  all routines having a name of the form MexP[A-Za-z]+
*
****************************************************************************
*/
FS1(void, MexPSelect,
    MARKEXPR *, markexpr)
{
    P(F, "(");
    MexPrint(markexpr->um.em.msub);
    P(F, ":");
    XprPrint(markexpr->um.em.esub);
    P(F, ")");
}

/***************************************************************************
*
* Function:  MexPInters
* Purpose:   This routine displays a marking expression that represents
*            an intersection.
* Arguments: markexpr - the expression
* Returns:
* See also:  all routines having a name of the form MexP[A-Za-z]+
*
****************************************************************************
*/
FS1(void, MexPInters,
    MARKEXPR *, markexpr)
{
    MexPBinary(markexpr, "&");
}
#endif

/***************************************************************************
*
* Function:  XprDepth
* Purpose:   This routine computes the depth needed for _xprVal in order to
*            evaluate a given integer expression.
* Arguments: expression - the expression
* Returns:   the depth
* See also:  MexDepth, all routines having a name of the form
*            MexO[A-Za-z]+ or XprO[A-Za-z]+
*
****************************************************************************
*/
FE1(unsigned, XprDepth,
    EXPRESSION *, expression)
{
    unsigned depth1, depth2;

    switch (XprKind(expression)) 
	{
	case EK_CONSTANT:
	case EK_NULLARY:
	    return(0);
	case EK_CARD:
	    return(MexDepth(expression->ue.me.msub1));
	case EK_MCOMP:
	    depth1 = MexDepth(expression->ue.me.msub1) + 1;
	    depth2 = MexDepth(expression->ue.me.msub2) + 1;
	    return((depth1 > depth2) ? depth1 : depth2);
	case EK_UNARY:
	    return(XprDepth(expression->ue.oe.sub1e));
	case EK_BINARY:
	    if ((expression->operator == EO_AND) ||
		(expression->operator == EO_OR))
		{
		    depth1 = 0;
		    depth2 = 0;
		}
	    else
		{
		    depth1 = 1;
		    depth2 = 1;
		}
	    depth1 += XprDepth(expression->ue.oe.sub1e);
	    depth2 += XprDepth(expression->ue.oe.sub2e);
	    return((depth1 > depth2) ? depth1 : depth2);
	case EK_TRINARY:
	    depth1 = XprDepth(expression->ue.oe.sub1e);
	    depth2 = XprDepth(expression->ue.oe.sub2e);
	    if (depth1 < depth2)
		depth1 = depth2;
	    depth2 = XprDepth(expression->ue.oe.sub3e);
	    return((depth1 > depth2) ? depth1 : depth2);
	}
    return(0);
}

/***************************************************************************
*
* Function:  XprEField
* Purpose:   This routine evaluates an integer expression that represents
*            a field of a known tuple.
* Arguments: expression - the expression
* Returns:   the field
* See also:  XprEField (formula.c)
*
****************************************************************************
*/
FS1(DOMAIN, XprEField,
    EXPRESSION *, expression)
{
    unsigned field;

    field = (unsigned) XprEval(expression->ue.oe.sub1e);
    if ((!FogGTuple()) || (TupGArity(FogGTuple()) <= field))
	{
	    if (errorFlag == NO_ERROR)
		{
		    (void) sprintf(errStr, "Illegal field reference: %u",
				   field);
		    error(errStr);
		    if (FogGTuple())
			TupDump(FogGTuple(), (unsigned) 0);
		    errorFlag = XPR_ERROR;
		}
	    return(0L);
	}
    return(TupGField(FogGTuple(), field));
}

/***************************************************************************
*
* Function:  XprEDiv
* Purpose:   This routine evaluates an integer expression that represents
*            the quotient of a division.
* Arguments: expression - the expression 
* Returns:   the quotient
* See also:  XprEDiv (formula.c)
*
****************************************************************************
*/
FS1(DOMAIN, XprEDiv,
    EXPRESSION *, expression)
{
    register DOMAIN divisor;

    divisor = XprEval(expression->ue.oe.sub2e);
    if (divisor)
	return(XprEval(expression->ue.oe.sub1e) / divisor);
    if (errorFlag == NO_ERROR)
	{
	    error("Division by zero");
	    errorFlag = XPR_ERROR;
	}
    return((DOMAIN) 0L);
}

/***************************************************************************
*
* Function:  XprERem
* Purpose:   This routine evaluates an integer expression that represents
*            the remainder of a division.
* Arguments: expression - the expression
* Returns:   the remainder
* See also:  XprERem (formula.c)
*
****************************************************************************
*/
FS1(DOMAIN, XprERem,
    EXPRESSION *, expression)
{
    register DOMAIN divisor;

    divisor = XprEval(expression->ue.oe.sub2e);
    if (divisor)
	return(XprEval(expression->ue.oe.sub1e) % divisor);
    if (errorFlag == NO_ERROR)
	{
	    error("Division by zero");
	    errorFlag = XPR_ERROR;
	}
    return((DOMAIN) 0L);
}

/***************************************************************************
*
* Function:  XprRDiv
* Purpose:   This routine reduces an integer expression that represents
*            the quotient of a division.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  XprRDiv (formula.c)
*
****************************************************************************
*/
FS1(unsigned, XprRDiv,
    EXPRESSION *, expression)
{
    unsigned rlevel1, rlevel2;
    DOMAIN value;

    rlevel1 = XprReduce(expression->ue.oe.sub1e);
    rlevel2 = XprReduce(expression->ue.oe.sub2e);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (rlevel2 == RE_CONSTANT)
		{
		    value = XprEval(expression);
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = value;
		    return(RE_CONSTANT);
		}
	    if (!XprEval(expression->ue.oe.sub1e))
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = (DOMAIN) 0L;
		    return(RE_CONSTANT);
		}
	    return(rlevel2);
	}
    if (rlevel2 == RE_CONSTANT)
	{
	    value = XprEval(expression->ue.oe.sub2e);
	    if (value == (DOMAIN) 0L)
		{
		    if (errorFlag == NO_ERROR)
			{
			    error("Division by zero");
			    errorFlag = XPR_ERROR;
			}
		    return(rlevel1);
		}
	    if (value == (DOMAIN) -1L)
		{
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_NEG;
		    return(rlevel1);
		}
	    if (value == (DOMAIN) 1L)
		{
		    XprDiscard(expression->ue.oe.sub2e);
		    XprForce(expression, expression->ue.oe.sub1e);
		    return(rlevel1);
		}
	    return(rlevel1);
	}
    return((rlevel2 > rlevel1) ? rlevel2 : rlevel1);
}

/***************************************************************************
*
* Function:  XprRRem
* Purpose:   This routine reduces an integer expression that represents
*            the remainder of a division.
* Arguments: expression - the expression
* Returns:   RE_CONSTANT if the result of the reduction is a constant
*            number, RE_LOCAL if the result is non-constant but can
*            be evaluated independently of context, and RE_GLOBAL if the
*            result is non-constant and needs context information in
*            evaluation.
* See also:  XprRRem (formula.c)
*
****************************************************************************
*/
FS1(unsigned, XprRRem,
    EXPRESSION *, expression)
{
    unsigned rlevel1, rlevel2;
    DOMAIN value;

    rlevel1 = XprReduce(expression->ue.oe.sub1e);
    rlevel2 = XprReduce(expression->ue.oe.sub2e);
    if (rlevel1 == RE_CONSTANT)
	{
	    if (rlevel2 == RE_CONSTANT) 
		{
		    value = XprEval(expression);
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = value;
		    return(RE_CONSTANT);
		}
	    if (!XprEval(expression->ue.oe.sub1e)) 
		{
		    XprDiscard(expression->ue.oe.sub1e);
		    XprDiscard(expression->ue.oe.sub2e);
		    expression->operator = EO_DOM;
		    expression->ue.value = (DOMAIN) 0L;
		    return(RE_CONSTANT);
		}
	    return(rlevel2);
	}
    if (rlevel2 == RE_CONSTANT)
	{
	    if ((XprEval(expression->ue.oe.sub2e) == (DOMAIN) 0L) &&
		(errorFlag == NO_ERROR))
		{
		    error("Division by zero");
		    errorFlag = XPR_ERROR;
		}
	    return(rlevel1);
	}
    return((rlevel2 > rlevel1) ? rlevel2 : rlevel1);
}

/***************************************************************************
*
* Function:  XprODom
* Purpose:   This routine produces code for evaluating an integer
*            expression that points to a constant number.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS1(void, XprODom,
    EXPRESSION *, expression)
{
    UDOMAIN uval;

    uval = (UDOMAIN) (expression->ue.value);
    P(F, "    _xprDom = (DOMAIN) ");
    P(F, (uval > (UDOMAIN) ~uval) ? DOMXFORMAT : UDOMFORMAT, uval);
    P(F, ";\n");
}

/***************************************************************************
*
* Function:  XprOField
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents a field of a known tuple.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOField,
    EXPRESSION *, expression,
    unsigned, depth)
{
    XprOut(expression->ue.oe.sub1e, depth);
    P(F, "    if ((!_xprTuple) || (TupGArity(_xprTuple) <= (unsigned) _xprDom))\n        {\n            (void) fprintf(_out, \"\\n\");\n            (void) fflush(_out);\n            (void) fprintf(_err, \"\\nIllegal field reference: %%u\\n\",\n                           (unsigned) _xprDom);\n            (void) fflush(_err);\n            if (_xprTuple)\n                TupDump(_xprTuple, (unsigned) 0);\n            abort();\n        }\n    _xprDom = TupGField(_xprTuple, (unsigned) _xprDom);\n");
}

/***************************************************************************
*
* Function:  XprOCard
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents the cardinality of a marking.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOCard,
    EXPRESSION *, expression,
    unsigned, depth)
{
    MexOut(expression->ue.me.msub1, depth);
    P(F, "    _xprDom = (DOMAIN) 0L;\n    (void) MarBegin(_xprMk);\n    while (MarThis(_xprMk))\n        {\n            _xprDom += MultupGCount(MarThis(_xprMk));\n            (void) MarStep(_xprMk);\n        }\n    MarDiscard(_xprMk);\n");
}

/***************************************************************************
*
* Function:  XprONeg
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents the opposite number of a number.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprONeg,
    EXPRESSION *, expression,
    unsigned, depth)
{
    XprOut(expression->ue.oe.sub1e, depth);
    /* Since unary minus is marginal, we can handle it in a pedantic way. */
    P(F, "    _xprDom = ((_xprDom == (DOMAIN) MINDOMAIN) ? _xprDom : ~(_xprDom - (DOMAIN) 1L));\n");
}

/***************************************************************************
*
* Function:  XprONot
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents a boolean negation.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprONot,
    EXPRESSION *, expression,
    unsigned, depth)
{
    XprOut(expression->ue.oe.sub1e, depth);
    P(F, "    _xprDom = !_xprDom;\n");
}

/***************************************************************************
*
* Function:  XprOBnot
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents a bitwise complementation.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOBnot,
    EXPRESSION *, expression,
    unsigned, depth)
{
    XprOut(expression->ue.oe.sub1e, depth);
    P(F, "    _xprDom = ~_xprDom;\n");
}

/***************************************************************************
*
* Function:  XprOTimes
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents a product.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOTimes,
    EXPRESSION *, expression,
    unsigned, depth)
{
    unsigned long goNumber;

    goNumber = pxGotoNumber++;
    XprOut(expression->ue.oe.sub1e, depth + 1);
    P(F, "    if (!_xprDom)\n        goto _r%lu;\n    _xprVal[%u].d = _xprDom;\n",
      goNumber, depth);
    XprOut(expression->ue.oe.sub2e, depth + 1);
    P(F, "    _xprDom *= _xprVal[%u].d;\n_r%lu: ;\n", depth, goNumber);
}

/***************************************************************************
*
* Function:  XprODiv
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents the quotient of a division.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprODiv,
    EXPRESSION *, expression,
    unsigned, depth)
{
    XprOut(expression->ue.oe.sub2e, depth + 1);
    P(F, "    if (!_xprDom)\n        {\n            (void) fprintf(_out, \"\\n\");\n            (void) fflush(_out);\n            P(_err, \"Division by zero\\n\");\n            (void) fflush(_err);\n            abort();\n        }\n    _xprVal[%u].d = _xprDom;\n", depth);
    XprOut(expression->ue.oe.sub1e, depth + 1);
    P(F, "    _xprDom /= _xprVal[%u].d;\n", depth);
}

/***************************************************************************
*
* Function:  XprORem
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents the remainder of a division.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprORem,
    EXPRESSION *, expression,
    unsigned, depth)
{
    XprOut(expression->ue.oe.sub2e, depth + 1);
    P(F, "    if (!_xprDom)\n        {\n            (void) fprintf(_out, \"\\n\");\n            (void) fflush(_out);\n            P(_err, \"Division by zero\\n\");\n            (void) fflush(_err);\n            abort();\n        }\n    _xprVal[%u].d = _xprDom;\n", depth);
    XprOut(expression->ue.oe.sub1e, depth + 1);
    P(F, "    _xprDom %%= _xprVal[%u].d;\n", depth);
}

/***************************************************************************
*
* Function:  XprOMinus
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents a subtraction.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOMinus,
    EXPRESSION *, expression,
    unsigned, depth)
{
    XprOut(expression->ue.oe.sub2e, depth + 1);
    P(F, "    _xprVal[%u].d = _xprDom;\n", depth);
    XprOut(expression->ue.oe.sub1e, depth + 1);
    P(F, "    _xprDom -= _xprVal[%u].d;\n", depth);
}

/***************************************************************************
*
* Function:  XprOPlus
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents a sum.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOPlus,
    EXPRESSION *, expression,
    unsigned, depth)
{
    XprOut(expression->ue.oe.sub1e, depth + 1);
    P(F, "    _xprVal[%u].d = _xprDom;\n", depth);
    XprOut(expression->ue.oe.sub2e, depth + 1);
    P(F, "    _xprDom += _xprVal[%u].d;\n", depth);
}

/***************************************************************************
*
* Function:  XprOLshift
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents a shifting of bits to the left.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOLshift,
    EXPRESSION *, expression,
    unsigned, depth)
{
    unsigned long goNumber;

    goNumber = pxGotoNumber++;
    XprOut(expression->ue.oe.sub1e, depth + 1);
    P(F, "    if (!_xprDom)\n        goto _r%lu;\n    _xprVal[%u].d = _xprDom;\n",
      goNumber, depth);
    XprOut(expression->ue.oe.sub2e, depth + 1);
    P(F, "    _xprDom = (_xprVal[%u].d << _xprDom);\n_r%lu: ;\n",
      depth, goNumber);
}

/***************************************************************************
*
* Function:  XprORshift
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents a shifting of bits to the right.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprORshift,
    EXPRESSION *, expression,
    unsigned, depth)
{
    unsigned long goNumber;

    goNumber = pxGotoNumber++;
    XprOut(expression->ue.oe.sub1e, depth + 1);
    P(F, "    if (!_xprDom)\n        goto _r%lu;\n    _xprVal[%u].d = _xprDom;\n",
      goNumber, depth);
    XprOut(expression->ue.oe.sub2e, depth + 1);
    P(F, "    _xprDom = (_xprVal[%u].d >> _xprDom);\n_r%lu: ;\n",
      depth, goNumber);
}

/***************************************************************************
*
* Function:  XprOLss
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents a "less than" comparison of numbers.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOLss,
    EXPRESSION *, expression,
    unsigned, depth)
{
    XprOut(expression->ue.oe.sub1e, depth + 1);
    P(F, "    _xprVal[%u].d = _xprDom;\n", depth);
    XprOut(expression->ue.oe.sub2e, depth + 1);
    P(F, "    _xprDom = (_xprVal[%u].d < _xprDom);\n", depth);
}

/***************************************************************************
*
* Function:  XprOLeq
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents a "less than or equal to"
*            comparison of numbers.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOLeq,
    EXPRESSION *, expression,
    unsigned, depth)
{
    XprOut(expression->ue.oe.sub1e, depth + 1);
    P(F, "    _xprVal[%u].d = _xprDom;\n", depth);
    XprOut(expression->ue.oe.sub2e, depth + 1);
    P(F, "    _xprDom = (_xprVal[%u].d <= _xprDom);\n", depth);
}

/***************************************************************************
*
* Function:  XprOEqu
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents an equality comparison of numbers.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOEqu,
    EXPRESSION *, expression,
    unsigned, depth)
{
    XprOut(expression->ue.oe.sub1e, depth + 1);
    P(F, "    _xprVal[%u].d = _xprDom;\n", depth);
    XprOut(expression->ue.oe.sub2e, depth + 1);
    P(F, "    _xprDom = (_xprVal[%u].d == _xprDom);\n", depth);
}

/***************************************************************************
*
* Function:  XprONeq
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents an inequality comparison
*            of numbers.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprONeq,
    EXPRESSION *, expression,
    unsigned, depth)
{
    XprOut(expression->ue.oe.sub1e, depth + 1);
    P(F, "    _xprVal[%u].d = _xprDom;\n", depth);
    XprOut(expression->ue.oe.sub2e, depth + 1);
    P(F, "    _xprDom = (_xprVal[%u].d != _xprDom);\n", depth);
}

/***************************************************************************
*
* Function:  XprOBand
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents a bitwise "and" operation.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOBand,
    EXPRESSION *, expression,
    unsigned, depth)
{
    unsigned long goNumber;

    goNumber = pxGotoNumber++;
    XprOut(expression->ue.oe.sub1e, depth + 1);
    P(F, "    if (!_xprDom)\n        goto _r%lu;\n    _xprVal[%u].d = _xprDom;\n",
      goNumber, depth);
    XprOut(expression->ue.oe.sub2e, depth + 1);
    P(F, "    _xprDom &= _xprVal[%u].d;\n_r%lu: ;\n", depth, goNumber);
}

/***************************************************************************
*
* Function:  XprOBxor
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents a bitwise "exclusive or" operation.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOBxor,
    EXPRESSION *, expression,
    unsigned, depth)
{
    XprOut(expression->ue.oe.sub1e, depth + 1);
    P(F, "    _xprVal[%u].d = _xprDom;\n", depth);
    XprOut(expression->ue.oe.sub2e, depth + 1);
    P(F, "    _xprDom ^= _xprVal[%u].d;\n", depth);
}

/***************************************************************************
*
* Function:  XprOBor
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents a bitwise "or" operation.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOBor,
    EXPRESSION *, expression,
    unsigned, depth)
{
    unsigned long goNumber;

    goNumber = pxGotoNumber++;
    XprOut(expression->ue.oe.sub1e, depth + 1);
    P(F, "    if (_xprDom == ~0L)\n        goto _r%lu;\n    _xprVal[%u].d = _xprDom;\n",
      goNumber, depth);
    XprOut(expression->ue.oe.sub2e, depth + 1);
    P(F, "    _xprDom |= _xprVal[%u].d;\n_r%lu: ;\n", depth, goNumber);
}

/***************************************************************************
*
* Function:  XprOAnd
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents a short-circuit boolean "and"
*            operation.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOAnd,
    EXPRESSION *, expression,
    unsigned, depth)
{
    unsigned long goNumber;

    goNumber = pxGotoNumber++;
    XprOut(expression->ue.oe.sub1e, depth);
    P(F, "    if (!_xprDom)\n        goto _r%lu;\n", goNumber);
    XprOut(expression->ue.oe.sub2e, depth);
    P(F, "    _xprDom = (_xprDom != (DOMAIN) 0L);\n_r%lu: ;\n", goNumber);
}

/***************************************************************************
*
* Function:  XprOOr
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents a short-circuit boolean "or"
*            operation.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOOr,
    EXPRESSION *, expression,
    unsigned, depth)
{
    unsigned long goNumber;

    goNumber = pxGotoNumber++;
    XprOut(expression->ue.oe.sub1e, depth);
    P(F, "    if (_xprDom != (DOMAIN) 0L)\n        goto _r%lu;\n", goNumber);
    XprOut(expression->ue.oe.sub2e, depth);
    P(F, "_r%lu: ;\n    _xprDom = (_xprDom != (DOMAIN) 0L);\n", goNumber);
}

/***************************************************************************
*
* Function:  XprOIf
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents an "if-then-else" operation.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOIf,
    EXPRESSION *, expression,
    unsigned, depth)
{
    unsigned long goNumber;

    goNumber = pxGotoNumber++;
    XprOut(expression->ue.oe.sub1e, depth);
    P(F, "    if (!_xprDom)\n        goto _r%lu;\n", goNumber);
    XprOut(expression->ue.oe.sub2e, depth);
    P(F, "    goto _s%lu;\n_r%lu: ;\n", goNumber, goNumber);
    XprOut(expression->ue.oe.sub3e, depth);
    P(F, "_s%lu: ;\n", goNumber);
}

/***************************************************************************
*
* Function:  XprOMls
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents a "less than" comparison of
*            markings.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOMls,
    EXPRESSION *, expression,
    unsigned, depth)
{
    unsigned long goNumber;

    goNumber = pxGotoNumber++;
    MexOut(expression->ue.me.msub2, depth + 1);
    P(F, "    if (MarEmpty(_xprMk))\n        {\n            _xprDom = (DOMAIN) 0L;\n            goto _r%lu;\n        }\n    _xprVal[%u].b.mb = _xprMk;\n",
      goNumber, depth);
    MexOut(expression->ue.me.msub1, depth + 1);
    P(F, "    _xprDom = (DOMAIN) MarLess(_xprMk, _xprVal[%u].b.mb);\n    MarDiscard(_xprVal[%u].b.mb);\n_r%lu: ;\n    MarDiscard(_xprMk);\n",
      depth, depth, goNumber);
}

/***************************************************************************
*
* Function:  XprOMlq
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents a "less than or equal to" comparison
*            of markings.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOMlq,
    EXPRESSION *, expression,
    unsigned, depth)
{
    unsigned long goNumber;

    goNumber = pxGotoNumber++;
    MexOut(expression->ue.me.msub1, depth + 1);
    P(F, "    if (MarEmpty(_xprMk))\n        {\n            _xprDom = 1L;\n            goto _r%lu;\n        }\n    _xprVal[%u].b.mb = _xprMk;\n",
      goNumber, depth);
    MexOut(expression->ue.me.msub2, depth + 1);
    P(F, "    _xprDom = (DOMAIN) MarLequal(_xprVal[%u].b.mb, _xprMk);\n    MarDiscard(_xprVal[%u].b.mb);\n_r%lu: ;\n    MarDiscard(_xprMk);\n",
      depth, depth, goNumber);
}

/***************************************************************************
*
* Function:  XprOMeq
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents an equality check of markings.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOMeq,
    EXPRESSION *, expression,
    unsigned, depth)
{
    MexOut(expression->ue.me.msub1, depth + 1);
    P(F, "    _xprVal[%u].b.mb = _xprMk;\n", depth);
    MexOut(expression->ue.me.msub2, depth + 1);
    P(F, "    _xprDom = (DOMAIN) MarEqual(_xprVal[%u].b.mb, _xprMk);\n    MarDiscard(_xprVal[%u].b.mb);\n    MarDiscard(_xprMk);\n",
      depth, depth);
}

/***************************************************************************
*
* Function:  XprOMnq
* Purpose:   This routine produces code for evaluating an integer
*            expression that represents an inequality check of markings.
* Arguments: expression - the expression
*            depth - this indicates how deep we are in the _xprVal stack
* Returns:
* See also:  all routines having a name of the form XprO[A-Za-z]+
*
****************************************************************************
*/
FS2(void, XprOMnq,
    EXPRESSION *, expression,
    unsigned, depth)
{
    MexOut(expression->ue.me.msub1, depth + 1);
    P(F, "    _xprVal[%u].b.mb = _xprMk;\n", depth);
    MexOut(expression->ue.me.msub2, depth + 1);
    P(F, "    _xprDom = (DOMAIN) !MarEqual(_xprVal[%u].b.mb, _xprMk);\n    MarDiscard(_xprVal[%u].b.mb);\n    MarDiscard(_xprMk);\n",
      depth, depth);
}

#ifdef DEBUG
/* All these XprP routines are trivial and analogous to the MexP routines. */

FS1(void, XprPDom,
    EXPRESSION *, expression)
{
    P(F, UDOMFORMAT, (UDOMAIN) (expression->ue.value));
}

FS1(void, XprPField,
    EXPRESSION *, expression)
{
    P(F, "field[");
    XprPrint(expression->ue.oe.sub1e);
    P(F, "]");
}

FS1(void, XprPCard,
    EXPRESSION *, expression)
{
    P(F, "card(");
    MexPrint(expression->ue.me.msub1);
    P(F, ")");
}

FS2(void, XprPUnary,
    EXPRESSION *, expression,
    char *, op)
{
    P(F, "(%s", op);
    XprPrint(expression->ue.oe.sub1e);
    P(F, ")");
}

FS1(void, XprPNeg,
    EXPRESSION *, expression)
{
    XprPUnary(expression, "-");
}

FS1(void, XprPNot,
    EXPRESSION *, expression)
{
    XprPUnary(expression, "!");
}

FS1(void, XprPBnot,
    EXPRESSION *, expression)
{
    XprPUnary(expression, "~");
}

FS2(void, XprPBinary,
    EXPRESSION *, expression,
    char *, op)
{
    P(F, "(");
    XprPrint(expression->ue.oe.sub1e);
    P(F, " %s ", op);
    XprPrint(expression->ue.oe.sub2e);
    P(F, ")");
}

FS1(void, XprPTimes,
    EXPRESSION *, expression)
{
    XprPBinary(expression, "*");
}

FS1(void, XprPDiv,
    EXPRESSION *, expression)
{
    XprPBinary(expression, "/");
}

FS1(void, XprPRem,
    EXPRESSION *, expression)
{
    XprPBinary(expression, "%");
}

FS1(void, XprPMinus,
    EXPRESSION *, expression)
{
    XprPBinary(expression, "-");
}

FS1(void, XprPPlus,
    EXPRESSION *, expression)
{
    XprPBinary(expression, "+");
}

FS1(void, XprPLshift,
    EXPRESSION *, expression)
{
    XprPBinary(expression, "<<");
}

FS1(void, XprPRshift,
    EXPRESSION *, expression)
{
    XprPBinary(expression, ">>");
}

FS1(void, XprPLss,
    EXPRESSION *, expression)
{
    XprPBinary(expression, "<");
}

FS1(void, XprPLeq,
    EXPRESSION *, expression)
{
    XprPBinary(expression, "<=");
}

FS1(void, XprPEqu,
    EXPRESSION *, expression)
{
    XprPBinary(expression, "==");
}

FS1(void, XprPNeq,
    EXPRESSION *, expression)
{
    XprPBinary(expression, "!=");
}

FS1(void, XprPBand,
    EXPRESSION *, expression)
{
    XprPBinary(expression, "&");
}

FS1(void, XprPBxor,
    EXPRESSION *, expression)
{
    XprPBinary(expression, "^");
}

FS1(void, XprPBor,
    EXPRESSION *, expression)
{
    XprPBinary(expression, "|");
}

FS1(void, XprPAnd,
    EXPRESSION *, expression)
{
    XprPBinary(expression, "&&");
}

FS1(void, XprPOr,
    EXPRESSION *, expression)
{
    XprPBinary(expression, "||");
}

FS1(void, XprPIf,
    EXPRESSION *, expression)
{
    P(F, "(");
    XprPrint(expression->ue.oe.sub1e);
    P(F, "?");
    XprPrint(expression->ue.oe.sub2e);
    P(F, ":");
    XprPrint(expression->ue.oe.sub3e);
    P(F, ")");
}

FS2(void, XprPMBin,
    EXPRESSION *, expression,
    char *, op)
{
    P(F, "(");
    MexPrint(expression->ue.me.msub1);
    P(F, " %s ", op);
    MexPrint(expression->ue.me.msub2);
    P(F, ")");
}

FS1(void, XprPMls,
    EXPRESSION *, expression)
{
    XprPMBin(expression, "<");
}

FS1(void, XprPMlq,
    EXPRESSION *, expression)
{
    XprPMBin(expression, "<=");
}

FS1(void, XprPMeq,
    EXPRESSION *, expression)
{
    XprPMBin(expression, "==");
}

FS1(void, XprPMnq,
    EXPRESSION *, expression)
{
    XprPMBin(expression, "!=");
}
#endif

/***************************************************************************
*
* Function:  ForSnoc
* Purpose:   This routine constructs a formula.
* Arguments: operator - the outermost operator of the formula
*            fp1 - an integer expression if the operator parameter
*                  is FO_EXPR, and otherwise the first maximal proper
*                  subformula
*            fp2 - the second maximal proper subformula if the operator
*                  parameter is not FO_EXPR
* Returns:   the formula
* See also:  ForCons (formula.c)
*
****************************************************************************
*/
FE3(FORMULA *, ForSnoc,
    unsigned, operator,
    FORMULA *, fp1,
    FORMULA *, fp2)
{
    FORMULA *formula;

    formula = ForAlloc();
    formula->operator = operator;
    if (operator == FO_EXPR)
	formula->uf.expression = (EXPRESSION *) fp1;
    else
	{
	    formula->uf.of.sub1f = fp1;
	    formula->uf.of.sub2f = fp2;
	}
    return(formula);
}

#ifdef DEBUG
static unsigned fxCnt = 0;
#endif

/***************************************************************************
*
* Function:  ForXua
* Purpose:   This routine is an interface to ForSnoc and is needed because
*            the construction of formulas in ltl.y is not immediately
*            stopped after a finding of a syntax error.
* Arguments: operator - as in ForSnoc
*            fp1 - as in ForSnoc
*            fp2 - as in ForSnoc
* Returns:   NULL in error cases, and otherwise the formula
* See also:  ForSnoc, ForAux (formula.c)
*
****************************************************************************
*/
FE3(FORMULA *, ForXua,
    unsigned, operator,
    FORMULA *, fp1,
    FORMULA *, fp2)
{
    FORMULA *fp3;
    FORMULA *fp4;
#ifdef DEBUG
    unsigned i;
#endif

    if (errorFlag == NO_ERROR)
	{
	    fp3 = ForSnoc(operator, fp1, fp2);
#ifdef DEBUG
	    pxFile = _err;
	    for (i = 0; i < fxCnt; i++)
		(void) fprintf(_err, " ");
	    (void) fprintf(_err, "before ForReduce\n");
	    ForPrint(fp3);
	    (void) fprintf(_err, "\n");
	    (void) fflush(_err);
	    fxCnt++;
#endif
	    fp4 = ForReduce(fp3);
#ifdef DEBUG
	    fxCnt--;
	    for (i = 0; i < fxCnt; i++)
		(void) fprintf(_err, " ");
	    (void) fprintf(_err, "after ForReduce\n");
	    ForPrint(fp4);
	    (void) fprintf(_err, "\n");
	    for (i = 0; i < fxCnt; i++)
		(void) fprintf(_err, " ");
	    (void) fprintf(_err, "----\n");
	    (void) fflush(_err);
#endif
	    return(fp4);
	}
    if (operator == FO_EXPR)
	{
	    if (fp1)
		XprDiscard((EXPRESSION *) fp1);
	}
    else
	{
	    if (fp1)
		ForDracsid(fp1);
	    if (fp2)
		ForDracsid(fp2);
	}
    return(NULL);
}

/***************************************************************************
*
* Function:  ForYpoc
* Purpose:   This routine makes a copy of a formula.
* Arguments: formula - the formula to be copied
* Returns:   the copy
* See also:  ForCopy (formula.c)
*
****************************************************************************
*/
FE1(FORMULA *, ForYpoc,
    FORMULA *, formula)
{
    FORMULA *copy;

    if (!formula)
	return(NULL);
    copy = ForAlloc();
    copy->operator = formula->operator;
    if (formula->operator == FO_EXPR)
	copy->uf.expression = XprCopy(formula->uf.expression);
    else
	{
	    copy->uf.of.sub1f = ForYpoc(formula->uf.of.sub1f);
	    copy->uf.of.sub2f = ForYpoc(formula->uf.of.sub2f);
	}
    return(copy);
}

/***************************************************************************
*
* Function:  ForDracsid
* Purpose:   This routine removes a formula from the memory.
* Arguments: formula - the formula
* Returns:
* See also:  ForDiscard (formula.c)
*
****************************************************************************
*/
FE1(void, ForDracsid,
    FORMULA *, formula)
{
    if (!formula)
	return;
    if (formula->operator == FO_EXPR)
	XprDiscard(formula->uf.expression);
    else
	{
	    ForDracsid(formula->uf.of.sub1f);
	    ForDracsid(formula->uf.of.sub2f);
	}
    Deallocate((char *) formula);
}

#ifdef DEBUG
/***************************************************************************
*
* Function:  ForPmud
* Purpose:   This routine displays a formula.
* Arguments: formula - the formula
*            depth - the output starts from the (depth + 1)th column
* Returns:
* See also:  ForPrint, ForDump (formula.c)
*
****************************************************************************
*/
FE2(void, ForPmud,
    FORMULA *, formula,
    unsigned, depth)
{
    DumpIndent(depth);
    if (formula)
	P(_err, "FORMULA: operator %u\n", formula->operator);
    else
	P(_err, "FORMULA false\n");
    (void) fflush(_err);
    if (formula)
	{
	    if (formula->operator == FO_EXPR)
		XprDump(formula->uf.expression, depth + 1);
	    else
		{
		    ForPmud(formula->uf.of.sub1f, depth + 1);
		    ForPmud(formula->uf.of.sub2f, depth + 1);
		}
	}
    DumpIndent(depth);
    P(_err, "End of FORMULA\n");
    (void) fflush(_err);
}

/***************************************************************************
*
* Function:  ForPrint
* Purpose:   This routine displays a formula in a somewhat more readable
*            form than ForPmud.
* Arguments: formula - the formula.
* Returns:
* See also:  ForPmud
*
****************************************************************************
*/
FE1(void, ForPrint,
    FORMULA *, formula)
{
    if (!formula)
	{
	    P(F, "_|_");
	    return;
	}
    if (formula->operator == FO_EXPR)
	{
	    XprPrint(formula->uf.expression);
	    return;
	}
    P(F, "(");
    ForPrint(formula->uf.of.sub1f);
    if (formula->operator == FO_UNTIL)
	P(F, " until ");
    else
	P(F, "=>");
    ForPrint(formula->uf.of.sub2f);
    P(F, ")");
}
#endif

/* A collection of routines for formulas to be used by
 * list routines of dynam.c and table routines of ptrbl.c. */
TOOLS forSloot =
{ (CFUNC) ForYpoc, ForDracsid
#ifdef DEBUG
, ForPmud
#endif
 };

/***************************************************************************
*
* Function:  ForIdentic
* Purpose:   This routine checks if two formulas are equivalent.
* Arguments: fp1 - a formula
*            fp2 - a formula
* Returns:   TRUE if the formulas are equivalent with respect to the
*            used rules, and FALSE otherwise
* See also:
*
****************************************************************************
*/
FE2(BOOL, ForIdentic,
    FORMULA *, fp1,
    FORMULA *, fp2)
{
    unsigned ev;

    if (fp1 == fp2)
	return(TRUE);
    if (!fp1)
	return(!fp2);
    if ((!fp2) || (fp1->operator != fp2->operator))
	return(FALSE);
    if (fp1->operator == FO_EXPR)
	return(XprIdentic(fp1->uf.expression, fp2->uf.expression));
    if (fp1->operator == FO_IMPLIES)
	{
	    if (!(fp1->uf.of.sub1f))
		return(!(fp2->uf.of.sub1f));
	    if (!(fp2->uf.of.sub1f))
		return(FALSE);
	}
    if (ForIdentic(fp1->uf.of.sub2f, fp2->uf.of.sub2f) &&
	ForIdentic(fp1->uf.of.sub1f, fp2->uf.of.sub1f))
	return(TRUE);
    if (fp1->operator != FO_IMPLIES)
	return(FALSE);
    ev = 0;
    if ((fp1->uf.of.sub1f->operator == FO_IMPLIES) &&
	(fp2->uf.of.sub1f->operator == FO_IMPLIES) &&
	(!(fp1->uf.of.sub1f->uf.of.sub2f)) && (!(fp2->uf.of.sub1f->uf.of.sub2f)) &&
	((ForIdentic(fp1->uf.of.sub1f->uf.of.sub1f, fp2->uf.of.sub2f) &&
	  ForIdentic(fp2->uf.of.sub1f->uf.of.sub1f, fp1->uf.of.sub2f)) ||
	 (fp1->uf.of.sub2f && (fp1->uf.of.sub2f->operator == FO_IMPLIES) &&
	  fp2->uf.of.sub2f && (fp2->uf.of.sub2f->operator == FO_IMPLIES) &&
	  ForIdentic(fp1->uf.of.sub1f->uf.of.sub1f, fp2->uf.of.sub2f->uf.of.sub2f) &&
	  ForIdentic(fp2->uf.of.sub1f->uf.of.sub1f, fp1->uf.of.sub2f->uf.of.sub2f) &&
	  !(ev = ForIdentic(fp1->uf.of.sub2f->uf.of.sub1f,
			    fp2->uf.of.sub2f->uf.of.sub1f) ? 0 : 1))))
	return(TRUE);
    if ((!(fp1->uf.of.sub2f)) || (fp1->uf.of.sub2f->operator != FO_IMPLIES) ||
	(!(fp2->uf.of.sub2f)) || (fp2->uf.of.sub2f->operator != FO_IMPLIES))
	return(FALSE);
    if (ForIdentic(fp1->uf.of.sub2f->uf.of.sub2f, fp2->uf.of.sub2f->uf.of.sub2f) &&
	(((ev |= ForIdentic(fp1->uf.of.sub1f, fp2->uf.of.sub2f->uf.of.sub1f) ? 12 : 4)
	  & 12) == 12) &&
	(((ev |= ForIdentic(fp2->uf.of.sub1f, fp1->uf.of.sub2f->uf.of.sub1f) ? 48 : 16)
	  & 48) == 48))
	return(TRUE);
    if ((!(fp1->uf.of.sub2f->uf.of.sub2f)) || (!(fp2->uf.of.sub2f->uf.of.sub2f)) ||
	(fp1->uf.of.sub2f->uf.of.sub2f->operator != FO_IMPLIES) ||
	(fp2->uf.of.sub2f->uf.of.sub2f->operator != FO_IMPLIES) ||
	!ForIdentic(fp1->uf.of.sub2f->uf.of.sub2f->uf.of.sub2f,
		    fp2->uf.of.sub2f->uf.of.sub2f->uf.of.sub2f))
	return(FALSE);
    return((ForIdentic(fp1->uf.of.sub1f, fp2->uf.of.sub2f->uf.of.sub2f->uf.of.sub1f)
	    &&
	    ((((((ev & 48) == 0)
		&&
		ForIdentic(fp2->uf.of.sub1f, fp1->uf.of.sub2f->uf.of.sub1f))
	       ||
	       ((ev & 48) == 48))
	      &&
	      ForIdentic(fp2->uf.of.sub2f->uf.of.sub1f,
			 fp1->uf.of.sub2f->uf.of.sub2f->uf.of.sub1f))
	     ||
	     ((((ev |=
		 ForIdentic(fp2->uf.of.sub1f,
			    fp1->uf.of.sub2f->uf.of.sub2f->uf.of.sub1f) ? 192 : 64)
		& 192) == 192)
	      &&
	      ((((ev & 3) == 0)
		&&
		ForIdentic(fp1->uf.of.sub2f->uf.of.sub1f,
			   fp2->uf.of.sub2f->uf.of.sub1f))
	       ||
	       ((ev & 3) == 3)))))
	   ||
	   (((((ev & 192) == 0)
	      &&
	      ForIdentic(fp2->uf.of.sub1f, fp1->uf.of.sub2f->uf.of.sub2f->uf.of.sub1f))
	     ||
	     ((ev & 192) == 192))
	    &&
	    ((((ev & 12) == 0)
	      &&
	      ForIdentic(fp1->uf.of.sub1f, fp2->uf.of.sub2f->uf.of.sub1f))
	     ||
	     ((ev & 12) == 12))
	    &&
	    ForIdentic(fp1->uf.of.sub2f->uf.of.sub1f,
		       fp2->uf.of.sub2f->uf.of.sub2f->uf.of.sub1f)));
}

/* All these MexI and XprI routines are trivial and analogous to ForIdentic. */

FS2(BOOL, MexIConstant,
    MARKEXPR *, mep1,
    MARKEXPR *, mep2)
{
    return(MarEqual(mep1->um.marking, mep2->um.marking));
}

FS2(BOOL, MexIPlace,
    MARKEXPR *, mep1,
    MARKEXPR *, mep2)
{
    return(mep1->um.place == mep2->um.place);
}

FS2(BOOL, MexITrange,
    MARKEXPR *, mep1,
    MARKEXPR *, mep2)
{
    XRANGE *xrg1;
    XRANGE *xrg2;

    xrg1 = (XRANGE *) LstBegin(mep1->um.trange);
    xrg2 = (XRANGE *) LstBegin(mep2->um.trange);
    while (xrg1 && xrg2)
	{
	    xrg1 = (XRANGE *) LstStep(mep1->um.trange);
	    xrg2 = (XRANGE *) LstStep(mep2->um.trange);
	}
    if (xrg1 || xrg2)
	return(FALSE);
    xrg1 = (XRANGE *) LstBegin(mep1->um.trange);
    xrg2 = (XRANGE *) LstBegin(mep2->um.trange);
    while (xrg1)
	{
	    if ((!XprIdentic(XrgGLower(xrg1), XrgGLower(xrg2))) ||
		!XprIdentic(XrgGUpper(xrg1), XrgGUpper(xrg2)))
		return(FALSE);
	    xrg1 = (XRANGE *) LstStep(mep1->um.trange);
	    xrg2 = (XRANGE *) LstStep(mep2->um.trange);
	}
    return(TRUE);
}

FS2(BOOL, MexITimes,
    MARKEXPR *, mep1,
    MARKEXPR *, mep2)
{
    return(XprIdentic(mep1->um.em.esub, mep2->um.em.esub) &&
	   MexIdentic(mep1->um.em.msub, mep2->um.em.msub));
}

FS2(BOOL, MexICartes,
    MARKEXPR *, mep1,
    MARKEXPR *, mep2)
{
    return((MexIdentic(mep1->um.mm.sub1m, mep2->um.mm.sub1m) &&
	    MexIdentic(mep1->um.mm.sub2m, mep2->um.mm.sub2m)) ||
	   (MexIdentic(mep1->um.mm.sub1m, mep2->um.mm.sub2m) &&
	    MexIdentic(mep1->um.mm.sub2m, mep2->um.mm.sub1m)));
}

FS2(BOOL, MexIDiff,
    MARKEXPR *, mep1,
    MARKEXPR *, mep2)
{
    return(MexIdentic(mep1->um.mm.sub1m, mep2->um.mm.sub1m) &&
	   MexIdentic(mep1->um.mm.sub2m, mep2->um.mm.sub2m));
}

FS2(BOOL, XprIDom,
    EXPRESSION *, ep1,
    EXPRESSION *, ep2)
{
    return(ep1->ue.value == ep2->ue.value);
}

FS2(BOOL, XprIField,
    EXPRESSION *, ep1,
    EXPRESSION *, ep2)
{
    return(XprIdentic(ep1->ue.oe.sub1e, ep2->ue.oe.sub1e));
}

FS2(BOOL, XprICard,
    EXPRESSION *, ep1,
    EXPRESSION *, ep2)
{
    return(MexIdentic(ep1->ue.me.msub1, ep2->ue.me.msub1));
}

FS2(BOOL, XprITimes,
    EXPRESSION *, ep1,
    EXPRESSION *, ep2)
{
    return((XprIdentic(ep1->ue.oe.sub1e, ep2->ue.oe.sub1e) &&
	    XprIdentic(ep1->ue.oe.sub2e, ep2->ue.oe.sub2e)) ||
	   (XprIdentic(ep1->ue.oe.sub1e, ep2->ue.oe.sub2e) &&
	    XprIdentic(ep1->ue.oe.sub2e, ep2->ue.oe.sub1e)));
}

FS2(BOOL, XprIDiv,
    EXPRESSION *, ep1,
    EXPRESSION *, ep2)
{
    return(XprIdentic(ep1->ue.oe.sub1e, ep2->ue.oe.sub1e) &&
	   XprIdentic(ep1->ue.oe.sub2e, ep2->ue.oe.sub2e));
}

FS2(BOOL, XprIIf,
    EXPRESSION *, ep1,
    EXPRESSION *, ep2)
{
    return(XprIdentic(ep1->ue.oe.sub1e, ep2->ue.oe.sub1e) &&
	   XprIdentic(ep1->ue.oe.sub2e, ep2->ue.oe.sub2e) &&
	   XprIdentic(ep1->ue.oe.sub3e, ep2->ue.oe.sub3e));
}

FS2(BOOL, XprIMls,
    EXPRESSION *, ep1,
    EXPRESSION *, ep2)
{
    return(MexIdentic(ep1->ue.me.msub1, ep2->ue.me.msub1) &&
	   MexIdentic(ep1->ue.me.msub2, ep2->ue.me.msub2));
}

FS2(BOOL, XprIMeq,
    EXPRESSION *, ep1,
    EXPRESSION *, ep2)
{
    return((MexIdentic(ep1->ue.me.msub1, ep2->ue.me.msub1) &&
	    MexIdentic(ep1->ue.me.msub2, ep2->ue.me.msub2)) ||
	   (MexIdentic(ep1->ue.me.msub1, ep2->ue.me.msub2) &&
	    MexIdentic(ep1->ue.me.msub2, ep2->ue.me.msub1)));
}

/***************************************************************************
*
* Function:  MexAffected
* Purpose:   This routine estimates if a high level transition could
*            directly affect the truth values of the non-temporal
*            subformulas of user's formula. Clearly, we need an upper
*            bound.
* Arguments: markexpr - a marking expression
*            transition - the transition
* Returns:   TRUE if the transition is connected to some place mentioned
*            in the expression, and FALSE otherwise
* See also:  XprAffected
*
****************************************************************************
*/
FE2(BOOL, MexAffected,
    MARKEXPR *, markexpr,
    TRANSITION *, transition)
{
    PLACE *place;
    ARC *arc;
    XRANGE *xrange;

    switch (MexKind(markexpr)) 
	{
	case MK_CONSTANT:
	    return(FALSE);
	case MK_PLACE:
	    place = (PLACE *) LstNth(PlaceLst, markexpr->um.place + 1);
	    if (place == monPlace)
		return(FALSE);
	    LSTFOR (TraGArcs(transition), arc, ARC)
		if (ArcGPlace(arc) == place)
		    return(TRUE);
	    return(FALSE);
	case MK_TRANGE:
	    TRGFOR (markexpr->um.trange, xrange) 
		if (XprAffected(xrange->lower, transition) ||
		    XprAffected(xrange->upper, transition))
		    return(TRUE);
	    return(FALSE);
	case MK_EXPRMARK:
	    return(XprAffected(markexpr->um.em.esub, transition) ||
		   MexAffected(markexpr->um.em.msub, transition));
	case MK_MARKMARK:
	    return(MexAffected(markexpr->um.mm.sub1m, transition) ||
		   MexAffected(markexpr->um.mm.sub2m, transition));
	}
    return(FALSE);
}

/***************************************************************************
*
* Function:  XprAffected
* Purpose:   This routine estimates if a high level transition could
*            directly affect the truth values of the non-temporal
*            subformulas of user's formula. Clearly, we need an upper
*            bound.
* Arguments: expression - an integer expression
*            transition - the transition
* Returns:   TRUE if the transition is connected to some place mentioned
*            in the expression, and FALSE otherwise
* See also:  MexAffected
*
****************************************************************************
*/
FE2(BOOL, XprAffected,
    EXPRESSION *, expression,
    TRANSITION *, transition)
{
    switch (XprKind(expression)) 
	{
	case EK_CONSTANT:
	case EK_NULLARY:
	    return(FALSE);
	case EK_CARD:
	    return(MexAffected(expression->ue.me.msub1, transition));
	case EK_MCOMP:
	    return(MexAffected(expression->ue.me.msub1, transition) ||
		   MexAffected(expression->ue.me.msub2, transition));
	case EK_UNARY:
	    return(XprAffected(expression->ue.oe.sub1e, transition));
	case EK_BINARY:
	    return(XprAffected(expression->ue.oe.sub1e, transition) ||
		   XprAffected(expression->ue.oe.sub2e, transition));
	case EK_TRINARY:
	    return(XprAffected(expression->ue.oe.sub1e, transition) ||
		   XprAffected(expression->ue.oe.sub2e, transition) ||
		   XprAffected(expression->ue.oe.sub3e, transition));
	}
    return(FALSE);
}

/***************************************************************************
*
* Function:  SetExMex
* Purpose:   This routine sets the prpp specific entries in
*            the expression evaluation and reduction function tables.
* Arguments:
* Returns:
* See also:  SetExMex (formula.c)
*
****************************************************************************
*/
FE0(void, SetExMex)
{
    SetMexEval(MO_PLACE, MexEPlace);
    SetMexReduce(MO_TRANGE, MexRTrange);
    SetXprEval(EO_FIELD, XprEField);
    SetXprEval(EO_DIV, XprEDiv);
    SetXprEval(EO_REM, XprERem);
    SetXprReduce(EO_DIV, XprRDiv);
    SetXprReduce(EO_REM, XprRRem);
}

#define MexISum MexICartes
#define MexISelect MexITimes
#define MexIInters MexICartes
#define XprINeg XprIField
#define XprINot XprIField
#define XprIBnot XprIField
#define XprIRem XprIDiv
#define XprIMinus XprIDiv
#define XprIPlus XprITimes
#define XprILshift XprIDiv
#define XprIRshift XprIDiv
#define XprILss XprIDiv
#define XprILeq XprIDiv
#define XprIEqu XprITimes
#define XprINeq XprITimes
#define XprIBand XprITimes
#define XprIBxor XprITimes
#define XprIBor XprITimes
#define XprIAnd XprIDiv
#define XprIOr XprIDiv
#define XprIMlq XprIMls
#define XprIMnq XprIMeq

#ifdef DEBUG
MEXOUT mexOutTable[SZ_MO] = 
{
    { MexOConstant, MexPConstant, MexIConstant },
    { MexOPlace, MexPPlace, MexIPlace },
    { MexOTrange, MexPTrange, MexITrange },
    { MexOTimes, MexPTimes, MexITimes },
    { MexOCartes, MexPCartes, MexICartes },
    { MexOSum, MexPSum, MexISum },
    { MexODiff, MexPDiff, MexIDiff },
    { MexOSelect, MexPSelect, MexISelect },
    { MexOInters, MexPInters, MexIInters }
};

XPROUT xprOutTable[SZ_EO - 3] = 
{
    { XprODom, XprPDom, XprIDom },
    { XprOField, XprPField, XprIField },
    { XprOCard, XprPCard, XprICard },
    { XprONeg, XprPNeg, XprINeg },
    { XprONot, XprPNot, XprINot },
    { XprOBnot, XprPBnot, XprIBnot },
    { XprOTimes, XprPTimes, XprITimes },
    { XprODiv, XprPDiv, XprIDiv },
    { XprORem, XprPRem, XprIRem },
    { XprOMinus, XprPMinus, XprIMinus },
    { XprOPlus, XprPPlus, XprIPlus },
    { XprOLshift, XprPLshift, XprILshift },
    { XprORshift, XprPRshift, XprIRshift },
    { XprOLss, XprPLss, XprILss },
    { XprOLeq, XprPLeq, XprILeq },
    { XprOEqu, XprPEqu, XprIEqu },
    { XprONeq, XprPNeq, XprINeq },
    { XprOBand, XprPBand, XprIBand },
    { XprOBxor, XprPBxor, XprIBxor },
    { XprOBor, XprPBor, XprIBor },
    { XprOAnd, XprPAnd, XprIAnd },
    { XprOOr, XprPOr, XprIOr },
    { XprOIf, XprPIf, XprIIf },
    { XprOMls, XprPMls, XprIMls },
    { XprOMlq, XprPMlq, XprIMlq },
    { XprOMeq, XprPMeq, XprIMeq },
    { XprOMnq, XprPMnq, XprIMnq }
};
#else
MEXOUT mexOutTable[SZ_MO] = 
{
    { MexOConstant, MexIConstant },
    { MexOPlace, MexIPlace },
    { MexOTrange, MexITrange },
    { MexOTimes, MexITimes },
    { MexOCartes, MexICartes },
    { MexOSum, MexISum },
    { MexODiff, MexIDiff },
    { MexOSelect, MexISelect },
    { MexOInters, MexIInters }
};

XPROUT xprOutTable[SZ_EO - 3] = 
{
    { XprODom, XprIDom },
    { XprOField, XprIField },
    { XprOCard, XprICard },
    { XprONeg, XprINeg },
    { XprONot, XprINot },
    { XprOBnot, XprIBnot },
    { XprOTimes, XprITimes },
    { XprODiv, XprIDiv },
    { XprORem, XprIRem },
    { XprOMinus, XprIMinus },
    { XprOPlus, XprIPlus },
    { XprOLshift, XprILshift },
    { XprORshift, XprIRshift },
    { XprOLss, XprILss },
    { XprOLeq, XprILeq },
    { XprOEqu, XprIEqu },
    { XprONeq, XprINeq },
    { XprOBand, XprIBand },
    { XprOBxor, XprIBxor },
    { XprOBor, XprIBor },
    { XprOAnd, XprIAnd },
    { XprOOr, XprIOr },
    { XprOIf, XprIIf },
    { XprOMls, XprIMls },
    { XprOMlq, XprIMlq },
    { XprOMeq, XprIMeq },
    { XprOMnq, XprIMnq }
};
#endif
