/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  ptnet.c
*
* Part of: PROD's programs araprod and "gphgen"
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     This module contains routines for reachability graph generation
*     of a low level net. For example, stubborn sets and sleep sets
*     are computed hete. DSL(HUT) reports A 26, B 12 and B 13 are needed
*     for understanding, as well the PN '94 paper and the CS&P '94
*     paper of the author, Valmari's CAV '93 paper, and the
*     article of Courcoubetis, Vardi, Wolper & Yannakakis in
*     Formal Methods in System Design 1 (1992) 2/3, pp. 275-288.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include <stdio.h>
#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "graph.h"
#include "mark.h"
#include "encode.h"
#include "ptnet.h"
#include "ptsymm.h"

unsigned long _nTrVisible; /* number of visible transitions */
PTRBL *_tblPlc;          /* a table of all low level places */
PTRBL *_dupTblPlc;       /* a copy of _tblPlc, needed in the symmetry method */
PTRBL **_tblTrn;         /* _tblTrn[i] contains all those low level transitions
			  * that are in precedence class i */
unsigned *_higher;       /* _higher[i] indicates if precedence class i
			  * prioritizes and/or is prioritized by other
			  * classes (or itself !!) */
BALTREE *_plcBalTree;    /* a tree of all low level places */
BALTREE **_trnBalTree;   /* _trnBalTree[i] contains all those low level
			  * transitions that are in precedence class i */
TRN *_lowTrn;            /* used in the unfolding phase if there are more
			  * than one precedence class */
unsigned _hiClass;       /* used in the unfolding phase if there are more
			  * than one precedence class */
BOOL _ptSy;              /* TRUE if and only if the symmetry method is used */
unsigned _ptVerbose;     /* indicates the level of verbosity in which the
			  * computed place symmetries are reported */
BOOL _ptStub;            /* TRUE if and only if the stubborn set method
			  * is used */
BOOL _ptCFFD;            /* TRUE if and only if the CFFD-preserving version
			  * of the stubborn set method is used */
BOOL _ptDefVis;          /* transition instances are visible by default if
			  * _ptDefVis is TRUE, and otherwise invisible
			  * by default */
BOOL _ptSleep;           /* this is TRUE if and only if the sleep set method
			  * is used */
BOOL _plRnd;             /* in the incremental algorithm of the stubborn set
			  * method, the disabling place of a disabled
			  * transition is selected randomly if _plRnd is
			  * TRUE, and otherwise the first disabling place
			  * is chosen */
BOOL _ptLazy;            /* in the incremental algorithm of the stubborn set
			  * method, the first stubborn set found is chosen
			  * if _ptLazy is TRUE, and otherwise a stubborn
			  * set having the least number of enabled transitions
			  * is chosen (however, the collection from which
			  * the winner is chosen may be a minor subset of
			  * the collection of all stubborn sets) */
BOOL _trRnd;             /* (a) in the deletion algorithm of the stubborn set
			  * method, a candidate to be deleted is chosen
			  * randomly if _trRnd is TRUE, and otherwise
			  * the first possible candidate is chosen;
			  * (b) in the incremental algorithm in the case
			  * that _ptLazy is TRUE, the starting transition
			  * is chosen randomly if _trRnd is TRUE, and otherwise
			  * the first possible transition is chosen */
BOOL _ptDelAlg;          /* this is TRUE if and only if the deletion algorithm
			  * of the stubborn set is used */
BOOL _ptTryAgain;        /* if this is TRUE and the deletion algorithm
			  * of the stubborn set is used, then the algorithm
			  * is applied repeatedly in order to find a stubborn
			  * set having the least number of enabled transitions
			  * (success is guaranteed if the least number is 1
			  * or if the number of enabled transitions is
			  * less than or equal to 5) */
LIST *_binOrList;        /* the BINOR vertices of an and/or-graph in the
			  * deletion algorithm of the stubborn set method */
LIST *_ptList;           /* a "Tarjan stack" used in the incremental algorithm
			  * of the stubborn set method */
unsigned long _ptSgCount; /* the number of strongly connected components
			   * found in a transition dependency graph in
			   * the incremental algorithm of the stubborn set
			   * method */
unsigned long _ptECnt;    /* the number of enabled transitions */
unsigned long _ptInvisCnt; /* the number of invisible enabled transitions */
unsigned long _ptNow;     /* the number of visited vertices in a transition
			   * dependency graph in the incremental algorithm
			   * of the stubborn set method */
unsigned long _ptBest;    /* the number of the strongly connected component
			   * of the above graph that, among the so far found
			   * components, has the least number of enabled
			   * transitions */
unsigned long _ptSzBest;  /* the size of the _ptBest component */
unsigned long _ptCounter; /* used for counting the starting transition
			   * in the incremental algorithm when _ptLazy and
			   * _trRnd are TRUE */
TRN *_ptKingTrn;
unsigned _ptClass;        /* a precedence class */
BOOL _ptRecurs;           /* when TRUE, the CFFD-preserving version of
			   * the stubborn set method may take a "recursion
			   * step" (a stack is used instead of actual
			   * recursion) */
BALTREE *_ptSlpSet[4];    /* _ptSlpSet[0] is the "stack sleep set" of the
			   * current marking. _ptSlpSet[1] is the possible
			   * "table sleep set" at the time when the marking
			   * was popped from the stack. _ptSlpSet[2]
			   * is either the "table sleep set" when the visit in
			   * question is over, or a plain auxiliary set not
			   * associated with any marking. _ptSlpSet[3] is the
			   * "stack sleep set" of some successor marking. */
PLC *_ptFlyPlc;           /* a place corresponding to a state
			   * of the tester (if there is a tester) */
BOOL ptDoIt = TRUE;
BOOL ptBranching;
TRN *ptFirst;
PLC *buPlc;               /* a place corresponding to a state of the
			   * Buchi automaton (if there is such) */
FOOL _buchiOK;
TRN *ptInvFirst;
unsigned long ptM;
TRN *ptTrn;
TRN *pt2Trn;
PLC *ptPlc;
TRARC *ptTrArc;
TRDRC *ptTrDrc;
PLARC *ptPlArc;
PLDRC *ptPlDrc;
BINOR *ptBinOr;
LIST **_visBlock;
char *_visColour;
BOOL _blocked;
PTRBL *visibles = NULL;
#ifdef SMODELS
BOOL _ptSmodels;
BOOL _ptLongPlay;
INEQ *ptIneq;
#endif

/***************************************************************************
*
* Function:  TrnEnblCheck
* Purpose:   This routine checks if a transition is enabled.
* Arguments: trn - the transition
* Returns:
* See also:  TrnChkEnbl
*
****************************************************************************
*/
FE1(BOOL, TrnEnblCheck,
    TRN *, trn)
{
    TRARC *trArc;
    TRDRC *trDrc;

    TrnReset(trn);
    if (Buchi() && ((TrnDummy(trn) && (PtGECnt() != 0L)) ||
		    ((TrnInitial(trn) == 0) ==
		     (PthEnd(nowPath) == GphStart(theGraph)))))
	return(FALSE);
    if (TrnGPurIn(trn))
	LSTFOR (TrnGPurIn(trn), trArc, TRARC)
	    if (PlcGCount(TrArcGPlc(trArc)) < TrArcGWeight(trArc))
		return(FALSE);
    if (TrnGNeutral(trn))
	LSTFOR (TrnGNeutral(trn), trArc, TRARC)
	    if (PlcGCount(TrArcGPlc(trArc)) < TrArcGWeight(trArc))
		return(FALSE);
    if (TrnGMosIn(trn))
	LSTFOR (TrnGMosIn(trn), trDrc, TRDRC)
	    if (PlcGCount(TrDrcGPlc(trDrc)) < TrDrcGIn(trDrc))
		return(FALSE);
    if (TrnGMosOut(trn))
	LSTFOR (TrnGMosOut(trn), trDrc, TRDRC)
	    if (PlcGCount(TrDrcGPlc(trDrc)) < TrDrcGIn(trDrc))
		return(FALSE);
    TrnAccept(trn);
    PtIECnt();
    if (!ptFirst)
	ptFirst = trn;
    if (!TrnVisible(trn))
	{
	    PtIInvisCnt();
	    if (!ptInvFirst)
		ptInvFirst = trn;
	}
    return(Buchi() && TrnInitial(trn));
}

/***************************************************************************
*
* Function:  FndEnbl
* Purpose:   This routine finds all enabled transitions in class 0.
* Arguments:
* Returns:
* See also:  FndEnbl
*
****************************************************************************
*/
FE0(void, FndEnbl)
{
    ClPtRecurs();
    PtCECnt();
    PtCInvisCnt();
    ptFirst = NULL;
    ptInvFirst = NULL; 
    (void) PtrBlApply(TblTrn(0), TrnEnblCheck);
}

/***************************************************************************
*
* Function:  AfterEnbl
* Purpose:   This routine checks if trn2 is enabled immediately after
*            the firing of trn1.
* Arguments: trn1 - an enabled transition
*            trn2 - a transition
* Returns:   TRUE if the "trn1 trn2" sequence is enabled, and
*            FALSE otherwise
* See also:
*
****************************************************************************
*/
FE2(BOOL, AfterEnbl,
    TRN *, trn1,
    TRN *, trn2)
{
    TRARC *trArc;
    TRDRC *trDrc;
    BOOL afterEnabled;

    if (TrnGPurIn(trn1))
	LSTFOR (TrnGPurIn(trn1), trArc, TRARC)
	    PlcDCount(TrArcGPlc(trArc), TrArcGWeight(trArc));
    if (TrnGPurOut(trn1))
	LSTFOR (TrnGPurOut(trn1), trArc, TRARC)
	    PlcICount(TrArcGPlc(trArc), TrArcGWeight(trArc));
    if (TrnGMosIn(trn1))
	LSTFOR (TrnGMosIn(trn1), trDrc, TRDRC)
	    PlcDCount(TrDrcGPlc(trDrc),
		      TrDrcGIn(trDrc) - TrDrcGOut(trDrc));
    if (TrnGMosOut(trn1))
	LSTFOR (TrnGMosOut(trn1), trDrc, TRDRC)
	    PlcICount(TrDrcGPlc(trDrc),
		      TrDrcGOut(trDrc) - TrDrcGIn(trDrc));
    if (TrnGPurIn(trn2))
	LSTFOR (TrnGPurIn(trn2), trArc, TRARC)
	    if (PlcGCount(TrArcGPlc(trArc)) < TrArcGWeight(trArc))
		goto FAIL;
    if (TrnGNeutral(trn2))
	LSTFOR (TrnGNeutral(trn2), trArc, TRARC)
	    if (PlcGCount(TrArcGPlc(trArc)) < TrArcGWeight(trArc))
		goto FAIL;
    if (TrnGMosIn(trn2))
	LSTFOR (TrnGMosIn(trn2), trDrc, TRDRC)
	    if (PlcGCount(TrDrcGPlc(trDrc)) < TrDrcGIn(trDrc))
		goto FAIL;
    if (TrnGMosOut(trn2))
	LSTFOR (TrnGMosOut(trn2), trDrc, TRDRC)
	    if (PlcGCount(TrDrcGPlc(trDrc)) < TrDrcGIn(trDrc))
		goto FAIL;
    afterEnabled = TRUE;
    goto RESTORE;
 FAIL: ;
    afterEnabled = FALSE;
 RESTORE: ;
    if (TrnGPurIn(trn1))
	LSTFOR (TrnGPurIn(trn1), trArc, TRARC)
	    PlcICount(TrArcGPlc(trArc), TrArcGWeight(trArc));
    if (TrnGPurOut(trn1))
	LSTFOR (TrnGPurOut(trn1), trArc, TRARC)
	    PlcDCount(TrArcGPlc(trArc), TrArcGWeight(trArc));
    if (TrnGMosIn(trn1))
	LSTFOR (TrnGMosIn(trn1), trDrc, TRDRC)
	    PlcICount(TrDrcGPlc(trDrc),
		      TrDrcGIn(trDrc) - TrDrcGOut(trDrc));
    if (TrnGMosOut(trn1))
	LSTFOR (TrnGMosOut(trn1), trDrc, TRDRC)
	    PlcDCount(TrDrcGPlc(trDrc),
		      TrDrcGOut(trDrc) - TrDrcGIn(trDrc));
    return(afterEnabled);
}

/***************************************************************************
*
* Function:  TrnBest
* Purpose:   This routine fires a transition if it is enabled and in the
*            stubborn set computed by the incremental algorithm.
* Arguments: trn - the transition
* Returns:
* See also:  TrnFire, TrnPrecCheck, TrnReCheck, PtOldSleep, TrnChkRe,
*            TrnDelCheck
*
****************************************************************************
*/
FE1(BOOL, TrnBest,
    TRN *, trn)
{
    if (TrnEnabled(trn) && ((TrnGVal(trn) == PtBest()) ||
			    (TrnGVal(trn) == PtrBlGCount(TblTrn(0)) + 1L)))
	TrnFire(trn);
    return(FALSE);
}

/***************************************************************************
*
* Function:  TrnPrecCheck
* Purpose:   This routine fires a transition if it is enabled and no
*            enabled transition has any priority over it.
* Arguments: trn - the transition
* Returns:
* See also:  TrnFire, TrnBest, TrnReCheck, PtOldSleep, TrnChkRe,
*            TrnDelCheck
*
****************************************************************************
*/
FE1(BOOL, TrnPrecCheck,
    TRN *, trn)
{
    if (TrnEnabled(trn))
	{
	    if (TrnGHigher(trn))
		{
		    TRN *htrn;
		    
		    LSTFOR (TrnGHigher(trn), htrn, TRN)
			{
			    if (TrnEnabled(htrn))
				break;
			}
		    if (!htrn)
			TrnFire(trn);
		}
	    else
		TrnFire(trn);
	}
    return(FALSE);
}

/***************************************************************************
*
* Function:  TrnReCheck
* Purpose:   This routine fires a transition if it is enabled.
* Arguments: trn - the transition
* Returns:
* See also:  TrnFire, TrnBest, TrnPrecCheck, PtOldSleep, TrnChkRe,
*            TrnDelCheck
*
****************************************************************************
*/
FE1(BOOL, TrnReCheck,
    TRN *, trn)
{
    if (TrnEnabled(trn))
	TrnFire(trn);
    return(FALSE);
}

/***************************************************************************
*
* Function:  PtNewSleep
* Purpose:   This routine inserts a transition into the stack sleep
*            set of a successor marking of ptTrn if the transition and
*            ptTrn commute at the current marking.
* Arguments: a - the number of the transition (that must be enabled)
* Returns:
* See also: 
*
****************************************************************************
*/
FE1(BOOL, PtNewSleep,
    unsigned long *, a)
{
    pt2Trn = (TRN *) PtrBlGet(TblTrn(0), *a);
    if (AfterEnbl(ptTrn, pt2Trn) && AfterEnbl(pt2Trn, ptTrn))
	(void) BalTrInsert(Pt4SlpSet(), (char *) LngCons(*a),
			   (char *) a, TRUE);
    return(FALSE);
}

/***************************************************************************
*
* Function:  PtOldSleep
* Purpose:   This routine fires an enabled transition and associates
*            a stack sleep set with the resulting marking, though only if
*            the current marking is being revisited and the transition
*            is in the table sleep set of the current marking but not
*            in the stack sleep set of the current marking.
* Arguments: a - the number of the transition to be fired (is in the
*                the "table sleep set" but not in the "stack sleep set")
* Returns:
* See also:  TrnFire, TrnBest, TrnPrecCheck, TrnReCheck, TrnChkRe,
*            TrnDelCheck
*
****************************************************************************
*/
FE1(BOOL, PtOldSleep,
    unsigned long *, a)
{
    if (BalTrElement(PtSlpSet(), (char *) a))
	{
	    (void) BalTrInsert(Pt3SlpSet(), (char *) LngCons(*a),
			       (char *) a, TRUE);
	    return(FALSE);
	}
    ptTrn = (TRN *) PtrBlGet(TblTrn(0), *a);
    PtS4SlpSet(BalTrCreate(&lngBalTools));
    (void) BalTrApply(PtSlpSet(), PtNewSleep);
    TrnFire(ptTrn);
    PtD4SlpSet();
    (void) BalTrInsert(PtSlpSet(), (char *) LngCons(*a), (char *) a, TRUE);
    return(FALSE);
}

/***************************************************************************
*
* Function:  TrnChkRe
* Purpose:   This routine fires an enabled transition and associates
*            a stack sleep set with the resulting marking, though only if
*            the current marking is being visited for the first time
*            and the transition is not in the stack sleep set of the
*            current marking. If the stubborn set method is used,
*            no transition outside of the computed stubborn set is fired.
* Purpose:   This routine is essentially similar to PtOldSleep, except
*            that the current marking is being visited for the first time.
* Arguments: trn - the transition
* Returns:
* See also:  TrnFire, TrnBest, TrnPrecCheck, TrnReCheck, PtOldSleep,
*            TrnDelCheck
*
****************************************************************************
*/
FE1(BOOL, TrnChkRe,
    TRN *, trn)
{
    if (!TrnEnabled(trn))
	return(FALSE);
    if (BalTrElement(PtSlpSet(), (char *) TrnANumber(trn)) ||
	(PtDelAlg() && TrnSatisfied(trn)) ||
	(PtStub() && (!PtDelAlg()) && (TrnGVal(trn) != PtBest()) &&
	 (TrnGVal(trn) != PtrBlGCount(TblTrn(0)) + 1L)))
	return(FALSE);
    PtS4SlpSet(BalTrCreate(&lngBalTools));
    ptTrn = trn;
    (void) BalTrApply(Pt3SlpSet(), PtNewSleep);
    TrnFire(trn);
    PtD4SlpSet();
    (void) BalTrInsert(Pt3SlpSet(), (char *) LngCons(TrnGNumber(trn)),
		       (char *) TrnANumber(trn), TRUE);
    return(FALSE);
}

/***************************************************************************
*
* Function:  TrnRedInc
* Purpose:   This routine is an interface to TrnLookStack in the
*            incremental algorithm
* Arguments: trn - a transition
* Returns:
* See also:  TrnLookStack, TrnRedDel, PtRedInc, PtRedDel
*
****************************************************************************
*/
FE1(BOOL, TrnRedInc,
    TRN *, trn)
{
    if (TrnEnabled(trn) && (TrnGVal(trn) == PtBest()))
	{
	    TrnLookStack(trn);
	    if (!PtRecurs())
		return(TRUE);
	}
    return(FALSE);
}

/***************************************************************************
*
* Function:  TrnRedDel
* Purpose:   This routine is an interface to TrnLookStack in the
*            deletion algorithm.
* Arguments: trn - a transition
* Returns:
* See also:  TrnLookStack, TrnRedInc, PtRedInc, PtRedDel
*
****************************************************************************
*/
FE1(BOOL, TrnRedDel,
    TRN *, trn)
{
    if (TrnEnabled(trn) && (!TrnSatisfied(trn)))
	{
	    TrnLookStack(trn);
	    if (!PtRecurs())
		return(TRUE);
	}
    return(FALSE);
}

/***************************************************************************
*
* Function:  PtGphUpdate
* Purpose:   This routine updates the graph information concerning the
*            current marking.
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE0(void, PtGphUpdate)
{
    if (PtSleep())
	{
	    if (Pt2SlpSet())
		{
		    if (BalTrGElCount(PtSlpSet()) !=
			BalTrGElCount(Pt2SlpSet()))
			GphUpdate(theGraph, PthEnd(nowPath), theAwCount,
				  theAwFirst, PtSlpSet());
		}
	    else if ((theAwCount == 0L) && (BalTrGElCount(PtSlpSet()) == 0L))
		GphAddTermNode(theGraph, PthEnd(nowPath));
	    else
		GphUpdate(theGraph, PthEnd(nowPath), theAwCount, theAwFirst,
			  PtSlpSet());
	    GphIncGenCount(theGraph);
	    if (theDots)
		DotOutput();
	}
    else if (Buchi() && PtStub())
	{
	    if (theAwCount == 0L)
		GphAddTermNode(the2Graph, PthEnd(thePath));
	    else
		GphUpdate(the2Graph, PthEnd(thePath), theAwCount, theAwFirst,
			  PtSlpSet());
	    GphIncGenCount(the2Graph);
	}
    else
	{
	    if (theAwCount == 0L)
		GphAddTermNode(theGraph, PthEnd(nowPath));
	    else
		GphUpdate(theGraph, PthEnd(nowPath), theAwCount, theAwFirst,
			  PtSlpSet());
	    GphIncGenCount(theGraph);
	    if (theDots)
		DotOutput();
	}
}

/***************************************************************************
*
* Function:  PtC1
* Purpose:   This on-the-fly verification routine handles a case
*            where an error of the form specified by the reject states of
*            the tester is found.
* Arguments: awAddr - a predecessor arrow of ndAddr (-1L if we are in
*                     the initial marking)
*            ndAddr - the node where the error was observed
* Returns:
* See also:  PtC2, PtC3, PtC4, PtOnTheFly, PtInfinite, PtBuchi,
*            IllTrace (inst.c), ExOnTheFly (graph.c)
*
****************************************************************************
*/
FE2(void, PtC1,
    unsigned long, awAddr,
    unsigned long, ndAddr)
{
    if (awAddr != (unsigned long) -1L)
	PtGphUpdate();
    if (Divergence())
	BalTrDestroy(theNodeTree, FALSE);
    if (Ara())
	LTSEnd();
    (void) fprintf(_out, "\nReject state reached\n");
    if (PtStub())
	GphSetSeed(theGraph, PtTerm());
    if (PthLast(nowPath))
	{
	    LstDestroy(PthGSteps(nowPath), TRUE);
	    PthSSteps(nowPath, LstCreate(&stpTools));
	}
    PthSEnd(nowPath, ndAddr);
    ExOnTheFly(theGraph, nowPath, FALSE);
}

/***************************************************************************
*
* Function:  TrnChkEnbl
* Purpose:   This routine checks if a transition is enabled, but only
*            in the case that no enabled transition has been found yet.
* Arguments: trn - 
* Returns:
* See also:  TrnEnblCheck
*
****************************************************************************
*/
FE1(BOOL, TrnChkEnbl,
    TRN *, trn)
{
    TRARC *trArc;
    TRDRC *trDrc;

    if (TrnGPurIn(trn))
	{
	    LSTFOR (TrnGPurIn(trn), trArc, TRARC)
		{
		    if (PlcGCount(TrArcGPlc(trArc)) < TrArcGWeight(trArc))
			return(FALSE);
		}
	}
    if (TrnGNeutral(trn))
	{
	    LSTFOR (TrnGNeutral(trn), trArc, TRARC)
		{
		    if (PlcGCount(TrArcGPlc(trArc)) < TrArcGWeight(trArc))
			return(FALSE);
		}
	}
    if (TrnGMosIn(trn))
	{
	    LSTFOR (TrnGMosIn(trn), trDrc, TRDRC)
		{
		    if (PlcGCount(TrDrcGPlc(trDrc)) < TrDrcGIn(trDrc))
			return(FALSE);
		}
	}
    if (TrnGMosOut(trn))
	{
	    LSTFOR (TrnGMosOut(trn), trDrc, TRDRC)
		{
		    if (PlcGCount(TrDrcGPlc(trDrc)) < TrDrcGIn(trDrc))
			return(FALSE);
		}
	}
    return(TRUE);
}

/***************************************************************************
*
* Function:  PtC2
* Purpose:   This on-the-fly verification routine handles a possible case
*            where an error of the form specified by the deadlock monitor
*            states of the tester is found.
* Arguments: awAddr - a predecessor arrow of ndAddr (-1L if we are in
*                     the initial marking)
*            ndAddr - the node where the possible error is observed
* Returns:
* See also:  PtC1, PtC3, PtC4, PtOnTheFly, PtInfinite, PtBuchi,
*            IllStabFail (inst.c), ExOnTheFly (graph.c)
*
****************************************************************************
*/
FE2(void, PtC2,
    unsigned long, awAddr,
    unsigned long, ndAddr)
{
    if (PtrBlApply(TblTrn(0), TrnChkEnbl))
	return;
    if (awAddr != (unsigned long) -1L)
	PtGphUpdate();
    GphAddTermNode(theGraph, ndAddr);
    GphIncGenCount(theGraph);
    if (theDots)
	DotOutput();
    if (Divergence())
	BalTrDestroy(theNodeTree, FALSE);
    if (Ara())
	LTSEnd();
    (void) fprintf(_out, "\nDeadlock reached\n");
    if (PtStub())
	GphSetSeed(theGraph, PtTerm());
    if (PthLast(nowPath))
	{
	    LstDestroy(PthGSteps(nowPath), TRUE);
	    PthSSteps(nowPath, LstCreate(&stpTools));
	}
    PthSEnd(nowPath, ndAddr);
    ExOnTheFly(theGraph, nowPath, FALSE);
}

/***************************************************************************
*
* Function:  PtC3
* Purpose:   This on-the-fly verification routine handles a case
*            where an error of the form specified by the livelock monitor
*            states of the tester is found.
* Arguments: awAddr - a predecessor arrow of ndAddr (-1L if we are in
*                     the initial marking)
*            ndAddr - the node where the error was observed
* Returns:
* See also:  PtC1, PtC2, PtC4, PtOnTheFly, PtInfinite, PtBuchi,
*            IllDivTrace (inst.c), ExOnTheFly (graph.c)
*
****************************************************************************
*/
FE2(void, PtC3,
    unsigned long, awAddr,
    unsigned long, ndAddr)
{
    BalTrDestroy(theNodeTree, FALSE);
    PthInsLast(thePath, awAddr, ndAddr);
    if (Ara())
	LTSEnd();
    (void) fprintf(_out, "\nLivelock reached\n");
    if (PtStub())
	GphSetSeed(theGraph, PtTerm());
    ExOnTheFly(theGraph, thePath, TRUE);
}

/***************************************************************************
*
* Function:  PtC4
* Purpose:   This on-the-fly verification routine handles a case
*            where an error of the form specified by the infinite path
*            monitor states of the tester is found.
* Arguments: awAddr - a predecessor arrow of ndAddr (-1L if we are in
*                     the initial marking)
*            ndAddr - the node where the error was observed
* Returns:
* See also:  PtC1, PtC2, PtC3, PtOnTheFly, PtInfinite, PtBuchi,
*            IllInfTrace (inst.c), ExOnTheFly (graph.c)
*
****************************************************************************
*/
FE2(void, PtC4,
    unsigned long, awAddr,
    unsigned long, ndAddr)
{
    PthDiscard(the2Path);
    if (Divergence())
	BalTrDestroy(theNodeTree, FALSE);
    PthInsLast(the3Path, awAddr, ndAddr);
    if (Ara())
	LTSEnd();
    (void) fprintf(_out, "\nIllegal infinite path found\n");
    if (PtStub())
	GphSetSeed(theGraph, PtTerm());
    ExOnTheFly(theGraph, the3Path, TRUE);
}

/***************************************************************************
*
* Function:  PtFollowers
* Purpose:   This routine is the main "state expansion" routine when
*            there is no on-the-fly verification and no CFFD-preserving
*            version of the stubborn set method.
* Arguments:
* Returns:
* See also:  PtOnTheFly, PtInfinite, PtBuchi, PtProCFFD
*
****************************************************************************
*/
FE0(void, PtFollowers)
{
    nowPath = thePath;
    theAwCount = 0L;
    theAwFirst = (unsigned long) -1L;
    if (PtSleep())
	{
	    if (Pt2SlpSet())
		{
		    theAwCount = GphAwCount(theGraph, PthEnd(thePath));
		    if (theAwCount != 0L)
			GphFlFirst(theGraph, PthEnd(thePath), &theAwFirst,
				   (unsigned long *) NULL);
		    PtS3SlpSet(BalTrCreate(&lngBalTools));
		    (void) BalTrApply(Pt2SlpSet(), PtOldSleep);
		    PtDSlpSet();
		    PtSSlpSet(Pt3SlpSet());
		}
	    else
		{
		    PtS3SlpSet(BalTrCopy(PtSlpSet(), TRUE));
		    PtExpand();
		    PtD3SlpSet();
		}
	    PtGphUpdate();
	    PtDSlpSet();
	    if (Pt2SlpSet())
		PtD2SlpSet();
	}
    else if (PtStub())
	{
	    FndEnbl();
	    if (
#ifdef SMODELS
		(
#endif
		 PtGECnt() <= 1L
#ifdef SMODELS
		 )
		|| (PtSmodels() && (OnTheFly() || PtCFFD()) &&
		    (PtGInvisCnt() == 0L))
#endif
		)
		{
		    (void) PtrBlApply(TblTrn(0), TrnReCheck);
		    PtGphUpdate();
		}
	    else if (PtDelAlg())
		PtDeletionAlg();
	    else
		PtIncrementalAlg();
	}
    else
	{
	    for (PtCClass(); PtClass() < IngGNofClasses(); PtIClass())
		{
		    (void) PtrBlApply(TblTrn(PtClass()), TrnEnblCheck);
		    if (!Dependent(PtClass()))
			(void) PtrBlApply(TblTrn(PtClass()), TrnReCheck);
		}
	    for (PtCClass(); PtClass() < IngGNofClasses(); PtIClass())
		if (Dependent(PtClass()) && !SelfHigher(PtClass()))
		    {
			if (Higher(PtClass()))
			    (void) PtrBlApply(TblTrn(PtClass()), TrnPrecCheck);
			else
			    (void) PtrBlApply(TblTrn(PtClass()), TrnReCheck);
		    }
	    PtGphUpdate();
	}
}

/***************************************************************************
*
* Function:  PtExpand
* Purpose:   This is the main routine of the sleep set method.
* Arguments:
* Returns:
* See also:
*
****************************************************************************
*/
FE0(void, PtExpand)
{
    ptDoIt = FALSE;
    FndEnbl();
    if (PtGECnt() == 0L)
	return;
    if ((OnTheFly() || PtCFFD()) && (PtGInvisCnt() == 0L))
	{
	    (void) PtrBlApply(TblTrn(0), TrnReCheck);
	    return;
	}
    if (PtDelAlg())
	PtDeletionAlg();
    else if (PtStub())
	PtIncrementalAlg();
    (void) PtrBlApply(TblTrn(0), TrnChkRe);
}

/***************************************************************************
*
* Function:  TrnCntCand
* Purpose:   This routine is used for computing the number of enabled
*            transitions that would remain if a certain transition
*            is deleted in the deletion algorithm of the stubborn set
*            method.
* Arguments: trn - a transition
* Returns:
* See also:  TrnCntICand
*
****************************************************************************
*/
FE1(BOOL, TrnCntCand,
    TRN *, trn)
{
    if (!trn)
	return(TRUE);
    if (TrnEnabled(trn) && (!TrnSatisfied(trn)) && (!TrnThirsty(trn)))
	ptM++;
    return(FALSE);
}

/***************************************************************************
*
* Function:  TrnCntICand
* Purpose:   This routine is like TrnCntCand, except that invisible
*            transitions are counted only.
* Arguments: trn - a transition
* Returns:
* See also:  TrnCntCand
*
****************************************************************************
*/
FE1(BOOL, TrnCntICand,
    TRN *, trn)
{
    if (!trn)
	return(TRUE);
    if (TrnEnabled(trn) && (!TrnVisible(trn)) &&
	(!TrnSatisfied(trn)) && (!TrnThirsty(trn)))
	ptM++;
    return(FALSE);
}

/***************************************************************************
*
* Function:  TrnVRemoved
* Purpose:   This routine is used for finding a visible
*            transition that would be deleted if a certain transition
*            is deleted in the deletion algorithm of the stubborn set
*            method.
* Arguments: trn - a transition
* Returns:   TRUE if and only if trn is a kind of a transition wanted
* See also:
*
****************************************************************************
*/
FE1(BOOL, TrnVRemoved,
    TRN *, trn)
{
    return((!trn) ||
	   (TrnVisible(trn) && (TrnSatisfied(trn) || TrnThirsty(trn))));
}

/***************************************************************************
*
* Function:  TrnWorseVis
* Purpose:   This routine is used for finding a visible
*            transitions that is not in the chosen strongly connected
*            component in the incremental algorithm of the stubborn
*            set method. (If there is no such transition, the stubborn set
*            contains all visible transitions. The converse does not
*            necessarily hold, but it would take a considerable amount
*            of time to compute a stubborn set that would contain the
*            transitions of the component and then as many disabled
*            transitions as possible.)
* Arguments: trn - a transition
* Returns:   TRUE if and only if trn is a kind of a transition wanted
* See also:
*
****************************************************************************
*/
FE1(BOOL, TrnWorseVis,
    TRN *, trn)
{
    return(TrnVisible(trn) && (TrnGVal(trn) != PtBest()));
}

/***************************************************************************
*
* Function:  TrnFndStrong
* Purpose:   This routine is used for checking whether any key
*            transition remains if a certain transition is deleted in
*            the deletion algorithm of the stubborn set method.
* Arguments: trn - a transition
* Returns:   TRUE if and only if trn is a key transition
* See also:
*
****************************************************************************
*/
FE1(BOOL, TrnFndStrong,
    TRN *, trn)
{
    if (!trn)
      return(TRUE);
    if (TrnSatisfied(trn) || TrnThirsty(trn))
      return(FALSE);
    if (TrnGPurIn(trn))
	LSTFOR (TrnGPurIn(trn), ptTrArc, TRARC)
	    {
		ptPlc = TrArcGPlc(ptTrArc);
		if (PlcGPurOut(ptPlc))
		    LSTFOR (PlcGPurOut(ptPlc), ptPlArc, PLARC)
			{
			    ptTrn = PlArcGTrn(ptPlArc);
			    if (TrnSatisfied(ptTrn) || TrnThirsty(ptTrn))
				return(FALSE);
			}
		if (PlcGMosOut(ptPlc))
		    LSTFOR (PlcGMosOut(ptPlc), ptPlDrc, PLDRC)
			{
			    ptTrn = PlDrcGTrn(ptPlDrc);
			    if (TrnSatisfied(ptTrn) || TrnThirsty(ptTrn))
				return(FALSE);
			}
	    }
    if (TrnGNeutral(trn))
	LSTFOR (TrnGNeutral(trn), ptTrArc, TRARC)
	    {
		ptPlc = TrArcGPlc(ptTrArc);
		if (PlcGPurOut(ptPlc))
		    LSTFOR (PlcGPurOut(ptPlc), ptPlArc, PLARC)
			{
			    ptTrn = PlArcGTrn(ptPlArc);
			    if (TrnSatisfied(ptTrn) || TrnThirsty(ptTrn))
				return(FALSE);
			}
		if (PlcGMosOut(ptPlc))
		    LSTFOR (PlcGMosOut(ptPlc), ptPlDrc, PLDRC)
			{
			    ptTrn = PlDrcGTrn(ptPlDrc);
			    if (TrnSatisfied(ptTrn) || TrnThirsty(ptTrn))
				return(FALSE);
			}
	    }
    if (TrnGMosIn(trn))
	LSTFOR (TrnGMosIn(trn), ptTrDrc, TRDRC)
	    {
		ptPlc = TrDrcGPlc(ptTrDrc);
		if (PlcGPurOut(ptPlc))
		    LSTFOR (PlcGPurOut(ptPlc), ptPlArc, PLARC)
			{
			    ptTrn = PlArcGTrn(ptPlArc);
			    if (TrnSatisfied(ptTrn) || TrnThirsty(ptTrn))
				return(FALSE);
			}
		if (PlcGMosOut(ptPlc))
		    LSTFOR (PlcGMosOut(ptPlc), ptPlDrc, PLDRC)
			{
			    ptTrn = PlDrcGTrn(ptPlDrc);
			    if (TrnSatisfied(ptTrn) || TrnThirsty(ptTrn))
				return(FALSE);
			}
	    }
    if (TrnGMosOut(trn))
	LSTFOR (TrnGMosOut(trn), ptTrDrc, TRDRC)
	    {
		ptPlc = TrDrcGPlc(ptTrDrc);
		if (PlcGPurOut(ptPlc))
		    LSTFOR (PlcGPurOut(ptPlc), ptPlArc, PLARC)
			{
			    ptTrn = PlArcGTrn(ptPlArc);
			    if (TrnSatisfied(ptTrn) || TrnThirsty(ptTrn))
				return(FALSE);
			}
		if (PlcGMosOut(ptPlc))
		    LSTFOR (PlcGMosOut(ptPlc), ptPlDrc, PLDRC)
			{
			    ptTrn = PlDrcGTrn(ptPlDrc);
			    if (TrnSatisfied(ptTrn) || TrnThirsty(ptTrn))
				return(FALSE);
			}
	    }
    return(TRUE);
}
