/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  ptsymm.c
*
* Part of: PROD's programs araprod and "gphgen"
*
* Author:  Kimmo Varpaaniemi
*
* Current version:
*     $Id$
*
* Description:
*     The routines in this module are needed for the computation
*     of place symmetries of an unfolded net. The used algorithm
*     is described in "On Computing Symmetries and Stubborn Sets",
*     the technical report B 12 (April 1994) of the Digital Systems
*     Laboratory of Helsinki University of Technology. As said in
*     the report, the algorithm is a straightforward modification
*     of the symmetry computation algorithm presented in "Symmetries
*     of Petri Nets", an article written by Karsten Schmidt and
*     published in Petri Net Newsletter 43 (1993), pp. 9-25.
*
*     The computation of place symmetries takes long, often too long.
*     It should be possible to implement the algorithm much more
*     efficiently than is done in this module. Instead of commenting
*     this module in detail, I suggest that someone designs something
*     strictly better from scratch.
*
* Version history:
*     $Log$
***************************************************************************/

/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include <stdio.h>
#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "gphutil.h"
#include "graph.h"
#include "mark.h"
#include "encode.h"
#include "ptnet.h"
#include "ptsymm.h"

LIST *_plcSyGenerators;
BALTREE *_plcSymmetries;
BOOL _plcIMkSymm;

#ifdef DEBUG
FE2(void, PtrDump,
    char *, ptr,
    unsigned, depth)
{
    DumpIndent(depth);
    (void) fprintf(_err, "pointer 0x%lx\n", (unsigned long) ptr);
    (void) fflush(_err);
}
#endif

FE2(int, PtrCompare,
    char *, ptr1,
    char *, ptr2)
{
    return((ptr1 == ptr2) ? 0 : ((ptr1 < ptr2) ? -1 : 1));
}

TOOLS ptrTools =
{ (CFUNC) NULL, NULL
#ifdef DEBUG
, PtrDump
#endif
 };

BALTOOLS ptrBalTools =
{ (CFUNC) NULL, NULL
#ifdef DEBUG
, PtrDump
#endif
, PtrCompare, (FOOL) NULL };

BALTREE *ptrTree;
BALTREE *ptrTree2;

FE1(BOOL, PtrCut,
    char *, ptr)
{
    if (BalTrElement(ptrTree2, ptr))
	(void) BalTrInsert(ptrTree, ptr, ptr, TRUE);
    return(FALSE);
}

FE1(BOOL, PtrMinus,
    char *, ptr)
{
    (void) BalTrDelete(ptrTree, (char *) NULL, ptr, FALSE);
    return(FALSE);
}

FE1(BOOL, PtrAdd,
    char *, ptr)
{
    (void) BalTrInsert(ptrTree, ptr, ptr, FALSE);
    return(FALSE);
}

FE2(BALTREE *, PtrTrCut,
    BALTREE *, ptrTr1,
    BALTREE *, ptrTr2)
{
    ptrTree = BalTrCreate(&ptrBalTools);
    ptrTree2 = ptrTr2;
    (void) BalTrApply(ptrTr1, PtrCut);
    return(ptrTree);
}

FE2(BALTREE *, PtrTrMinus,
    BALTREE *, ptrTr1,
    BALTREE *, ptrTr2)
{
    ptrTree = BalTrCopy(ptrTr1, FALSE);
    (void) BalTrApply(ptrTr2, PtrMinus);
    return(ptrTree);
}

FE2(int, PlcDegCompare,
    PLC *, plc1,
    PLC *, plc2)
{
    static unsigned count1, count2;
    static PLARC *plArc1, *plArc2;
    static PLDRC *plDrc1, *plDrc2;
    static BALTREE *tree1, *tree2;
    static int cmpVal;

    if (plc1 == plc2)
	return(0);
    if (PlcIMkSymm())
	{
	    if (PlcGCount(plc1) > PlcGCount(plc2))
		return(1);
	    if (PlcGCount(plc2) > PlcGCount(plc1))
		return(-1);
	}
    if (PlcGPurIn(plc1) && !PlcGPurIn(plc2))
	return(1);
    if (PlcGPurIn(plc2) && !PlcGPurIn(plc1))
	return(-1);
    if (PlcGPurOut(plc1) && !PlcGPurOut(plc2))
	return(1);
    if (PlcGPurOut(plc2) && !PlcGPurOut(plc1))
	return(-1);
    if (PlcGNeutral(plc1) && !PlcGNeutral(plc2))
	return(1);
    if (PlcGNeutral(plc2) && !PlcGNeutral(plc1))
	return(-1);
    if (PlcGMosIn(plc1) && !PlcGMosIn(plc2))
	return(1);
    if (PlcGMosIn(plc2) && !PlcGMosIn(plc1))
	return(-1);
    if (PlcGMosOut(plc1) && !PlcGMosOut(plc2))
	return(1);
    if (PlcGMosOut(plc2) && !PlcGMosOut(plc1))
	return(-1);
    if (PlcGPurIn(plc1))
	{
	    count1 = LstSize(PlcGPurIn(plc1));
	    count2 = LstSize(PlcGPurIn(plc2));
	    if (count1 > count2)
		return(1);
	    if (count2 > count1)
		return(-1);
	}
    if (PlcGPurOut(plc1))
	{
	    count1 = LstSize(PlcGPurOut(plc1));
	    count2 = LstSize(PlcGPurOut(plc2));
	    if (count1 > count2)
		return(1);
	    if (count2 > count1)
		return(-1);
	}
    if (PlcGNeutral(plc1))
	{
	    count1 = LstSize(PlcGNeutral(plc1));
	    count2 = LstSize(PlcGNeutral(plc2));
	    if (count1 > count2)
		return(1);
	    if (count2 > count1)
		return(-1);
	}
    if (PlcGMosIn(plc1))
	{
	    count1 = LstSize(PlcGMosIn(plc1));
	    count2 = LstSize(PlcGMosIn(plc2));
	    if (count1 > count2)
		return(1);
	    if (count2 > count1)
		return(-1);
	}
    if (PlcGMosOut(plc1))
	{
	    count1 = LstSize(PlcGMosOut(plc1));
	    count2 = LstSize(PlcGMosOut(plc2));
	    if (count1 > count2)
		return(1);
	    if (count2 > count1)
		return(-1);
	}
    if (PlcGPurIn(plc1))
	{
	    tree1 = BalTrCreate(&plArcBalTools);
	    LSTFOR (PlcGPurIn(plc1), plArc1, PLARC)
		(void) BalTrInsert(tree1, (char *) plArc1,
				   (char *) plArc1, FALSE);
	    tree2 = BalTrCreate(&plArcBalTools);
	    LSTFOR (PlcGPurIn(plc2), plArc2, PLARC)
		(void) BalTrInsert(tree2, (char *) plArc2,
				   (char *) plArc2, FALSE);
	    plArc1 = (PLARC *) BalTrFirst(tree1);
	    plArc2 = (PLARC *) BalTrFirst(tree2);
	    while (plArc1)
		{
		    if ((cmpVal = PlArcCompare(plArc1, plArc2)))
			break;
		    plArc1 = (PLARC *)
			BalTrNext(tree1, (char *) plArc1,
				  (char *) plArc1, (char *) NULL);
		    plArc2 = (PLARC *)
			BalTrNext(tree2, (char *) plArc2,
				  (char *) plArc2, (char *) NULL);
		}
	    BalTrDestroy(tree1, FALSE);
	    BalTrDestroy(tree2, FALSE);
	    if (plArc1)
		return(cmpVal);
	}
    if (PlcGPurOut(plc1))
	{
	    tree1 = BalTrCreate(&plArcBalTools);
	    LSTFOR (PlcGPurOut(plc1), plArc1, PLARC)
		(void) BalTrInsert(tree1, (char *) plArc1,
				   (char *) plArc1, FALSE);
	    tree2 = BalTrCreate(&plArcBalTools);
	    LSTFOR (PlcGPurOut(plc2), plArc2, PLARC)
		(void) BalTrInsert(tree2, (char *) plArc2,
				   (char *) plArc2, FALSE);
	    plArc1 = (PLARC *) BalTrFirst(tree1);
	    plArc2 = (PLARC *) BalTrFirst(tree2);
	    while (plArc1)
		{
		    if ((cmpVal = PlArcCompare(plArc1, plArc2)))
			break;
		    plArc1 = (PLARC *)
			BalTrNext(tree1, (char *) plArc1,
				  (char *) plArc1, (char *) NULL);
		    plArc2 = (PLARC *)
			BalTrNext(tree2, (char *) plArc2,
				  (char *) plArc2, (char *) NULL);
		}
	    BalTrDestroy(tree1, FALSE);
	    BalTrDestroy(tree2, FALSE);
	    if (plArc1)
		return(cmpVal);
	}
    if (PlcGNeutral(plc1))
	{
	    tree1 = BalTrCreate(&plArcBalTools);
	    LSTFOR (PlcGNeutral(plc1), plArc1, PLARC)
		(void) BalTrInsert(tree1, (char *) plArc1,
				   (char *) plArc1, FALSE);
	    tree2 = BalTrCreate(&plArcBalTools);
	    LSTFOR (PlcGNeutral(plc2), plArc2, PLARC)
		(void) BalTrInsert(tree2, (char *) plArc2,
				   (char *) plArc2, FALSE);
	    plArc1 = (PLARC *) BalTrFirst(tree1);
	    plArc2 = (PLARC *) BalTrFirst(tree2);
	    while (plArc1)
		{
		    if ((cmpVal = PlArcCompare(plArc1, plArc2)))
			break;
		    plArc1 = (PLARC *)
			BalTrNext(tree1, (char *) plArc1,
				  (char *) plArc1, (char *) NULL);
		    plArc2 = (PLARC *)
			BalTrNext(tree2, (char *) plArc2,
				  (char *) plArc2, (char *) NULL);
		}
	    BalTrDestroy(tree1, FALSE);
	    BalTrDestroy(tree2, FALSE);
	    if (plArc1)
		return(cmpVal);
	}
    if (PlcGMosIn(plc1))
	{
	    tree1 = BalTrCreate(&plDrcBalTools);
	    LSTFOR (PlcGMosIn(plc1), plDrc1, PLDRC)
		(void) BalTrInsert(tree1, (char *) plDrc1,
				   (char *) plDrc1, FALSE);
	    tree2 = BalTrCreate(&plDrcBalTools);
	    LSTFOR (PlcGMosIn(plc2), plDrc2, PLDRC)
		(void) BalTrInsert(tree2, (char *) plDrc2,
				   (char *) plDrc2, FALSE);
	    plDrc1 = (PLDRC *) BalTrFirst(tree1);
	    plDrc2 = (PLDRC *) BalTrFirst(tree2);
	    while (plDrc1)
		{
		    if ((cmpVal = PlDrcCompare(plDrc1, plDrc2)))
			break;
		    plDrc1 = (PLDRC *)
			BalTrNext(tree1, (char *) plDrc1,
				  (char *) plDrc1, (char *) NULL);
		    plDrc2 = (PLDRC *)
			BalTrNext(tree2, (char *) plDrc2,
				  (char *) plDrc2, (char *) NULL);
		}
	    BalTrDestroy(tree1, FALSE);
	    BalTrDestroy(tree2, FALSE);
	    if (plDrc1)
		return(cmpVal);
	}
    if (PlcGMosOut(plc1))
	{
	    tree1 = BalTrCreate(&plDrcBalTools);
	    LSTFOR (PlcGMosOut(plc1), plDrc1, PLDRC)
		(void) BalTrInsert(tree1, (char *) plDrc1,
				   (char *) plDrc1, FALSE);
	    tree2 = BalTrCreate(&plDrcBalTools);
	    LSTFOR (PlcGMosOut(plc2), plDrc2, PLDRC)
		(void) BalTrInsert(tree2, (char *) plDrc2,
				   (char *) plDrc2, FALSE);
	    plDrc1 = (PLDRC *) BalTrFirst(tree1);
	    plDrc2 = (PLDRC *) BalTrFirst(tree2);
	    while (plDrc1)
		{
		    if ((cmpVal = PlDrcCompare(plDrc1, plDrc2)))
			break;
		    plDrc1 = (PLDRC *)
			BalTrNext(tree1, (char *) plDrc1,
				  (char *) plDrc1, (char *) NULL);
		    plDrc2 = (PLDRC *)
			BalTrNext(tree2, (char *) plDrc2,
				  (char *) plDrc2, (char *) NULL);
		}
	    BalTrDestroy(tree1, FALSE);
	    BalTrDestroy(tree2, FALSE);
	    if (plDrc1)
		return(cmpVal);
	}
    return(0);
}

BALTOOLS clpBalTools =
{ (CFUNC) NULL, NULL
#ifdef DEBUG
, PlcDump
#endif
, PlcDegCompare, (FOOL) NULL };

FE2(int, TrnDegCompare,
    TRN *, trn1,
    TRN *, trn2)
{
    static unsigned count1, count2;
    static TRARC *trArc1, *trArc2;
    static TRDRC *trDrc1, *trDrc2;
    static BALTREE *tree1, *tree2;
    static int cmpVal;

    if (trn1 == trn2)
	return(0);
    if (TrnGPurIn(trn1) && !TrnGPurIn(trn2))
	return(1);
    if (TrnGPurIn(trn2) && !TrnGPurIn(trn1))
	return(-1);
    if (TrnGPurOut(trn1) && !TrnGPurOut(trn2))
	return(1);
    if (TrnGPurOut(trn2) && !TrnGPurOut(trn1))
	return(-1);
    if (TrnGNeutral(trn1) && !TrnGNeutral(trn2))
	return(1);
    if (TrnGNeutral(trn2) && !TrnGNeutral(trn1))
	return(-1);
    if (TrnGMosIn(trn1) && !TrnGMosIn(trn2))
	return(1);
    if (TrnGMosIn(trn2) && !TrnGMosIn(trn1))
	return(-1);
    if (TrnGMosOut(trn1) && !TrnGMosOut(trn2))
	return(1);
    if (TrnGMosOut(trn2) && !TrnGMosOut(trn1))
	return(-1);
    if (TrnGPurIn(trn1))
	{
	    count1 = LstSize(TrnGPurIn(trn1));
	    count2 = LstSize(TrnGPurIn(trn2));
	    if (count1 > count2)
		return(1);
	    if (count2 > count1)
		return(-1);
	}
    if (TrnGPurOut(trn1))
	{
	    count1 = LstSize(TrnGPurOut(trn1));
	    count2 = LstSize(TrnGPurOut(trn2));
	    if (count1 > count2)
		return(1);
	    if (count2 > count1)
		return(-1);
	}
    if (TrnGNeutral(trn1))
	{
	    count1 = LstSize(TrnGNeutral(trn1));
	    count2 = LstSize(TrnGNeutral(trn2));
	    if (count1 > count2)
		return(1);
	    if (count2 > count1)
		return(-1);
	}
    if (TrnGMosIn(trn1))
	{
	    count1 = LstSize(TrnGMosIn(trn1));
	    count2 = LstSize(TrnGMosIn(trn2));
	    if (count1 > count2)
		return(1);
	    if (count2 > count1)
		return(-1);
	}
    if (TrnGMosOut(trn1))
	{
	    count1 = LstSize(TrnGMosOut(trn1));
	    count2 = LstSize(TrnGMosOut(trn2));
	    if (count1 > count2)
		return(1);
	    if (count2 > count1)
		return(-1);
	}
    if (TrnGPurIn(trn1))
	{
	    tree1 = BalTrCreate(&trArcBalTools);
	    LSTFOR (TrnGPurIn(trn1), trArc1, TRARC)
		(void) BalTrInsert(tree1, (char *) trArc1,
				   (char *) trArc1, FALSE);
	    tree2 = BalTrCreate(&trArcBalTools);
	    LSTFOR (TrnGPurIn(trn2), trArc2, TRARC)
		(void) BalTrInsert(tree2, (char *) trArc2,
				   (char *) trArc2, FALSE);
	    trArc1 = (TRARC *) BalTrFirst(tree1);
	    trArc2 = (TRARC *) BalTrFirst(tree2);
	    while (trArc1)
		{
		    if ((cmpVal = TrArcCompare(trArc1, trArc2)))
			break;
		    trArc1 = (TRARC *)
			BalTrNext(tree1, (char *) trArc1,
				  (char *) trArc1, (char *) NULL);
		    trArc2 = (TRARC *)
			BalTrNext(tree2, (char *) trArc2,
				  (char *) trArc2, (char *) NULL);
		}
	    BalTrDestroy(tree1, FALSE);
	    BalTrDestroy(tree2, FALSE);
	    if (trArc1)
		return(cmpVal);
	}
    if (TrnGPurOut(trn1))
	{
	    tree1 = BalTrCreate(&trArcBalTools);
	    LSTFOR (TrnGPurOut(trn1), trArc1, TRARC)
		(void) BalTrInsert(tree1, (char *) trArc1,
				   (char *) trArc1, FALSE);
	    tree2 = BalTrCreate(&trArcBalTools);
	    LSTFOR (TrnGPurOut(trn2), trArc2, TRARC)
		(void) BalTrInsert(tree2, (char *) trArc2,
				   (char *) trArc2, FALSE);
	    trArc1 = (TRARC *) BalTrFirst(tree1);
	    trArc2 = (TRARC *) BalTrFirst(tree2);
	    while (trArc1)
		{
		    if ((cmpVal = TrArcCompare(trArc1, trArc2)))
			break;
		    trArc1 = (TRARC *)
			BalTrNext(tree1, (char *) trArc1,
				  (char *) trArc1, (char *) NULL);
		    trArc2 = (TRARC *)
			BalTrNext(tree2, (char *) trArc2,
				  (char *) trArc2, (char *) NULL);
		}
	    BalTrDestroy(tree1, FALSE);
	    BalTrDestroy(tree2, FALSE);
	    if (trArc1)
		return(cmpVal);
	}
    if (TrnGNeutral(trn1))
	{
	    tree1 = BalTrCreate(&trArcBalTools);
	    LSTFOR (TrnGNeutral(trn1), trArc1, TRARC)
		(void) BalTrInsert(tree1, (char *) trArc1,
				   (char *) trArc1, FALSE);
	    tree2 = BalTrCreate(&trArcBalTools);
	    LSTFOR (TrnGNeutral(trn2), trArc2, TRARC)
		(void) BalTrInsert(tree2, (char *) trArc2,
				   (char *) trArc2, FALSE);
	    trArc1 = (TRARC *) BalTrFirst(tree1);
	    trArc2 = (TRARC *) BalTrFirst(tree2);
	    while (trArc1)
		{
		    if ((cmpVal = TrArcCompare(trArc1, trArc2)))
			break;
		    trArc1 = (TRARC *)
			BalTrNext(tree1, (char *) trArc1,
				  (char *) trArc1, (char *) NULL);
		    trArc2 = (TRARC *)
			BalTrNext(tree2, (char *) trArc2,
				  (char *) trArc2, (char *) NULL);
		}
	    BalTrDestroy(tree1, FALSE);
	    BalTrDestroy(tree2, FALSE);
	    if (trArc1)
		return(cmpVal);
	}
    if (TrnGMosIn(trn1))
	{
	    tree1 = BalTrCreate(&trDrcBalTools);
	    LSTFOR (TrnGMosIn(trn1), trDrc1, TRDRC)
		(void) BalTrInsert(tree1, (char *) trDrc1,
				   (char *) trDrc1, FALSE);
	    tree2 = BalTrCreate(&trDrcBalTools);
	    LSTFOR (TrnGMosIn(trn2), trDrc2, TRDRC)
		(void) BalTrInsert(tree2, (char *) trDrc2,
				   (char *) trDrc2, FALSE);
	    trDrc1 = (TRDRC *) BalTrFirst(tree1);
	    trDrc2 = (TRDRC *) BalTrFirst(tree2);
	    while (trDrc1)
		{
		    if ((cmpVal = TrDrcCompare(trDrc1, trDrc2)))
			break;
		    trDrc1 = (TRDRC *)
			BalTrNext(tree1, (char *) trDrc1,
				  (char *) trDrc1, (char *) NULL);
		    trDrc2 = (TRDRC *)
			BalTrNext(tree2, (char *) trDrc2,
				  (char *) trDrc2, (char *) NULL);
		}
	    BalTrDestroy(tree1, FALSE);
	    BalTrDestroy(tree2, FALSE);
	    if (trDrc1)
		return(cmpVal);
	}
    if (TrnGMosOut(trn1))
	{
	    tree1 = BalTrCreate(&trDrcBalTools);
	    LSTFOR (TrnGMosOut(trn1), trDrc1, TRDRC)
		(void) BalTrInsert(tree1, (char *) trDrc1,
				   (char *) trDrc1, FALSE);
	    tree2 = BalTrCreate(&trDrcBalTools);
	    LSTFOR (TrnGMosOut(trn2), trDrc2, TRDRC)
		(void) BalTrInsert(tree2, (char *) trDrc2,
				   (char *) trDrc2, FALSE);
	    trDrc1 = (TRDRC *) BalTrFirst(tree1);
	    trDrc2 = (TRDRC *) BalTrFirst(tree2);
	    while (trDrc1)
		{
		    if ((cmpVal = TrDrcCompare(trDrc1, trDrc2)))
			break;
		    trDrc1 = (TRDRC *)
			BalTrNext(tree1, (char *) trDrc1,
				  (char *) trDrc1, (char *) NULL);
		    trDrc2 = (TRDRC *)
			BalTrNext(tree2, (char *) trDrc2,
				  (char *) trDrc2, (char *) NULL);
		}
	    BalTrDestroy(tree1, FALSE);
	    BalTrDestroy(tree2, FALSE);
	    if (trDrc1)
		return(cmpVal);
	}
    return(0);
}

BALTOOLS nrtBalTools =
{ (CFUNC) NULL, NULL
#ifdef DEBUG
, TrnDump
#endif
, TrnDegCompare, (FOOL) NULL };

unsigned long multipl;

FE1(BOOL, PlcPre,
    PLC *, plc)
{
    PLARC *plArc;
    PLDRC *plDrc;
    TRN *trn;

    if (PlcGPurIn(plc))
	LSTFOR (PlcGPurIn(plc), plArc, PLARC)
	    {
		if (PlArcGWeight(plArc) == multipl)
		    {
			trn = PlArcGTrn(plArc);
			(void) BalTrInsert(ptrTree, (char *) trn, (char *) trn,
					   TRUE);
		    }
	    }
    if (PlcGNeutral(plc))
	LSTFOR (PlcGNeutral(plc), plArc, PLARC)
	    {
		if (PlArcGWeight(plArc) == multipl)
		    {
			trn = PlArcGTrn(plArc);
			(void) BalTrInsert(ptrTree, (char *) trn, (char *) trn,
					   TRUE);
		    }
	    }
    if (PlcGMosIn(plc))
	LSTFOR (PlcGMosIn(plc), plDrc, PLDRC)
	    {
		if (PlDrcGIn(plDrc) == multipl)
		    {
			trn = PlDrcGTrn(plDrc);
			(void) BalTrInsert(ptrTree, (char *) trn, (char *) trn,
					   TRUE);
		    }
	    }
    if (PlcGMosOut(plc))
	LSTFOR (PlcGMosOut(plc), plDrc, PLDRC)
	    {
		if (PlDrcGIn(plDrc) == multipl)
		    {
			trn = PlDrcGTrn(plDrc);
			(void) BalTrInsert(ptrTree, (char *) trn, (char *) trn,
					   TRUE);
		    }
	    }
    return(FALSE);
}

FE1(BOOL, PlcPost,
    PLC *, plc)
{
    PLARC *plArc;
    PLDRC *plDrc;
    TRN *trn;

    if (PlcGPurOut(plc))
	LSTFOR (PlcGPurOut(plc), plArc, PLARC)
	    {
		if (PlArcGWeight(plArc) == multipl)
		    {
			trn = PlArcGTrn(plArc);
			(void) BalTrInsert(ptrTree, (char *) trn, (char *) trn,
					   TRUE);
		    }
	    }
    if (PlcGNeutral(plc))
	LSTFOR (PlcGNeutral(plc), plArc, PLARC)
	    {
		if (PlArcGWeight(plArc) == multipl)
		    {
			trn = PlArcGTrn(plArc);
			(void) BalTrInsert(ptrTree, (char *) trn, (char *) trn,
					   TRUE);
		    }
	    }
    if (PlcGMosIn(plc))
	LSTFOR (PlcGMosIn(plc), plDrc, PLDRC)
	    {
		if (PlDrcGOut(plDrc) == multipl)
		    {
			trn = PlDrcGTrn(plDrc);
			(void) BalTrInsert(ptrTree, (char *) trn, (char *) trn,
					   TRUE);
		    }
	    }
    if (PlcGMosOut(plc))
	LSTFOR (PlcGMosOut(plc), plDrc, PLDRC)
	    {
		if (PlDrcGOut(plDrc) == multipl)
		    {
			trn = PlDrcGTrn(plDrc);
			(void) BalTrInsert(ptrTree, (char *) trn, (char *) trn,
					   TRUE);
		    }
	    }
    return(FALSE);
}

FE1(BOOL, TrnPre,
    TRN *, trn)
{
    TRARC *trArc;
    TRDRC *trDrc;
    PLC *plc;

    if (TrnGPurIn(trn))
	LSTFOR (TrnGPurIn(trn), trArc, TRARC)
	    {
		if (TrArcGWeight(trArc) == multipl)
		    {
			plc = TrArcGPlc(trArc);
			(void) BalTrInsert(ptrTree, (char *) plc, (char *) plc,
					   TRUE);
		    }
	    }
    if (TrnGNeutral(trn))
	LSTFOR (TrnGNeutral(trn), trArc, TRARC)
	    {
		if (TrArcGWeight(trArc) == multipl)
		    {
			plc = TrArcGPlc(trArc);
			(void) BalTrInsert(ptrTree, (char *) plc, (char *) plc,
					   TRUE);
		    }
	    }
    if (TrnGMosIn(trn))
	LSTFOR (TrnGMosIn(trn), trDrc, TRDRC)
	    {
		if (TrDrcGIn(trDrc) == multipl)
		    {
			plc = TrDrcGPlc(trDrc);
			(void) BalTrInsert(ptrTree, (char *) plc, (char *) plc,
					   TRUE);
		    }
	    }
    if (TrnGMosOut(trn))
	LSTFOR (TrnGMosOut(trn), trDrc, TRDRC)
	    {
		if (TrDrcGIn(trDrc) == multipl)
		    {
			plc = TrDrcGPlc(trDrc);
			(void) BalTrInsert(ptrTree, (char *) plc, (char *) plc,
					   TRUE);
		    }
	    }
    return(FALSE);
}

FE1(BOOL, TrnPost,
    TRN *, trn)
{
    TRARC *trArc;
    TRDRC *trDrc;
    PLC *plc;

    if (TrnGPurOut(trn))
	LSTFOR (TrnGPurOut(trn), trArc, TRARC)
	    {
		if (TrArcGWeight(trArc) == multipl)
		    {
			plc = TrArcGPlc(trArc);
			(void) BalTrInsert(ptrTree, (char *) plc, (char *) plc,
					   TRUE);
		    }
	    }
    if (TrnGNeutral(trn))
	LSTFOR (TrnGNeutral(trn), trArc, TRARC)
	    {
		if (TrArcGWeight(trArc) == multipl)
		    {
			plc = TrArcGPlc(trArc);
			(void) BalTrInsert(ptrTree, (char *) plc, (char *) plc,
					   TRUE);
		    }
	    }
    if (TrnGMosIn(trn))
	LSTFOR (TrnGMosIn(trn), trDrc, TRDRC)
	    {
		if (TrDrcGOut(trDrc) == multipl)
		    {
			plc = TrDrcGPlc(trDrc);
			(void) BalTrInsert(ptrTree, (char *) plc, (char *) plc,
					   TRUE);
		    }
	    }
    if (TrnGMosOut(trn))
	LSTFOR (TrnGMosOut(trn), trDrc, TRDRC)
	    {
		if (TrDrcGOut(trDrc) == multipl)
		    {
			plc = TrDrcGPlc(trDrc);
			(void) BalTrInsert(ptrTree, (char *) plc, (char *) plc,
					   TRUE);
		    }
	    }
    return(FALSE);
}

FE2(BALTREE *, PrePlcTr,
    BALTREE *, plcTree,
    unsigned long, i)
{
    multipl = i;
    ptrTree = BalTrCreate(&ptrBalTools);
    (void) BalTrApply(plcTree, PlcPre);
    return(ptrTree);
}

FE2(BALTREE *, PostPlcTr,
    BALTREE *, plcTree,
    unsigned long, i)
{
    multipl = i;
    ptrTree = BalTrCreate(&ptrBalTools);
    (void) BalTrApply(plcTree, PlcPost);
    return(ptrTree);
}

FE2(BALTREE *, PreTrnTr,
    BALTREE *, trnTree,
    unsigned long, i)
{
    multipl = i;
    ptrTree = BalTrCreate(&ptrBalTools);
    (void) BalTrApply(trnTree, TrnPre);
    return(ptrTree);
}

FE2(BALTREE *, PostTrnTr,
    BALTREE *, trnTree,
    unsigned long, i)
{
    multipl = i;
    ptrTree = BalTrCreate(&ptrBalTools);
    (void) BalTrApply(trnTree, TrnPost);
    return(ptrTree);
}

FE1(BOOL, MltPlcPre,
    PLC *, plc)
{
    PLARC *plArc;
    PLDRC *plDrc;
    unsigned long *lng;

    if (PlcGPurIn(plc))
	LSTFOR (PlcGPurIn(plc), plArc, PLARC)
	    {
		lng = PlArcAWeight(plArc);
		(void) BalTrInsert(ptrTree, (char *) lng, (char *) lng, TRUE);
	    }
    if (PlcGNeutral(plc))
	LSTFOR (PlcGNeutral(plc), plArc, PLARC)
	    {
		lng = PlArcAWeight(plArc);
		(void) BalTrInsert(ptrTree, (char *) lng, (char *) lng, TRUE);
	    }
    if (PlcGMosIn(plc))
	LSTFOR (PlcGMosIn(plc), plDrc, PLDRC)
	    {
		lng = PlDrcAIn(plDrc);
		(void) BalTrInsert(ptrTree, (char *) lng, (char *) lng, TRUE);
	    }
    if (PlcGMosOut(plc))
	LSTFOR (PlcGMosOut(plc), plDrc, PLDRC)
	    {
		lng = PlDrcAIn(plDrc);
		(void) BalTrInsert(ptrTree, (char *) lng, (char *) lng, TRUE);
	    }
    return(FALSE);
}

FE1(BOOL, MltPlcPost,
    PLC *, plc)
{
    PLARC *plArc;
    PLDRC *plDrc;
    unsigned long *lng;

    if (PlcGPurOut(plc))
	LSTFOR (PlcGPurOut(plc), plArc, PLARC)
	    {
		lng = PlArcAWeight(plArc);
		(void) BalTrInsert(ptrTree, (char *) lng, (char *) lng, TRUE);
	    }
    if (PlcGNeutral(plc))
	LSTFOR (PlcGNeutral(plc), plArc, PLARC)
	    {
		lng = PlArcAWeight(plArc);
		(void) BalTrInsert(ptrTree, (char *) lng, (char *) lng, TRUE);
	    }
    if (PlcGMosIn(plc))
	LSTFOR (PlcGMosIn(plc), plDrc, PLDRC)
	    {
		lng = PlDrcAOut(plDrc);
		(void) BalTrInsert(ptrTree, (char *) lng, (char *) lng, TRUE);
	    }
    if (PlcGMosOut(plc))
	LSTFOR (PlcGMosOut(plc), plDrc, PLDRC)
	    {
		lng = PlDrcAOut(plDrc);
		(void) BalTrInsert(ptrTree, (char *) lng, (char *) lng, TRUE);
	    }
    return(FALSE);
}
