/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  redirect.c
*
* Part of: PROD's program probe
*
* Authors: Keijo Heljanko
*
* Current version:
*     $Id: redirect.c,v 3.1 1997/04/17 15:44:07 kepa Exp $
*
* Description:
*     This module handles redirection of PROBE output to a file
*     or to a filter prgram.
*
* Version history:
*     $Log: redirect.c,v $
 * Revision 3.1  1997/04/17  15:44:07  kepa
 * *** empty log message ***
 *
***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef UNIX
#include <unistd.h>
#endif
#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "ptrbl.h"
#include "baltree.h"
#include "bdd.h"
#include "mbufs.h"
#include "mark.h"
#include "expr.h"
#include "gphutil.h"
#include "graph.h"
#include "formula.h"
#include "comline.h"
#include "redirect.h"

static FILE *oldout;
static FILE *f;

#if (defined(UNIX) || defined(_WIN32))
static char *tempname = NULL;
#endif

/***************************************************************************
*
* Function:  InitRedirect
* Purpose:
* Arguments: r - a pointer to a redirect struct
* Returns:   BOOL - Was the redirection successful
* See also:
*
****************************************************************************
*/
FE1(BOOL, InitRedirect,
    REDIRECT *, r)
{
#if (defined(UNIX) || defined(_WIN32))
    register char *regTemp;
#endif
    oldout = _out;

    if(r->type != REDIRECT_NONE) {

#ifdef DEBUG_REDIRECT
	(void) fprintf(_err, "Redirection: %d \"%s\"\n",
		r->type,
		r->name);
	if(r->args != NULL) {
	    (void) fprintf(_err, "Args: \"%s\"\n",
		    r->args);
	}
#endif
	
	if(r->name == NULL) {
		r->type = REDIRECT_NONE;
		return FALSE;
	}
	
	switch(r->type) {

	  case REDIRECT_FILE:
            if(!(f = fopen(r->name, "w"))) {
		(void) fprintf(_err,
			       "Could not open file \"%s\" for redirection.\n",
			       r->name);
		(void) fflush(_err);
		r->type = REDIRECT_NONE;
		return FALSE;
	    }
	    _out = f;
	    break;
	    
	  case REDIRECT_FILE_APPEND:
            if(!(f = fopen(r->name, "a"))) {
		(void) fprintf(_err,
			       "Could not open file \"%s\" for redirection.\n",
			       r->name);
		(void) fflush(_err);
		r->type = REDIRECT_NONE;
		return FALSE;
	    }
	    
	    _out = f;
	    break;

	  case REDIRECT_FILTER:
	    
#if (defined(UNIX) || defined(_WIN32))
   	    regTemp = tmpnam((char *) NULL);
   	    tempname = StrCopy(regTemp);

            if((f = fopen(tempname, "w")) == NULL) {
		(void) fprintf(_err,
			       "Could not fopen temporary file for filter redirection.\n");
		(void) fflush(_err);
		r->type = REDIRECT_NONE;
		return FALSE;
	    }
	    
	    _out = f;
#else
		(void) fprintf(_err, 
			       "Redirecting to a filter not supported on this system.\n");
		(void) fflush(_err); 
#endif
	    break;
	    
	  case REDIRECT_NONE:
	  default:
	    break;
	}
    }

    return TRUE;
}

/***************************************************************************
*
* Function:  CloseRedirect
* Purpose:   Closes the redirection file and in case of a filter
*            feeds all the output through a forked filter process.
*            Also restores the original output destination.
* Arguments: r - a pointer to a redirect struct
* Returns:   BOOL - was the redirection closing successful
* See also:
*
****************************************************************************
*/
FE1(BOOL, CloseRedirect,
    REDIRECT *, r)
{
    _out = oldout;

    switch(r->type) {
      case REDIRECT_FILE:
      case REDIRECT_FILE_APPEND:
	(void) fflush(f);
	(void) fclose(f);
	break;
      case REDIRECT_FILTER:
#if (defined(UNIX) || defined(_WIN32))
	(void) fflush(f);

	if((/* status = */ fseek(f, 0L, 0))) {
	    (void) fprintf(_err, 
			   "Couldn't seek to the beginning of redirection temporary file.\n");
	    (void) fflush(_err);
	    (void) fclose(f);

	    /* But not so if we have created it ourselves. */
	
	    (void) unlink(tempname);
	    Deallocate(tempname);
	    tempname = NULL;
	    
	    r->type = REDIRECT_NONE;
	    return FALSE;
	}

	/* Close the file. */

	(void) fclose(f);

	/* status = */ (void) RunFilterSystem(r->name, r->args);
	
	/* But not so if we have created it ourselves. */
	
	(void) unlink(tempname);
	Deallocate(tempname);
	tempname = NULL;
#endif
	break;
      case REDIRECT_NONE:
      default:
	break;
    }
    
    return TRUE;
}

/***************************************************************************
*
* Function:  CleanupRedirect
* Purpose:   Deallocates all memory and other resources used in redirection.
*            Cleans the redirect struct.
* Arguments: 
* Returns:
* See also:
*
****************************************************************************
*/
FE1(BOOL, CleanupRedirect,
    REDIRECT *, r)
{

    r->type = REDIRECT_NONE;

    if(r->name != NULL) {
	Deallocate(r->name);
	r->name = NULL;
    }
    if(r->args != NULL) {
	Deallocate(r->args);
	r->args = NULL;
    }

    return TRUE;
}

#if (defined(UNIX) || defined(_WIN32))

/***************************************************************************
*
* Function:  RunFilterSystem
* Purpose:   Feeds the contents of a temporary file through a filter
*            and back to _out. The filter arguments used are in string
*            cmdline.
*
* Arguments: temp - a file descriptor to a temporary file,
*                   must point at the beginning
*            filtername - Name of the filter program to be run, it must
*                         exist in the user search path.
*            cmdline - A string containing the commmandline used as
*                      arguments to the filter. NULL if none.
*
* Returns:   TRUE if the filter operation was successful, FALSE
*            otherwise.
* See also:
*
****************************************************************************
*/
FE2(BOOL, RunFilterSystem,
    char *, filtername,
    char *, cmdline)
{
    char *temp;

    if(filtername == NULL) {
	filtername = "";
    }
    if(cmdline == NULL) {
	cmdline = "";
    }
    temp = (char *) Allocate(strlen(filtername) + 3 + strlen(tempname) + 1
			     + strlen(cmdline) + 1);
    (void) strcpy(temp, filtername);
    (void) strcat(temp, " < ");
    (void) strcat(temp, tempname);
    (void) strcat(temp, " ");
    (void) strcat(temp, cmdline);

#ifdef DEBUG_REDIRECT
    (void) fprintf(_err, "System: \"%s\"\n", temp);
    (void) fflush(_err);
#endif

    /* We make a string which is of the form:

       temp = "filtername < /tmp/temfile cmdline-args"

       it is then executed by "/bin/sh -c temp" on Unix
       and by its moral equivalent in WIN32. */
    
    (void) system(temp);

    Deallocate(temp);
    
    return TRUE;
}
#endif
