/***************************************************************************
*
* Copyright (c) 1989 -- 2000
* by Helsinki University of Technology, Digital Systems Laboratory,
* a.k.a. Laboratory for Theoretical Computer Science.
* All rights reserved.
*
* This software is provided "as is" without express or implied warranty.
*
* Module:  temporal.c
*
* Part of: PROD's program probe
*
* Authors: Keijo Heljanko
*
* Current version:
*     $Id: temporal.c,v 3.1 1997/04/17 15:44:14 kepa Exp $
*
* Description:
*     This module implements model checking for basic temporal CTL formulas.
*     The real "meat" of the CTL model checker is here.
*
* Version history:
*     $Log: temporal.c,v $
 * Revision 3.1  1997/04/17  15:44:14  kepa
 * *** empty log message ***
 *
***************************************************************************/


/***************************************************************************
*
* Static variable rcsid is not used anywhere in the code. The only reason
* for rcsid is the Revision Control System (RCS). RCS fills the template
* string with the latest revision information. C compiler puts the
* resulting literal string into the object file, and linker puts it into
* the executable file. It is then possible to check the revision
* information of the object or executable file. (There is a little
* problem. An optimizing compiler or linker may eliminate the unused
* static variable.)
*
****************************************************************************
*/
#ifdef RCS
static char rcsid[] = "$Id$";
#endif

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "absolute.h"
#include "prod.h"
#include "alloc.h"
#include "dynam.h"
#include "dassert.h"
#include "bitarr.h"
#include "stackpro.h"
#include "temporal.h"



/***************************************************************************
*
* Function:  CheckEX
* Purpose:   Check whether the EX formula f is TRUE or FALSE. If the
*            argument top_level is TRUE, provide a counter-example path.
*            Otherwise label all states with the formula truth value.
*
* Arguments: s -      Starting state.
*
*            f1 -     Subformula on entry.
*            f2 -     Overwritten by f on exit.
*
*            g -      Pointer to graph successor data for finding the
*                     successors of a state.
*            num_states - Number of states in the graph.
*
*            top_level - Does the caller want a conter-example path or to
*                     label all the states in the graph.
*
*            path -   Stack containing the witness path on exit,
*                     if witness requested and found.
*                     Empty stack on exit otherwise.
*
* Returns:   TRUE or FALSE, depending on the truth value of the formula f
*            in state s. If the top_level argument is TRUE and the return
*            value is TRUE, the stack path will contain a witness path.
*            This path consists of two states, s and the witness state s.
*
* See also:
*
****************************************************************************
*/
FE7(BOOL, CheckEX,
    NODE, s,
    BITARRAY *, f1,
    BITARRAY *, f2,
    GRAPH_SUCC *, g,
    unsigned long, num_states,
    BOOL, top_level,
    STACK *, path)
{
    unsigned long i;
    SUCC_NUM j;
    NODE t;
    
    if(top_level == FALSE) {

	/* No witness wanted, just label all the states in some order. */
	for(i=0; i < num_states; i++) {

	    /* Assume f2 = EX f1 to be FALSE */
	    j = 0;
	    ClearBit(f2, i);

	    /* Special case for dead markings. */

	    if(GetSuccessor(g, i, 0) == NO_SUCCESSOR) {
		if(GetBit(f1, i)) {
		    SetBit(f2, i);
		}		
		continue;
	    }
	    
	    /* Loop through all successors, trying to find one
	       in which f1 is TRUE. */
	    
	    while((t = GetSuccessor(g, i, j)) != NO_SUCCESSOR) {
		if(GetBit(f1, t)) {

		    /* A witness found, no need to continue
		       -> the formula is TRUE. */
		    
		    SetBit(f2, i);
		    break;
		}
		
		j++;
	    }
	}

    } else {

	ClearStack(path);

	/* Witness wanted, we must start labeling from state s.
	   Note that all the top-level formula states will NOT have their
	   truth values calculated. */

	/* Assume f2 = EX f1 to be FALSE */
	j = 0;
	ClearBit(f2, s);

	    /* Special case for dead markings. */

	if(GetSuccessor(g, s, 0) == NO_SUCCESSOR) {
	    if(GetBit(f1, s)) {
		SetBit(f2, s);

		/* Prepare the witness. */

		PushStack(path, s);
		PushStack(path, s);
	    }		
	}
	
	while((t = GetSuccessor(g, s, j)) != NO_SUCCESSOR) {
	    if(GetBit(f1, t)) {

		/* A witness found, no need to continue
		   -> he formula is TRUE. */
		    
		SetBit(f2, s);

		/* Prepare the witness. */

		PushStack(path, t);
		PushStack(path, s);

		break;
	    }
	    
	    j++;
	}
    }

    return GetBit(f2, s);
}


/***************************************************************************
*
* Function:  CheckAX
* Purpose:   Check whether the AX formula f is TRUE or FALSE. If the
*            argument top_level is TRUE, provide a counter-example path.
*            Otherwise label all states with the formula truth value.
*
* Arguments: s -      Starting state.
*
*            f1 -     Subformula on entry.
*            f2 -     Overwritten by f on exit.
*
*            g -      Pointer to graph successor data for finding the
*                     successors of a state.
*            num_states - Number of states in the graph.
*
*            top_level - Does the caller want a conter-example path or to
*                     label all the states in the graph.
*
*            path -   Stack containing the counter-example path on exit,
*                     if counter-example requested and found.
*                     Empty stack on exit otherwise.
*
* Returns:   TRUE or FALSE, depending on the truth value of the formula f
*            in state s. If the top_level argument is TRUE and the return
*            value is FALSE, the stack path will contain a counter-example path.
*            This path consists of two states, s and the counter-example state s.
*
* See also:
*
****************************************************************************
*/
FE7(BOOL, CheckAX,
    NODE, s,
    BITARRAY *, f1,
    BITARRAY *, f2,
    GRAPH_SUCC *, g,
    unsigned long, num_states,
    BOOL, top_level,
    STACK *, path)
{
    unsigned long i;
    SUCC_NUM j;
    NODE t;
    
    if(top_level == FALSE) {

	/* No counter-example wanted, just label all the states in some order. */
	for(i=0; i < num_states; i++) {

	    /* Assume f2 = AX f1 to be TRUE */
	    j = 0;
	    SetBit(f2, i);

	    /* Special case for dead markings. */

	    if(GetSuccessor(g, i, 0) == NO_SUCCESSOR) {
		if(!GetBit(f1, i)) {
		    ClearBit(f2, i);
		}		
		continue;
	    }
	    
	    /* Loop through all successors, trying to find one
	       in which f1 is FALSE. */
	    
	    while((t = GetSuccessor(g, i, j)) != NO_SUCCESSOR) {
		if(!GetBit(f1, t)) {

		    /* A counter-example found, no need to continue
		       -> the formula is FALSE. */
		    
		    ClearBit(f2, i);
		    break;
		}
		
		j++;
	    }
	}

    } else {

	ClearStack(path);

	/* Counter-example wanted, we must start labeling from state s.
	   Note that all the top-level formula states will NOT have their
	   truth values calculated. */

	/* Assume f2 = AX f1 to be TRUE */
	j = 0;
	SetBit(f2, s);

	/* Special case for dead markings. */

	if(GetSuccessor(g, s, 0) == NO_SUCCESSOR) {
	    if(!GetBit(f1, s)) {
		ClearBit(f2, s);

		/* Prepare the counter-example. */

		PushStack(path, s);
		PushStack(path, s);
	    }		
	}
	
	while((t = GetSuccessor(g, s, j)) != NO_SUCCESSOR) {
	    if(!GetBit(f1, t)) {

		/* A counter-example found, no need to continue
		   -> the formula is FALSE. */
		    
		ClearBit(f2, s);

		/* Prepare the counter-example. */

		PushStack(path, t);
		PushStack(path, s);

		break;
	    }
	    
	    j++;
	}
    }

    return GetBit(f2, s);
}


/***************************************************************************
*
* Function:  InitBits
* Purpose:   Overwrite the left argument f1 with marked and
*            right argument f2 with f.
* Arguments: 
*            f1 -     Leftmost formula on entry, overwritten by marked on exit.
*            f2 -     Righmost formula on entry, overwritten by f on exit.
*
* Returns:   
* See also:
*
****************************************************************************
*/
FE2(void, InitBits,
    BITARRAY *, f1,
    BITARRAY *, f2)
{
    /*
       The reasoning below is valid for both cases of formulas:
       
       f = A f1 U f2

             or
	     
       f = E f1 U f2

       ---------------------------------------------------------------------

       We want the marked to be 1 for all states for which we don't have to
       look into successor states to decide the truth value of the formula,
       and 0 if visiting that state during depth-first search is needed.

       The f should be initialized to the truth value if it is known,
       and 0 otherwise.

       To reach this goal use the following truth-table to intialize
       both f and marked:

       f1    f2    f1'    f2'
       left  right marked f
       -----------------------
       0     0     1      0
       0     1     1      1
       1     0     0      0
       1     1     1      1

       =>

       f1' = ~f1 | f2
       f2' = f2

       */


    /* f1' = ~f1 | f2 */
    BitArrayNegateS1OrS2OverwriteS1(f1, f2);

    /* f2' = f2 */
    /* Nothing needs to be done. */

}


/***************************************************************************
*
* Function:  CheckAU
* Purpose:   Check whether the AU formula f is TRUE or FALSE. If the
*            argument top_level is TRUE, provide a counter-example path.
*            Otherwise label all states with the formula truth value.
*
* Arguments: s -      Starting state.
*
*            f1 -     Leftmost formula on entry, overwritten by marked on exit.
*            f2 -     Righmost formula on entry, overwritten by f on exit.
*
*            g -      Pointer to graph successor data for finding the
*                     successors of a state.
*            num_states - Number of states in the graph.
*
*            top_level - Does the caller want a conter-example path or to
*                     label all the states in the graph.
*
*            path -   Stack containing the counter-example path on exit,
*                     if counter-example requested and found.
*                     Empty stack on exit otherwise.
*
* Returns:   TRUE or FALSE, depending on the truth value of the formula f
*            in state s. If the top_level argument is TRUE and the return
*            value is TRUE, the stack path will contain a witness path.
*
* See also:
*
****************************************************************************
*/
FE7(BOOL, CheckAU,
    NODE, s,
    BITARRAY *, f1,
    BITARRAY *, f2,
    GRAPH_SUCC *, g,
    unsigned long, num_states,
    BOOL, top_level,
    STACK *, path)
{
    NODE start;
    unsigned long i;
    BOOL result;
    STACK *st;
    STACK *sn;
    
    result = TRUE;
    start = s;

    /* Overwrite leftmost and rightmost subformula by marked and f. */
    
    InitBits(f1, f2);

    /* Allocate search stacks. */

    st = CreateStack();
    sn = CreateStack();


    if(top_level == FALSE) {

	/* No counter-example wanted, just label all the states in some order. */

	for(i=0; i < num_states; i++) {
	    if(!GetBit(f1, i)) {

		/* If the state isn't marked, the truth value of f in it
		   isn't known. */

		result = VisitAU(i, f2, f1, st, sn, g);

		if(result == FALSE) {

		    /* Clear the successor number stack first. */

		    ClearStack(sn);

		    /* We don't need the counter-example path,
		       everything else is cleared nicely also. */

		    ClearStack(st);

		}

	    }

	}
	
    } else {

	ClearStack(path);

	/* Counter-example wanted, we must start labeling from state s, otherwise
	   the counter-example will not be a complete path. Note that all the
	   top-level formula states will NOT have their truth values calculated. */

	result = VisitAU(start, f2, f1, st, sn, g);
	
	if(result == FALSE) {

	    /* Now the stack st contains a counter-example path.
	       We'd like to have it in reverse order though. */

	    /* Clear the successor number stack first. */

	    ClearStack(sn);

	    /* Reverse the counter-example path. */
	    
	    while(!EmptyStack(st)) {
		s = PopStack(st);

		/* Do the reversing. */
		PushStack(path, s);
	    }


	}

    }

    DeleteStack(sn);
    DeleteStack(st);
    
    return GetBit(f2, start);
}



/***************************************************************************
*
* Function:  VisitAU
* Purpose:   Check whether the AU formula f is TRUE in state s and all
*            other states visited during the depth-first search 
* Arguments: s -      Starting state.
*
*            f -      Result bitarray. Initialized to represent truth values
*                     that can be computed without consulting successor
*                     states, otherwise 0.
*
*            marked - Initalized to 1 if the truth value of the formula can
*                     be decided without consulting successor states,
*                     0 otherwise. All states visited will be marked after
*                     return from this procedure.
*
*            st -     Empty search stack for state numbers.
*            sn -     Empty search stack for successor numbers.
*
*            g -      Pointer to graph successor data for finding the
*                     successors of a state.
*
* Returns:   TRUE or FALSE. If the formula is FALSE, the stack st
*            contains a conterexample path starting from s.
*
* See also:
*
****************************************************************************
*/
FE6(BOOL, VisitAU,
    NODE, s,
    BITARRAY *, f,
    BITARRAY *, marked,
    STACK *, st,
    STACK *, sn,
    GRAPH_SUCC *, g)
{
    NODE t;
    SUCC_NUM succ_num;

    for (;;) {

	for (;;) {

	    if(!GetBit(marked, s)) {

		/* A new unmarked state found, mark it first. */

		SetBit(marked, s);


		/* Check that all the children satisfy the formula before
		   marking a state as a satisfying one. */
	    
		succ_num = 0;

		/* Get the first child, which is guaranteed to exist
		   in a total Kripke structure. */

		t = GetSuccessor(g, s, succ_num);

		/* With prod there can be states in which there is
		   no successor, so visit the same node again in that
		   case. */

		PushStack(st, s);
		PushStack(sn, succ_num);

		if(t != NO_SUCCESSOR) {
		    s = t;
		} else {

		    /* There are no successors and the state is not marked
		       => the state does not satisfy righmost subformula
		       and has no successors that could satisfy it 
		       => the formula is FALSE. Visit the state again
		       to notice this fact. */
		    ;
		}
	    
	    } else {

		/* If a state is marked, we already know the truth of the
		   formula in it. We want to find a counterexample with
		   the depth first search. If it isn't found, we have to
		   continue the depth first search trying to find it. */

		if(!GetBit(f, s)) {

		    /* The formula is FALSE -> we have found a counterexample.
		       The stack st contains now the counter-example path,
		       when we append s to it. All states on the path are
		       initalized to FALSE, and have been marked, so there
		       is nothing for us to do. */

		    PushStack(st, s);

		    /* Push a zero on sn just to make it as long as st. */

		    PushStack(sn, 0);

		    return(FALSE);

		} else {

		    /* The formula is TRUE -> we don't have to search the
		       successors of the current state. Therefore we
		       return to the parent state in the depth-first search
		       by doing nothing. */
		
		    ;
		
		}

		break;

	    }

	}

	for (;;) {

	    /* Check if the depth-first search has ending without giving
	       a counterexample. -> The formula is TRUE. */

	    if(EmptyStack(st)) {

		/* All states visited are marked and the formula is
		   set to TRUE in all of them. There is nothing else
		   to do except return. */

		goto RETTRUE;
	    }

	    /* Get parent and the successor number from stack. */
	    
	    s = PopStack(st);
	    succ_num = PopStack(sn);

	    /* Visit the next successor, if there is any. */

	    succ_num++;

	    if((t = GetSuccessor(g, s, succ_num)) != NO_SUCCESSOR) {

		/* Successor found, push the depth-first state on stack. */

		PushStack(st, s);
		PushStack(sn, succ_num);

		/* Visit the successor by exiting the loop. */

		s = t;
		break;
		
	    } else {

		/* All successors visited and no counter-example found so
		   far -> the formula must be TRUE. */

		SetBit(f, s);

		/* Continue the search for a counterexample. */
	    }

	    
	}


    }
 RETTRUE: ;
    return(TRUE);
}



/***************************************************************************
*
* Function:  CheckEU
* Purpose:   Check whether the EU formula f is TRUE or FALSE. If the
*            argument top_level is TRUE, provide a witness path. Otherwise
*            label all states with the formula truth value.
*
* Arguments: s -      Starting state.
*
*            f1 -     Leftmost formula on entry, overwritten by marked on exit.
*            f2 -     Righmost formula on entry, overwritten by f on exit.
*
*            g -      Pointer to graph successor data for finding the
*                     successors of a state.
*            num_states - Number of states in the graph.
*
*            top_level - Does the caller want an witness path or to
*                     label all the states in the graph.
*
*            path -   Stack containing the witness path on exit, if witness
*                     requested and found. Empty stack on exit otherwise.
*
* Returns:   TRUE or FALSE, depending on the truth value of the formula f
*            in state s. If the top_level argument is TRUE and the return
*            value is TRUE, the stack path will contain a witness path.
*
* See also:
*
****************************************************************************
*/
FE7(BOOL, CheckEU,
    NODE, s,
    BITARRAY *, f1,
    BITARRAY *, f2,
    GRAPH_SUCC *, g,
    unsigned long, num_states,
    BOOL, top_level,
    STACK *, path)
{
    NODE start;
    unsigned long i;
    BOOL result;
    STACK *st;
    STACK *sn;
    STACK *cst;
    DFSARRAY *root;
    BITARRAY *modify;
    DFS_NUM *p;
    
    result = TRUE;
    start = s;

    /* Overwrite leftmost and rightmost subformula by marked and f. */
    
    InitBits(f1, f2);

    /* Allocate search stacks. */

    st = CreateStack();
    sn = CreateStack();
    cst = CreateStack();

    /* Allocate dfs array. */

    p = (DFS_NUM *) HallocArray(num_states, sizeof(DFS_NUM));
    root = (DFSARRAY *) p;
    for (i = num_states; i--; *(p++) = (DFS_NUM) 0L)
	;

    /* Allocate modify bitmap for scc-search. Also initialize it to all 0:s. */

    modify = CreateBitArray(num_states);
    
    
    if(top_level == FALSE) {

	/* No witness wanted, just label all the states in some order. */

	for(i=0; i < num_states; i++) {
	    if(!GetBit(f1, i)) {

		/* If the state isn't marked, the truth value of f in it
		   isn't known. */

		result = VisitEU(i, f2, f1, modify, st, sn, cst, root, g);

		if(result != FALSE) {

		    /* Clear the successor number stack first. */

		    ClearStack(sn);

		    /* And then the witness path. Everything
		       has been cleaned up for in the VisitEU
		       procedure. */

		    ClearStack(st);
		
		}

	    }

	}
	
    } else {

	ClearStack(path);

	/* Witness wanted, we must start labeling from state s, otherwise
	   the witness will not be a complete path. Note that all the
	   top-level formula states will NOT have their truth values
	   calculated. */

	result = VisitEU(start, f2, f1, modify, st, sn, cst, root, g);

	if(result != FALSE) {

	    /* Now the stack st contains a witness path.
	       We'd like to have it in reverse order though. */

	    /* Clear the successor number stack first. */

	    ClearStack(sn);

	    /* We want only the witness path but do also the
	       cleaning up just to be nice. */

	    /* Remove the witness state first, because it
	       has already been cleaned up for. */

	    s = PopStack(st);
	    PushStack(path, s);

	    /* Now cleanup the other states. */
	    
	    while(!EmptyStack(st)) {
		s = PopStack(st);

		/* Do the reversing. */
		PushStack(path, s);

		/* No cleanups needed, VisitEU has taken care of
		   them. */
	    }

	}

    }

    DeleteBitArray(modify);
    Dehallocate((char *) root);

    DeleteStack(cst);
    DeleteStack(sn);
    DeleteStack(st);
    
    return GetBit(f2, start);
}

/***************************************************************************
*
* Function:  VisitEU
* Purpose:   Check whether the EU formula f is TRUE in state s and all
*            other states visited during the depth-first search 
* Arguments: s -      Starting state.
*
*            f -      Result bitarray. Initialized to represent truth values
*                     that can be computed without consulting successor
*                     states, otherwise 0.
*
*            marked - Initalized to 1 if the truth value of the formula can
*                     be decided without consulting successor states,
*                     0 otherwise. All states visited will be marked after
*                     return from this procedure.
*
*            modify - Initialized to 0 and left that way after the
*                     routine.
*
*            st -     Empty search stack for state numbers.
*            sn -     Empty search stack for successor numbers.
*
*            cst -    Empty search stack for storing states in strongly
*                     connected component computation.
*
*            root -   Array for storing the deth-first search numbers
*                     for computing strongly connected components.
*                     Initialized to 0 and left that way after the
*                     routine.
*
*            g -      Pointer to graph successor data for finding the
*                     successors of a state.
*
* Returns:   TRUE or FALSE. If the formula is TRUE, the stack st
*            contains a witness path starting from s. All states
*            in stack st must be labeled TRUE after exiting this routine.
*
* See also:
*
****************************************************************************
*/
FE9(BOOL, VisitEU,
    NODE, s,
    BITARRAY *, f,
    BITARRAY *, marked,
    BITARRAY *, modify,
    STACK *, st,
    STACK *, sn,
    STACK *, cst,
    DFSARRAY *, root,
    GRAPH_SUCC *, g)
{
    NODE t;
    SUCC_NUM succ_num;
    DFS_NUM dfs;
    DFS_NUM succ_root;
    
    dfs = DFS_START;
    
    for (;;) {

	for (;;) {

	    if(!GetBit(marked, s)) {

		/* A new unmarked state found, mark it first. */

		SetBit(marked, s);

		/* Check that all the children satisfy the formula before
		   marking a state as a satisfying one. */
	    
		succ_num = 0;

		/* Get the first child, which is guaranteed to exist
		   in a total Kripke structure. */

		t = GetSuccessor(g, s, succ_num);

		/* With prod there can be states in which there is
		   no successor, so visit the same node again in that
		   case. */


		SetDfsArray(root, s, dfs);
		
		/* Clearing the modify bit is done for us beforehand.
		   The modify bits all all guaranteed to be zero
		   on entry to this procedure. */
/*
                ClearBit(modify, s);
*/
		/* Push the state on stack for Tarjan's algorithm. */

		PushStack(cst, s);

		dfs += DFS_ONE;
	    
		PushStack(st, s);
		PushStack(sn, succ_num);

		if(t != NO_SUCCESSOR) {
		    s = t;
		} else {
		    /* There is no successor state => No witness can be
		       found => the formula is FALSE. Because f is
		       initialize to FALSE, there is nothing to do. */
		    ;
		}
		    
	    } else {

		/* If a state is marked, we already know the truth of the
		   formula in it. We want to find a witness with
		   the depth first search. If it isn't found, we have to
		   continue the depth first search trying to find it. */

		if(GetBit(f, s)) {

		    /* The formula is TRUE -> we have found a witness.
		       The stack st contains now the witness-example path,
		       when we append s to it. */

		    PushStack(st, s);

		    /* Push a zero on sn just to make it as long as st. */

		    PushStack(sn, 0);


		    /* We rely on the facts that all the states which are on
		       the cst stack are such that there is a path from them
		       to the witness state s. Thefore they shoul be labeled
		       TRUE.
		       
		       This is a property of the Tarjan's algorithm for strongly
		       connected components. I'll try to come by with a more
		       convincing proof somehow.
		       */

		    while(!EmptyStack(cst)) {
			s = PopStack(cst);

			/* Clean up scc-search structures. */

			SetDfsArray(root, s, 0);
			ClearBit(modify, s);

			/* Set the formula to TRUE. */
			SetBit(f, s);
		    }

		    /* Use the stack st as a counter-example path or
		       destroy it if it is not needed. */
		    
		    goto RETTRUE;

		} else {

		    /* The formula is FALSE -> we don't have to search the
		       successors of the current state. Therefore we
		       return to the parent state in the depth-first search
		       by doing nothing. */
		
		    ;
		
		}

		break;

	    }

	}
	
	for (;;) {

	    /* Check if the depth-first search has ending without giving
	       a witness. -> The formula is FALSE. */

	    if(EmptyStack(st)) {

		/* All states reachable from start state visited and no
		   witness found. The formula must be FALSE in all visited
		   states. Because that was the initialized truth value,
		   there is nothing to do.

		   Also the cst search stack is empty and everything
		   else cleaned up. There is nothing else
		   to do except return. */

		return(FALSE);
	    }

	    /* Get the current child root-number for Tarjan's
	       scc-algorithm. */

	    succ_root = GetDfsArray(root, s);
	    
	    /* Get parent and the successor number from stack. */
	    
	    s = PopStack(st);
	    succ_num = PopStack(sn);


	    /* Check if we can reach a state with a smaller
	       dfs-number through successors. */

	    if((succ_root != 0) &&
	       (GetDfsArray(root, s) > succ_root)) {

		/* We can reach state with dfs number succ_root
		   which was found before the current state in the
		   depth-first search. We record this state
		   and mark that this state can't be a
		   root of a strongly connected component. */
		
		SetDfsArray(root, s, succ_root);
		SetBit(modify, s);
	    }

	    
	    /* Visit the next successor, if there is any. */

	    succ_num++;

	    if((GetSuccessor(g, s, 0) != NO_SUCCESSOR) &&
	       ((t = GetSuccessor(g, s, succ_num)) != NO_SUCCESSOR)) {

		/* Successor found, push the depth-first state on stack. */

		PushStack(st, s);
		PushStack(sn, succ_num);

		/* Visit the successor by exiting the loop. */

		s = t;
		break;
		
	    } else {

		/* All successors visited and no witness found so
		   far -> the formula must be FALSE. Because the
		   array is initialized to FALSE, and all visited
		   states are marked, there is nothing to do. */

		if(!GetBit(modify, s)) {

		    /* We have found a root of a
		       strongly connected component. */

/*		    SetDfsArray(root, s, 0); */

		    /* Pop all states which belong to this component
		       off the stack and do cleanup for them.

		       The formula is FALSE in all states belonging
		       to this strongly connected component, because
		       no witness state could be found amongst states
		       reachable from it. */

		    do {
			t = PopStack(cst);
			SetDfsArray(root, t, 0);
			ClearBit(modify,t);
		    } while(t != s);
		    
		    /* All the states in this strongly connected component
		       are marked and labeled FALSE. Also the root
		       array is cleaned for them. */

		} else {

		    /* All children visited, but s is not a scc-root.
		       Therefore from s one can reach a state found
		       earlier in the depth-first search. Through this
		       successor state s can reach the current search
		       state as long as the strongly connected component
		       s belongs to is not popped off the cst stack. */

		    ;
		}

		/* Continue the search for a witness. */
	    }

	    
	}


    }
 RETTRUE: ;
    return(TRUE);
}
