/**
 *
 */
package fr.lip6.move.pnml.framework.utils;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import fr.lip6.move.pnml.framework.hlapi.HLAPIRootClass;
import fr.lip6.move.pnml.framework.utils.exception.InvalidIDException;
import fr.lip6.move.pnml.framework.utils.exception.VoidRepositoryException;

/**
 * This class gives you access to the different models under processing. Be
 * careful this is a singleton and so NOT thread safe!
 *
 * The usage is simple, at a given moment only one document is under use, to
 * work with another one you have to switch to another opened document.
 *
 * @author Guillaume Giffo
 *
 */
public final class ModelRepository {
    /**
     * the singleton instance.
     */
    private static ModelRepository instance;

    /**
     * the map which contains document workspace's ids and associated document.
     */
    private final Map<String, ModelRepositoryUnit> documents;

    /**
     * The current default ID.
     */
    private String currentDocWSId;

    /**
     * build the ModelRepository if necessary.
     */
    private ModelRepository() {
        documents = new HashMap<String, ModelRepositoryUnit>();
    }

    /**
     * Sets the current selected model.
     *
     * @param documentWSId
     *            the id of the Document Workspace to be set as current.
     * @throws InvalidIDException
     *             if the model id doesn't exists
     * @deprecated use changeCurrentDocWorkspace instead
     */
    public void setCurrentModel(String documentWSId) throws InvalidIDException {
        setCurrentDocWorkspace(documentWSId);
    }

    /**
     * Sets the current selected document workspace.
     *
     * @param documentWSId
     *            the id of the Document Workspace to be set as current.
     * @throws InvalidIDException
     *             if the model id doesn't exists
     * @deprecated use changeCurrentDocWorkspace instead
     */
    public void setCurrentDocWorkspace(String documentWSId)
            throws InvalidIDException {

        changeCurrentDocWorkspace(documentWSId);
    }

    /**
     * Sets the current selected document workspace.
     *
     * @param documentWSId
     *            the id of the Document Workspace to be set as current.
     * @throws InvalidIDException
     *             if the model id doesn't exists
     */
    public void changeCurrentDocWorkspace(String documentWSId)
            throws InvalidIDException {
        if (!documents.containsKey(documentWSId)) {
            throw new InvalidIDException("the Id " + documentWSId
                    + " is not registered");
        }
        currentDocWSId = documentWSId;
    }

    /**
     * Gives the id of the currently used model.
     *
     * @return an id
     * @deprecated use getCurrentDocumentId() instead
     */
    public String getCurrentModelId() {
        return getCurrentDocWSId();
    }

    /**
     * Gives the id of the currently used document workspace.
     *
     * @return an id
     */
    public String getCurrentDocWSId() {
        return currentDocWSId;
    }

    /**
     * Return the ModelRepository singleton.
     *
     * @return the singleton
     */
    public static synchronized ModelRepository getInstance() { // NOPMD by
        // ggiffo on
        // 7/17/08 5:49
        // PM
        if (instance == null) {
            instance = new ModelRepository();
        }
        return instance;
    }

    /**
     * Creates a new model workspace into the repository, this model became the
     * current one.
     *
     * @param modelId
     *            the model Id
     * @return the new model Id
     * @throws InvalidIDException
     *             if the id is already used
     * @deprecated use createDocumentWorkspace(String modelId)
     */
    public String createModelWorkspace(String modelId)
            throws InvalidIDException {
        return createDocumentWorkspace(modelId);
    }

    /**
     * Creates a new model workspace into the repository, this model became the
     * current one.
     *
     * @param docWorkspaceId
     *            the model Id
     * @return the new model Id
     * @throws InvalidIDException
     *             if the id is already used
     */
    public String createDocumentWorkspace(String docWorkspaceId)
            throws InvalidIDException {
        if (documents.containsKey(docWorkspaceId)) {
            throw new InvalidIDException("the Id " + docWorkspaceId
                    + " is already used");
        }
        final IdRepository idrep = new IdRepository();
        final ModelRepositoryUnit mep = new ModelRepositoryUnit(idrep, null);
        documents.put(docWorkspaceId, mep);
        setCurrentModel(docWorkspaceId);
        return docWorkspaceId;
    }

    /**
     * Creates a new model workspace into the repository, this model became the
     * current one.
     *
     * @param modelId
     *            the model Id
     * @param petriNetDoc
     *            the petriNetdoc object, root of the registered model.
     * @return the model Id
     * @throws InvalidIDException
     *             if the id is already used
     * @deprecated
     */
    public String createModelWorkspace(String modelId,
            HLAPIRootClass petriNetDoc) throws InvalidIDException {
        return createDocumentWorkspace(modelId, petriNetDoc);
    }

    /**
     * Creates a new model workspace into the repository, this model becomes the
     * current one.
     *
     * @param docWorkspaceId
     *            an id to give to this document workspace
     * @param petriNetDoc
     *            the petriNetdoc object, root of the registered model.
     * @return the model Id
     * @throws InvalidIDException
     *             if the id is already used
     */
    public String createDocumentWorkspace(String docWorkspaceId,
            HLAPIRootClass petriNetDoc) throws InvalidIDException {
        if (documents.containsKey(docWorkspaceId)) {
            throw new InvalidIDException("the Id " + docWorkspaceId
                    + " is already used");
        }
        final IdRepository idrep = new IdRepository();
        final ModelRepositoryUnit mep = new ModelRepositoryUnit(idrep,
                petriNetDoc);
        documents.put(docWorkspaceId, mep);
        setCurrentModel(docWorkspaceId);
        return docWorkspaceId;
    }

    /**
     * Destroys current modelWorkspace data about a registered model, the current
     * model ID becomes null.
     *
     * @return the number of left opened models.
     * @throws VoidRepositoryException
     *             if the repository is void
     * @deprecated use destroyCurrentDocumentWorkspace() instead
     */
    public int destroyCurrentModelWorkspace() throws VoidRepositoryException {
        return destroyCurrentWorkspace();
    }

    /**
     * Destroys current modelWorkspace data about a registered model, the current
     * model ID becomes null.
     *
     * @return the number of left opened models.
     * @throws VoidRepositoryException
     *             if the repository is void
     */
    public int destroyCurrentWorkspace() throws VoidRepositoryException {
        final String modelId = getCurrentDocWSId();
        if (documents == null || documents.isEmpty()) {
            throw new VoidRepositoryException(
                    "there is no registered models in the repository"); // NOPMD
            // by
            // ggiffo
            // on
            // 9/9/08
            // 10:49
            // AM
        }
        documents.remove(modelId);
        currentDocWSId = null; // NOPMD by ggiffo on 7/17/08 5:48 PM
        return documents.size();
    }

    /**
     * Returns the IdRepository associated to the current model.
     *
     * @return the model IdRepository
     * @throws VoidRepositoryException
     *             if the repository is empty
     * @see fr.lip6.move.pnml.framework.utils.IdRepository
     */
    public IdRepository getCurrentIdRepository() throws VoidRepositoryException {
        if (documents == null || documents.isEmpty()) {
            throw new VoidRepositoryException(
                    "there is no registered models in the repository");
        }
        return documents.get(getCurrentDocWSId()).getIdrep();
    }

    /**
     * Returns the root class associated to the current model.
     *
     * @return the model root object (null if not set).
     * @throws VoidRepositoryException
     *             if the repository is void
     * @see fr.lip6.move.pnml.framework.hlapi.HLAPIRootClass
     */
    public HLAPIRootClass getCurrentHLAPIRootClass()
            throws VoidRepositoryException {
        if (documents == null || documents.isEmpty()) {
            throw new VoidRepositoryException(
                    "there is no registered models in the repository");
        }
        return documents.get(getCurrentDocWSId()).getPetriNetDoc();
    }

    /**
     * Returns the root class associated to the current model.
     *
     * @param hrc :
     *            the root class
     * @throws VoidRepositoryException
     *             if the repository is void
     * @see fr.lip6.move.pnml.framework.hlapi.HLAPIRootClass
     */
    public void setCurrentHLAPIRootClass(HLAPIRootClass hrc)
            throws VoidRepositoryException {
        if (documents == null || documents.isEmpty()) {
            throw new VoidRepositoryException(
                    "there is no registered models in the repository");
        }
        documents.get(getCurrentDocWSId()).setPetriNetDoc(hrc);
    }

    /**
     * Gives all registered id's.
     *
     * @return a set of registered ids.
     */
    public Set<String> getRegisteredModelsId() {
        return documents.keySet();
    }

    /** *********************pretty print tools*********************** */

    /**
     * Sets pretty print use.
     *
     * @param status
     *            a boolean.
     */
    public void setPrettyPrintStatus(boolean status) {
        documents.get(getCurrentDocWSId()).setPrettyPrint(status);
    }

    /**
     * Gets pretty print use status.
     *
     * @return the boolean representing the status.
     */
    public boolean isPrettyPrintActive() {
        return documents.get(getCurrentDocWSId()).isPrettyPrint();
    }

    /**
     * Gets Pretty Print Data.
     *
     * @return PrettyPrintData
     */
    public PrettyPrintData getPrettyPrintData() {
        return documents.get(getCurrentDocWSId()).getPrettyPrinter();
    }

    /** *****************end of pretty print tools******************* */

    /**
     * Gives the number of currently opened WorkSpaces.
     *
     * @return this number.
     */
    public static int size() {
        return getInstance().documents.size();
    }

    /**
     * Destroys the singleton.
     */
    public static void reset() {
        // TODO avoid null usage
        instance = null; // NOPMD by ggiffo on 7/17/08 5:48 PM
    }
}
